/* Copyright (C) 2000-2001 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
#include "InjectorControl_i.h"
#include <CECInetInfo.h>
#include <process.hpp>
#include <ipMutex.hpp>
#include <iostream>
#include <fstream>
#include <assert.h>
#include "TimeService.h"
#include "HTTPInjectorStatus_i.h"

const char* strXMLPrefix="WebRelay";

extern HTTPInjectorStatusImpl* InjectorStatus;

using namespace std;

void* InjectorControlImpl::InjectorControlImpl_TECHECKER(void* args)
{
    InjectorControlImpl* pIC = reinterpret_cast<InjectorControlImpl*>(args);
    assert(pIC);

    CosNaming::NamingContext_var context;
    try
    {
        CORBA::ORB_var orb;
        int i = 0;
        orb = CORBA::ORB_init( i, NULL, "omniORB3" );
        CORBA::Object_var initServ;
        initServ = orb->resolve_initial_references("NameService");
        context = CosNaming::NamingContext::_narrow(initServ);
        if (CORBA::is_nil(context))
        {
            return NULL;
        }
    }
    catch(CORBA::ORB::InvalidName )
    {
    }

    while (true)
    {
        string strTERegName;
        {
            CTLockGuard<CTMutex> lock(pIC->m_TEMutex);

            // Loop through all contexts and get all objects.
            for(TestExecuters_t::const_iterator ppTE = pIC->m_TEs.begin(); ppTE != pIC->m_TEs.end(); ++ppTE)
            {
                if((*ppTE)->m_bTEStopping)
                {
                    // The Test Executer process is stopping. Wait for it to stop,
                    // store the results files and perform cleanup.
                    strTERegName = (*ppTE)->m_strRegName;
                    pIC->WriteLogMsg(20,"Detected that Test Executer '" + strTERegName
                        +"' is stopping");
                    break;
                }
                else
                {
                    // Check whether the Test Executer process has unregistered its name,
                    // in which case simulate a OnExecuterStopping().
                    string strRegName = (*ppTE)->m_strRegName;
                    if(strRegName.empty())
                        continue;
                    CosNaming::Name cosName;
                    cosName.length( 1 );
                    cosName[0].id = CORBA::string_dup(strRegName.c_str());   // string copied
                    cosName[0].kind = CORBA::string_dup(TestExecuter::CLASS_NAME); // string copied

                    CORBA::Object_var obj;
                    try
                    {
                        obj = context->resolve( cosName );
                        if(obj->_non_existent())
                        {
                            context->unbind(  cosName );
                            pIC->OnExecuterStopping(strRegName.c_str());
                        }
                    }
                    catch( CORBA::Exception &)
                    {
                        pIC->OnExecuterStopping(strRegName.c_str());
                    }
                    catch( ... )
                    {
                        // Catch all other exceptions
                    }
                }
            }
        }

        if(!strTERegName.empty())
        {
            // There is a Test Executer process to wait for.
            pIC->TEComplete(strTERegName.c_str());
        }

#ifdef _WIN32
        ::Sleep(2000);
#else
        CTCurrentThread::sleep(2000);
#endif
    }
    return NULL;
}

InjectorControlImpl::InjectorControlImpl(CORBAServer* corba)
    : m_lStopTimeout(30000), m_corba(corba), m_bActivated(false),
      m_pRepository(NULL), m_offset(0)
{
	m_bLocal = true; //assume it's local to start with
	m_nTraceLevel=10;
    // Create log file with name ICLog_IPADR.txt, where IPADR is the
    // IP address (with dots replaced by underscores) of the daemon host.
    char * pOpenStaHome = getenv("OPENSTA_HOME") ;
    if (!pOpenStaHome)
        return ;
    string strLocalPath = pOpenStaHome;

    // Get IP address (with dots replaced by underscores) of daemon host.
    CECInetInfo inet;
    m_registered_name=inet.GetIPAddressAsString();
    string strICLogFilNam = "ICLog_" + m_registered_name + ".txt";
#ifdef _WIN32
    strLocalPath += "Engines\\" + strICLogFilNam;
#else
    strLocalPath += "Engines/" + strICLogFilNam;
#endif

    m_pLogFile = new ofstream( strLocalPath.c_str(), std::ios::binary|std::ios::trunc|std::ios::out);
     
    // Register the interface with the name service
    WriteLogMsg(TRCLEV_NONE,"InjectorControlImpl::InjectorControlImpl - Register IC with name " + m_registered_name);
    m_corba->ActivateObject(this);    
    m_corba->NameObject( this->_this(),InjectorControl::CLASS_NAME,m_registered_name.c_str());
    WriteLogMsg(TRCLEV_NONE,"InjectorControlImpl::InjectorControlImpl - Register IC with name: complete");

    // Start Schedulder thread
    m_pTEChecker = new omni_thread(&InjectorControlImpl_TECHECKER ,this);
    m_pTEChecker->start();

}

InjectorControlImpl::~InjectorControlImpl()
{
    // De-register the interface
     WriteLogMsg(TRCLEV_NONE,"InjectorControlImpl::~InjectorControlImpl - DeRegister IC");
    m_corba->UnNameObject( InjectorControl::CLASS_NAME,m_registered_name.c_str());
    WriteLogMsg(TRCLEV_NONE,"InjectorControlImpl::~InjectorControlImpl - DeRegister : complete");
    delete m_pLogFile;
    m_pLogFile = NULL;
}

CORBA::Boolean InjectorControlImpl::Activate(const char* strManagerRegString, const char* strReposHost,
                            const char* strTestRunString, const char* strTestName)
{
    WriteLogMsg(TRCLEV_NONE,"InjectorControlImpl::Activate");
    m_strTestRunString = strTestRunString;
    m_strManagerRegString = strManagerRegString;
    m_strReposHost = strReposHost;
    WriteLogMsg(TRCLEV_NONE,"TestRunString = " + m_strTestRunString);

    // Only one test can be run at a time.
    if (m_bActivated)
    {
        WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::Activate: Already activated");
        return false;
    }

    try
    {
		// try LAN
		m_pTestManager = NULL;
		m_pXMLTMRelay = NULL;
		m_pXMLTRRelay = NULL;

        m_strTestName = strTestName;

        // Narrow + Store the pointer to TestManager object
        WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::Activate: Find TM on LAN");

#ifndef _WIN32
        CORBA::Object_ptr tmObj = m_corba->FindObjectWithName(TestManager::CLASS_NAME, strManagerRegString);
#else
        CORBA::Object_var tmObj = m_corba->FindObjectWithName(TestManager::CLASS_NAME, strManagerRegString);
#endif
        m_pTestManager = TestManager::_narrow(tmObj);   
		m_bLocal = true; //it's on LAN
    }
    catch(...)
    {
		try
		{
			//try WEB
			m_pTestManager = NULL;
			m_pXMLTMRelay = NULL;
			m_pXMLTRRelay = NULL;
			// Narrow + Store the pointer to the web relay
			WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::Activate: Find TM XMLRelay for web");
				
			CORBA::Object_var XMLobj=m_corba->FindObjectWithName(XMLRelayTestManager::CLASS_NAME, strXMLPrefix); 
			m_pXMLTMRelay = XMLRelayTestManager::_narrow(XMLobj);
			m_bLocal = false; //it's via the WEB
		}
		catch(...)
		{
			WriteLogMsg(TRCLEV_NONE,"Exception: Activate - Failed to find the TestManager anywhere. Resetting for next attempt.");
			Reset();
			return false;
		}
    }

    try
    {
		if(m_bLocal==true)
		{
			// Narrow + Store the pointer to TestRun object on the LAN
			WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::Activate: Find CyrStTestRun on LAN");
			CORBA::Object_var trObj = m_corba->FindObjectWithName(CyrStTestRun::CLASS_NAME, strTestRunString);
			m_pTestRun = CyrStTestRun::_narrow(trObj);
		}
		else 
		{
			// do something via the web
			WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::Activate: Find XMLRelayTestRun via the WEB");
			CORBA::Object_var trObj = m_corba->FindObjectWithName(XMLRelayTestRun::CLASS_NAME, "WebRelay");
			m_pXMLTRRelay = XMLRelayTestRun::_narrow(trObj);
		}
    }
    catch(...)
    {
        WriteLogMsg(TRCLEV_NONE,"Exception: Activate - Failed to find the TestRun anywhere. Resetting for next attempt.");
        Reset();
        return false;
    }

    try
    {
			// Create a connection to the repository on specified host on LAN
			string strTmp = m_strTestRunString;
			int nPos = strTmp.find_last_of('~');
			strTmp.erase(0, nPos+1);
			m_strTestRunFolder = strTmp;

			WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::Activate: Connect to repository on LAN");
			m_pRepository = new CRemoteRepository(m_corba, m_strReposHost, m_strTestName, m_strTestRunFolder);
    }
    catch (...)
    {
        WriteLogMsg(TRCLEV_NONE,"Exception: Activate - Failed to initialize the Repository Connection from the test run pointer. Resetting for next attempt.");
        if(!CORBA::is_nil(m_pTestRun))
        {
            CECErrorLog msg(TRCLEV_NONE,strTestName,InjectorControl::CLASS_NAME, "Activate: Failed to initialize the Repository Connection anywhere. Resetting for next attempt.");
            m_pTestRun->AddTestOutput(CECErrorLog::OUTPUT_TYPE, msg.ToString().c_str());
        }
        Reset();
        return false;
    }

	try
	{
		//get the trace level setting from the repository
		WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::Activate: Get Trace Level LAN");
		string tracelevel=m_pRepository->GetTraceLevel("General");
		m_nTraceLevel=atoi((char*)tracelevel.c_str());
		if(m_nTraceLevel==0 && tracelevel!="0")
		{
			//atoi error use default of 10
			m_nTraceLevel=10;
		}
		WriteLogMsg(TRCLEV_LOW,"InjectorControlImpl::Activate: Trace Level = "+tracelevel);
	}
	catch(...)
	{
		m_nTraceLevel=10;
		WriteLogMsg(TRCLEV_NONE,"InjectorControlImpl::Activate: Exception couldn't get trace level from anywhere");
	}
	
    try
    {
		if(m_bLocal==true)
		{
			// Get the time offset from LAN
			WriteLogMsg(TRCLEV_MEDIUM,"Finding time offset on LAN");
			CECInetInfo inet;
			CORBA::Object_ptr tsObj = m_corba->FindObjectWithName(TimeService::CLASS_NAME, inet.GetIPAddressAsString(strReposHost).c_str());
			TimeService_var TS = TimeService::_narrow(tsObj);
		   long lSvrTime = TS->GetTime();

		   if(lSvrTime == -1)
		   {
				// -1 indicates failure
				WriteLogMsg(TRCLEV_MEDIUM,"Time offset from LAN Error, defaulting to 0");
				m_offset=0;
		   }
		   else
		   {
				//valid time returned
			    CTTime tm;
			    tm.now();
			    m_offset = tm.getSeconds() - lSvrTime;

			    char strOffset[32];
			    sprintf(strOffset, "Time Offset from LAN = %d", m_offset);
			    WriteLogMsg(TRCLEV_MEDIUM,strOffset);
		   }
		}
		else
		{
			// Get the time offset from the web	
			WriteLogMsg(TRCLEV_MEDIUM,"Finding time offset via WEB");

			CORBA::Object_var obj=m_corba->FindObjectWithName(XMLRelayTimeService::CLASS_NAME, strXMLPrefix); 
			XMLRelayTimeService_var TS = XMLRelayTimeService::_narrow(obj);
			long lSvrTime = TS->GetTime(strReposHost);
			if(lSvrTime == -1)
			{
				// -1 indicates failure
				WriteLogMsg(TRCLEV_MEDIUM,"Time offset from WEB Error, defaulting to 0");
				m_offset=0;
			}
			else
			{
				//valid time returned
				CTTime tm;
				tm.now();
				m_offset = tm.getSeconds() - lSvrTime;

				char strOffset[32];
				sprintf(strOffset, "Time Offset via web = %d", m_offset);
				WriteLogMsg(TRCLEV_MEDIUM,strOffset);
			}
		}
	}	
	catch(...)
	{
        WriteLogMsg(TRCLEV_NONE,"Exception: Activate - Failed to get time offset from anywhere");
    }

    m_bActivated = true;
    WriteLogMsg(TRCLEV_NONE,"InjectorControlImpl::Activate: Activated successfully");
    return true;
}

CORBA::Boolean InjectorControlImpl::StopTest(CORBA::Boolean bForce, long lTimeout)
{
    WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::StopTest");

#ifdef _WIN32
    m_lStopTimeout = (lTimeout == 0 ? INFINITE : lTimeout);
#else
    m_lStopTimeout = (lTimeout == 0 ? 30000 : lTimeout);
#endif
    TestExecuters_t TEs;
    
    {
        CTLockGuard<CTMutex> lock(m_TEMutex);
        TEs = m_TEs;
    }

	long lTotalTEs = 0;
    // Loop through all contexts and get all objects.
    for(TestExecuters_t::iterator ppTE = TEs.begin(); ppTE != TEs.end(); ++ppTE)
    {
		lTotalTEs++;
        try
        {
            if(bForce)
            {
	            WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::StopTest: Force Kill TE");
                // Kill the Test Executer process.
#ifdef _WIN32
                HANDLE hProcess = (HANDLE)(*ppTE)->m_pid;
                if(!TerminateProcess(hProcess, 0))
                    return false;
                CloseHandle(hProcess);
#else
                int iPid = (int)(*ppTE)->m_pid;
				kill(iPid, SIGKILL);
#endif
                // Notify the Test Manager.
                OnExecuterStopping((*ppTE)->m_strRegName.c_str());
            }
            else
            {
                // Ask the Test Executer for a graceful shutdown.
	            WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::StopTest: Calling Stop on TE");
				if (!CORBA::is_nil((*ppTE)->m_pTE))
				{
	                (*ppTE)->m_pTE->Stop();
				}
            }
        }
        catch(...)
        {
            WriteLogMsg(TRCLEV_NONE,"Exception: StopTest");
        }
    }

	if(lTotalTEs==0)
	{
		//then as there should be some the "OnInjectorStopped()" call will not be made and this injector will not stop
		CECInetInfo inet;
        WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::StopTest: Number of executers == 0");
        try
        {
			if((m_bLocal==true)&&(m_pTestManager))
			{
				WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::StopTest: call TM OnInjStopped on LAN");
				m_pTestManager->OnInjectorStopped(inet.GetHostName().c_str());
			}
			else if((m_bLocal==false)&&(m_pXMLTMRelay))
			{
				WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::StopTest: call TM OnInjStopped via WEB");
				m_pXMLTMRelay->OnInjectorStopped(m_strReposHost.c_str(),m_strManagerRegString.c_str(),inet.GetHostName().c_str());
			}
			else
			{
				WriteLogMsg(TRCLEV_NONE,"InjectorControlImpl::StopTest: call TM OnInjStopped: No pointer to TM on LAN or WEB");
			}
        }
        catch(...)
        {
        }
        WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::StopTest: call Reset");
        Reset();
    }

    return false;
}

CORBA::Boolean InjectorControlImpl::StartTaskGroup(TaskGroupExecInfo& tgExec)
{
    WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::StartTaskGroup");

    CTLockGuard<CTMutex> lock(m_TEMutex);

    try
    {
        SharedPtr<TEInfo> pTE;

        // For web task groups we add them to the same executer. So, check for an existing
        // executer in our internal list.
        string strTarget(tgExec.strTarget);
        if (strTarget == "HTP")
        {
            WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::StartTaskGroup: Search for existing HTP TE");

            for(TestExecuters_t::const_iterator ppTE = m_TEs.begin(); ppTE != m_TEs.end(); ++ppTE)
            {
                if ((*ppTE)->m_strTarget == strTarget)
                {
                    WriteLogMsg(TRCLEV_MEDIUM,"Found an existing HTP executer.");
                    pTE = *ppTE;
                    break;
                }
            }
        }

        if (pTE.IsNull())
        {
			string strTrcLev = "10";
			if(m_bLocal==true)
			{
				string strTrcLev = m_pRepository->GetTraceLevel(strTarget);
			}
            string strMsg = "InjectorControlImpl::StartTaskGroup: Trace Level For Target "
                + strTarget + "=" + strTrcLev;
            WriteLogMsg(TRCLEV_MEDIUM,strMsg);
            // The Test Executer object is still nil so we have to start a new executer
            string cmdline = "\"";
            cmdline += getenv("OPENSTA_HOME");
            cmdline +="Engines/TExecuter_";
            cmdline += strTarget + "\"";
            cmdline += " -v " + strTrcLev + " -n ";
            cmdline += m_strTestName;
            cmdline += " -r ";
            cmdline += m_strReposHost;
            cmdline += " -t ";
            cmdline += "\""+m_strTestRunString+"\"";

            // Create a signal with auto-generated name.
            CIPSignal condition(NULL, 30000);
            if(!condition.isValid())
                return false;
            cmdline += " -s " + condition.getName();

            WriteLogMsg(TRCLEV_LOW,"InjectorControlImpl::StartTaskGroup: Start TE: "+cmdline);

            CTProcess proc(cmdline, false);

            // Construct the expected registered name and attempt to narrow the TestExecuter
            if (proc.start(true) == 0 && proc.isRunning())
            {
                if(condition.wait() != CT_WAIT_OK)
                    return false;

                char strReg[64];
                CECInetInfo inet;
#ifdef _WIN32
                sprintf(strReg, "%s_%8.8lX", inet.GetIPAddressAsString().c_str(), proc.getActualPid());
#else
                sprintf(strReg, "%s_%8.8lX", inet.GetIPAddressAsString().c_str(), proc.getPid());
#endif
                WriteLogMsg(TRCLEV_MEDIUM,"***Expected executer regname = "+string(strReg));

                WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::StartTaskGroup: Narrow TE obj");
                CORBA::Object_var teObj = m_corba->FindObjectWithName(TestExecuter::CLASS_NAME, strReg);

                TEInfo* pTei = new TEInfo();
                pTei->m_pTE = TestExecuter::_narrow(teObj);
                pTei->m_strTarget = tgExec.strTarget;
                pTei->m_tgtType = tgExec.tgtType;
                pTei->m_strRegName = strReg;
                pTei->m_bTEStopping = false;
#ifdef _WIN32
                pTei->m_hProcess = (HANDLE)proc.getPid();
                pTei->m_pid = proc.getActualPid();
#else
                pTei->m_pid = (long)proc.getPid();
#endif

                SharedPtr<TEInfo> sp(pTei);
                pTE = sp;
                m_TEs.push_back(pTE);
            }
            else
            {
                WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::StartTaskGroup: failed to start the Test Executer process.");
                return false;
            }
        }

        // The correct executer for this task group is now running so we can now invoke it
        // so start the execution of the task group
        WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::StartTaskGroup: Call StartTaskGroup on TE obj");
        bool bRet = pTE->m_pTE->StartTaskGroup(tgExec);
        if (bRet)
        {
            m_TaskGroupToTE.insert(TaskGroupToTE_t::value_type(tgExec.lTaskGroupId, pTE));
            WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::StartTaskGroup: TaskGroup started.");
        }
        else
        {
            WriteLogMsg(TRCLEV_LOW,"InjectorControlImpl::StartTaskGroup: TaskGroup failed.");
        }
        return bRet;
    }
    catch(...)
    {
        WriteLogMsg(TRCLEV_NONE,"InjectorControlImpl::StartTaskGroup: Exception");
    }

    return false;
}

CORBA::Boolean InjectorControlImpl::StopTaskGroup(long nTaskGroupID)
{
    char strMsg[64];
    sprintf(strMsg, "StopTaskGroup called for TG %d", nTaskGroupID);
    WriteLogMsg(TRCLEV_LOW,strMsg);
    try
    {
        WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::StopTaskGroup");

        CTLockGuard<CTMutex> lock(m_TEMutex);

        TaskGroupToTE_t::iterator pTEI = m_TaskGroupToTE.find(nTaskGroupID);
        if(pTEI != m_TaskGroupToTE.end())
        {
            // Call StopTaskGroup on this executer
            WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::StopTaskGroup: Call StopTaskGroup on TE");
            pTEI->second->m_pTE->StopTaskGroup(nTaskGroupID);
        }
    }
    catch(...)
    {
        WriteLogMsg(TRCLEV_NONE,"InjectorControlImpl::StopTaskGroup: Exception");
    }

    return true;
}

TestExecuter_ptr InjectorControlImpl::GetTaskGroupExecuter(long nTaskGroupID)
{
    try
    {
        WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::GetTaskGroupExecuter: Search for TE");

        CTLockGuard<CTMutex> lock(m_TEMutex);

        TaskGroupToTE_t::iterator pTEI = m_TaskGroupToTE.find(nTaskGroupID);
        if(pTEI != m_TaskGroupToTE.end())
        {
            WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::GetTaskGroupExecuter: Found TE");
            return TestExecuter::_duplicate(pTEI->second->m_pTE);
        }
    }
    catch(...)
    {
        WriteLogMsg(TRCLEV_NONE,"InjectorControlImpl::GetTaskGroupExecuter: Exception");
    }
    return TestExecuter::_nil();
}

void InjectorControlImpl::OnTaskGroupStopped(long nTaskGroupID, TaskGroupStopReason& eReason)
{
    char strMsg[64];
    sprintf(strMsg, "OnTaskGroupStopped called for TG %d", nTaskGroupID);
    WriteLogMsg(TRCLEV_MEDIUM,strMsg);
    try
    {
        WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::OnTaskGroupStopped: Find TE");
        bool bStop = false;

        {
            CTLockGuard<CTMutex> lock(m_TEMutex);

            TaskGroupToTE_t::iterator pTEI = m_TaskGroupToTE.find(nTaskGroupID);
            if(pTEI != m_TaskGroupToTE.end())
            {
                WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::OnTaskGroupStopped: Erase Info");
                m_TaskGroupToTE.erase(pTEI);
                bStop = true;
            }
        }

        if(bStop)
        {
			//the pointers are defined in the activate function
			if((m_bLocal==true)&&(m_pTestManager))
			{
				WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::OnTaskGroupStopped: Call TM OnTGStopped on LAN");
				m_pTestManager->OnTaskGroupStopped(nTaskGroupID, eReason);
			}
			else if((m_bLocal==false)&&(m_pXMLTMRelay))
			{
				WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::OnTaskGroupStopped: Call TM OnTGStopped via WEB");
				m_pXMLTMRelay->OnTaskGroupStopped(m_strReposHost.c_str(),m_strManagerRegString.c_str(), nTaskGroupID, eReason);
			}
			else
			{
				WriteLogMsg(TRCLEV_NONE,"InjectorControlImpl::OnTaskGroupStopped: No pointer to TM on LAN or WEB");
			}
        }
    }
    catch(...)
    {
        WriteLogMsg(TRCLEV_NONE,"InjectorControlImpl::OnTaskGroupStopped: Exception");
    }
}

void InjectorControlImpl::OnExecuterStopping(const char* strExecuterRegName)
{
    WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::OnExecuterStopping");
    {
        CTLockGuard<CTMutex> lock(m_TEMutex);

        for(TestExecuters_t::const_iterator ppTE = m_TEs.begin(); ppTE != m_TEs.end(); ++ppTE)
        {
            if((*ppTE)->m_strRegName == strExecuterRegName)
            {
                // Mark the Test Executer as stopping.
                (*ppTE)->m_bTEStopping = true;
                break;
            }
        }
    }
}

long InjectorControlImpl::GetTestVUs()
{
    WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::GetTestVUs");
	long retVal = 0;
	try
	{
		if((m_bLocal==true)&&(m_pTestManager))
		{
			WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::GetTestVUs: on LAN");
			retVal = m_pTestManager->GetTestVUs();
		}
		else if((m_bLocal==false)&&(m_pXMLTMRelay))
		{
			WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::GetTestVUs: via WEB");
			retVal = m_pXMLTMRelay->GetTestVUs(m_strReposHost.c_str(),m_strManagerRegString.c_str());
		}
		else
		{
			WriteLogMsg(TRCLEV_NONE,"InjectorControlImpl::GetTestVUs:: No pointer to TM on LAN or WEB");
		}
		
	}
	catch(...)
	{
	    WriteLogMsg(TRCLEV_NONE,"Exception: GetTestVUs");
	}
	return retVal;
}

CORBA::Boolean InjectorControlImpl::GetTofFile(const char* strTofFile)
{
	WriteLogMsg(TRCLEV_NONE,"**** In GetTofFile");
    if (!m_bActivated || !m_pRepository)
	{
		 WriteLogMsg(TRCLEV_LOW,"Not Activated.");
        return false;
	}

	WriteLogMsg(TRCLEV_NONE,"**** Call Repos GetTofFile");
    bool bRet = false;
	
	bRet = m_pRepository->CheckoutTofFile(strTofFile);
    if (bRet)
		WriteLogMsg(TRCLEV_NONE,"**** GetTofFile: return true");
	else
		WriteLogMsg(TRCLEV_NONE,"**** GetTofFile: return false");

	return bRet;
}

CORBA::Boolean InjectorControlImpl::GetProfileFile(const char* strProfileFile)
{
    WriteLogMsg(TRCLEV_NONE,"**** In GetProfileFile");
    if (!m_bActivated)
	{
	    WriteLogMsg(TRCLEV_LOW,"Not Activated.");
        return false;
	}

    WriteLogMsg(TRCLEV_NONE,"**** Call Repos CheckoutProfileFile");
    bool bRet = false;
	
	bRet = m_pRepository->CheckoutProfileFile(strProfileFile);
    if (bRet)
		WriteLogMsg(TRCLEV_NONE,"**** GetProfileFile: return true");
	else
		WriteLogMsg(TRCLEV_NONE,"**** GetProfileFile: return false");

	return bRet;
}

CORBA::Boolean InjectorControlImpl::GetDataFile(const char* strDataFile)
{
    WriteLogMsg(TRCLEV_NONE,"**** In GetDataFile");
    if (!m_bActivated || !m_pRepository)
        return false;

    return m_pRepository->CheckoutDataFile(strDataFile);
}

CORBA::Boolean InjectorControlImpl::SetResultsFile(const char* strFilename)
{
    WriteLogMsg(TRCLEV_MEDIUM,"SetResultsFile: "+string(strFilename));

    CTLockGuard<CTMutex> lock(m_OpenResultsFilesListMutex);
    m_OpenResultsFilesList.push_back(strFilename);

    return true;
}

CORBA::Boolean InjectorControlImpl::StoreResultsFile(const char* strLocalResultsFile)
{
    WriteLogMsg(TRCLEV_NONE,"***StoreResultsFile: " + string(strLocalResultsFile));
    bool bRet = m_pRepository->StoreResultsFile(strLocalResultsFile);

    // Erase the file name from the "file to store" collection.
    {
        CTLockGuard<CTMutex> lock(m_OpenResultsFilesListMutex);
        for (int i=0; i<(int)m_OpenResultsFilesList.size(); i++)
        {
            if (m_OpenResultsFilesList[i] == strLocalResultsFile)
            {
                m_OpenResultsFilesList.erase(m_OpenResultsFilesList.begin() + i);
                break;
            }
        }
    }

    const char* const zMsg = bRet ? "StoreResultsFile succeeded for file '"
        : "Failed to store file '";
    WriteLogMsg(TRCLEV_NONE,string(zMsg) + strLocalResultsFile + "'.");

    return bRet;
}

void InjectorControlImpl::Reset()
{
    m_offset = 0;

    m_pTestRun = CyrStTestRun_var();
#ifndef _WIN32
    CORBA::release(m_pTestManager);
#else
	m_pTestManager = TestManager_var();
#endif

	m_pXMLTMRelay = XMLRelayTestManager_var();
	//set null?
    if (m_pRepository)
    {
        delete m_pRepository;
        m_pRepository = NULL;
    }

    m_bActivated = false;
	m_bLocal = true; //reset it back to local LAN

    return;   
}

void InjectorControlImpl::WriteLogMsg(int severity,const string& msg)
{
	/*
	TRCLEV_NONE          0              // Fatal Errors only
	TRCLEV_LOW          10              // Important Trace error
	TRCLEV_MEDIUM       20              // Normal Trace Level
	TRCLEV_HIGH         50              // High Trace level
    */

    static CTMutex mutex;

    CTLockGuard<CTMutex> lock(mutex);

    // m_pLogFile may be NULL for a short period between closing the log file
    // for one test-run and opening one for the next, so ignore any messages
    // that the IC attempts to log inbetween.
    if (m_pLogFile == NULL) return; // Cannot write, if no log file object, so ignore

	if(severity>m_nTraceLevel)
	{
		//it's too minor
		return;
	}

    if(m_pLogFile->good())
    {
		CTTime t;
        t.now();
        string strTim(t.Format("%Y-%m-%d %H:%M:%S "));
        m_pLogFile->write(strTim.c_str(), strTim.length());

        m_pLogFile->write( msg.c_str(), msg.length() );
        m_pLogFile->write( "\r\n", 2 );
        m_pLogFile->flush();
    }
}

void InjectorControlImpl::AddTestOutput(const char* strType, const char* strOutput)
{
	try
	{
	    static CTMutex mutex;
	    CTLockGuard<CTMutex> lock(mutex);

		if (m_bLocal)
		{
			if (!CORBA::is_nil(m_pTestRun))
			    m_pTestRun->AddTestOutput(strType, strOutput);
		}
		else
		{
			if (!CORBA::is_nil(m_pXMLTRRelay))
			    m_pXMLTRRelay->AddTestOutput(m_strReposHost.c_str(), m_strTestRunString.c_str(), strType, strOutput);
		}
	}
	catch(...)
	{
	    WriteLogMsg(TRCLEV_NONE,"Exception: AddTestOutput");
	}
}

CORBA::Boolean InjectorControlImpl::IsActivated(const char* strManagerRegString)
{
    return m_bActivated && m_strManagerRegString == strManagerRegString;
}

void InjectorControlImpl::TEComplete(const char* strExecuterRegName)
{
    WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::TEComplete for '" + string(strExecuterRegName)
        +"'.");
    try
    {
        // Cleanup.

        bool bNoMoreTEs = false;
        long lPid;

        {
            CTLockGuard<CTMutex> lock(m_TEMutex);

            // Make sure that all task groups entries belonging to this Test Executer are
            // erased.
            for(TaskGroupToTE_t::iterator pTG = m_TaskGroupToTE.begin(); pTG != m_TaskGroupToTE.end(); )
            {
                if (pTG->second->m_strRegName == strExecuterRegName)
                    m_TaskGroupToTE.erase(pTG++);
                else
                    ++pTG;
            }

            // Erase the Test Executer entry.
            for(TestExecuters_t::iterator ppTE = m_TEs.begin(); ppTE != m_TEs.end(); ++ppTE)
            {
                if((*ppTE)->m_strRegName == strExecuterRegName)
                {
#ifdef _WIN32
                    lPid = (long)(*ppTE)->m_hProcess;
#else
                    lPid = (*ppTE)->m_pid;
#endif
                    m_TEs.erase(ppTE);
                    break;
                }
            }

            bNoMoreTEs = m_TEs.empty();
        }

        // Wait for the actual process to terminate.
#ifdef _WIN32
        DWORD dwRet = WaitForSingleObject((HANDLE)lPid, (DWORD)m_lStopTimeout);
        CloseHandle((HANDLE)lPid);
        if(dwRet == WAIT_FAILED)
            WriteLogMsg(TRCLEV_LOW,"Error: failed to wait for Test Executer process to complete.");
        else if(dwRet == WAIT_TIMEOUT)
            WriteLogMsg(TRCLEV_LOW,"Error: timed out while waiting for Test Executer process to complete.");
#else
        int child_sts;
        waitpid(lPid, &child_sts, 0);
#endif

        // If all execters/task groups have finished,
        // inform manager that we have finished and reset for net run
        WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::OnExecuterStopping");
        if (bNoMoreTEs)
        {

			CECInetInfo inet;
            WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::TEComplete: gTEInfoList.size() == 0");
            try
            {
				if((m_bLocal==true)&&(m_pTestManager))
				{
					WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::TEComplete: call TM OnInjStopped on LAN");
					m_pTestManager->OnInjectorStopped(inet.GetHostName().c_str());
				}
				else if((m_bLocal==false)&&(m_pXMLTMRelay))
				{
					WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::TEComplete: call TM OnInjStopped via WEB");
					m_pXMLTMRelay->OnInjectorStopped(m_strReposHost.c_str(),m_strManagerRegString.c_str(),inet.GetHostName().c_str());
				}
				else
				{
					WriteLogMsg(TRCLEV_NONE,"InjectorControlImpl::TEComplete: call TM OnInjStopped: No pointer to TM on LAN or WEB");
				}
            }
            catch(...)
            {
            }
            WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::TEComplete: call Reset");
            Reset();
        }
    }
    catch(...)
    {
        WriteLogMsg(TRCLEV_NONE,"InjectorControlImpl::TEComplete: Exception");
    }
}

long InjectorControlImpl::TransferResults(const char* strTestName, 
										const char* strReposHost, const char* strRunFolder)
{
	//stop the httpstats file in IS being written
	try
	{
		InjectorStatus->StopMonitoring();
	}
	catch(...)
	{ 
		WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::Couldn't stop HTTPStats file being written by HTTP IS");
	}
	long nFilesTransfered = 0;
	try
    {
        // Create a connection to the repository on specified host
        CRemoteRepository* pRepository = new CRemoteRepository(m_corba, strReposHost, strTestName, strRunFolder);

        {
		    CTLockGuard<CTMutex> lock(m_OpenResultsFilesListMutex);

		    while (m_OpenResultsFilesList.size() > 0)
		    {
			    string strFile = m_OpenResultsFilesList[0];
			    WriteLogMsg(TRCLEV_MEDIUM,"InjectorControlImpl::TransferResults: Store file: "+strFile);

				for (int nTry = 0; nTry<=3; nTry++)
				{
				    if(pRepository->StoreResultsFile(strFile.c_str()))
					{
						m_OpenResultsFilesList.erase(m_OpenResultsFilesList.begin());
					    WriteLogMsg(TRCLEV_MEDIUM,"StoreResultsFile succeeded for file '" + strFile + "'.");
					    nFilesTransfered++;
						break;
					}
					else
					{
					    WriteLogMsg(TRCLEV_MEDIUM,"Failed to store file '" + strFile + "'.");
					}

					if (nTry == 3)
					{
						m_OpenResultsFilesList.erase(m_OpenResultsFilesList.begin());
					    CECInetInfo inet;
						string strMsg;
						strMsg = "All attempts to store results file "+strFile+" from machine "+inet.GetHostName()+" have failed.";
						CECErrorLog msg(TRCLEV_NONE,m_strTestName.c_str(),InjectorControl::CLASS_NAME, strMsg.c_str());
					    WriteLogMsg(TRCLEV_NONE,msg.ToString().c_str());
					}
				}
		    }
        }

		// Also, store the IC Log file in the Repository.
		m_pLogFile->flush();
		delete m_pLogFile;
		m_pLogFile = NULL;
		string strICLogFilNam = "ICLog_" + m_registered_name + ".txt";
		if(pRepository->StoreResultsFile(strICLogFilNam.c_str()))
			nFilesTransfered++;

	    // Create new IC Log file for next test-run.
	    // Create log file with name ICLog_IPADR.txt, where IPADR is the
	    // IP address (with dots replaced by underscores) of the daemon host.
	    char * pOpenStaHome = getenv("OPENSTA_HOME") ;
	    string strLocalPath = pOpenStaHome;
	
#ifdef _WIN32
	    strLocalPath += "Engines\\" + strICLogFilNam;
#else
	    strLocalPath += "Engines/" + strICLogFilNam;
#endif

		delete pRepository;
		pRepository = NULL;

	    m_pLogFile = new ofstream( strLocalPath.c_str(), std::ios::binary|std::ios::trunc|std::ios::out);
		return nFilesTransfered;
	}
	catch(...)
	{
	    WriteLogMsg(TRCLEV_NONE,"Exception: TransferResults");
	}
	return nFilesTransfered;
}


int InjectorControlImpl::GetTraceLevel()
{
	//the IS asks for this trace setting "General"
	return m_nTraceLevel;
}
