/* Copyright (C) 2000-2001 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
#include "StepByStep_Control_i.h"
#include <CECInetInfo.h>

StepByStepControlImpl::StepByStepControlImpl(CORBAServer* corba, InjectorControlImpl& ic)
    : m_corba(corba), m_ic(ic)
{
    char * pOpenStaHome = getenv("OPENSTA_HOME") ;
    if(!pOpenStaHome)
        return;

    m_LogFile = pOpenStaHome;

    // Get IP address (with dots replaced by underscores) of daemon host.
    CECInetInfo inet;
    m_registered_name=inet.GetIPAddressAsString();
    m_LogFileBaseName = "SBSCLog_" + m_registered_name + ".txt";
#ifdef WIN32
    m_LogFile += "Engines\\" + m_LogFileBaseName;
#else
    m_LogFile += "Engines/" + m_LogFileBaseName;
#endif
    m_pLogFile = new std::ofstream(m_LogFile.c_str(), std::ios::binary|std::ios::trunc|std::ios::out);
    if(m_pLogFile)
        m_ic.SetResultsFile(m_LogFileBaseName.c_str());

    // Register the interface with the name service.
    WriteLogMsg("StepByStepControlImpl::StepByStepControlImpl - Register SBSC with name " + m_registered_name);
    m_corba->ActivateObject(this);    
    m_corba->NameObject(this->_this(), StepByStepControl::CLASS_NAME, m_registered_name.c_str());
    WriteLogMsg("StepByStepControlImpl::StepByStepControlImpl - Register SBSC with name: complete");
}

StepByStepControlImpl::~StepByStepControlImpl()
{
    // De-register the interface.
    m_corba->UnNameObject(StepByStepControl::CLASS_NAME, m_registered_name.c_str());

    delete m_pLogFile;
    m_pLogFile = NULL;
}

void StepByStepControlImpl::SetCommander(StepByStepCommander_ptr refCommander)
{
    if(CORBA::is_nil(refCommander))
        WriteLogMsg(">>>SetCommander(nil)");
    else
        WriteLogMsg(">>>SetCommander");
    m_refCommander = StepByStepCommander::_duplicate(refCommander);
}

void StepByStepControlImpl::SetTestExecuter(StepByStepScriptExecuter_ptr refTestExecuter)
{
    m_refTestExecuter = StepByStepScriptExecuter::_duplicate(refTestExecuter);

    if(CORBA::is_nil(refTestExecuter))
    {
        if(m_pLogFile)
        {
            WriteLogMsg(">>>SetTestExecuter(nil)");
		    m_pLogFile->flush();
		    delete m_pLogFile;
		    m_pLogFile = NULL;
        }
    }
    else
    {
        WriteLogMsg(">>>SetTestExecuter");
        if(!m_pLogFile)
        {
            m_pLogFile = new std::ofstream(m_LogFile.c_str(), std::ios::binary|std::ios::trunc|std::ios::out);
            if(m_pLogFile)
                m_ic.SetResultsFile(m_LogFileBaseName.c_str());
        }
    }
}

CORBA::Boolean StepByStepControlImpl::Next(CORBA::Long lTaskGroupID, CORBA::Long lVUId)
{
    CORBA::Boolean bRet = 0;

    GetLogStream() << ">>>Next: task group = " << lTaskGroupID
        << ", VU = " << lVUId
        << "\r\n";
    ReleaseLogStream();

    // Forward to Test Executer.
    if(!CORBA::is_nil(m_refCommander))
    {
        try {
            m_refTestExecuter->Next(lTaskGroupID, lVUId);
            bRet = 1;
        } catch(...)
        {
        }
    }

    return bRet;
}

CORBA::Boolean StepByStepControlImpl::IsStepByStep()
{
    // Forward to Commander.
    CORBA::Boolean bRet = 0;
    if(!CORBA::is_nil(m_refCommander))
    {
        try {
            bRet = m_refCommander->IsStepByStep();
        } catch(...)
        {
        }
    }

    GetLogStream() << ">>>IsStepByStep: returns " << (bRet ? "YES" : "NO")
                   << "\r\n";
    ReleaseLogStream();

    return bRet;
}

void StepByStepControlImpl::NextRequest(const char* ScriptName, CORBA::Long lScriptLineNumber, CORBA::Long lTaskGroupID, CORBA::Long lVUId, CORBA::Long lConId, const char* strRequest)
{
    GetLogStream() << ">>>NextRequest: task group ID = " << lTaskGroupID
                   << ", VU = " << lVUId
                   << ", conid = " << lConId
                   << ", script = " << ScriptName
                   << ", line = " << lScriptLineNumber
                   << ", request =\n" << strRequest
                   << "\r\n";
    ReleaseLogStream();

    // Forward to Commander.
    if(!CORBA::is_nil(m_refCommander))
    {
        try {
            m_refCommander->NextRequest(ScriptName, lScriptLineNumber,
                lTaskGroupID, lVUId, lConId, strRequest);
        } catch(...)
        {
        }
    }
}

void StepByStepControlImpl::RequestComplete(CORBA::Long lTaskGroupID, CORBA::Long lVUId,
    CORBA::Long lConId, CORBA::Boolean bIsBinaryData, const char* strHeaders,
    const char* strBody, CORBA::Long lBodyLength)
{
    GetLogStream() << ">>>RequestComplete: task group ID = "
                   << lTaskGroupID
                   << ", VU = " << lVUId
                   << ", conid = " << lConId
                   << (bIsBinaryData ? ", binary data" : "")
                   << "\nheaders =\n" << strHeaders
                   << "\nbody(" << lBodyLength << ") =\n"
                   << strBody
                   << "\r\n";
    ReleaseLogStream();

    // Forward to Commander.
    if(!CORBA::is_nil(m_refCommander))
    {
        try {
            m_refCommander->RequestComplete(lTaskGroupID, lVUId,
                lConId, bIsBinaryData, strHeaders, strBody, lBodyLength);
        } catch(...)
        {
        }
    }
}

void StepByStepControlImpl::WriteLogMsg(const std::string& msg)
{
    CTLockGuard<CTMutex> lock(m_logFileMutex);

    if(m_pLogFile && m_pLogFile->good())
    {
        m_pLogFile->write( msg.c_str(), msg.length() );
        *m_pLogFile << "\r\n";
        m_pLogFile->flush();
    }
}
