/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrStApi CyrWrapLib
// @module  Test Output Wrappers | This module contains the implementation of
//          any simple Test Output types.  These output values are in effect
//          strings when they enter the OpenSTARep database or when they are
//          transmitted via CORBA.  The objects are created to hide this
//          functioanlity from the developer.
//
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <CyrWrapLib/CECTestOutput.h>
///////////////////////////////////////////////////////////////////////////////
// Constants
#define COMMA ","
const char CECTimerOutput::OUTPUT_TYPE    [] = "Timer.txt";
const char CECTestLog::OUTPUT_TYPE        [] = "TestLog.txt";
const char CECTestReport::OUTPUT_TYPE     [] = "TestRep.txt";
const char CECTestDiagnostic::OUTPUT_TYPE [] = "TestDiag.txt";
const char CECTestHistory::OUTPUT_TYPE    [] = "TestHis.txt";
const char CECOtaSimpleQuery::OUTPUT_TYPE [] = "OtaStat.txt";
const char CECOtaHeader::OUTPUT_TYPE      [] = "OtaHeader.txt";
const char CECVULogOutput::OUTPUT_TYPE    [] = "VUsersLog.txt";
const char CECNTPQuery::OUTPUT_TYPE		  [] = "NTPStat.txt";
const char CECNTPHeader::OUTPUT_TYPE      [] = "NTPHeader.txt";
const char CECSNMPQuery::OUTPUT_TYPE	  [] = "SNMPStat.txt";
const char CECSNMPHeader::OUTPUT_TYPE     [] = "SNMPHeader.txt";
const char CECCollector1Data::OUTPUT_TYPE [] = "DC1Data.txt";
const char CECCollector1Header::OUTPUT_TYPE [] = "DC1Header.txt";
const char CECErrorLog::OUTPUT_TYPE       [] = "ErrLog.txt";

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This base constructor simply generates a time value for the
//          creation of this object.  It should only be used when creating a
//          completely new test output object.
CECTestOutput::CECTestOutput(long t_offsets)
{
    // Generate current time
    m_time.now();
	m_time=CTTime(m_time.getSeconds()-t_offsets);
}

CECTestOutput::~CECTestOutput()
{}


CTTime CECTestOutput::GetCreatedTime() const
{
    return m_time;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor creates a brand new instance of a Timer output.
//          The time of creation is logged internally, but all other values are
//          passed in.
// @parm    The description for this timer output.
// @parm    The value for this timer output.
// @parm    The number of active threads.
CECTimerOutput::CECTimerOutput( long t_offsets, const char* desc, long value, long active )
        : CECTestOutput(t_offsets)
{
    // Assign member variables
    m_desc = desc;
    m_active = active;
    m_value = value;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor recreates a timer output based on the passed
//          string.  This string will be parsed in an attempt to find the
//          relevant strings.
// @parm    The string to be used uring recreation
CECTimerOutput::CECTimerOutput( const char* str ) : CECTestOutput(0)
{
    string str_parse(str);
    string::size_type startpos=0;
    string::size_type pipe_pos=startpos;
    // Extract fields
    startpos=str_parse.find_first_of(COMMA ,pipe_pos);
    m_time = CTTime( atol( str_parse.substr(pipe_pos,startpos-pipe_pos).c_str()));
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA ,pipe_pos);
    m_value = atol(str_parse.substr(pipe_pos,startpos-pipe_pos).c_str());
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA ,pipe_pos);
    m_active = atol(str_parse.substr(pipe_pos,startpos-pipe_pos).c_str());
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA ,pipe_pos);
    m_desc = str_parse.substr(pipe_pos,startpos-pipe_pos);
}

CECTimerOutput::~CECTimerOutput()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method serializes the timer output fields into a single ','
//          separated string.  This string can then be used to store the data
//          in a file.
// @rdesc   The generated string.

string CECTimerOutput::ToString()
{
    // Serialize all values into a single string
    string ret( "" );
    char buf[20];

    // Time value
    sprintf( buf, "%d", m_time.getSeconds() );
    ret += buf;
    ret += COMMA;

    // Timer value
    sprintf( buf, "%d", m_value );
    ret += buf;
    ret += COMMA;

    // Number of active threads
    sprintf( buf, "%d", m_active );
    ret += buf;
    ret += COMMA;

    // Description
    ret += m_desc;
    ret += COMMA;

    return ret;
}

string CECTimerOutput::GetType()
{
    return string( OUTPUT_TYPE );
}

string CECTimerOutput::GetDescription() const
{
    return m_desc;
}

long CECTimerOutput::GetValue() const
{
    return m_value;
}

long CECTimerOutput::GetActiveThreads() const
{
    return m_active;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This is the one and only constructor for this type.  As there are
//          no distinguishing variables between a creation and retrieval for
//          this type, a boolean is added to the parameters.  By default this
//          boolean is true and the constructor will create a completely new
//          test log entry.  If set to fals then the passed string is parsed
//          and the relevant fields are extracted.
// @parm    The text to add are parse.
// @parm    Determines whether to parse text or not.
CECTestLog::CECTestLog( long t_offsets, const char* text, bool create /*=true*/ )
        : CECTestOutput(t_offsets)
{
    // If this is a brand new test log entry then simply assign the
    // text value to the member variable.  otherwise we need to parse the
    // information out of the text variable.
    if ( create == true )
        m_text = text;
    else
    {
        string str_parse(text);
        string::size_type startpos=0;
        string::size_type pipe_pos=startpos;
        // Extract fields
        startpos=str_parse.find_first_of(COMMA ,pipe_pos);
        m_time = CTTime( atol( str_parse.substr(pipe_pos,startpos-pipe_pos).c_str()));
        pipe_pos=++startpos;
        m_text= str_parse.substr(pipe_pos,str_parse.length()-1-pipe_pos);
    }
}

CECTestLog::~CECTestLog()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method serializes the test log fields into a single ','
//          separated string.  This string can then be used to store the data
//          in a file.
// @rdesc   The generated string.

string CECTestLog::ToString()
{
    // Serialize all values into a single string
    string ret( "" );
    char buf[20];

    // Time value
    sprintf( buf, "%d", m_time.getSeconds() );
    ret += buf;
    ret += COMMA;

    // Text string
    ret += m_text;
    ret += COMMA;

    return ret;
}

string CECTestLog::GetType()
{
    return string( OUTPUT_TYPE );
}

string CECTestLog::GetText() const
{
    return m_text;
}

CECTestReport::CECTestReport( long t_offsets, const char* text, bool create )
        : CECTestLog( t_offsets, text, create )
{}


string CECTestReport::GetType()
{
    return string( OUTPUT_TYPE );
}

string CECTestDiagnostic::GetType()
{
    return string( OUTPUT_TYPE );
}

CECTestHistory::CECTestHistory( long t_offsets, const char* text, bool create )
        : CECTestLog( t_offsets, text, create )
{}


string CECTestHistory::GetType()
{
    return string( OUTPUT_TYPE );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor takes a line read in from the test output file
//          or transferred via CORBA and converts it into an instance of the
//          class.
// @parm    A string containing all information for construction of object.
CECOtaSimpleQuery::CECOtaSimpleQuery( const char* str )
        : CECTestOutput(0)
{
    // Extract the node name and other strings
    string str_parse(str);
    string::size_type startpos=0;
    string::size_type pipe_pos=startpos;
    // Extract fields
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    m_node = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string index = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string start = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string stop = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string result = str_parse.substr(pipe_pos,startpos-pipe_pos);

    m_index = atoi(index.c_str());
    m_start = atoi(start.c_str());
    m_stop = atoi(stop.c_str());
    m_result = atoi(result.c_str());
}


///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor creates a new data object.
// @parm    Start time.
// @parm    Stop time.
// @parm    Result.
CECOtaSimpleQuery::CECOtaSimpleQuery( long t_offsets, const char* node, int index, int start, int stop, int result )
        : CECTestOutput(t_offsets)
{
    // Assign member variables
    m_node = node;
    m_index = index;
    m_start = start - t_offsets;
    m_stop = stop - t_offsets;
    m_result = result;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor tidies up the allocated memory for the result array.
CECOtaSimpleQuery::~CECOtaSimpleQuery()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the name of the node where the query was
//          executed.
// @rdesc   The node name.

string CECOtaSimpleQuery::GetNodeName()
{
    return m_node;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method serializes the query output fields into a single ','
//          separated string.  This string can then be used to store the data
//          in a file.
// @rdesc   The generated string.
string CECOtaSimpleQuery::ToString()
{
    // Serialize all data into a string
    string ret( "" );
    char buf[64];
    buf[0] = '\0';
    sprintf( buf + strlen(buf),
             "%s%s%d%s%d%s%d%s%d%s",
             m_node.c_str(), COMMA,
             m_index, COMMA,
             m_start, COMMA,
             m_stop, COMMA,
             m_result, COMMA );
    ret += buf;

    return ret;
}

string CECOtaSimpleQuery::GetType()
{
    return string( OUTPUT_TYPE );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the description of the query.
// @rdesc   The query description.
string CECOtaSimpleQuery::GetDescription()
{
    return m_desc;
}

int CECOtaSimpleQuery::GetStartTime()
{
    return m_start;
}

int CECOtaSimpleQuery::GetStopTime()
{
    return m_stop;
}

int CECOtaSimpleQuery::GetResult()
{
    return m_result;
}

int CECOtaSimpleQuery::GetIndex()
{
    return m_index;
}

CECOtaHeader::CECOtaHeader( const char* str )
        : CECTestOutput(0)
{
    // Extract the node name and other strings
    string str_parse(str);
    string::size_type startpos=0;
    string::size_type pipe_pos=startpos;
    // Extract fields
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    m_node = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string strIndex = str_parse.substr(pipe_pos,startpos-pipe_pos);
    m_index = atoi(strIndex.c_str());
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    m_desc = str_parse.substr(pipe_pos,startpos-pipe_pos);


}

CECOtaHeader::CECOtaHeader( long t_offsets, const char* node, int index, const char* desc )
        : CECTestOutput(t_offsets)
{
    // Assign member variables
    m_index = index;
    m_node = node;
    m_desc = desc;
}

CECOtaHeader::~CECOtaHeader()
{}


string CECOtaHeader::GetType()
{
    return string( OUTPUT_TYPE );
}

string CECOtaHeader::ToString()
{
    string ret( "" );
    char buf[64];
    buf[0] = '\0';
    sprintf( buf + strlen(buf),
             "%s%s%d%s%s%s",
             m_node.c_str(), COMMA,
             m_index, COMMA,
             m_desc.c_str(), COMMA );
    ret += buf;

    return ret;
}

string CECOtaHeader::GetNodeName()
{
    return m_node;
}

string CECOtaHeader::GetDescription()
{
    return m_desc;
}

int CECOtaHeader::GetIndex()
{
    return m_index;
}

//////////////////////////////////
CECNTPHeader::CECNTPHeader( const char* str )
        : CECTestOutput(0)
{
    // Extract the node name and other strings
    string str_parse(str);
    string::size_type startpos=0;
    string::size_type pipe_pos=startpos;
    // Extract fields
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string strTG = str_parse.substr(pipe_pos,startpos-pipe_pos);
	m_nTG = atoi(strTG.c_str());
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string strIndex = str_parse.substr(pipe_pos,startpos-pipe_pos);
    m_index = atoi(strIndex.c_str());
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    m_desc = str_parse.substr(pipe_pos,startpos-pipe_pos);


}

CECNTPHeader::CECNTPHeader( long t_offsets, int nTaskGroupID, int index, const char* desc )
        : CECTestOutput(t_offsets)
{
    // Assign member variables
    m_index = index;
    m_nTG = nTaskGroupID;
    m_desc = desc;
}

CECNTPHeader::~CECNTPHeader()
{}


string CECNTPHeader::GetType()
{
    return string( OUTPUT_TYPE );
}

string CECNTPHeader::ToString()
{
    string ret( "" );
    char buf[64];
    buf[0] = '\0';
    sprintf( buf + strlen(buf),
             "%d%s%d%s%s%s",
             m_nTG, COMMA,
             m_index, COMMA,
             m_desc.c_str(), COMMA );
    ret += buf;

    return ret;
}

int CECNTPHeader::GetTaskGroupID()
{
    return m_nTG;
}

string CECNTPHeader::GetDescription()
{
    return m_desc;
}

int CECNTPHeader::GetIndex()
{
    return m_index;
}


///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor takes a line read in from the test output file
//          or transferred via CORBA and converts it into an instance of the
//          class.
// @parm    A string containing all information for construction of object.
CECNTPQuery::CECNTPQuery( const char* str )
        : CECTestOutput(0)
{
    // Extract the node name and other strings
    string str_parse(str);
    string::size_type startpos=0;
    string::size_type pipe_pos=startpos;
    // Extract fields
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string strTG = str_parse.substr(pipe_pos,startpos-pipe_pos);
	m_nTG = atoi(strTG.c_str());
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string index = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string start = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string stop = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string result = str_parse.substr(pipe_pos,startpos-pipe_pos);

    m_index = atoi(index.c_str());
    m_start = atoi(start.c_str());
    m_stop = atoi(stop.c_str());
    m_result = atoi(result.c_str());
}


///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor creates a new data object.
// @parm    Start time.
// @parm    Stop time.
// @parm    Result.
CECNTPQuery::CECNTPQuery( long t_offsets, int nTaskGroupID, int index, int start, int stop, int result )
        : CECTestOutput(t_offsets)
{
    // Assign member variables
    m_nTG = nTaskGroupID;
    m_index = index;
    m_start = start - t_offsets;
    m_stop = stop - t_offsets;
    m_result = result;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor tidies up the allocated memory for the result array.
CECNTPQuery::~CECNTPQuery()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the name of the node where the query was
//          executed.
// @rdesc   The node name.

int CECNTPQuery::GetTaskGroupID()
{
    return m_nTG;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method serializes the query output fields into a single ','
//          separated string.  This string can then be used to store the data
//          in a file.
// @rdesc   The generated string.
string CECNTPQuery::ToString()
{
    // Serialize all data into a string
    string ret( "" );
    char buf[64];
    buf[0] = '\0';
    sprintf( buf + strlen(buf),
             "%d%s%d%s%d%s%d%s%d%s",
             m_nTG, COMMA,
             m_index, COMMA,
             m_start, COMMA,
             m_stop, COMMA,
             m_result, COMMA );
    ret += buf;

    return ret;
}

string CECNTPQuery::GetType()
{
    return string( OUTPUT_TYPE );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the description of the query.
// @rdesc   The query description.
string CECNTPQuery::GetDescription()
{
    return m_desc;
}

int CECNTPQuery::GetStartTime()
{
    return m_start;
}

int CECNTPQuery::GetStopTime()
{
    return m_stop;
}

int CECNTPQuery::GetResult()
{
    return m_result;
}

int CECNTPQuery::GetIndex()
{
    return m_index;
}

////////////////////////////////////////////////////
CECSNMPHeader::CECSNMPHeader( const char* str )
        : CECTestOutput(0)
{
    // Extract the node name and other strings
    string str_parse(str);
    string::size_type startpos=0;
    string::size_type pipe_pos=startpos;
    // Extract fields
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string strTG = str_parse.substr(pipe_pos,startpos-pipe_pos);
	m_nTG = atoi(strTG.c_str());
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string strIndex = str_parse.substr(pipe_pos,startpos-pipe_pos);
    m_index = atoi(strIndex.c_str());
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    m_desc = str_parse.substr(pipe_pos,startpos-pipe_pos);


}

CECSNMPHeader::CECSNMPHeader( long t_offsets, int nTaskGroupID, int index, const char* desc )
        : CECTestOutput(t_offsets)
{
    // Assign member variables
    m_index = index;
    m_nTG = nTaskGroupID;
    m_desc = desc;
}

CECSNMPHeader::~CECSNMPHeader()
{}


string CECSNMPHeader::GetType()
{
    return string( OUTPUT_TYPE );
}

string CECSNMPHeader::ToString()
{
    string ret( "" );
    char buf[64];
    buf[0] = '\0';
    sprintf( buf + strlen(buf),
             "%d%s%d%s%s%s",
             m_nTG, COMMA,
             m_index, COMMA,
             m_desc.c_str(), COMMA );
    ret += buf;

    return ret;
}

int CECSNMPHeader::GetTaskGroupID()
{
    return m_nTG;
}

string CECSNMPHeader::GetDescription()
{
    return m_desc;
}

int CECSNMPHeader::GetIndex()
{
    return m_index;
}


///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor takes a line read in from the test output file
//          or transferred via CORBA and converts it into an instance of the
//          class.
// @parm    A string containing all information for construction of object.
CECSNMPQuery::CECSNMPQuery( const char* str )
        : CECTestOutput(0)
{
    // Extract the node name and other strings
    string str_parse(str);
    string::size_type startpos=0;
    string::size_type pipe_pos=startpos;
    // Extract fields
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string strTG = str_parse.substr(pipe_pos,startpos-pipe_pos);
	m_nTG = atoi(strTG.c_str());
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string index = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string start = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string stop = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string result = str_parse.substr(pipe_pos,startpos-pipe_pos);

    m_index = atoi(index.c_str());
    m_start = atoi(start.c_str());
    m_stop = atoi(stop.c_str());
    m_result = atoi(result.c_str());
}


///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor creates a new data object.
// @parm    Start time.
// @parm    Stop time.
// @parm    Result.
CECSNMPQuery::CECSNMPQuery( long t_offsets, int nTaskGroupID, int index, int start, int stop, int result )
        : CECTestOutput(t_offsets)
{
    // Assign member variables
    m_nTG = nTaskGroupID;
    m_index = index;
    m_start = start - t_offsets;
    m_stop = stop - t_offsets;
    m_result = result;
}
///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor tidies up the allocated memory for the result array.
CECSNMPQuery::~CECSNMPQuery()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the name of the node where the query was
//          executed.
// @rdesc   The node name.

int CECSNMPQuery::GetTaskGroupID()
{
    return m_nTG;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method serializes the query output fields into a single ','
//          separated string.  This string can then be used to store the data
//          in a file.
// @rdesc   The generated string.
string CECSNMPQuery::ToString()
{
    // Serialize all data into a string
    string ret( "" );
    char buf[64];
    buf[0] = '\0';
    sprintf( buf + strlen(buf),
             "%d%s%d%s%d%s%d%s%d%s",
             m_nTG, COMMA,
             m_index, COMMA,
             m_start, COMMA,
             m_stop, COMMA,
             m_result, COMMA );
    ret += buf;

    return ret;
}

string CECSNMPQuery::GetType()
{
    return string( OUTPUT_TYPE );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the description of the query.
// @rdesc   The query description.
string CECSNMPQuery::GetDescription()
{
    return m_desc;
}

int CECSNMPQuery::GetStartTime()
{
    return m_start;
}

int CECSNMPQuery::GetStopTime()
{
    return m_stop;
}

int CECSNMPQuery::GetResult()
{
    return m_result;
}

int CECSNMPQuery::GetIndex()
{
    return m_index;
}

////////////////////////////////////////////////////
////////////////////////////////////////////////////
CECCollector1Header::CECCollector1Header( const char* str )
        : CECTestOutput(0)
{
    // Extract the node name and other strings
    string str_parse(str);
    string::size_type startpos=0;
    string::size_type pipe_pos=startpos;
    // Extract fields
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string strTimestamp = str_parse.substr(pipe_pos,startpos-pipe_pos);
	m_timestamp = atoi(strTimestamp.c_str());
    pipe_pos=++startpos;
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string strIndex = str_parse.substr(pipe_pos,startpos-pipe_pos);
    m_index = atoi(strIndex.c_str());
}

CECCollector1Header::CECCollector1Header(long t_offset, int timestamp, int index, const char* prfnam)
        : CECTestOutput(t_offset)
{
    // Assign member variables
    m_timestamp = timestamp;
    m_index = index;
	m_prfnam = prfnam;
}

CECCollector1Header::~CECCollector1Header()
{}


string CECCollector1Header::GetType()
{
    return string( OUTPUT_TYPE );
}

string CECCollector1Header::ToString()
{
    string ret( "" );
    char buf[2048];
    buf[0] = '\0';
    sprintf( buf + strlen(buf),
             "%d%s%d%s%s%s",
             m_timestamp, COMMA,
             m_index, COMMA,
			 m_prfnam.c_str(), COMMA);
    ret += buf;

    return ret;
}

int CECCollector1Header::GetIndex()
{
    return m_index;
}

int CECCollector1Header::GetTimestamp()
{
    return m_timestamp;
}

string CECCollector1Header::GetProfileName()
{
    return m_prfnam;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor takes a line read in from the test output file
//          or transferred via CORBA and converts it into an instance of the
//          class. This assumes that fields appear in the following order in
//          the input string:
//
//          <PRE>
//          1.  begtim (int)        Anomaly begin time
//          2.  modtim (int)        Anomaly last modified time
//          3.  endtim (int)        Anomaly end time
//          4.  anotim (int)        Anomaly duration
//          5.  anotyp (string)     Anomaly type
//          6.  anoval (int)        Anomaly value
//          7.  pid (int)           Anomaly process ID
//          8.  anoid (int)         Anomaly ID
//          9.  dbnam (string)      Anomaly DB name
//          10. dbobjnam (string)   Anomaly DB object name
//          11. prfnam (string)     Profile name
//          </PRE>
//
// @parm    A string containing all information for construction of object.
//
CECCollector1Data::CECCollector1Data(const char* str)
        : CECTestOutput(0)
{
    // Prepare to extract fields from the input string.
    string str_parse(str);
    string::size_type startpos=0;
    string::size_type pipe_pos=startpos;

    // Extract fields from the input string
    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string begtim = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;

    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string modtim = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;

    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string endtim = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;

    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string anotim = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;

    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    m_anotyp = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;

    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string anosev = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;

    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string anoval = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;

    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string pid = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;

    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    string anoid = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;

    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    m_dbnam = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;

    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    m_dbobjnam = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;

    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    m_prfnam = str_parse.substr(pipe_pos,startpos-pipe_pos);

    // Convert integer fields from string to integer.
    m_begtim = atoi(begtim.c_str());
    m_modtim = atoi(modtim.c_str());
    m_endtim = atoi(endtim.c_str());
    m_anotim = atoi(anotim.c_str());
    m_anosev = atoi(anosev.c_str());
	m_anoval = atoi(anoval.c_str());
    m_pid = atoi(pid.c_str());
    m_anoid = atoi(anoid.c_str());
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor creates a new data object.
//
// @parm    t_offsets.
// @parm    Anomaly type string.
// @parm    Anomaly value.
// @parm    Anomaly PID.
// @parm    Anomaly ID.
// @parm    DB name.
// @parm    DB object name.
// @parm    Profile name
// @parm    Anomaly begin time.
// @parm    Anomaly last modified time.
// @parm    Anomaly end time.
// @parm    Anomaly duration.
//
CECCollector1Data::CECCollector1Data(long t_offsets, const char* anotyp, int anosev,
    int anoval, int pid, int anoid, const char* dbnam, const char* dbobjnam,
    const char* prfnam, int begtim, int modtim, int endtim, int anotim)
    : CECTestOutput(t_offsets)
{
    // Assign member variables
    m_begtim = begtim - t_offsets;
    m_modtim = modtim - t_offsets;
    m_endtim = endtim - t_offsets;
    m_anotim = anotim;
    m_anotyp = anotyp;
	m_anosev = anosev ;
    m_anoval = anoval;
    m_pid = pid;
    m_anoid = anoid;
    m_dbnam = dbnam;
    m_dbobjnam = dbobjnam;
    m_prfnam = prfnam;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor tidies up the allocated memory for the result array.
//
CECCollector1Data::~CECCollector1Data()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method serializes the fields of collected data into a single
//          delimited string. This string can then be used to store the data
//          in a file.
//
// @rdesc   The generated string.
//
string CECCollector1Data::ToString()
{
    // Set up the return string in a temporary buffer.
    char buf[2048];
    buf[0] = '\0';
    sprintf(buf + strlen(buf),
            "%d%s%d%s%d%s%d%s%s%s%d%s%d%s%d%s%d%s%s%s%s%s%s%s",
            m_begtim, COMMA,
            m_modtim, COMMA,
            m_endtim, COMMA,
            m_anotim, COMMA,
            m_anotyp.c_str(), COMMA,
            m_anosev, COMMA,
            m_anoval, COMMA,
            m_pid, COMMA,
            m_anoid, COMMA,
            m_dbnam.c_str(), COMMA,
            m_dbobjnam.c_str(), COMMA,
            m_prfnam.c_str(), COMMA);

    // Copy data to return string.
    string ret("");
    ret += buf;

    return ret;
}

string CECCollector1Data::GetType()
{
    return string(OUTPUT_TYPE);
}

int CECCollector1Data::GetBeginTime()
{
    return m_begtim;
}

int CECCollector1Data::GetModTime()
{
    return m_modtim;
}

int CECCollector1Data::GetEndTime()
{
    return m_endtim;
}

int CECCollector1Data::GetAnomalyTime()
{
    return m_anotim;
}

string CECCollector1Data::GetAnomalyType()
{
    return m_anotyp;
}

int CECCollector1Data::GetAnomalySeverity()
{
    return m_anosev;
}

int CECCollector1Data::GetAnomalyValue()
{
    return m_anoval;
}

int CECCollector1Data::GetPid()
{
    return m_pid;
}

int CECCollector1Data::GetAnomalyId()
{
    return m_anoid;
}

string CECCollector1Data::GetDBName()
{
    return m_dbnam;
}

string CECCollector1Data::GetDBObjName()
{
    return m_dbobjnam;
}

string CECCollector1Data::GetProfileName()
{
    return m_prfnam;
}

/////////////////////////////////////////////////////////////////////////////////////
CECVULogOutput::CECVULogOutput(long t_offsets, int nVUsers)
			: CECTestOutput(t_offsets)
{
	m_nVUsers = nVUsers;
}

CECVULogOutput::CECVULogOutput(const char* strEntry)
			: CECTestOutput(0)
{
        string str_parse(strEntry);
        string::size_type startpos=0;
        string::size_type pipe_pos=startpos;
        // Extract fields
        startpos=str_parse.find_first_of(COMMA ,pipe_pos);
        m_time = CTTime( atol( str_parse.substr(pipe_pos,startpos-pipe_pos).c_str()));
        pipe_pos=++startpos;
        startpos=str_parse.find_first_of(COMMA ,pipe_pos);
        string strVUsers= str_parse.substr(pipe_pos,startpos-pipe_pos);

		m_nVUsers = atoi(strVUsers.c_str());

}

CECVULogOutput::~CECVULogOutput()
{
}

CTTime CECVULogOutput::GetTime()
{
	return m_time;
}

int CECVULogOutput::GetVUsers()
{
	return m_nVUsers;
}

string CECVULogOutput::GetType()
{
    return string( OUTPUT_TYPE );
}

string CECVULogOutput::ToString()
{
    // Serialize all values into a single string
    string ret( "" );
    char buf[20];

    // Time value
    sprintf( buf, "%d,%d,", m_time.getSeconds(), m_nVUsers );
	ret += buf;
    return ret;
}


CECErrorLog::CECErrorLog(long t_offsets, const char* strTestName, const char* strLocation,
						 const char* strMsg)
     : CECTestOutput(t_offsets)

{
	m_strTestName = strTestName;
	m_strLocation = strLocation;
	m_strMsg = strMsg;
}

CECErrorLog::CECErrorLog(const char* strEntry)
     : CECTestOutput(0)
{
    string str_parse(strEntry);
    string::size_type startpos=0;
    string::size_type pipe_pos=startpos;
    // Extract fields
    startpos=str_parse.find_first_of(COMMA ,pipe_pos);
    m_time = CTTime( atol( str_parse.substr(pipe_pos,startpos-pipe_pos).c_str()));
    pipe_pos=++startpos;

    startpos=str_parse.find_first_of(COMMA ,pipe_pos);
    m_strTestName= str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;

    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    m_strLocation = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;

    startpos=str_parse.find_first_of(COMMA,pipe_pos);
    m_strMsg = str_parse.substr(pipe_pos,startpos-pipe_pos);
    pipe_pos=++startpos;
}

CECErrorLog::~CECErrorLog()
{
}

CTTime CECErrorLog::GetTime()
{
	return m_time;
}

string CECErrorLog::GetTestName()
{
	return m_strTestName;
}

string CECErrorLog::GetLocation()
{
	return m_strLocation;
}

string CECErrorLog::GetMessage()
{
	return m_strMsg;
}

string CECErrorLog::GetType()
{
    return string( OUTPUT_TYPE );
}

string CECErrorLog::ToString()
{
    // Serialize all values into a single string
    string ret( "" );
    char buf[20];

    // Time value
    sprintf( buf, "%d", m_time.getSeconds() );
    ret += buf;
    ret += COMMA;

    // Text string
    ret += m_strTestName;
    ret += COMMA;
    // Text string
    ret += m_strLocation;
    ret += COMMA;
    // Text string
    ret += m_strMsg;
    ret += COMMA;

    return ret;

}
