/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     ProviderReg
// @module  Provider Register | This module contains the class ProviderReg.
//          This class is responsible for maintaining a list of all CORBA
//          classes that can be created by a Factory module.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <Factory/ProviderReg.h>
#include <algorithm>
///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor is responsible for deleting any remaining memory.
//          This should never occur as each Provider is responsible for tidying
//          up its own memory, but we know how forgetful some programmers are..
ProviderReg::~ProviderReg()
{
    // Tidy memory
    for( ProviderList::iterator i = m_listProvider.begin(); i !=m_listProvider.end(); i++ )
        delete *i;
    m_listProvider.clear();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method registers a new Provider class entry.
// @parm    Name of the class.
// @parm    Name of the Provider class.
// @parm    Type of the class within the Provider.
void ProviderReg::Register(const char* name, const char* provider, int type )
{
    // Create new provider element and assign values to elements
    PPROV_S p = new PROV_S;
    p->name     = name;
    p->provider = provider;
    p->type     = type;

    // Add new element to list
    m_listProvider.push_back( p );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method removes a single entry from the list of registered
//          classes.  Note that calling this method will invalidate any
//          searches currently being done.
// @parm    Name of the class.
// @parm    Name of the provider module.
void ProviderReg::Deregister(const char* name, const char* provider )
{
    // Loop through elements and remove the one we're looking for
    for ( ProviderList::iterator i = m_listProvider.begin(); i != m_listProvider.end(); i++ )
    {
        // If name and provider module match then remove it and throw it away.
        if (( (*i)->name == name ) &&
                ( (*i)->provider == provider ))
        {
            delete *i;
            *i = NULL;
        }
    }
    m_listProvider.remove(NULL);
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method removes all classes registered with a particular
//          Provider module.  Note that calling this method will invalidate any
//          searches currently being done.
// @parm    Name of provider module.
void ProviderReg::DeregisterAll(const char* provider )
{
    // Loop through all registered provider information
    for ( ProviderList::iterator i = m_listProvider.begin(); i!= m_listProvider.end(); i++ )
    {
        // If provider matches then get rid of entry
        const PPROV_S ps = *i;
        if ( ps->provider == provider )
        {
            delete ps;
            *i = NULL;
        }
    }
    m_listProvider.remove( NULL );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method searches for an element in the list and returns true if
//          is present or false if it isn't.  This particular version doesn't
//          care about type or module.
// @rdesc   True or false depending on whether element is found.
// @parm    The name of the class to find.
bool ProviderReg::InSearchList(const char* entry )
{
    // If the entry matches then return true.
    if ( std::find(m_listSearch.begin(),m_listSearch.end(), entry )!=m_listSearch.end())
        return true;
    else
        return false;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method populates the search list with all the class names that
//          the provider knows about.  Entries are not repeated. Using the
//          overloaded versions will allow specific filtering to be done on
//          Provider and type.
// @syntax  string GetFirstClass()
// @syntax  string GetFirstClass( string provider )
// @syntax  string GetFirstClass( int type )
//
// @parm    const char* | provider | Name of the provider module to search.
// @parm    int | type | Type of class to search for.
//
// @rdesc   The first element in the list.
//
// @xref    <mf ProviderReg::GetFirstProvider>, <mf ProviderReg::GetNext>
string ProviderReg::GetFirstClass()
{
    // Empty current list
    m_listSearch.clear();

    // Loop through registration list
    for ( ProviderList::iterator i = m_listProvider.begin(); i != m_listProvider.end(); i++ )
    {
        // If entry is not already in list then we need to add it.
        if ( InSearchList( ((*i)->name).c_str() ) == false )
            m_listSearch.push_back( (*i)->name );
    }

    // Return first entry in list
#ifdef WIN32
    m_searchIndex = 0 ;
#else
    m_searchIndex = m_listSearch.begin();
#endif
    if ( m_listSearch.empty() )
        return "";
    return m_listSearch.front();
}

string ProviderReg::GetFirstClass(const char* provider )
{
    // Empty current list
    m_listSearch.clear();

    // Loop through registration list
    for ( ProviderList::iterator i = m_listProvider.begin(); i != m_listProvider.end(); i++ )
    {
        // Does entry match provider?
        if ( (*i)->provider==string(provider))
        {
            // Is entry already in list? If not then add it
            if ( InSearchList( ((*i)->name).c_str() ) == false )
                m_listSearch.push_back( (*i)->name );
        }
    }

    // Return first entry in list
#ifdef WIN32
    m_searchIndex = 0 ;
#else
    m_searchIndex = m_listSearch.begin();
#endif
    if ( m_listSearch.empty() )
        return "";
    return m_listSearch.front();
}

string ProviderReg::GetFirstClass(int type)
{
    // Empty current list
    m_listSearch.clear();

    // Loop through registration list
    for ( ProviderList::iterator i = m_listProvider.begin(); i != m_listProvider.end(); i++ )
    {
        // Does entry match type?
        if ( (*i)->type == type )
        {
            // Is entry already in list? If not then add it
            if ( InSearchList( ((*i)->name).c_str() ) == false )
                m_listSearch.push_back( (*i)->name );
        }
    }

    // Return first entry in list
#ifdef WIN32
    m_searchIndex = 0 ;
#else
    m_searchIndex = m_listSearch.begin();
#endif
    if ( m_listSearch.empty() )
        return "";
    return m_listSearch.front();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method populates the search list with all the Provider modules
//          that contain a particular class.
// @syntax  string GetFirstProvider()
// @syntax  string GetFirstProvider( string name )
//
// @rdesc   The first element in the list.
// @parm    string | name | The name of the class to search for.
//
// @xref    <mf ProviderReg::GetFirstClass>, <mf ProviderReg::GetNext>
string ProviderReg::GetFirstProvider()
{
    // Empty current list
    m_listSearch.clear();

    // Loop through registration list
    for ( ProviderList::iterator i = m_listProvider.begin(); i != m_listProvider.end(); i++ )
    {
        // Is entry already in list?  If not then add it
        if ( InSearchList( ((*i)->provider).c_str() ) == false )
            m_listSearch.push_back( (*i)->provider );
    }

    // Return first entry in list
#ifdef WIN32
    m_searchIndex = 0 ;
#else
    m_searchIndex = m_listSearch.begin();
#endif
    if ( m_listSearch.empty() )
        return "";
    return m_listSearch.front();
}

string ProviderReg::GetFirstProvider( const char* name )
{
    // Empty current list
    m_listSearch.clear();

    // Loop through registration list
    for ( ProviderList::iterator i = m_listProvider.begin(); i != m_listProvider.end(); i++ )
    {
        // Does entry match class?
        if ( (*i)->name==string(name))
        {
            // Is entry already in list?  If not then add it
            if ( InSearchList( ((*i)->provider).c_str() ) == false )
                m_listSearch.push_back( (*i)->provider );
        }
    }

    // Return first entry in list
#ifdef WIN32
    m_searchIndex = 0 ;
#else
    m_searchIndex = m_listSearch.begin();
#endif
    if ( m_listSearch.empty() )
        return "";
    return m_listSearch.front();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves the next element of the search list.
// @rdesc   The next element in the list or an empty string if at end of list.
//
// @xref    <mf ProviderReg::GetFirstClass>, <mf ProviderReg::GetFirstProvider>
string ProviderReg::GetNext()
{
    string s = "";

    // Check to see if we are at end of list
    if ( (++m_searchIndex)!= m_listSearch.end() )
        s = *m_searchIndex;

    // Return next element
    return s;
}









