/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrStProvider
// @module  OpenSTARep Interface Provider | This module contains the
//          implementation of the CyrStProvider class.  This Provider allows
//          creation and management of OpenSTARep COM objects via a CORBA
//          interface.  AS COM is used this Provider should only be
//          instantiated on a Windows platform.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#ifdef _WIN32
#   pragma warning (disable : 4503)
#endif

#include <CyrException.h>
#include <Repository/CyrStProv.h>
#include <Repository/CyrStServer_i.h>
#include <Repository/CyrStProject_i.h>
#include <Repository/CyrStFolder_i.h>
#include <Repository/CyrStFile_i.h>
#include <Repository/CyrStTestRun_i.h>
#include <Repository/CyrStFileIt_i.h>

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The constructor simply registers the objects that this Provider
//          can create and starts the COM services.
// @parm    Pointer to CORBA helper functionality
// @parm    The name of this Provider object.
// @parm    The name of the Factory object that owns this Provider.
CyrStProvider::CyrStProvider( CORBAServer* pServer, const char* name,
                              const char* owner )
        :ProviderImpl( pServer, name, owner )
{
    pServer->ActivateObject( this);
    pServer->NameObject( this->_this(), Provider::CLASS_NAME, name);

    // Register CORBA objects that this Provider can create
    Register( CyrStServer::CLASS_NAME,      name, PROVIDER_DATA );
    Register( CyrStProject::CLASS_NAME,     name, PROVIDER_DATA );
    Register( CyrStFolder::CLASS_NAME,      name, PROVIDER_DATA );
    Register( CyrStFile::CLASS_NAME,        name, PROVIDER_DATA );
    Register( CyrStTestRun::CLASS_NAME,     name, PROVIDER_DATA );
    Register( CyrStFileItImpl::CLASS_NAME,  name, PROVIDER_DATA );

    m_beans = new BeanArray();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor shuts down the COM services and empties the array
//          of OpenSTARep interfaces.
void CyrStProvider::Shutdown()
{
    // Workaround

    // Must destroy all objects here.  The base destructor also calls this
    // method, but due to the COM interface shutting down before then causes
    // exceptions when destroying the COM pointers.
    EmptyList( 0, GetObjectList()->size(), false );

    // Tidy memory
    m_beans->clear();
    delete m_beans;

    ProviderImpl::Shutdown();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method creates a new instance of a CORBA object based on the
//          passed parameter list.
// @rdesc   A pointer to the newly created object or a nil reference.
// @parm    Reference to an array of constructor parameters.
CORBA::Object_ptr CyrStProvider::New( ParamList_ptr params )
{
    // Get default parameters out of list
    CORBA::String_var cls = params->GetValue( CyrFactory::PARAM_CLASS );
    CORBA::String_var pd = params->GetValue( CyrFactory::PARAM_PERIOD );
    long period = atol( pd );

    try
    {
        // Determine type of object and create an instance of that object
        if ( strcmp( cls, CyrStServer::CLASS_NAME ) == 0 )
        {
            // Create new Server object and make sure it is ready for use
            CyrStServerImpl* pServer = new CyrStServerImpl( this, params );
			PortableServer::ObjectId* objid = m_pServer->ActivateObject( pServer );

            // Store it so that it can be freed later
            CORBA::String_var regName = pServer->GetRegisteredName();
            Store( objid, pServer, cls, regName, period );

            // Return reference to client
            return pServer->_this();
        }
        else if ( strcmp( cls, CyrStProject::CLASS_NAME ) == 0 )
        {
            // Create new Project object and make sure it is ready for use
            CyrStProjectImpl* pProject = new CyrStProjectImpl( this, params );
            PortableServer::ObjectId* objid = m_pServer->ActivateObject( pProject );

            // Store it so that it can be freed later
            CORBA::String_var regName = pProject->GetRegisteredName();
            Store( objid, pProject, cls, regName, period );

            // Return reference to client
            return pProject->_this();
        }
        else if ( strcmp( cls, CyrStFolder::CLASS_NAME ) == 0 )
        {
            // Create a new Folder object and make sure it is ready for use
            CyrStFolderImpl* pFolder = new CyrStFolderImpl( this, params );
            PortableServer::ObjectId* objid = m_pServer->ActivateObject( pFolder );

            // Store object so that it can be freed later
            CORBA::String_var regName = pFolder->GetRegisteredName();
            Store( objid, pFolder, cls, regName, period );

            // Return reference to client
            return pFolder->_this();
        }
        else if ( strcmp( cls, CyrStFile::CLASS_NAME ) == 0 )
        {
            // Create a new File object and make sure it is ready for use
            CyrStFileImpl* pFile = new CyrStFileImpl( this, params );
            PortableServer::ObjectId* objid = m_pServer->ActivateObject( pFile );

            // Store object so that it can be freed later
            CORBA::String_var regName = pFile->GetRegisteredName();
            Store( objid, pFile, cls, regName, period );

            // Return reference to client
            return pFile->_this();
        }
        else if ( strcmp( cls, CyrStTestRun::CLASS_NAME ) == 0 )
        {
            // Create a new Test Run object and make sure it is ready for use
            CyrStTestRunImpl* pTestRun = new CyrStTestRunImpl( this, params );
            PortableServer::ObjectId* objid = m_pServer->ActivateObject( pTestRun );

            // Store object so that it can be freed later
            CORBA::String_var regName = pTestRun->GetRegisteredName();
            Store( objid, pTestRun, cls, regName, period );

            // Return referencce to client
            return pTestRun->_this();
        }
        else if ( strcmp( cls, CyrStFileItImpl::CLASS_NAME ) == 0 )
        {
            // Extract filename parameter
            CORBA::String_var path = params->GetValue( CyrFactory::PARAM_NAME );
            if ( strlen( path ) > 0 )
            {
                // Create a new file iterator
                CyrStFileItImpl* pFileIt = new CyrStFileItImpl( path );
                PortableServer::ObjectId* objid = m_pServer->ActivateObject( pFileIt );

                // Store object so that it can be freed later
                Store( objid, pFileIt, cls, path, period );

                // Return reference to client
                return pFileIt->_this();
            }
        }
    }
    catch( CyrRepositoryException ex )
    {
        // Expected exception caught here.  Must have been thrown due to an
        // error during creation of a OpenSTARep object.
        string error( "Repository exception occured: " );
        error += ex.what();
        error += "\nIn method: ";
        error += ex.where();
        LogEvent( 1, error.c_str() );
    }
    catch( CORBA::Exception& ex )
    {
        string error = CorbaExceptionString( ex );
        error += " caught whilst trying to create class: ";
        error += cls;
        LogEvent( 1, error.c_str() );
    }
    /*    catch( _com_error ex )
        {
            string error( "COM exception occurred: " );
            error += ex.Description();
            error += " whilst trying to create class: ";
            error += cls;
            LogEvent( 1, error );
        }*/
    catch( ... )
    {
        // Unknown exception caught
        string error( "Exception caught whilst trying to create class: " );
        error += cls;
        LogEvent( 1, error.c_str() );
    }

    // Return a nil object as we haven't found one we can create
    return CORBA::Object::_nil();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method stores a OpenSTARep COM interface with a unique key.
//          This key is generated by the method and returned to the storing
//          code.  The key is based on the name of the object and also the
//          name of the user logged on.  If the key is found already in the
//          array then a letter is added and an attempt is made again.
// @rdesc   The generated key for the stored object.
// @parm    The pointer to the OpenSTARep COM object to store.
// @parm    The name of the object to store.
// @parm    The name of the user trying to store the object.
string CyrStProvider::StorePtr( CECFFBean* bean, const char* name, const char* user )
{
    // Generate a new key for this value
    string key( name );
    key += "_";
    key += user;

    // Check if key already exists.  If it does then just add a letter to
    // end of key until it no longer exists.  It is unlikely that this will
    // continue for too long as it requires the user to constantly create the
    // same object and this is a stupid thing to do as it will require huge
    // amounts of memory on the server machine.
    while ( m_beans->find( key )!=m_beans->end())
        key += 'A';

    // Add bean to array
    m_beans->insert( BeanArray::value_type(key, bean ));

    // Return what should be a unique handle to the calling code.
    // This handle should be used whenever access is required to the
    // underlying bean.
    return key;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves a specified object from the array of beans.
//          If the entry cannot be found then NULL will be returned.
// @rdesc   Pointer to stored OpenSTARep COM object.
// @parm    The key of the object to find.
CECFFBean* CyrStProvider::GetPtr( string key )
{
    // Retrieve bean from array
    return (*m_beans)[key];
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc  This method removes the specified COM interface object from the
//         array of beans.  As the interface is removed it will be destroyed.
// @parm   The key of the object to remove.
void CyrStProvider::RemovePtr( string key )
{
    // Remove all mention of bean from array.
    m_beans->erase( key );
}

