/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef __SOCKET__
#include "socket.hpp"
#endif

#include <iostream>
#include <cstdio>

#ifndef __MTHREAD__
#include "mthread.hpp"
#endif

#ifndef __XALLOC__
#include "xalloc.hpp"
#endif

#ifdef WIN32
bool winsockInitialize()
{
    WORD wVersionRequested;
    WSADATA wsaData;
    int iError;

    // Must support Windows sockets 2.2.
    wVersionRequested = MAKEWORD(2, 2);
    iError = WSAStartup(wVersionRequested, &wsaData);
    if (iError != 0)
    {
        /* Tell the user that we couldn't find a useable */
        /* winsock.dll.     */
        if (iError != 0)
            return false;
    }

    /* The Windows Sockets DLL is acceptable. Proceed. */

    return true;
}

void winsockTerminate()
{
    WSACleanup();
}
#endif

//---------------------------------------------------------------------------
// SocketsInitialize
//---------------------------------------------------------------------------
// Initialize the Socket library.
// Calling this function more than once from the same process (perhaps in
// different modules) does not do any harm, as long as each call have
// a matching SocketsUninitialize() call after using socket features.
//---------------------------------------------------------------------------
// Arguments :
// Return    :  Status (TRUE or FALSE).
//---------------------------------------------------------------------------
bool SocketsInitialize()
{
#ifdef WIN32
    return winsockInitialize();
#else
    return true;
#endif
}

//---------------------------------------------------------------------------
// SocketsUninitialize
//---------------------------------------------------------------------------
// Uninitialize the Socket library.
//---------------------------------------------------------------------------
// Arguments :
// Return    :  Status (TRUE or FALSE).
//---------------------------------------------------------------------------
bool SocketsUninitialize()
{
#ifdef WIN32
    winsockTerminate();
    return true;
#else
    return true;
#endif
}

//---------------------------------------------------------------------------
// The hostname's length
//---------------------------------------------------------------------------
const int CTInetAddress::iMaxHostNameLen_ = 64;

//---------------------------------------------------------------------------
// The number of pending connections.
//---------------------------------------------------------------------------
const int CTSocket::iMaxListenEntries_ = 5;

//---------------------------------------------------------------------------
// operator<<
//---------------------------------------------------------------------------
// Output an inet address into an ostream
//---------------------------------------------------------------------------
// Arguments :  os          : the reference to the output stream
//              inetAddress : the inet address
// Return    :  reference to the output stream.
//---------------------------------------------------------------------------
ostream& operator<< (ostream& os, const CTInetAddress& inetAddress)
{
    long lAddr = inetAddress.getAddress();

    os << '<' << ((CTInetAddress&)inetAddress).getHostName()
    << ',' << inetAddress.getFamily()
    << ',' << ((lAddr & 0xff000000) >> 24)
    << '.' << ((lAddr & 0xff0000) >> 16)
    << '.' << ((lAddr & 0xff00) >> 8)
    << '.' << (lAddr & 0xff)
    <<'>';

    return os;
}

#ifdef WIN32
//---------------------------------------------------------------------------
// operator<<
//---------------------------------------------------------------------------
// Output an inet address into a CTOLogFileList.
//---------------------------------------------------------------------------
// Arguments :  os          : the reference to the CTOLogFileList
//              inetAddress : the inet address
// Return    :  reference to the CTOLogFileList.
//---------------------------------------------------------------------------
CTOLogFileList& operator<< (CTOLogFileList& os, const CTInetAddress& inetAddress)
{
    long lAddr = inetAddress.getAddress();

    os << '<' << ((CTInetAddress&)inetAddress).getHostName()
    << ',' << inetAddress.getFamily()
    << ',' << ((lAddr & 0xff000000) >> 24)
    << '.' << ((lAddr & 0xff0000) >> 16)
    << '.' << ((lAddr & 0xff00) >> 8)
    << '.' << (lAddr & 0xff)
    <<'>';

    return os;
}
#endif 

//---------------------------------------------------------------------------
// CTInetAddress::CTInetAddress
//---------------------------------------------------------------------------
// Default Constructor
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  None.
//---------------------------------------------------------------------------
CTInetAddress::CTInetAddress()
        : hostName_(), lAddress_(0L), sFamily_(0)
{
    *this = getLocalHostName();
}

//---------------------------------------------------------------------------
// CTInetAddress::CTInetAddress
//---------------------------------------------------------------------------
// Copy Constructor
//---------------------------------------------------------------------------
// Arguments :  address : an inet address
// Return    :  None.
//---------------------------------------------------------------------------
CTInetAddress::CTInetAddress(const CTInetAddress& address)
{
    hostName_ = address.hostName_;
    lAddress_ = address.lAddress_;
    sFamily_  = address.sFamily_;
}

//---------------------------------------------------------------------------
// CTInetAddress::CTInetAddress
//---------------------------------------------------------------------------
// Constructor with an hostname
//---------------------------------------------------------------------------
// Arguments :  host : the hostname.
// Return    :  None.
//---------------------------------------------------------------------------
CTInetAddress::CTInetAddress(const string& host)
        : hostName_(), lAddress_(0L), sFamily_(0)
{
    try
    {
        *this = getByName(host);
    }
    catch(CTSocketException& /*exception*/)
    {
        // ### TBD ###
    }

    catch(...)
    {
        throw;
    }
}

//---------------------------------------------------------------------------
// CTInetAddress::getByName
//---------------------------------------------------------------------------
// To retrieve host information by the hostname
//---------------------------------------------------------------------------
// Arguments :  host : the hostname.
// Return    :  an inet address.
//---------------------------------------------------------------------------
CTInetAddress CTInetAddress::getByName(const string& host)
{
    string hostName;
    short sFamily = AF_INET;
    long lAddress;

    lAddress = inet_addr(host.c_str());
#ifdef SUN
    if (lAddress == -1)
#else
    if (lAddress == INADDR_NONE)
#endif
    {
        hostent *hp = gethostbyname(host.c_str());
        if (hp)
        {
            memcpy(&lAddress,(char *)hp->h_addr, hp->h_length);
            lAddress = ntohl(lAddress);
            hostName = hp->h_name;
            sFamily = hp->h_addrtype;
        }
        else
        {
            if (host == "localhost")
            {
                lAddress = inet_addr("127.0.0.1");
#ifdef SUN
	        if (lAddress != -1)
#else
                if (lAddress != INADDR_NONE)
#endif
                {
                    lAddress = ntohl(lAddress);
                    hostName = host;
                    sFamily = AF_INET;
                }
                else
                    throw CTSocketException(string ("host information not found for 127.0.0.1"), __FILE__, __LINE__);
            }
            else
                throw CTSocketException(string ("host information not found for ") + host, __FILE__, __LINE__);
        }
    }
    else
    {
        lAddress = ntohl(lAddress);
        hostName = host;
        sFamily = AF_INET;
    }

    return CTInetAddress(hostName, lAddress, sFamily);
}

//---------------------------------------------------------------------------
// CTInetAddress::getHostName
//---------------------------------------------------------------------------
// To retrieve the hostname
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  the hostname.
//---------------------------------------------------------------------------
const string& CTInetAddress::getHostName()
{
    if (hostName_.empty())
    {
        try
        {
            if (lAddress_ == 0L)
                hostName_ = CTInetAddress::getLocalHostName();
            else
            {
                struct in_addr inAddress ;
                inAddress.s_addr = htonl (lAddress_) ;
                hostName_ = inet_ntoa (inAddress);
            }
        }
        catch (...)
        {
            throw CTSocketException(string ("Can't find host name ") + hostName_, __FILE__, __LINE__);
        }
    }

    return hostName_;
}

//---------------------------------------------------------------------------
// CTInetAddress::operator==
//---------------------------------------------------------------------------
// Equality Operator between self and an inet address
//---------------------------------------------------------------------------
// Arguments :  address : an inet address.
// Return    :  false if false, true otherwise.
//---------------------------------------------------------------------------
bool CTInetAddress::operator==(const CTInetAddress& address) const
{
    return getAddress() == address.getAddress();
}

//---------------------------------------------------------------------------
// CTInetAddress::getHostAddress
//---------------------------------------------------------------------------
// Returns the formated host address (ie: 192.9.200.232)
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  a string.
//---------------------------------------------------------------------------
string CTInetAddress::getHostAddress() const
{
    string str;
    char acAddr[16];
    union addr {
        long addr_i;
        unsigned char addr_c[4];
    } *addrp;

    long lAddress = htonl(lAddress_);
    memset(&acAddr[0], 0, sizeof(acAddr));
    addrp = (union addr *)&lAddress;
    sprintf(acAddr,"%u.%u.%u.%u",
            (*addrp).addr_c[0],
            (*addrp).addr_c[1],
            (*addrp).addr_c[2],
            (*addrp).addr_c[3]);
    str = acAddr;

    return str;
}

//---------------------------------------------------------------------------
// CTInetAddress::getLocalHostName
//---------------------------------------------------------------------------
// Returns the local host name.
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  a string.
//---------------------------------------------------------------------------
string CTInetAddress::getLocalHostName()
{
    static char acHostname[iMaxHostNameLen_];

#ifdef sun
    if (sysinfo(SI_HOSTNAME, acHostname, iMaxHostNameLen_) != 0)
#else
if (gethostname(acHostname, iMaxHostNameLen_) != 0)
#endif
        strcpy(acHostname, "localhost");
    return acHostname;
}

//---------------------------------------------------------------------------
// CTInetAddress::getPortByName
//---------------------------------------------------------------------------
// Returns the port number associated to a service name.
//---------------------------------------------------------------------------
// Arguments :  serviceName : the name of the requested service.
// Return    :  the port number.
//---------------------------------------------------------------------------
port_t CTInetAddress::getPortByName(const string& serviceName)
{
    struct servent  *sp;

    if (serviceName.empty())
        throw CTSocketException("NullPointerException", __FILE__, __LINE__);

    if ((sp = getservbyname(serviceName.c_str(), "tcp")) == NULL)
        throw CTSocketException("UnknownServiceException", __FILE__, __LINE__);

    return (port_t) ntohs((short) sp->s_port);
}

//---------------------------------------------------------------------------
// CTInetAddress::getHostByAddr
//---------------------------------------------------------------------------
// Returns the hostname associated to a 32 bits address.
//---------------------------------------------------------------------------
// Arguments :  lAddress : the name of the requested service.
// Return    :  the hostname.
//---------------------------------------------------------------------------
string CTInetAddress::getHostByAddr(long lAddress)
{
    string hostname;
#if 0
    struct hostent *hp;

    lAddress = htonl (lAddress) ;
    hp = gethostbyaddr((char *)&lAddress, sizeof (lAddress), AF_INET);
    if (hp == NULL)
        throw CTSocketException("host information not found", __FILE__, __LINE__);
    else
        hostname = hp->h_name;
#endif
    struct in_addr inAddress ;
    inAddress.s_addr = htonl (lAddress) ;
    hostname = inet_ntoa (inAddress);

    return hostname;
}

//---------------------------------------------------------------------------
// CTInetAddress::operator=
//---------------------------------------------------------------------------
// Assignment Operator between self and an inet address.
//---------------------------------------------------------------------------
// Arguments :  address : an inet address.
// Return    :  None.
//---------------------------------------------------------------------------
void CTInetAddress::operator= (const CTInetAddress& address)
{
    hostName_ = address.hostName_;
    lAddress_ = address.lAddress_;
    sFamily_ = address.sFamily_;
}

//---------------------------------------------------------------------------
// operator<<
//---------------------------------------------------------------------------
// Output a socket into an ostream
//---------------------------------------------------------------------------
// Arguments :  os      : the reference to the output stream
//              socket  : a socket reference
// Return    :  reference to the output stream.
//---------------------------------------------------------------------------
ostream& operator<< (ostream& os, const CTSocket& socket)
{
    os << socket.getInetAddress();
    os << " on port ";
    os << socket.getPort();
    return os;
}

#ifndef SUN
//---------------------------------------------------------------------------
// operator<<
//---------------------------------------------------------------------------
// Output a socket into a CTOLogFileList.
//---------------------------------------------------------------------------
// Arguments :  os      : the reference to the CTOLogFileList
//              socket  : a socket reference
// Return    :  reference to the CTOLogFileList.
//---------------------------------------------------------------------------
CTOLogFileList& operator<< (CTOLogFileList& os, const CTSocket& socket)
{
    os << socket.getInetAddress();
    os << " on port ";
    os << socket.getPort();
    return os;
}
#endif

//---------------------------------------------------------------------------
// CTSocket::setNonBlockingMode
//---------------------------------------------------------------------------
// Set non-blocking mode.
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  None.
//---------------------------------------------------------------------------
void CTSocket::setNonBlockingMode()
{
    int flags;

#ifdef WIN32
    unsigned long ulMode = 1;
    flags = ioctlsocket(iSock_, FIONBIO, &ulMode);
#else
    flags = fcntl (iSock_, F_GETFL, 0);
    if (flags == -1)
        throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, getErrorNum());
    flags = fcntl (iSock_, F_SETFL, flags | O_NONBLOCK);
#endif

    if (flags == -1)
        throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, getErrorNum());

    eBlockedMode_ = CTSocket::NONBLOCKED;
}

//---------------------------------------------------------------------------
// CTSocket::setBlockingMode
//---------------------------------------------------------------------------
// Set blocking mode.
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  None.
//---------------------------------------------------------------------------
void CTSocket::setBlockingMode()
{
    int flags;

#ifdef WIN32
    unsigned long ulMode = 0;
    flags = ioctlsocket(iSock_, FIONBIO, &ulMode);
#else
    flags = fcntl (iSock_, F_GETFL, 0);
    if (flags == -1)
        throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, getErrorNum());
    flags = fcntl (iSock_, F_SETFL, flags & ~O_NONBLOCK);
#endif

    if (flags == -1)
        throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, getErrorNum());

    eBlockedMode_ = CTSocket::BLOCKED;
}

//---------------------------------------------------------------------------
// CTSocket::CTSocket
//---------------------------------------------------------------------------
// Constructor for internal usage
//---------------------------------------------------------------------------
// Arguments :  iNewsock    : the socket descriptor
//              tPort       : the port number
//              eBlockedMode    : BLOCKED(default) or NONBLOCKED
//              eCreationMode   : DGRAM or STREAM(default)
//              host        : the hostname
//              sFamily     : the family class
//              lAddress    : the 32 bits address
//              peerHost    : the peer hostname
//              sPeerFamily : the peer family class
//              lPeerAddress: the peer 32 bits address
//              tPeerPort   : the peer port number
// Return    :  None.
//---------------------------------------------------------------------------
CTSocket::CTSocket(int iNewsock,
                   port_t tPort,
                   CTBlockedMode eBlockedMode,
                   CTCreationMode eCreationMode,
                   const string& host,
                   short sFamily,
                   long lAddress,
                   const string& peerHost,
                   short sPeerFamily,
                   long lPeerAddress,
                   port_t tPeerPort)
{
    iSock_ = iNewsock;
    address_.setHostName(host);
    address_.setFamily(sFamily);
    address_.setAddress(lAddress);
    tPort_ = tPort;
    peerAddress_.setHostName(peerHost);
    peerAddress_.setFamily(sPeerFamily);
    peerAddress_.setAddress(lPeerAddress);
    tPeerPort_ = tPeerPort ;
    eBlockedMode_ = eBlockedMode;
    eCreationMode_ = eCreationMode;
    iRef_ = 1;
}

//---------------------------------------------------------------------------
// CTSocket::CTSocket
//---------------------------------------------------------------------------
// Constructor for server usage
//---------------------------------------------------------------------------
// Arguments :  mode : the mode the socket will be created,
//                  CTSocket::STREAM, stream mode, default
//                  CTSocket::DGRAM, datagram mode
//              bReuseAddress : Allow the socket to be bound to an address that is already in use.
// Return    :  None.
//---------------------------------------------------------------------------
CTSocket::CTSocket(CTCreationMode eCreationMode, bool bReuseAddress /* = true */)
        : iSock_(-1), iRef_(1), eBlockedMode_(CTSocket::BLOCKED), eCreationMode_(eCreationMode), tPeerPort_ (0)
{
    create(eCreationMode, bReuseAddress);
}

//---------------------------------------------------------------------------
// CTSocket::CTSocket
//---------------------------------------------------------------------------
// Constructor for client usage
//---------------------------------------------------------------------------
// Arguments :  address : the address (we want) to be connected
//              tPort   : the port number to be connected
//              mode    : the used protocol
//              bReuseAddress : Allow the socket to be bound to an address that is already in use.
// Return    :  None.
//---------------------------------------------------------------------------
CTSocket::CTSocket(const CTInetAddress& address, port_t tPort, CTCreationMode eCreationMode, bool bReuseAddress /* = true */)
        : iSock_(-1), iRef_(1), eBlockedMode_(CTSocket::BLOCKED), eCreationMode_(eCreationMode), tPeerPort_ (0)
{
    create(eCreationMode, bReuseAddress);
    connect(address, tPort);
}

//---------------------------------------------------------------------------
// CTSocket::CTSocket
//---------------------------------------------------------------------------
// Copy Constructor
//---------------------------------------------------------------------------
// Arguments :  socket : a previously defined socket
// Return    :  None.
//---------------------------------------------------------------------------
CTSocket::CTSocket(const CTSocket& socket)
{
    address_ = socket.address_;
    tPort_ = socket.tPort_;
    peerAddress_ = socket.peerAddress_;
    tPeerPort_ = socket.tPeerPort_;
    iSock_ = socket.iSock_;
    eCreationMode_ = socket.eCreationMode_;
    eBlockedMode_ = socket.eBlockedMode_;
    iRef_ = socket.iRef_;
}

//---------------------------------------------------------------------------
// CTSocket::CTSocket
//---------------------------------------------------------------------------
// Constructor for client usage
//---------------------------------------------------------------------------
// Arguments :  host    : the host name (we want) to be connected
//              tPort   : the port number to be connected
//              mode    : the used protocol
//              bReuseAddress : Allow the socket to be bound to an address that is already in use.
// Return    :  None.
//---------------------------------------------------------------------------
CTSocket::CTSocket(const string& host, port_t tPort, CTCreationMode eCreationMode, bool bReuseAddress /* = true */)
        : iSock_(-1), iRef_(1), eBlockedMode_(CTSocket::BLOCKED), eCreationMode_(eCreationMode), tPeerPort_ (0)
{
    CTInetAddress address = CTInetAddress::getByName(host);
    create(eCreationMode, bReuseAddress);
    connect(address, tPort);
}

//---------------------------------------------------------------------------
// CTSocket::CTSocket
//---------------------------------------------------------------------------
// Constructor for a null socket
//---------------------------------------------------------------------------
// Arguments :  Ignored.
// Return    :  None.
//---------------------------------------------------------------------------
CTSocket::CTSocket(const char *)
{
    iSock_ = -1;
    tPort_ = 0;
    tPeerPort_ = 0;
    iRef_ = 1;
}

//---------------------------------------------------------------------------
// CTSocket::~CTSocket
//---------------------------------------------------------------------------
// Destructor
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  None.
//---------------------------------------------------------------------------
CTSocket::~CTSocket()
{
    if (iSock_ != -1)
        close();
    iSock_ = -1;
}

//---------------------------------------------------------------------------
// CTSocket::makeInetSocket
//---------------------------------------------------------------------------
// Creates a socket with the requested type specification
//---------------------------------------------------------------------------
// Arguments :  iType   :   SOCK_STREAM, connection-based byte stream
//                          SOCK_DGRAM, datagrams
// Return    :  a descriptor referencing the new socket, -1 if failed.
//---------------------------------------------------------------------------
int CTSocket::makeInetSocket(int iType)
{
    return socket(AF_INET, iType, 0);
}

//---------------------------------------------------------------------------
// CTSocket::makeStreamSocket
//---------------------------------------------------------------------------
// Creates a stream socket
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  a descriptor referencing the new socket, -1 if failed.
//---------------------------------------------------------------------------
int CTSocket::makeStreamSocket()
{
    int s = makeInetSocket(SOCK_STREAM);
    if (s != -1)
    {
        int iOption = 0 ;
        setsockopt (s, SOL_SOCKET, SO_OOBINLINE, (const char*) &iOption, sizeof (iOption) ) ;
#if 0
        linger lin;

        // Set to linger time of zero
        lin.l_onoff=1;
        lin.l_linger=0;

        // And set the socket up to receive these linger option
        setsockopt(s, SOL_SOCKET, SO_LINGER, (const char *) &lin, sizeof(lin));
#endif
    }
    return s ;
}

//---------------------------------------------------------------------------
// CTSocket::makeInetSocket
//---------------------------------------------------------------------------
// Creates a datagram socket
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  a descriptor referencing the new socket, -1 if failed.
//---------------------------------------------------------------------------
int CTSocket::makeDgramSocket()
{
    return makeInetSocket(SOCK_DGRAM);
}

//---------------------------------------------------------------------------
// setPort
//---------------------------------------------------------------------------
// Assigns the field port of an inet address
//---------------------------------------------------------------------------
// Arguments :  pSockaddr   : a pointer to an internet address
//              tPort       : a port number
// Return    :  None.
//---------------------------------------------------------------------------
static void setPort(sockaddr_in *pSockaddr, port_t tPort)
{
    pSockaddr->sin_port = htons((short) tPort);
}

//---------------------------------------------------------------------------
// CTSocket::create
//---------------------------------------------------------------------------
// Creates a new socket according to the choosed protocol
//---------------------------------------------------------------------------
// Arguments :  mode : the underlaying used protocol,
//                  CTSocket::STREAM, stream protocol
//                  CTSocket::DGRAM, datagram protocol
//              bReuseAddress : Allow the socket to be bound to an address that is already in use.
// Return    :  None.
//---------------------------------------------------------------------------
void CTSocket::create(CTCreationMode eCreationMode, bool bReuseAddress /* = true */)
{
    iSock_ = (eCreationMode == CTSocket::STREAM) ? makeStreamSocket() : makeDgramSocket();
    if (iSock_ == -1)
        throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, getErrorNum());

    ;
    setsockopt (iSock_, SOL_SOCKET, SO_REUSEADDR, (const char*) &bReuseAddress, sizeof (bReuseAddress) ) ;
}

//---------------------------------------------------------------------------
// CTSocket::reuseAddress
//---------------------------------------------------------------------------
// Allow the socket to be bound to an address that is already in use.
//---------------------------------------------------------------------------
// Arguments :  bReuseAddress : Allow the socket to be bound to an address that is already in use.
// Return    :  None.
//---------------------------------------------------------------------------
void CTSocket::reuseAddress(bool bReuseAddress)
{
    setsockopt (iSock_, SOL_SOCKET, SO_REUSEADDR, (const char*) &bReuseAddress, sizeof (bReuseAddress) ) ;
}

void CTSocket::setLinger(const char *my_linger, int linger_size)
{
    setsockopt (iSock_, SOL_SOCKET, SO_LINGER, my_linger, linger_size) ;
}


//---------------------------------------------------------------------------
// CTSocket::bindSocket
//---------------------------------------------------------------------------
// Binds the socket to the specified address on the specified port
//---------------------------------------------------------------------------
// Arguments :  address       : the inet address
//              tPort         : the port number
// Return    :  None.
//---------------------------------------------------------------------------
void CTSocket::bindSocket(const CTInetAddress &address, port_t tPort)
{
    sockaddr_in sin;

    memset((char *) &sin, 0, sizeof(sockaddr_in));
    setPort(&sin, tPort);
    sin.sin_addr.s_addr = (unsigned long) htonl (address.getAddress() ) ;
    sin.sin_family = address.getFamily();
    if (::bind(iSock_, (sockaddr *) &sin, sizeof(sin)) == -1)
        throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, getErrorNum());
}

//---------------------------------------------------------------------------
// CTSocket::bindToPort
//---------------------------------------------------------------------------
// Binds the socket on the specified port
//---------------------------------------------------------------------------
// Arguments :  tPort   : the port number
// Return    :  None.
//---------------------------------------------------------------------------
void CTSocket::bindToPort(port_t tPort)
{
    sockaddr_in sin = {AF_INET};

    setPort(&sin, tPort);
    if (::bind(iSock_, (sockaddr *) &sin, sizeof(sin)) == -1)
    {
        throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, getErrorNum());
    }
    getPortFromConnectedSocket();
}

//---------------------------------------------------------------------------
// CTSocket::getPortFromConnectedSocket
//---------------------------------------------------------------------------
// Gets the port number associated with the current socket.
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  None.
//---------------------------------------------------------------------------
void CTSocket::getPortFromConnectedSocket()
{
    /* Now that we're a connected CTSocket, let's extract the port number
       that the system chose for us and store it in the CTSocket object. */
    sockaddr_in  me;
    int iLen = sizeof(sockaddr_in);

    if (getsockname(iSock_, (sockaddr *) &me, &iLen) == -1)
        throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, getErrorNum());

    tPort_ = ntohs(me.sin_port);
}

//---------------------------------------------------------------------------
// CTSocket::bindAnonymously
//---------------------------------------------------------------------------
// Binds anonymously the current socket on the specified address.
// Retrieves port number from the system.
//---------------------------------------------------------------------------
// Arguments :  address :   the inet address.
// Return    :  None.
//---------------------------------------------------------------------------
void CTSocket::bindAnonymously(const CTInetAddress &address)
{
    bindSocket(address, 0);
    getPortFromConnectedSocket();
}

//---------------------------------------------------------------------------
// CTSocket::connect
//---------------------------------------------------------------------------
// Connects the socket to the specified port on the specified host
// Retrieves port number from the system.
//---------------------------------------------------------------------------
// Arguments :  hostname    :   the host name
//              tPort       :   the port number
// Return    :  None.
//---------------------------------------------------------------------------
void CTSocket::connect(const string &hostname, port_t tPort)
{
    CTInetAddress address = CTInetAddress::getByName(hostname);
    connect(address, tPort);
    iRef_ = 1;
}

//---------------------------------------------------------------------------
// CTSocket::connect
//---------------------------------------------------------------------------
// Connects the socket to the specified port on the specified host
// Retrieves port number from the system.
//---------------------------------------------------------------------------
// Arguments :  address :   the inet address
//              tPort   :   th port number
// Return    :  None.
//---------------------------------------------------------------------------
void CTSocket::connect(const CTInetAddress &dest, port_t tPort)
{
    struct sockaddr_in  him;

    him.sin_family = dest.getFamily();
    setPort(&him, tPort);
    him.sin_addr.s_addr = (unsigned long) htonl(dest.getAddress());

    if (iSock_ == -1)
        create(eCreationMode_);
#if 0
    linger lin;

    // Set to linger time of zero
    lin.l_onoff=1;
    lin.l_linger=0;

    // And set the socket up to receive these linger option
    setsockopt(iSock_, SOL_SOCKET, SO_LINGER, (const char *) &lin, sizeof(lin));
#endif
    if (::connect(iSock_, (sockaddr *) &him, sizeof(him)) < 0)
        throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, getErrorNum());

    getPortFromConnectedSocket();

    struct sockaddr_in  peer;
    int                 iPeerSize = sizeof(peer) ;
    if (::getpeername(iSock_, (sockaddr *) &peer, &iPeerSize) < 0)
        throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, getErrorNum());

    if (((CTInetAddress&)dest).getHostName() == inet_ntoa(peer.sin_addr))
    {
        peerAddress_.setHostName(((CTInetAddress&)dest).getHostName());
    }
    else
    {
        peerAddress_.setHostName(CTInetAddress::getHostByAddr(ntohl(peer.sin_addr.s_addr)));
    }
    peerAddress_.setFamily(peer.sin_family);
    peerAddress_.setAddress( (unsigned long) ntohl(peer.sin_addr.s_addr));
    tPeerPort_ = ntohs(peer.sin_port);
}

//---------------------------------------------------------------------------
// CTSocket::listen
//---------------------------------------------------------------------------
// Specifies how many connection requests the system will queue up
// while waiting for accept().
//---------------------------------------------------------------------------
// Arguments :  iCount  :   the backlog value
// Return    :  None.
//---------------------------------------------------------------------------
void CTSocket::listen(int iCount)
{
    if (::listen(iSock_, iCount) < 0)
        throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, getErrorNum());
}

//---------------------------------------------------------------------------
// CTSocket::accept
//---------------------------------------------------------------------------
// Accepts a connection.
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  a new socket.
//---------------------------------------------------------------------------
CTSocket CTSocket::accept()
{
    sockaddr_in him;
    int iLen = sizeof(sockaddr_in);
    int iNewsock = ::accept(iSock_, (sockaddr *) &him, &iLen);

    if (iNewsock < 0)
        throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, getErrorNum());

    int iOption = 0 ;
    setsockopt (iNewsock, SOL_SOCKET, SO_OOBINLINE, (const char*) &iOption, sizeof (iOption) ) ;

    string himHost ;
    struct in_addr inAddress ;
    inAddress.s_addr = htonl (address_.getAddress()) ;
    if (address_.getHostName() == inet_ntoa(inAddress))
    {
        himHost = inet_ntoa (him.sin_addr);
    }
    else
    {
        himHost = CTInetAddress::getHostByAddr(ntohl(him.sin_addr.s_addr));
    }

    struct sockaddr_in  me;
    int                 iMeSize = sizeof(me) ;
    if (::getsockname(iNewsock, (sockaddr *) &me, &iMeSize) < 0)
        throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, getErrorNum());

    string meHost ;
    if (address_.getHostName() == inet_ntoa (inAddress))
    {
        meHost = inet_ntoa (me.sin_addr);
    }
    else
    {
        meHost = CTInetAddress::getHostByAddr(ntohl(me.sin_addr.s_addr));
    }

    return CTSocket(iNewsock, ntohs(me.sin_port), eBlockedMode_, eCreationMode_, meHost, me.sin_family, ntohl(me.sin_addr.s_addr), himHost, him.sin_family, ntohl(him.sin_addr.s_addr), ntohs(him.sin_port) );
}

//---------------------------------------------------------------------------
// CTSocket::select
//---------------------------------------------------------------------------
// Selects a connection.
//---------------------------------------------------------------------------
// Arguments :  timeout     :   to specify a time to wait..
//        eVent     :   to specify event to select
// Return    :  CTSelection::TIMEDOUT on time out,
//              CTSelection::RECEIVE if data is available on reception.
//              CTSelection::SEND    if socket is available for emission.
//              CTSelection::EXCEPTION if exception is notified.
//---------------------------------------------------------------------------
CTSocket::CTSelection CTSocket::select(timeval &timeout, int iEvent)
{
    timeval *pTimeOut = &timeout;
    return doSelect(pTimeOut, iEvent);
}

//---------------------------------------------------------------------------
// CTSocket::select
//---------------------------------------------------------------------------
// Selects a connection.
//---------------------------------------------------------------------------
// Arguments :  iEvent      :   to specify event to select
// Return    :  CTSelection::TIMEDOUT on time out,
//              CTSelection::RECEIVE if data is available on reception.
//              CTSelection::SEND    if socket is available for emission.
//              CTSelection::EXCEPTION if exception is notified.
//---------------------------------------------------------------------------
CTSocket::CTSelection CTSocket::select(int iEvent)
{
    timeval *pTimeOut = (timeval *)NULL;
    return doSelect(pTimeOut, iEvent);
}

//---------------------------------------------------------------------------
// CTSocket::poll
//---------------------------------------------------------------------------
// Polls on a connection.
//---------------------------------------------------------------------------
// Arguments :  iEvent      :   to specify event to poll
// Return    :  CTSelection::TIMEDOUT on time out,
//              CTSelection::RECEIVE if data is available on reception.
//              CTSelection::SEND    if socket is available for emission.
//              CTSelection::EXCEPTION if exception is notified.
//---------------------------------------------------------------------------
CTSocket::CTSelection CTSocket::poll(int iEvent)
{
    timeval TimeOut;
    TimeOut.tv_sec = 0;
    TimeOut.tv_usec = 0;
    return doSelect(&TimeOut, iEvent);
}

//---------------------------------------------------------------------------
// CTSocket::doSelect
//---------------------------------------------------------------------------
// Selects a connection. internal usage
//---------------------------------------------------------------------------
// Arguments :  pTimeOut    :   to specify a time to wait..
//              eVent       :   to specify event to select
// Return    :  CTSelection::TIMEDOUT on time out,
//              CTSelection::RECEIVE if data is available on reception.
//              CTSelection::SEND    if socket is available for emission.
//              CTSelection::EXCEPTION if exception is notified.
//---------------------------------------------------------------------------
CTSocket::CTSelection CTSocket::doSelect(timeval *pTimeOut, int iEvent)
{
    int iNfds = 32;
    fd_set readfds, writefds, exceptfds;
    fd_set *preadfds =  NULL, *pwritefds = NULL, *pexceptfds = NULL;

#ifdef  WIN32
#pragma warning(disable : 4127)
#endif  // WIN32

    if (iEvent & _S_RECEIVE_)
    {
        preadfds = &readfds;
        FD_ZERO(preadfds);
        FD_SET(iSock_, preadfds);
    }

    if (iEvent & _S_SEND_)
    {
        pwritefds = &writefds;
        FD_ZERO(pwritefds);
        FD_SET(iSock_, pwritefds);
    }

    if (iEvent & _S_EXCEPTION_)
    {
        pexceptfds = &exceptfds;
        FD_ZERO(pexceptfds);
        FD_SET(iSock_, pexceptfds);
    }

#ifdef  WIN32
#pragma warning(default : 4127)
#endif  // WIN32

    int iSelect = ::select(iNfds, preadfds, pwritefds, pexceptfds, pTimeOut);

    if (iSelect == -1)
        /* Error */
        throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, getErrorNum());
    else if (iSelect == 0)
        /* Time limit expired. */
        return CTSocket::TIMEDOUT;

    if (iEvent & _S_RECEIVE_)
    {
        if (FD_ISSET(iSock_,preadfds))
            return CTSocket::RECEIVE;
    }
    else if (iEvent & _S_SEND_)
    {
        if (FD_ISSET(iSock_,pwritefds))
            return CTSocket::SEND;
    }
    else if (iEvent & _S_EXCEPTION_)
    {
        if (FD_ISSET(iSock_,pexceptfds))
            return CTSocket::EXCEPTION;
    }
    else
    {
        throw CTSocketException("select on bad socket descriptor", __FILE__, __LINE__, getErrorNum());
    }

    return CTSocket::UNKNOWN;

}

//---------------------------------------------------------------------------
// CTSocket::close
//---------------------------------------------------------------------------
// Closes the socket.
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  None.
//---------------------------------------------------------------------------
void CTSocket::close()
{
    int ret;

    iRef_--;

    if (iRef_ == 0)
    {
#ifdef WIN32
        ret = ::closesocket(iSock_);
#else
        ret = ::close(iSock_);
#endif
        iSock_ = -1;
        if (ret != 0)
            throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, getErrorNum());
    }
}

//---------------------------------------------------------------------------
// CTSocket::shutdown
//---------------------------------------------------------------------------
// Terminate gracefully send or/and receive on the socket.
//---------------------------------------------------------------------------
// Arguments :  .
// Return    :  None.
//---------------------------------------------------------------------------
void CTSocket::shutdown(CTTermination eTermination)
{
    int iHow = -1;

    switch (eTermination)
    {
    case RECEPTION:
        iHow = 0;
        break;
    case TRANSMISSION:
        iHow = 1;
        break;
    case BOTH:
        iHow = 2;
        break;
    default:
        throw CTSocketException("Invalid termination mode specified", __FILE__, __LINE__);
    }

    if (::shutdown(iSock_, iHow) != 0)
        throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, getErrorNum());
}

//---------------------------------------------------------------------------
// CTSocket::swap
//---------------------------------------------------------------------------
// swap all bytes of a double.
//---------------------------------------------------------------------------
// Arguments :  d       :   a double
// Return    :  a double with all bytes swapped.
//---------------------------------------------------------------------------
double CTSocket::swap(double d)
{
#ifdef WIN32
    char acBuffer[8];
    char *pcPtr = (char *)&d;
    acBuffer[0] = pcPtr[7];
    acBuffer[1] = pcPtr[6];
    acBuffer[2] = pcPtr[5];
    acBuffer[3] = pcPtr[4];
    acBuffer[4] = pcPtr[3];
    acBuffer[5] = pcPtr[2];
    acBuffer[6] = pcPtr[1];
    acBuffer[7] = pcPtr[0];

    return *(double *)acBuffer;
#else
    return d;
#endif
}

//---------------------------------------------------------------------------
// CTSocket::swap
//---------------------------------------------------------------------------
// swap all bytes of a float.
//---------------------------------------------------------------------------
// Arguments :  f       :   a float
// Return    :  a float with all bytes swapped.
//---------------------------------------------------------------------------
float CTSocket::swap(float f)
{
#ifdef WIN32
    char acBuffer[4];
    char *pcPtr = (char *)&f;
    acBuffer[0] = pcPtr[3];
    acBuffer[1] = pcPtr[2];
    acBuffer[2] = pcPtr[1];
    acBuffer[3] = pcPtr[0];

    return *(float *)acBuffer;
#else
    return f;
#endif
}

//---------------------------------------------------------------------------
// CTSocket::send
//---------------------------------------------------------------------------
// Sends a string over the network.
//---------------------------------------------------------------------------
// Arguments :  data    :   a reference to the data to be sent
// Return    :  the number of bytes sent.
//---------------------------------------------------------------------------
long CTSocket::send(const string& data)
{
    const char *pcDataPtr;
    long lDataLen;

    pcDataPtr = data.c_str();
    lDataLen = data.length();
    (*this) << lDataLen;
    if (lDataLen > 0)      // We don't want to send a null buffer
    {
        sendBuffer((void *)pcDataPtr, int(sizeof(char) *lDataLen), "unable to send a char*");
    }

    return lDataLen;
}

//---------------------------------------------------------------------------
// CTSocket::write
//---------------------------------------------------------------------------
// Sends a string over the network.
// A synonym to the previous method (CTSocket::send).
//---------------------------------------------------------------------------
// Arguments :  data    :   a reference to the data to be sent
// Return    :  the number of bytes sent.
//---------------------------------------------------------------------------
long CTSocket::write(const string& data)
{
    return send(data);
}

//---------------------------------------------------------------------------
// CTSocket::operator<<
// A synonym to the previous method (CTSocket::write).
//---------------------------------------------------------------------------
// Receives a string from the network.
//---------------------------------------------------------------------------
// Arguments :  data    :   a reference to the received data
// Return    :  a self reference.
//---------------------------------------------------------------------------
CTSocket& CTSocket::operator<<(const string &data)
{
    send(data);
    return *this;
}

//---------------------------------------------------------------------------
// CTSocket::recv
//---------------------------------------------------------------------------
// Receives a string from the network.
//---------------------------------------------------------------------------
// Arguments :  data    :   a reference to the received data
// Return    :  the number of bytes received.
//---------------------------------------------------------------------------
long CTSocket::recv(string& data)
{
    const char *pcDataPtr;
    long lDataLen = 0;

    (*this) >> lDataLen;
    try
    {
        data.resize(int(lDataLen));
    }
    catch(...)
    {
        throw CTxalloc("Can't resize string", lDataLen);
    }
    pcDataPtr = data.c_str();
    if (lDataLen > 0)
    {
        recvBuffer((void *)pcDataPtr, int(sizeof(char) * lDataLen), "unable to receive a char*");
    }

    return lDataLen;
}

//---------------------------------------------------------------------------
// CTSocket::read
// A synonym to the previous method (CTSocket::recv).
//---------------------------------------------------------------------------
// Receives a string from the network.
//---------------------------------------------------------------------------
// Arguments :  data    :   a reference to the received data
// Return    :  the number of bytes received.
//---------------------------------------------------------------------------
long CTSocket::read(string &data)
{
    return recv(data);
}

//---------------------------------------------------------------------------
// CTSocket::operator>>
// A synonym to the previous method (CTSocket::recv).
//---------------------------------------------------------------------------
// Receives a string from the network.
//---------------------------------------------------------------------------
// Arguments :  data    :   a reference to the received data
// Return    :  a self reference.
//---------------------------------------------------------------------------
CTSocket& CTSocket::operator>>(string &data)
{
    recv(data);

    return *this;
}

//---------------------------------------------------------------------------
// CTSocket::sendTo
//---------------------------------------------------------------------------
// Sends data to a specific destination.
//---------------------------------------------------------------------------
// Arguments :  pv      :   a void pointer to the data to be sent
//              len     :   length of the buffer to be sent
//              address :   the target inet address
//              tPort   :   the target port number
// Return    :  the number of bytes sent.
//---------------------------------------------------------------------------
long CTSocket::sendTo(void* pv, int iLen, const CTInetAddress& address, port_t tPort)
{
    const char *pcDataBuffer;
    int iDataLen, iCharSent, iOffset;

    pcDataBuffer = (char *)pv;
    iDataLen = iLen;

    sockaddr_in  it;
    it.sin_family = address.getFamily();
    setPort(&it, tPort);
    it.sin_addr.s_addr = (unsigned long) htonl(address.getAddress());
    int iSockLen = sizeof(it);

    if (iDataLen <= 0)
        throw CTSocketException( "ArrayIndexOutOfBoundsException", __FILE__, __LINE__);

    iOffset = 0;
    while (iDataLen > 0)
    {
        iCharSent = ::sendto(iSock_, pcDataBuffer + iOffset, iDataLen, 0, (sockaddr*)&it, iSockLen);
        if (iCharSent == -1)
        {
            long lErrorNum = getErrorNum();
#ifdef WIN32
            if (!isBlocking() && lErrorNum == WSAEWOULDBLOCK)
#else
            if (!isBlocking() && lErrorNum == EAGAIN)
#endif
                continue;
            throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, lErrorNum);
        }
        iOffset += iCharSent;
        iDataLen -= iCharSent;
    }

    return iOffset;
}

//---------------------------------------------------------------------------
// CTSocket::recvFrom
//---------------------------------------------------------------------------
// Receives data from a specific source.
//---------------------------------------------------------------------------
// Arguments :  data    :   a reference to the data to be sent
//              address :   the source inet address
//              tPort   :   the source port number
// Return    :  the number of bytes received.
//---------------------------------------------------------------------------
long CTSocket::recvFrom(void* pv, int iLen, CTInetAddress& address, port_t& tPort)
{
    char *pcDataBuffer;
    int iDataLen, iReceived, iOffset;

    pcDataBuffer = (char *)pv;
    iDataLen = iLen;
    iOffset = 0;

    sockaddr_in it;
    int iSockLen = sizeof(it);

    while (iDataLen > 0)
    {
        iReceived = ::recvfrom(iSock_, pcDataBuffer + iOffset, iDataLen, 0, (sockaddr*)&it, &iSockLen);
        if (iReceived == -1)
        {
            long lErrorNum = getErrorNum();
#ifdef WIN32
            if (!isBlocking() && lErrorNum == WSAEWOULDBLOCK)
#else
            if (!isBlocking() && lErrorNum == EAGAIN)
#endif
                continue;

            throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, lErrorNum);
        }
        iOffset += iReceived;
        iDataLen -= iReceived;
    }

    address.setFamily(it.sin_family);
    address.setAddress( (unsigned long) ntohl(it.sin_addr.s_addr));

    struct in_addr inAddress ;
    inAddress.s_addr = htonl (address_.getAddress()) ;
    if (address_.getHostName() == inet_ntoa (inAddress))
    {
        address.setHostName(inet_ntoa(it.sin_addr));
    }
    else
    {
        address.setHostName(CTInetAddress::getHostByAddr(ntohl(it.sin_addr.s_addr)));
    }

    tPort = ntohs(it.sin_port);

    return iOffset;
}

/***********************************************
** Here, are the methods used to receive data. *
** Most of them use the 'recvBuffer' method.   *
***********************************************/

//---------------------------------------------------------------------------
// CTSocket::recvBuffer
//---------------------------------------------------------------------------
// Gets the error message from the internal error number.
//---------------------------------------------------------------------------
// Arguments :  pv          :   a void pointer to the buffer to store the data
//              iLen        :   the size of the buffer
//              szMessage   :   the message if failure during the receiption
//              bUrgent     :   a boolean to indicate if data is urgent
// Return    :  the number of bytes received.
//---------------------------------------------------------------------------
long CTSocket::recvBuffer(void* pv, int iLen, const char* szMessage, bool bUrgent /* = false */)
{
    int iOffset = 0, iCharRecv = 0;
    int iDataLen = iLen;
    char *pcDataBuffer = (char *)pv;

    if (iDataLen < 0)
        throw CTSocketException(szMessage, __FILE__, __LINE__);

    do
    {
        iCharRecv = ::recv(iSock_, pcDataBuffer + iOffset, iDataLen, bUrgent ? MSG_OOB : 0);
        if (iCharRecv == -1)
        {

            long lErrorNum = getErrorNum();
#ifdef WIN32
            if (!isBlocking() && lErrorNum == WSAEWOULDBLOCK)
#else
            if (!isBlocking() && lErrorNum == EAGAIN)
#endif
                continue;
            string errorMessage = szMessage;
            errorMessage += "\n";
            errorMessage += getErrorMsg();
            throw CTSocketException(errorMessage, __FILE__, __LINE__, lErrorNum);
        }
        else if (iCharRecv == 0)
            break;
        else
        {
            iOffset += iCharRecv;
            iDataLen -= iCharRecv;
        }

        CTCurrentThread::sleep (1) ;
    }
    while (iDataLen > 0);

    return iOffset;
}

//---------------------------------------------------------------------------
// CTSocket::recvAvailableData
//---------------------------------------------------------------------------
// Receive as much data as is currently available
// (up to the size of the buffer supplied).
// this the standard.
//---------------------------------------------------------------------------
// Arguments :  pv          :   a void pointer to the buffer to store the data
//              iMaxLen     :   maximum data size (size of the buffer)
//              bUrgent     :   a boolean to indicate if data is urgent
// Return    :  the number of bytes received.
//---------------------------------------------------------------------------
long CTSocket::recvAvailableData(void* pBuffer, int iMaxLen, bool bUrgent /* = false */)
{
    int     iCharRecv, iOffset;
    bool    bMoreData = false ;

    if (iMaxLen <= 0)
        throw CTSocketException("iMaxLen must be greater than zero!", __FILE__, __LINE__);

    iOffset = 0;
    do
    {
        iCharRecv = ::recv(iSock_, (char*) pBuffer + iOffset, iMaxLen - iOffset, bUrgent ? MSG_OOB : 0);
        if (iCharRecv == -1)
        {
            long lErrorNum = getErrorNum();
#ifdef WIN32
            if (!isBlocking() && lErrorNum == WSAEWOULDBLOCK)
#else
            if (!isBlocking() && lErrorNum == EAGAIN)
#endif
            {
                iCharRecv = 0 ;
            }
            else
            {
                throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, lErrorNum);
            }
        }

        CTCurrentThread::sleep (1) ;

        iOffset += iCharRecv;

        int     iNfds = 32;
        fd_set  fds ;
        timeval timeOut ;

        timeOut.tv_sec = timeOut.tv_usec = 0;

#ifdef  WIN32
#pragma warning(disable : 4127)
#endif  // WIN32
        FD_ZERO(&fds);
        FD_SET(iSock_, &fds);
#ifdef  WIN32
#pragma warning(default : 4127)
#endif  // WIN32

        if (!isBlocking() || (iCharRecv == 0) )
        {
            break ;
        }

        int iSelect = ::select(iNfds, (fd_set*) (bUrgent ? NULL : &fds), NULL, (fd_set*) (bUrgent ? &fds : NULL), &timeOut);

        switch (iSelect)
        {
        case 0 :
            bMoreData = false ;
            break ;
        case 1 :
            bMoreData = true ;
            break ;
        default :
            throw CTSocketException(getErrorMsg(), __FILE__, __LINE__, getErrorNum());
        }
    }
    while ( (bMoreData) && (iOffset < iMaxLen) ) ;

    return iOffset;
}

long CTSocket::get(char& c)
{
    return recvBuffer((void *)&c, sizeof(c), "unable to receive a char");
}

long CTSocket::get(unsigned char& uc)
{
    return recvBuffer((void *)&uc, sizeof(uc), "unable to receive an unsigned char");
}

long CTSocket::get(char* p, size_t N)
{
    return recvBuffer((void *)p, sizeof(*p) * N, "unable to receive a fixed char array");
}

long CTSocket::get(unsigned char* p, size_t N)
{
    return recvBuffer((void *)p, sizeof(*p) * N, "unable to receive a fixed unsigned char array");
}

long CTSocket::get(double* p, size_t N)
{
    long lCharRecv = 0;

    if (N > 0)
    {
        lCharRecv = recvBuffer((void *)p, sizeof(*p) * N, "unable to receive a fixed double array");
        for (unsigned int i = 0; i < N ; i++)
            p[i] = swap(p[i]);
    }

    return lCharRecv;
}

long CTSocket::get(float* p, size_t N)
{
    long lCharRecv = 0;

    if (N > 0)
    {
        lCharRecv = recvBuffer((void *)p, sizeof(*p) * N, "unable to receive a fixed float array");
        for (unsigned int i = 0; i < N ; i++)
            p[i] = swap(p[i]);
    }

    return lCharRecv;
}

long CTSocket::get(int*, size_t)
{
    throw CTSocketException("unallowed to receive a fixed int array", __FILE__, __LINE__);
    return -1;
}

long CTSocket::get(unsigned int*, size_t)
{
    throw CTSocketException("unallowed to receive a fixed unsigned int array", __FILE__, __LINE__);
    return -1;
}

long CTSocket::get(long* p, size_t N)
{
    long lCharRecv = 0;

    if (N > 0)
    {
        lCharRecv = recvBuffer((void *)p, sizeof(*p) * N, "unable to receive a fixed long array");
        for (unsigned int i = 0; i < N ; i++)
            p[i] = (long)ntohl(p[i]);
    }

    return lCharRecv;
}

long CTSocket::get(unsigned long* p, size_t N)
{
    long lCharRecv = 0;

    if (N > 0)
    {
        lCharRecv = recvBuffer((void *)p, sizeof(*p) * N, "unable to receive a fixed unsigned long array");
        for (unsigned int i = 0; i < N ; i++)
            p[i] = (unsigned long)ntohl(p[i]);
    }

    return lCharRecv;
}

long CTSocket::get(short* p, size_t N)
{
    long lCharRecv = 0;

    if (N > 0)
    {
        lCharRecv = recvBuffer((void *)p, sizeof(*p) * N, "unable to receive a fixed short array");
        for (unsigned int i = 0; i < N ; i++)
            p[i] = (short)ntohs(p[i]);
    }

    return lCharRecv;
}

long CTSocket::get(unsigned short* p, size_t N)
{
    long lCharRecv = 0;

    if (N > 0)
    {
        lCharRecv = recvBuffer((void *)p, sizeof(*p) * N, "unable to receive a fixed unsigned short array");
        for (unsigned int i = 0; i < N ; i++)
            p[i] = (short)ntohs(p[i]);
    }

    return lCharRecv;
}

//---------------------------------------------------------------------------
// CTSocket::operator>>
//---------------------------------------------------------------------------
// Receives a null terminated string.
//---------------------------------------------------------------------------
// Arguments :  szStr   :   a pointer to a NEW null terminated string
// Return    :  a self reference.
//---------------------------------------------------------------------------
CTSocket& CTSocket::operator>>(char*& szStr)
{
    string string;
    char c = 0x1;
    try
    {
        while (c != '\0')
        {
            recvBuffer((void *)&c, sizeof(c) , "unable to receive a single char");
            string += c;
        }
        if (!szStr)
            delete []szStr;
        try
        {
            szStr = new char[string.length()];
        }
        catch (CTxalloc&)
        {
            throw CTSocketException ("operator>>::new char[string.length()] failed.", __FILE__, __LINE__) ;
        }
        memcpy(szStr, string.c_str(), string.length());
    }
    catch (CTSocketException&)
    {
        throw CTSocketException("unable to receive a char*", __FILE__, __LINE__, getErrorNum());
    }
    return *this;
}

long CTSocket::getString(char* p, size_t tMaxLen)
{
    int iDataLen = tMaxLen, iCharRecv = -1;

    if (iDataLen <= 0)
        throw CTSocketException("unable to receive a fixed char array", __FILE__, __LINE__);

    while ((iCharRecv = ::recv(iSock_, p, iDataLen, 0)) == -1 )
    {
        long lErrorNum = getErrorNum();
#ifdef WIN32
        if (!isBlocking() && lErrorNum == WSAEWOULDBLOCK)
#else
        if (!isBlocking() && lErrorNum == EAGAIN)
#endif
            continue;

        string errorMessage = "unable to receive a fixed char array";
        errorMessage += "\n";
        errorMessage += getErrorMsg();
        throw CTSocketException(errorMessage, __FILE__, __LINE__, getErrorNum());
    }

    return iCharRecv;
}

CTSocket& CTSocket::operator>>(char& c)
{
    get(c);
    return *this;
}

CTSocket& CTSocket::operator>>(double& d)
{
    recvBuffer(&d, sizeof(d), "unable to receive a double");
    d = swap(d);
    return *this;
}

CTSocket& CTSocket::operator>>(float& f)
{
    recvBuffer((void*)&f, sizeof(f), "unable to receive a float");
    f = swap(f);
    return *this;
}

CTSocket& CTSocket::operator>>(int&)
{
    throw CTSocketException("unallowed to receive an int", __FILE__, __LINE__);
    return *this;
}

CTSocket& CTSocket::operator>>(bool& b)
{
    unsigned long ul = 0;
    recvBuffer((void*)&ul, sizeof(ul), "unable to receive a bool");
    ul = (unsigned long)ntohl(ul);
    if (ul > 0)
        b = true ;
    else
        b = false ;
    return *this;
}

CTSocket& CTSocket::operator>>(long&  l)
{
    recvBuffer((void*)&l, sizeof(l), "unable to receive a long");
    l = (long)ntohl(l);
    return *this;
}

CTSocket& CTSocket::operator>>(short& s)
{
    recvBuffer((void*)&s, sizeof(s), "unable to receive a short");
    s = (short)ntohs(s);
    return *this;
}

CTSocket& CTSocket::operator>>(unsigned char& uc)
{
    get(uc);
    return *this;
}

CTSocket& CTSocket::operator>>(unsigned int&)
{
    throw CTSocketException("unallowed to receive an unsigned int", __FILE__, __LINE__);
    return *this;
}

CTSocket& CTSocket::operator>>(unsigned long& l)
{
    recvBuffer((void*)&l, sizeof(l), "unable to receive an unsigned long");
    l = ntohl(l);
    return *this;
}

CTSocket& CTSocket::operator>>(unsigned short& s)
{
    recvBuffer((void*)&s, sizeof(s), "unable to receive an unsigned short");
    s = (short)ntohs(s);
    return *this;
}

/**********************************************
** Here, are the methods used to send data.  **
** Most of them use the 'sendBuffer' method. **
**********************************************/

//---------------------------------------------------------------------------
// CTSocket::sendBuffer
//---------------------------------------------------------------------------
// Gets the error message from the internal error number.
//---------------------------------------------------------------------------
// Arguments :  pv          :   a void pointer to the data to be sent
//              iLen        :   the size of the buffer
//              szMessage   :   the message if failure during the sent
//              bUrgent     :   a boolean to indicate if data is urgent
// Return    :  number of bytes sent
//---------------------------------------------------------------------------
long CTSocket::sendBuffer(void* pv, int iLen, const char* szMessage, bool bUrgent /* = false */)
{
    const char *pcDataBuf;
    int iDataLen, iCharSend, iOffset;

    pcDataBuf = (char *)pv;
    iDataLen = iLen;

    if (iDataLen < 0)
        throw CTSocketException( "ArrayIndexOutOfBoundsException", __FILE__, __LINE__);

    iOffset = 0;
    do
    {
        iCharSend = ::send(iSock_, pcDataBuf + iOffset, iDataLen, bUrgent ? MSG_OOB : 0);
        if (iCharSend == -1)
        {
            long lErrorNum = getErrorNum();
#ifdef WIN32
            if (!isBlocking() && lErrorNum == WSAEWOULDBLOCK)
#else
            if (!isBlocking() && lErrorNum == EAGAIN)
#endif
                continue;
            string errorMessage = szMessage;
            errorMessage += "\n";
            errorMessage += getErrorMsg();
            throw CTSocketException( errorMessage, __FILE__, __LINE__, lErrorNum);
        }
        else if (iCharSend == 0)
            break;
        else
        {
            iOffset += iCharSend;
            iDataLen -= iCharSend;
        }
    }
    while (iDataLen > 0);

    return iOffset;
}

CTSocket& CTSocket::operator<<(const char* szStr)
{
    sendBuffer((void*)szStr, strlen(szStr)+1, "unable to send a char*");
    return *this;
}

CTSocket& CTSocket::operator<<(char c)
{
    sendBuffer(&c, sizeof(c), "unable to send a char");
    return *this;
}

CTSocket& CTSocket::operator<<(unsigned char uc)
{
    sendBuffer(&uc, sizeof(uc), "unable to send an unsigned char");
    return *this;
}

CTSocket& CTSocket::operator<<(double d)
{
    d = swap(d);
    sendBuffer(&d, sizeof(d), "unable to send a double");
    return *this;
}

CTSocket& CTSocket::operator<<(float f)
{
    f = swap(f);
    sendBuffer(&f, sizeof(f), "unable to send a float");
    return *this;
}

CTSocket& CTSocket::operator<<(int)
{
    throw CTSocketException("unallowed to send an int", __FILE__, __LINE__);
    return *this;
}

CTSocket& CTSocket::operator<<(unsigned int)
{
    throw CTSocketException("unallowed to send an unsigned int", __FILE__, __LINE__);
    return *this;
}

CTSocket& CTSocket::operator<<(long l)
{
    l = (long)htonl(l);
    sendBuffer(&l, sizeof(l), "unable to send a long");
    return *this;
}

CTSocket& CTSocket::operator<<(unsigned long ul)
{
    ul = (unsigned long)htonl(ul);
    sendBuffer(&ul, sizeof(ul), "unable to send an unsigned long");
    return *this;
}

CTSocket& CTSocket::operator<<(bool b)
{
    unsigned long ul = 0;
    if (b)
        ul = 1;
    ul = (unsigned long)htonl(ul);
    sendBuffer(&ul, sizeof(ul), "unable to send a bool");
    return *this;
}

CTSocket& CTSocket::operator<<(short s)
{
    s = (short)htons(s);
    sendBuffer(&s, sizeof(s), "unable to send a short");
    return *this;
}

CTSocket& CTSocket::operator<<(unsigned short us)
{
    us = (short)htons(us);
    sendBuffer(&us, sizeof(us), "unable to send an unsigned short");
    return *this;
}

long CTSocket::put(char c)
{
    return sendBuffer(&c, sizeof(c), "unable to send a char");
}

long CTSocket::put(const char *p, size_t N)
{
    return sendBuffer((void *)p, sizeof(*p)*N, "unable to send an array of char");
}

long CTSocket::put(unsigned char uc)
{
    return sendBuffer(&uc, sizeof(uc), "unable to send an unsigned char");
}

long CTSocket::put(const unsigned char *p, size_t N)
{
    return sendBuffer((void *)p, sizeof(*p)*N, "unable to send a fixed char array");
}

long CTSocket::put(const short *p, size_t N)
{
    long lCharSent = 0;

    if (N > 0)
    {
        short * psArray = NULL ;
        try
        {
            psArray = new short[N];
        }
        catch (CTxalloc&)
        {
            throw CTSocketException ("put:new short[N] failed.", __FILE__, __LINE__) ;
        }
        for (unsigned int i = 0; i < N ; i++)
        {
            psArray[i] = (short)htons(p[i]);
        }
        lCharSent = sendBuffer((void *)psArray, sizeof(*psArray)*N, "unable to send a fixed short array");
        delete []psArray;
    }
    return lCharSent;
}

long CTSocket::put(const unsigned short *p, size_t N)
{
    long lCharSent = 0;

    if (N > 0)
    {
        unsigned short * pusArray = NULL ;
        try
        {
            pusArray = new unsigned short[N];
        }
        catch (CTxalloc&)
        {
            throw CTSocketException ("put:new unsigned short[N] failed.", __FILE__, __LINE__) ;
        }
        for (unsigned int i = 0; i < N ; i++)
        {
            pusArray[i] = (unsigned short)htons(p[i]);
        }
        lCharSent = sendBuffer((void *)pusArray, sizeof(*pusArray)*N, "unable to send a fixed unsigned short array");
        delete []pusArray;
    }
    return lCharSent;
}

long CTSocket::put(const int *, size_t)
{
    throw CTSocketException("unallowed to send a fixed int array", __FILE__, __LINE__);
    return -1;
}

long CTSocket::put(const unsigned int *, size_t)
{
    throw CTSocketException("unallowed to send a fixed unsigned int array", __FILE__, __LINE__);
    return -1;
}

long CTSocket::put(const long *p, size_t N)
{
    long lCharSent = 0;

    if (N > 0)
    {
        long * plArray = NULL ;
        try
        {
            plArray = new long[N];
        }
        catch (CTxalloc&)
        {
            throw CTSocketException ("put:new long[N] failed.", __FILE__, __LINE__) ;
        }
        for (unsigned int i = 0; i < N ; i++)
        {
            plArray[i] = (long)htonl(p[i]);
        }
        lCharSent = sendBuffer((void *)plArray, sizeof(*plArray)*N, "unable to send a fixed long array");
        delete []plArray;
    }
    return lCharSent;
}

long CTSocket::put(const unsigned long *p, size_t N)
{
    long lCharSent = 0;

    if (N > 0)
    {
        unsigned long * pulArray = NULL ;
        try
        {
            pulArray = new unsigned long[N];
        }
        catch (CTxalloc&)
        {
            throw CTSocketException ("put:new unsigned long[N] failed.", __FILE__, __LINE__) ;
        }
        for (unsigned int i = 0; i < N ; i++)
        {
            pulArray[i] = (unsigned long)htonl(p[i]);
        }
        lCharSent = sendBuffer((void *)pulArray, sizeof(*pulArray)*N, "unable to send a fixed unsigned long array");
        delete []pulArray;
    }
    return lCharSent;
}

long CTSocket::put(const float* p, size_t N)
{
    long lCharSent = 0;

    if (N > 0)
    {
        float * pfArray = NULL ;
        try
        {
            pfArray = new float[N];
        }
        catch (CTxalloc&)
        {
            throw CTSocketException ("put:new float[N] failed.", __FILE__, __LINE__) ;
        }
        for (unsigned int i = 0; i < N ; i++)
            pfArray[i] = swap(p[i]);
        lCharSent = sendBuffer((void *)pfArray, sizeof(*pfArray)*N, "unable to send a fixed float array");
        delete []pfArray;
    }

    return lCharSent;
}

long CTSocket::put(const double* p, size_t N)
{
    long lCharSent = 0;

    if (N > 0)
    {
        double *pdArray = NULL ;
        try
        {
            pdArray = new double[N];
        }
        catch (CTxalloc&)
        {
            throw CTSocketException ("put:new double[N] failed.", __FILE__, __LINE__) ;
        }
        for (unsigned int i = 0; i < N ; i++)
            pdArray[i] = swap(p[i]);
        lCharSent = sendBuffer((void *)pdArray, sizeof(*pdArray)*N, "unable to send a fixed double array");
        delete []pdArray;
    }

    return lCharSent;
}

//---------------------------------------------------------------------------
// CTSocket::getErrorMsg
//---------------------------------------------------------------------------
// Gets the error message from the internal error number.
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  the error message.
//---------------------------------------------------------------------------
string CTSocket::getErrorMsg()
{
    static string msg;
#ifdef WIN32
    int num = WSAGetLastError();

    switch (num)
    {
    case WSAEACCES:
        msg = "\
              WSAEACCES,\
              Permission denied .,\
              An attempt was made to access a socket in a way forbidden by its access permissions. An example is using a broadcast address for sendto without broadcast permission being set using setsockopt(SO_BROADCAST).\
              ";
        break;
    case WSAEADDRINUSE:
        msg = "\
              WSAEADDRINUSE,\
              Address already in use.,\
              Only one usage of each socket address (protocol/IP address/port) is normally permitted. This error occurs if an application attempts to bind a socket to an IP address/port that has already been used for an existing socket, or a socket that wasn't closed properly, or one that is still in the process of closing. For server applications that need to bind multiple sockets to the same port number, consider using setsockopt(SO_REUSEADDR). Client applications usually need not call bind at all - connectwill choose an unused port automatically.\
              ";
        break;
    case WSAEADDRNOTAVAIL:
        msg = "\
              WSAEADDRNOTAVAIL,\
              Cannot assign requested address.,\
              The requested address is not valid in its context. Normally results from an attempt to bind to an address that is not valid for the local machine, or connect/sendtoan address or port that is not valid for a remote machine (e.g. port 0).\
              ";
        break;
    case WSAEAFNOSUPPORT:
        msg = "\
              WSAEAFNOSUPPORT,\
              Address family not supported by protocol family.,\
              An address incompatible with the requested protocol was used. All sockets are created with an associated \"address family\" (i.e. AF_INET for Internet Protocols) and a generic protocol type (i.e. SOCK_STREAM). This error will be returned if an incorrect protocol is explicitly requested in the socket call, or if an address of the wrong family is used for a socket, e.g. in sendto.\
              ";
        break;
    case WSAEALREADY:
        msg = "\
              WSAEALREADY,\
              Operation already in progress.,\
              An operation was attempted on a non-blocking socket that already had an operation in progress - i.e. calling connecta second time on a non-blocking socket that is already connecting, or canceling an asynchronous request (WSAAsyncGetXbyY) that has already been canceled or completed.\
              ";
        break;
    case WSAECONNABORTED:
        msg = "\
              WSAECONNABORTED,\
              Software caused connection abort.,\
              An established connection was aborted by the software in your host machine, possibly due to a data transmission timeout or protocol error.\
              ";
        break;
    case WSAECONNREFUSED:
        msg = "\
              WSAECONNREFUSED,\
              Connection refused.,\
              No connection could be made because the target machine actively refused it. This usually results from trying to connect to a service that is inactive on the foreign host - i.e. one with no server application running.\
              ";
        break;
    case WSAECONNRESET:
        msg = "\
              WSAECONNRESET,\
              Connection reset by peer.,\
              A existing connection was forcibly closed by the remote host. This normally results if the peer application on the remote host is suddenly stopped, the host is rebooted, or the remote host used a \"hard close\" (see setsockopt for more information on the SO_LINGERoption on the remote socket.)\
              ";
        break;
    case WSAEDESTADDRREQ:
        msg = "\
              WSAEDESTADDRREQ,\
              Destination address required.,\
              A required address was omitted from an operation on a socket. For example, this error will be returned if sendtois called with the remote address of ADDR_ANY.\
              ";
        break;
    case WSAEFAULT:
        msg = "\
              WSAEFAULT,\
              Bad address.,\
              The system detected an invalid pointer address in attempting to use a pointer argument of a call. This error occurs if an application passes an invalid pointer value, or if the length of the buffer is too small. For instance, if the length of an argument which is a struct sockaddr is smaller than sizeof(struct sockaddr).\
              ";
        break;
    case WSAEHOSTDOWN:
        msg = "\
              WSAEHOSTDOWN,\
              Host is down.,\
              A socket operation failed because the destination host was down. A socket operation encountered a dead host. Networking activity on the local host has not been initiated. These conditions are more likely to be indicated by the error WSAETIMEDOUT.\
              ";
        break;
    case WSAEHOSTUNREACH:
        msg = "\
              WSAEHOSTUNREACH,\
              No route to host.,\
              A socket operation was attempted to an unreachable host. See WSAENETUNREACH \
              ";
        break;
    case WSAEINPROGRESS:
        msg = "\
              WSAEINPROGRESS,\
              Operation now in progress.,\
              A blocking operation is currently executing. Windows Sockets only allows a single blocking operation to be outstanding per task (or thread), and if any other function call is made (whether or not it references that or any other socket) the function fails with the WSAEINPROGRESS error.\
              ";
        break;
    case WSAEINTR:
        msg = "\
              WSAEINTR,\
              Interrupted function call.,\
              A blocking operation was interrupted by a call to WSACancelBlockingCall.\
              ";
        break;
    case WSAEINVAL:
        msg = "\
              WSAEINVAL,\
              Invalid argument.,\
              Some invalid argument was supplied (for example, specifying an invalid level to the setsockopt function). In some instances, it also refers to the current state of the socket - for instance, calling accept on a socket that is not listening.\
              ";
        break;
    case WSAEISCONN:
        msg = "\
              WSAEISCONN,\
              Socket is already connected.,\
              A connect request was made on an already connected socket. Some implementations also return this error if sendto is called on a connected SOCK_DGRAM socket (For SOCK_STREAM sockets, the to parameter in sendtois ignored), although other implementations treat this as a legal occurrence.\
              ";
        break;
    case WSAEMFILE:
        msg = "\
              WSAEMFILE,\
              Too many open files.,\
              Too many open sockets. Each implementation may have a maximum number of socket handles available, either globally, per process or per thread.\
              ";
        break;
    case WSAEMSGSIZE:
        msg = "\
              WSAEMSGSIZE,\
              Message too long.,\
              A message sent on a datagram socket was larger than the internal message buffer or some other network limit, or the buffer used to receive a datagram into was smaller than the datagram itself.\
              ";
        break;
    case WSAENETDOWN:
        msg = "\
              WSAENETDOWN,\
              Network is down.,\
              A socket operation encountered a dead network. This could indicate a serious failure of the network system (i.e. the protocol stack that the WinSock DLL runs over), the network interface, or the local network itself.\
              ";
        break;
    case WSAENETRESET:
        msg = "\
              WSAENETRESET,\
              Network dropped connection on reset.,\
              The host you were connected to crashed and rebooted. May also be returned by setsockoptif an attempt is made to set SO_KEEPALIVE on a connection that has already failed.\
              ";
        break;
    case WSAENETUNREACH:
        msg = "\
              WSAENETUNREACH,\
              Network is unreachable.,\
              A socket operation was attempted to an unreachable network. This usually means the local software knows no route to reach the remote host.\
              ";
        break;
    case WSAENOBUFS:
        msg = "\
              WSAENOBUFS,\
              No buffer space available.,\
              An operation on a socket could not be performed because the system lacked sufficient buffer space or because a queue was full.\
              ";
        break;
    case WSAENOPROTOOPT:
        msg = "\
              WSAENOPROTOOPT,\
              Bad protocol option.,\
              An unknown, invalid or unsupported option or level was specified in a getsockopt or setsockoptcall.\
              ";
        break;
    case WSAENOTCONN:
        msg = "\
              WSAENOTCONN,\
              Socket is not connected.,\
              A request to send or receive data was disallowed because the socket is not connected and (when sending on a datagram socket using sendto) no address was supplied. Any other type of operation might also return this error - for example, setsockoptsetting SO_KEEPALIVE if the connection has been reset.\
              ";
        break;
    case WSAENOTSOCK:
        msg = "\
              WSAENOTSOCK,\
              Socket operation on non-socket.,\
              An operation was attempted on something that is not a socket. Either the socket handle parameter did not reference a valid socket, or for select, a member of an fd_set was not valid.\
              ";
        break;
    case WSAEOPNOTSUPP:
        msg = "\
              WSAEOPNOTSUPP,\
              Operation not supported.,\
              The attempted operation is not supported for the type of object referenced. Usually this occurs when a socket descriptor to a socket that cannot support this operation, for example, trying to accept a connection on a datagram socket.\
              ";
        break;
    case WSAEPFNOSUPPORT:
        msg = "\
              WSAEPFNOSUPPORT,\
              Protocol family not supported.,\
              The protocol family has not been configured into the system or no implementation for it exists. Has a slightly different meaning to WSAEAFNOSUPPORT, but is interchangeable in most cases, and all Windows Sockets functions that return one of these specify WSAEAFNOSUPPORT.\
              ";
        break;
    case WSAEPROCLIM:
        msg = "\
              WSAEPROCLIM,\
              Too many processes.,\
              A Windows Sockets implementation may have a limit on the number of applications that may use it simultaneously. WSAStartupmay fail with this error if the limit has been reached.\
              ";
        break;
    case WSAEPROTONOSUPPORT:
        msg = "\
              WSAEPROTONOSUPPORT,\
              Protocol not supported.,\
              The requested protocol has not been configured into the system, or no implementation for it exists. For example, a socketcall requests a SOCK_DGRAM socket, but specifies a stream protocol.\
              ";
        break;
    case WSAEPROTOTYPE:
        msg = "\
              WSAEPROTOTYPE,\
              Protocol wrong type for socket. A protocol was specified in the socketfunction call that does not support the semantics of the socket type requested. For example, the ARPA Internet UDP protocol cannot be specified with a socket type of SOCK_STREAM.\
              ";
        break;
    case WSAESHUTDOWN:
        msg = "\
              WSAESHUTDOWN,\
              Cannot send after socket shutdown.,\
              A request to send or receive data was disallowed because the socket had already been shut down in that direction with a previous shutdown call. By calling shutdowna partial close of a socket is requested, which is a signal that sending or receiving or both has been discontinued.\
              ";
        break;
    case WSAESOCKTNOSUPPORT:
        msg = "\
              WSAESOCKTNOSUPPORT,\
              Socket type not supported.,\
              The support for the specified socket type does not exist in this address family. For example, the optional type SOCK_RAW might be selected in a socketcall, and the implementation does not support SOCK_RAW sockets at all.\
              ";
        break;
    case WSAETIMEDOUT:
        msg = "\
              WSAETIMEDOUT,\
              Connection timed out.,\
              A connection attempt failed because the connected party did not properly respond after a period of time, or established connection failed because connected host has failed to respond.\
              ";
        break;
    case WSAEWOULDBLOCK:
        msg = "\
              WSAEWOULDBLOCK,\
              Resource temporarily unavailable.,\
              This error is returned from operations on non-blocking sockets that cannot be completed immediately, for example recv when no data is queued to be read from the socket. It is a non-fatal error, and the operation should be retried later. It is normal for WSAEWOULDBLOCK to be reported as the result from calling connecton a non-blocking SOCK_STREAM socket, since some time must elapse for the connection to be established.\
              ";
        break;
    case WSAHOST_NOT_FOUND:
        msg = "\
              WSAHOST_NOT_FOUND,\
              Host not found.,\
              No such host is known. The name is not an official hostname or alias, or it cannot be found in the database(s) being queried. This error may also be returned for protocol and service queries, and means the specified name could not be found in the relevant database.\
              ";
        break;
#if 0
    case WSA_INVALID_HANDLE:
        msg = "\
              WSA_INVALID_HANDLE,\
              (OS dependent),\
              Specified event object handle is invalid.,\
              An application attempts to use an event object, but the specified handle is not valid.\
              ";
        break;
    case WSA_INVALID_PARAMETER:
        msg = "\
              WSA_INVALID_PARAMETER,\
              (OS dependent),\
              One or more parameters are invalid.,\
              An application used a Windows Sockets function which directly maps to a Win32 function. The Win32 function is indicating a problem with one or more parameters.\
              ";
        break;
    case WSAINVALIDPROCTABLE:
        msg = "\
              WSAINVALIDPROCTABLE,\
              (OS dependent),\
              Invalid procedure table from service provider.,\
              A service provider returned a bogus proc table to WS2_32.DLL. (Usually caused by one or more of the function pointers being NULL.) \
              ";
        break;
    case WSAINVALIDPROVIDER:
        msg = "\
              WSAINVALIDPROVIDER,\
              (OS dependent),\
              Invalid service provider version number.,\
              A service provider returned a version number other than 2.0.\
              ";
        break;
    case WSA_IO_PENDING:
        msg = "\
              WSA_IO_PENDING,\
              (OS dependent),\
              Overlapped operations will complete later.,\
              The application has initiated an overlapped operation which cannot be completed immediately. A completion indication will be given at a later time when the operation has been completed.\
              ";
        break;
    case WSA_IO_INCOMPLETE:
        msg = "\
              WSA_IO_INCOMPLETE,\
              (OS dependent),\
              Overlapped I/O event object not in signaled state.,\
              The application has tried to determine the status of an overlapped operation which is not yet completed. Applications that use WSAWaitForMultipleEventsin a polling mode to determine when an overlapped operation has completed will get this error code until the operation is complete.\
              ";
        break;
    case WSA_NOT_ENOUGH_MEMORY:
        msg = "\
              WSA_NOT_ENOUGH_MEMORY,\
              (OS dependent),\
              Insufficient memory available.,\
              An application used a Windows Sockets function which directly maps to a Win32 function. The Win32 function is indicating a lack of required memory resources.\
              ";
        break;
#endif
    case WSANOTINITIALISED:
        msg = "\
              WSANOTINITIALISED,\
              Successful WSAStartup not yet performed.,\
              Either the application hasn't called WSAStartup, or WSAStartup failed. The application may be accessing a socket which the current active task does not own (i.e. trying to share a socket between tasks), or WSACleanuphas been called too many times.\
              ";
        break;
    case WSANO_DATA:
        msg = "\
              WSANO_DATA,\
              Valid name, no data record of requested type.,\
              The requested name is valid and was found in the database, but it does not have the correct associated data being resolved for. The usual example for this is a hostname -> address translation attempt (using gethostbyname or WSAAsyncGetHostByName) which uses the DNS (Domain Name Server), and an MX record is returned but no A record - indicating the host itself exists, but is not directly reachable.\
              ";
        break;
    case WSANO_RECOVERY:
        msg = "\
              WSANO_RECOVERY,\
              This is a non-recoverable error.,\
              This indicates some sort of non-recoverable error occurred during a database lookup. This may be because the database files (e.g. BSD-compatible HOSTS, SERVICES or PROTOCOLS files) could not be found, or a DNS request was returned by the server with a severe error.\
              ";
        break;
#if 0
    case WSAPROVIDERFAILEDINIT:
        msg = "\
              WSAPROVIDERFAILEDINIT,\
              (OS dependent),\
              Unable to initialize a service provider.,\
              Either a service provider's DLL could not be loaded (LoadLibrary failed) or the provider's WSPStartup/NSPStartupfunction failed.\
              ";
        break;
    case WSASYSCALLFAILURE:
        msg = "\
              WSASYSCALLFAILURE,\
              (OS dependent),\
              System call failure.,\
              Returned when a system call that should never fail does. For example, if a call to WaitForMultipleObjectsfails or one of the registry functions fails trying to manipulate theprotocol/namespace catalogs.\
              ";
        break;
#endif
    case WSASYSNOTREADY:
        msg = "\
              WSASYSNOTREADY,\
              Network subsystem is unavailable.,\
              This error is returned by WSAStartup if the Windows Sockets implementation cannot function at this time because the underlying system it uses to provide network services is currently unavailable. Users should check:,\n\
                 that the WINSOCK.DLL file is in the current path,\n\
                 that the WINSOCK.DLL file is from the same vendor as the underlying protocol stack. They cannot be mixed and matched (WinSock DLLs must be supplied by the same vendor that provided the underlying protocol stack).\n\
                 that they are not trying to use more than one Windows Sockets implementation simultaneously. If there is more than one WINSOCK DLL on your system, be sure the first one in the path is appropriate for the network subsystem currently loaded.\n\
                 the Windows Sockets implementation documentation to be sure all necessary components are currently installed and configured correctly.\
              ";
        break;
    case WSATRY_AGAIN:
        msg = "\
              WSATRY_AGAIN,\
              Non-authoritative host not found.,\
              This is usually a temporary error during hostname resolution and means that the local server did not receive a response from an authoritative server. A retry at some time later may be successful.\
              ";
        break;
    case WSAVERNOTSUPPORTED:
        msg = "\
              WSAVERNOTSUPPORTED,\
              WINSOCK.DLL version out of range.,\
              The current Wndows Sockets implementation does not support the Windows Sockets specification version requested by the application. Check that no old WINSOCK.DLL files are being accessed, or contact the stack vendor to see if an updated WINSOCK.DLL exists.\
              ";
        break;
    case WSAEDISCON:
        msg = "\
              WSAEDISCON,\
              Graceful shutdown in progress.,\
              Returned by recv, WSARecvto indicate the remote party has initiated a graceful shutdown sequence.\
              ";
        break;
#if 0
    case WSA_OPERATION_ABORTED:
        msg = "\
              WSA_OPERATION_ABORTED,\
              (OS dependent)\
              Overlapped operation aborted.\
              An overlapped operation was canceled due to the closure of the socket, or the execution of the SIO_FLUSH command in WSAIoctl.\
              ";
        break;
#endif
    default:
        {
            char buffer[32];
            msg = itoa(num, buffer, 10);
            msg += " unknown error code";
        }
        break;
    }

#else
    msg = strerror(errno);
#endif

    return msg;
}

//---------------------------------------------------------------------------
// CTSocket::getErrorNumber
//---------------------------------------------------------------------------
// Gets the error number from the internal error number.
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  the error number.
//---------------------------------------------------------------------------
long CTSocket::getErrorNum()
{
#ifdef WIN32
    return WSAGetLastError();
#else
    return errno;
#endif
}

//---------------------------------------------------------------------------
// CTSocketSelectorItem::CTSocketSelectorItem
//---------------------------------------------------------------------------
// Constructor.
//---------------------------------------------------------------------------
// Arguments :  socket  : a reference to a valid CTSocket,
//              iEvent  : one or many events to select to.
// Return    :  None.
//---------------------------------------------------------------------------
CTSocketSelectorItem::CTSocketSelectorItem(CTSocket& socket, int iEvent)
        : socket_(socket), iEvent_(iEvent)
{}

//---------------------------------------------------------------------------
// CTSocketSelectorItem::operator =
//---------------------------------------------------------------------------
// Operator Copy.
//---------------------------------------------------------------------------
// Arguments :  item    : a reference to a CTSocketSelectorItem,
// Return    :  *this.
//---------------------------------------------------------------------------
CTSocketSelectorItem& CTSocketSelectorItem::operator = (const CTSocketSelectorItem& item)
{
    socket_ = item.socket_ ;
    iEvent_ = item.iEvent_ ;
    return *this;
}

//---------------------------------------------------------------------------
// CTSocketSelectorItem::operator ==
//---------------------------------------------------------------------------
// Equality operator.
//---------------------------------------------------------------------------
// Arguments :  item    : a reference to a CTSocketSelectorItem,
// Return    :  true or false.
//---------------------------------------------------------------------------
bool CTSocketSelectorItem::operator == (const CTSocketSelectorItem& item)
{
    return socket_.iSock_ == ((CTSocketSelectorItem&)item).getSocket().iSock_;
}

//---------------------------------------------------------------------------
// CTSocketSelectorItem::getSocket
//---------------------------------------------------------------------------
// Returns a reference to the socket we want to select.
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  a reference to the socket.
//---------------------------------------------------------------------------
CTSocket& CTSocketSelectorItem::getSocket()
{
    return socket_;
}

//---------------------------------------------------------------------------
// CTSocketSelectorItem::getEvent
//---------------------------------------------------------------------------
// Returns the list of events we associated with the socket.
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  list of events we want ot select.
//---------------------------------------------------------------------------
int CTSocketSelectorItem::getEvent()
{
    return iEvent_;
}

//---------------------------------------------------------------------------
// CTSocketSelectorItem::testFunction
//---------------------------------------------------------------------------
// Returns if the socket is the searched socket.
//---------------------------------------------------------------------------
// Arguments :  selectorItem : selector item.
// Return    :  return true or false.
//---------------------------------------------------------------------------
bool CTSocketSelectorItem::testFunction (CTSocketSelectorItem* selectorItem, CTSocket* pSocket)
{
    return (& (selectorItem->getSocket() ) == pSocket) ;
}

//---------------------------------------------------------------------------
// CTSocketSelectorList::CTSocketSelectorList
//---------------------------------------------------------------------------
// Default constructor.
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  None.
//---------------------------------------------------------------------------
CTSocketSelectorList::CTSocketSelectorList()
{
    preadfds_ = pwritefds_ = pexceptfds_ = NULL;
    pTimeOut_ = NULL ;
}

//---------------------------------------------------------------------------
// CTSocketSelectorList::~CTSocketSelectorList
//---------------------------------------------------------------------------
// Destructor, destroys the inner list associated with this object.
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  None..
//---------------------------------------------------------------------------
CTSocketSelectorList::~CTSocketSelectorList()
{
    CTSelectorItemList::iterator iter;
    for (iter = list_.begin(); iter != list_.end(); iter++)
    {
        delete *iter;
    }
    list_.clear();
    if (pTimeOut_)
    {
        delete pTimeOut_ ;
        pTimeOut_ = NULL ;
    }
}

//---------------------------------------------------------------------------
// CTSocketSelectorList::operator()
//---------------------------------------------------------------------------
// Function operator
//---------------------------------------------------------------------------
// Arguments :  iTypeOfEvent : on exit specify what kind of event had been received
//                              _S_RECEIVE_, _S_SEND_, _S_EXCEPTION, _S_TIMEDOUT_
// Return    :  a reference to the socket.
//---------------------------------------------------------------------------
CTSocket& CTSocketSelectorList::operator()(int& iTypeOfEvent)
{
    static CTSocket nullSocket("I'm very null!!");

    int iNfds = 32;
    preadfds_ = pwritefds_ = pexceptfds_ = NULL;

    CTSelectorItemList::iterator iter;

    for (iter = list_.begin(); iter != list_.end(); iter++)
    {
        CTSocketSelectorItem& item = **iter;
        CTSocket &socket = item.getSocket();
        int iEvent = item.getEvent();

#ifdef  WIN32
#pragma warning(disable : 4127)
#endif

        if (iEvent & _S_RECEIVE_)
        {
            if (preadfds_ == NULL)
            {
                preadfds_ = &readfds_;
                FD_ZERO(preadfds_);
            }
            FD_SET(socket.iSock_, preadfds_);
        }

        if (iEvent & _S_SEND_)
        {
            if (pwritefds_ == NULL)
            {
                pwritefds_ = &writefds_;
                FD_ZERO(pwritefds_);
            }
            FD_SET(socket.iSock_, pwritefds_);
        }

        if (iEvent & _S_EXCEPTION_)
        {
            if (pwritefds_ == NULL)
            {
                pexceptfds_ = &exceptfds_;
                FD_ZERO(pexceptfds_);
            }
            FD_SET(socket.iSock_, pexceptfds_);
        }

#ifdef  WIN32
#pragma warning(default : 4127)
#endif
    }


    iTypeOfEvent = 0;
    int iSelect = ::select(iNfds, preadfds_, pwritefds_, pexceptfds_, pTimeOut_);

    if (iSelect == -1)
        /* Error */
        throw CTSocketException(CTSocket::getErrorMsg(), __FILE__, __LINE__, CTSocket::getErrorNum());

    if (iSelect == 0)
        /* Time limit expired. */
        iTypeOfEvent = _S_TIMEDOUT_;
    else
    {
        CTSelectorItemList::iterator iter;

        for (iter = list_.begin(); iter != list_.end(); iter++)
        {
            CTSocketSelectorItem& item = **iter;
            CTSocket &socket = item.getSocket();
            int iEvent = item.getEvent();

            if (iEvent & _S_RECEIVE_)
            {
                if (FD_ISSET(socket.iSock_,preadfds_))
                {
                    iTypeOfEvent = _S_RECEIVE_;
                    return socket;
                }
            }
            else if (iEvent & _S_SEND_)
            {
                if (FD_ISSET(socket.iSock_,pwritefds_))
                {
                    iTypeOfEvent = _S_SEND_;
                    return socket;
                }
            }
            else if (iEvent & _S_EXCEPTION_)
            {
                if (FD_ISSET(socket.iSock_,pexceptfds_))
                {
                    iTypeOfEvent = _S_EXCEPTION_;
                    return socket;
                }
            }
        }

        throw CTSocketException("None event on all selected ports", __FILE__, __LINE__);
    }

    return (CTSocket&)nullSocket;
}

//---------------------------------------------------------------------------
// CTSocketSelectorList::operator()
//---------------------------------------------------------------------------
// Function operator
//---------------------------------------------------------------------------
// Arguments :  resultList   : on exit receive the sockets witch are selected.
//              iTypeOfEvent : on exit specify what kind of event had been received
//                              _S_RECEIVE_, _S_SEND_, _S_EXCEPTION, _S_TIMEDOUT_
// Return    :  indicate if any sockets is seleted or not.
//---------------------------------------------------------------------------
bool CTSocketSelectorList::operator()(CTSocketSelectorList& resultList, int& iTypeOfEvent)
{
    bool    bSomeSocketsSelected ;
    int     iNfds = 32 ;
    preadfds_ = pwritefds_ = pexceptfds_ = NULL;

    CTSelectorItemList::iterator iter;

    for (iter = list_.begin(); iter != list_.end(); iter++)
    {
        CTSocketSelectorItem& item = **iter;
        CTSocket &socket = item.getSocket();
        int iEvent = item.getEvent();

#ifdef  WIN32
#pragma warning(disable : 4127)
#endif

        if (iEvent & _S_RECEIVE_)
        {
            if (preadfds_ == NULL)
            {
                preadfds_ = &readfds_;
                FD_ZERO(preadfds_);
            }
            FD_SET(socket.iSock_, preadfds_);
        }

        if (iEvent & _S_SEND_)
        {
            if (pwritefds_ == NULL)
            {
                pwritefds_ = &writefds_;
                FD_ZERO(pwritefds_);
            }
            FD_SET(socket.iSock_, pwritefds_);
        }

        if (iEvent & _S_EXCEPTION_)
        {
            if (pexceptfds_ == NULL)
            {
                pexceptfds_ = &exceptfds_;
                FD_ZERO(pexceptfds_);
            }
            FD_SET(socket.iSock_, pexceptfds_);
        }

#ifdef  WIN32
#pragma warning(default : 4127)
#endif
    }

    iTypeOfEvent = 0;
    int iSelect = ::select(iNfds, preadfds_, pwritefds_, pexceptfds_, pTimeOut_);

    if (iSelect == -1)
        /* Error */
        throw CTSocketException(CTSocket::getErrorMsg(), __FILE__, __LINE__, CTSocket::getErrorNum());

    resultList.reset () ;

    if (iSelect == 0)
    {
        bSomeSocketsSelected = false ;
        /* Time limit expired. */
        iTypeOfEvent = _S_TIMEDOUT_;
    }
    else
    {
        bSomeSocketsSelected = true ;
        CTSelectorItemList::iterator iter;

        for (iter = list_.begin(); iter != list_.end(); iter++)
        {
            CTSocketSelectorItem& item = **iter;
            CTSocket &socket = item.getSocket();
            int iMaskEvent = item.getEvent();
            int iSocketEvent = 0 ;

            if ( (iMaskEvent & _S_RECEIVE_) && (FD_ISSET(socket.iSock_,preadfds_) ) )
            {
                iSocketEvent |= _S_RECEIVE_;
            }
            if ( (iMaskEvent & _S_SEND_) && (FD_ISSET(socket.iSock_,pwritefds_) ) )
            {
                iSocketEvent |= _S_SEND_;
            }
            if ( (iMaskEvent & _S_EXCEPTION_) && (FD_ISSET(socket.iSock_,pexceptfds_) ) )
            {
                iSocketEvent |= _S_EXCEPTION_;
            }

            if (iSocketEvent)
            {
                iTypeOfEvent |= iSocketEvent ;
                resultList.add (socket, iSocketEvent) ;
            }
        }
    }

    return bSomeSocketsSelected ;
}

//---------------------------------------------------------------------------
// CTSocketSelectorList::setTimeOut()
//---------------------------------------------------------------------------
// Sets the time out we want to apply to the select function.
//---------------------------------------------------------------------------
// Arguments :  timeOut : a timeout to the select function
// Return    :  None.
//---------------------------------------------------------------------------
void CTSocketSelectorList::setTimeOut(const timeval *pTimeOut)
{
    if (pTimeOut)
    {
        if (pTimeOut_ == NULL)
        {
            try
            {
                pTimeOut_ = new timeval ;
            }
            catch (CTxalloc&)
            {
                throw CTSocketException ("setTimeOut:new timeval failed.", __FILE__, __LINE__) ;
            }
        }
        pTimeOut_->tv_sec = pTimeOut->tv_sec;
        pTimeOut_->tv_usec = pTimeOut->tv_usec;
    }
    else
    {
        if (pTimeOut_)
        {
            pTimeOut_ = NULL ;
        }
    }
}

//---------------------------------------------------------------------------
// CTSocketSelectorList::setPollingMode
//---------------------------------------------------------------------------
// Polls on the selected sockets.
//---------------------------------------------------------------------------
// Arguments :  timeOut : a timeout to the select function
// Return    :  None.
//---------------------------------------------------------------------------
void CTSocketSelectorList::setPollingMode()
{
    if (pTimeOut_ == NULL)
    {
        try
        {
            pTimeOut_ = new timeval ;
        }
        catch (CTxalloc&)
        {
            throw CTSocketException ("setPollingMode:new timeval failed.", __FILE__, __LINE__) ;
        }
    }
    pTimeOut_->tv_sec = 0;
    pTimeOut_->tv_usec = 0;
}

//---------------------------------------------------------------------------
// CTSocketSelectorList::reset()
//---------------------------------------------------------------------------
// Function operator
//---------------------------------------------------------------------------
// Arguments :  None.
// Return    :  None.
//---------------------------------------------------------------------------
void CTSocketSelectorList::reset()
{
    CTSelectorItemList::iterator iter;
    for (iter = list_.begin(); iter != list_.end(); iter++)
    {
        delete *iter;
    }
    list_.clear();
    if (pTimeOut_)
    {
        delete pTimeOut_ ;
        pTimeOut_ = NULL ;
    }
    if (preadfds_ != NULL)
        FD_ZERO(preadfds_);
    if (pwritefds_ != NULL)
        FD_ZERO(pwritefds_);
    if (pexceptfds_ != NULL)
        FD_ZERO(pexceptfds_);
    preadfds_ = pwritefds_ = pexceptfds_ = NULL;
}

//---------------------------------------------------------------------------
// CTSocketSelectorList::add
//---------------------------------------------------------------------------
// Adds a tuple we want to select for.
//---------------------------------------------------------------------------
// Arguments :  socket          : a reference to a socket we want to be notified.
//              iTypeOfEvent    : specify the kind of event to be notified
//                              _S_RECEIVE_, _S_SEND_, _S_EXCEPTION, _S_TIMEDOUT_
// Return    :  None.
//---------------------------------------------------------------------------
void CTSocketSelectorList::add(CTSocket& socket, int iTypeOfEvent)
{
    CTSelectorItemList::iterator iter;
    for(iter=list_.begin();iter!=list_.end();iter++)
    {
        if (&((*iter)->getSocket()) == &socket) break;
    }
    if (iter  != list_.end())
    {
        int iActualTypeOfEvent = (*iter)->getEvent () ;
        iActualTypeOfEvent |= iTypeOfEvent ;
        (*iter)->setEvent (iActualTypeOfEvent) ;
    }
    else
    {
        try
        {
            list_.push_back(new CTSocketSelectorItem(socket, iTypeOfEvent));
        }
        catch (CTxalloc&)
        {
            throw CTSocketException ("add:new CTSocketSelectorItem(socket, iTypeOfEvent) failed.", __FILE__, __LINE__) ;
        }
    }
}

//---------------------------------------------------------------------------
// CTSocketSelectorList::remove
//---------------------------------------------------------------------------
// Removes a tuple we want to select for.
//---------------------------------------------------------------------------
// Arguments :  socket          : a reference to a socket.
//              iTypeOfEvent    : specify the kind of event to be removed
//                              _S_RECEIVE_, _S_SEND_, _S_EXCEPTION, _S_TIMEDOUT_
// Return    :  None.
//---------------------------------------------------------------------------
void CTSocketSelectorList::remove(CTSocket& socket, int iTypeOfEvent)
{
    CTSocketSelectorItem* pItemToFind = NULL ;
    try
    {
        pItemToFind = new CTSocketSelectorItem(socket, iTypeOfEvent) ;
    }
    catch (CTxalloc&)
    {
        throw CTSocketException ("remove:new CTSocketSelectorItem(socket, iTypeOfEvent) failed.", __FILE__, __LINE__) ;
    }

    CTSelectorItemList::iterator iter;
    for(iter=list_.begin();iter!=list_.end();iter++)
    {
        if((*iter)->getSocket().iSock_ == socket.iSock_) break;
    }
    int iEvent ;
    if (iter  != list_.end())
    {
        CTSocketSelectorItem& item=**iter;
        iEvent = item.getEvent();
        list_.erase(iter);
    }


#ifdef  WIN32
#pragma warning(disable : 4127 4018)
#endif  // WIN32
    if (iEvent & _S_RECEIVE_)
    {
        if (preadfds_ != NULL)
        {
            FD_CLR(socket.iSock_, preadfds_);
            iEvent &= ~_S_RECEIVE_;
        }
    }

    if (iEvent & _S_SEND_)
    {
        if (pwritefds_ == NULL)
        {
            FD_CLR(socket.iSock_, pwritefds_);
            iEvent &= ~_S_SEND_;
        }
    }

    if (iEvent & _S_EXCEPTION_)
    {
        if (pwritefds_ == NULL)
        {
            FD_CLR(socket.iSock_, pexceptfds_);
            iEvent &= ~_S_EXCEPTION_;
        }
    }
#ifdef  WIN32
#pragma warning(default : 4127 4018)
#endif  // WIN32

    if (iEvent)
    {
        try
        {
            list_.push_back(new CTSocketSelectorItem(socket, iEvent));
        }
        catch (CTxalloc&)
        {
            throw CTSocketException ("add:new CTSocketSelectorItem(socket, iEvent) failed.", __FILE__, __LINE__) ;
        }
    }
}

//---------------------------------------------------------------------------
// CTSocketSelectorList::contains
//---------------------------------------------------------------------------
// Verify if a socket is in the list.
//---------------------------------------------------------------------------
// Arguments :  socket       : a reference to a socket.
//              iTypeOfEvent : on exit, indicates the events on the socket.
// return    :  indicates if the socket is in the list or not.
//---------------------------------------------------------------------------
bool CTSocketSelectorList::contains (CTSocket& socket, int& iTypeOfEvent)
{
    CTSelectorItemList::iterator iter;
    for(iter=list_.begin();iter!=list_.end();iter++)
    {
        if (&((*iter)->getSocket()) == &socket) break;
    }
    if (iter  != list_.end())
    {
        iTypeOfEvent = (*iter)->getEvent () ;
    }

    return (iter  != list_.end()) ;
}

bool CTSocketSelectorList::isEmpty(void) const
{
    return list_.empty();
}
