/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*            **** C O M M O N ****                  */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_MATCH_WILD                      */
/*          ============                         */
/*                                       */
/*  Description:                                 */
/*  ============                                 */
/*                                       */
/*  This routine will compare a given string against a boilerplate       */
/*  string (possibly containing wildcards) to see if it matches.         */
/*                                       */
/*===========================================================================*/

/*
 *  Standard ANSI header files.
 */
#include <string.h>

/*
 *  PSL Macros.
 */
#include <psl_macros.h>

/*
 *  Function Prototypes.
 */
#include psl_common_file (psl_rtndef.h)    /* PSL Routines */

/*
 *  Structure Definitions.
 */
#include psl_common_file (shrmsg.h)     /* PSL Messages */


/************************************************************************/
/*                                  */
/*  PSL_MATCH_WILD - Compare string against boilerplate             */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_MATCH_WILD (string_ptr, boilerplate_ptr)      */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret      = Unsigned Long PSL Routine Return Status:     */
/*                                  */
/*          PSL_NORMAL:   String matches Boilerplate    */
/*          PSL_NOMATCH:  String was not matched, but is    */
/*                    lower than the 'highest' string   */
/*                    which might be matched.       */
/*          PSL_ENDMATCH: String was not matched, and is    */
/*                    higher than the 'highest' string  */
/*                    which might be matched        */
/*                                  */
/*     string_ptr      = Pointer to string to match (null-term) */
/*     boilerplate_ptr = Pointer to match string against (null-term)*/
/*                                  */
/*  The routine will check for the wildcard characters. Such    */
/*  wildcard characters are platform-specific.          */
/*                                  */
/************************************************************************/

unsigned long   PSL_MATCH_WILD (const char *string_ptr,
                                const char *boilerplate_ptr)
{
    /*
     *  Local data.
     */
    char    string_char;       /* String Character */
    char    boiler_char;       /* Boilerplate Character */
    char    space_string_char; /* String Character (SPACE Chk) */
    long    string_len;        /* Length of String */
    long    boiler_len;        /* Length of Boilerplate String */
    long    string_offset;     /* String Offset */
    long    boiler_offset;     /* Boilerplate String Offset */
    long    saved_boiler_len;  /* Saved Boilerplate String Length */
    long    saved_string_offset;   /* Saved String Offset */
    long    saved_boiler_offset;   /* Saved Boilerplate String Offset */
    long    space_string_len;  /* Saved Length of String (SPACE Chk) */
    long    space_string_offset;   /* Saved String Offset (SPACE Chk) */
    long    wildcard_offset;   /* Wildcarded match offset */
    Boolean wild_seen;     /* Wildcard seen? */

    /*
     *  Get the length of both strings.
     */
    string_len = strlen (string_ptr);
    boiler_len = strlen (boilerplate_ptr);

    /*
     *  Setup the offsets for the strings.
     */
    string_offset = 0;
    boiler_offset = 0;

    /*
     *  Initialize wildcard offset and flag.
     */
    wildcard_offset = 0;
    wild_seen = PFALSE;

    /*
     *  CHKNXT_BOILERPLATE:
     *
     *  Check the next character in the boilerplate string.  At this point we
     *  need to determine if we have reached the last character in the string.
     *  Decrement the length.
     */
chknxt_boilerplate:
    boiler_len -= 1;
    if (boiler_len < 0) goto end_boilerplate;

    /*
     *  We still have some characters left - get the next one and increment the
     *  offset.
     */
    boiler_char = *(boilerplate_ptr + boiler_offset);
    boiler_offset += 1;

    /*
     *  Is this character a multi-wildcard?
     */
    if (boiler_char == '*')
    {

        /*
         *  We hit a multi-character wild-card - set up for recursive search.
         *  If this character is the last in the boilerplate string then it must
         *  match.
         */
        if (boiler_len == 0) goto string_matches;

        /*
         *  Save the string and boilerplate values.
         */
        wildcard_offset = string_len;
        saved_string_offset = string_offset;
        saved_boiler_len = boiler_len;
        saved_boiler_offset = boiler_offset;

        /*
         *  Indicate that we have enountered a wildcard and check the next character
         */
        wild_seen = PTRUE;
        goto chknxt_boilerplate;
    }

    /*
     *  Flag character low in case we haven't any characters left in match string.
     */
    string_char = '\0';

    /*
     *  Decrement the string length and determine if there are any characters left.
     */
    string_len -= 1;
    if (string_len < 0) goto unmatched_char;

    /*
     *  We still have some characters left - get the next one and increment the
     *  offset.
     */
    string_char = *(string_ptr + string_offset);
    string_offset += 1;

    /*
     *  Is this a single-character wildcard in the boilerplate?
     */
#ifdef VMS
    if (boiler_char == '%')
#else
    if (boiler_char == '?')
#endif
    {

        /*
         *  We have a single-character wildcard.  Flag that we know about it and
         *  get the next boilerplate character.
         */
        wild_seen = PTRUE;
        goto chknxt_boilerplate;
    }

    /*
     *  At this point we have two characters to compare.  Do they match?
     *  If so get the next character from the boilerplate.
     */
    if (boiler_char == string_char) goto chknxt_boilerplate;

    /*
     *  UNMATCHED_CHAR:
     *
     *  At this point the current characters do not match.  If we are in a wildcard
     *  then back out - else exit with error.
     */
unmatched_char:
    wildcard_offset -= 1;
    if (wildcard_offset < 0) goto match_failed;

    /*
     *  We were in a wildcard sequence.  Reset the values back and continue.
     *  First set to the next character in the string.
     */
    saved_string_offset += 1;

    /*
     *  Restore the string length and offset.
     */
    string_len = wildcard_offset;
    string_offset = saved_string_offset;

    /*
     *  Restore the boilerplate length and offset, then check the next boilerplate.
     */
    boiler_len = saved_boiler_len;
    boiler_offset = saved_boiler_offset;

    goto chknxt_boilerplate;

    /*
     *  END_BOILERPLATE:
     *
     *  We ran out of boilerplate string.  See if we ran out of source string
     *  as well and if so exit with success, else check for trailing spaces.
     */
end_boilerplate:
    if (string_len == 0) goto string_matches;

    /*
     *  First save the string length and offset.
     */
    space_string_len = string_len;
    space_string_offset = string_offset;

    /*
     *  CHKNXT_SOURCE:
     *
     *  Check the next source string for a space.
     */
chknxt_source:
    space_string_char = *(string_ptr + string_offset);
    string_offset += 1;

    /*
     *  Is this character a SPACE?
     */
    if (space_string_char == ' ')
    {

        /*
         *  This character is a SPACE.  If we can, what's the previous character?
         */
        string_len -= 1;
        if (string_len > 0)
        {
            goto chknxt_source;
        }
        else
        {
            goto string_matches;
        }
    }

    /*
     *  There was something other than nulls and spaces at the end - see
     *  if we were doing a wild-card and, if so, unwind it.
     *
     *  First, Restore the values saved earlier.
     */
    string_len = space_string_len;
    string_offset = space_string_offset;

    /*
     *  Are we doing a wildcard sequence?  If so, unwind it.
     */
    if (wildcard_offset > 0) goto unmatched_char;

    /*
     *  Any wild-cards encountered?
     */
    if (wild_seen) goto no_match;

    /*
     *  The match failed and the source string was longer than the
     *  boilerplate string.  If we have no wild-carding at all then
     *  we might have a situation such as:
     *  
     *  Source - "ABC"   Boilerplate - "A"
     *  
     *  in which case we want to return ENDMATCH instead of NOMATCH.
     *  The normal check (at MATCH_FAILED) would return NOMATCH in this case
     *  as it has to allow for the opposite as well:
     *  
     *  Source - "A"   Boilerplate - "ABC"
     *  
     *  We therefore do our own, separate, check here.  Note that, to be
     *  safe, we insist that the next character in the source string is
     *  greater than a space character.
     *
     *  Determine if we should return either NOMATCH or ENDMATCH.  ENDMATCH is
     *  returned purley for performance reasons (so that the utility does not
     *  call us again if the next set of strings, in alphabetical sequence,
     *  will not match).
     */
    if (string_char < boiler_char)
    {
        goto no_match;
    }
    else if (string_char <= ' ')
    {
        goto no_match;
    }
    else
    {
        goto end_match;
    }

    /*
     *  STRING_MATCHES:
     *
     *  All remaining characters in the source string were blank or
     *  null so we accept them
     */
string_matches:
    return PSL_NORMAL;

    /*
     *  MATCH_FAILED:
     *
     *  The match failed - see if we have hit any wildcards and, if not,
     *  check for 'end of possible matches'.
     */
match_failed:
    if (wild_seen) goto no_match;

    if (string_char <= boiler_char) goto no_match;

    /*
     *  END_MATCH:
     *
     *  This string does not match the boilerplate and any subsequent strings,
     *  in alphabetical sequence, will not match either.
     */
end_match:
    return PSL_ENDMATCH;

    /*
     *  NO_MATCH:
     *
     *  This string does not match the boilerplate but subsequent strings,
     *  in alphabetical sequence, may match.
     */
no_match:
    return PSL_NOMATCH;
}
