/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*               **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_TRACE_RTNS              */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  PSL Trace Routines.                          */
/*                                       */
/*===========================================================================*/

/*
 *  PSL Macros.
 */
#include <psl_macros.h>

/*
 *  Standard ANSI header files.
 */
#include <stdio.h>          /* Standard I/O */
#include <stdlib.h>         /* Standard Library Routines */
#include <string.h>         /* String functions */
#include <stdarg.h>         /* Variable Argument List handling */



/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>     /* PSL Routines */
#include <trc_rtndef.h>     /* PSL Trace Routines */

static unsigned long GET_TRACE (char *filename_ptr);

/*
 *  VMS-Specific Structures
 */
#ifdef VMS
#pragma nostandard          /* Keep Alpha Happy! */
#ifdef ALPHA
#include <afrdef.h>         /* Alignment Fault Symbols */
#endif /* ALPHA */
#include <ssdef.h>          /* SS Symbols */
#include <descrip.h>            /* VAX Descriptors */
#include <lnmdef.h>         /* $TRNLNM Symbols */
#include <starlet.h>            /* VMS System Services */
#include <lib$routines.h>       /* VMS Library Symbols */
#pragma standard
#endif /* VMS */
#ifdef WIN32
#include <windows.h>
#endif


/*
 *  Structure Definitions.
 */
#include psl_common_file (shrmsg.h) /* PSL messages */
#include psl_common_file (itemlist.h)   /* Itemlist structure */
#include psl_common_file (psldate_prmdef.h)
/* PSL Date/Time Symbols */
#include psl_common_file (psl_trc_prmdef.h)
/* Trace Symbols */
#include psl_common_file (psl_trcdef.h) /* Trace Structure */

/*
 *  Trace Structure.
 */
static struct psl_trc psl_trc;


/************************************************************************/
/*                                  */
/*  PSL_INIT_TRACE - Initialize Tracing             */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     PSL_INIT_TRACE ();                       */
/*                                  */
/*  Where:                              */
/*                                  */
/*                                  */
/*  This routine initializes PSL Tracing. It tries to obtain the    */
/*  name of the Trace file. If this succeeds, the routine opens */
/*  the Trace file and sets a flag to indicate that tracing is  */
/*  taking place to a file. If it fails to obtain the Trace File    */
/*  name, Tracing takes place to stdout.                */
/*                                  */
/************************************************************************/

void    PSL_INIT_TRACE (void)
{
    /*
     *  Local Data.
     */
    unsigned long       iret;       /* Return Status */
    char       filename [256]; /* Trace Filename */
    unsigned long       date;       /* Current Date */
    unsigned long       time;       /* Current Time */
    char       buffer [132];   /* Output Buffer */

    /*
     *  Get Name of Trace File.
     */
    iret = GET_TRACE (filename);

    /*
     *  If it succeeded, Open Trace File for Write Access.
     *  If it failed we use stdout.
     */
    if (psl_test_status (iret))     /* If we got a name */
    {
        psl_trc.psl_trc_file_ptrpf = fopen (filename, "w");
        /* Open the file */

        if (psl_trc.psl_trc_file_ptrpf == NULL)
            /* If it failed */
        {
            (void) printf ("Error Opening Trace File - using stdout\n");
            /* Tell the user */
        }
    }

    /*
     *  Get the current Date and Time.
     */
    PSL_GET_DATTIM (&date, &time);

    /*
     *  Format the Date and Time into the output buffer.
     */
    (void) memset (buffer, '\0', sizeof (buffer));
    (void) PSL_FMT_DATTIM (date,    /* Date */
                           time,       /* Time */
                           buffer,     /* Into this buffer */
                           sizeof (buffer),    /* ... which is this big */
                           PSDTYP_PSL, /* PSL Date Type */
                           PSDFMT_DATE | PSDFMT_TIME | PSDFMT_CENTURY);
    /* Flags */
    /*
     *  Output it.
     */
    PSL_OUTPUT_TRACE ("Tracing Started on %s\n", buffer);

    return;
}


/************************************************************************/
/*                                  */
/*  PSL_SET_TRC_ON - Set Tracing Option             */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     PSL_SET_TRC_ON (option);                 */
/*                                  */
/*  Where:                              */
/*                                  */
/*     option     = Tracing Option to set               */
/*                                  */
/*  This routine Sets a PSL Tracing option.             */
/*                                  */
/************************************************************************/

void    PSL_SET_TRC_ON (unsigned long option)
{
    /*
     *  Set option flag and corresponding Class flag.
     */
    switch (option)
    {
        /*
         *  General Options.
         */
    case PTRC_GEN_COMMAND:
        {
            psl_trc.psl_trc_genflgl |= PTGN_COMMAND;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_GEN_COMMAND Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_GEN_TIMING:
        {
            psl_trc.psl_trc_genflgl |= PTGN_TIMING;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_GEN_TIMING Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_GEN_ALIGN:
        {
            psl_trc.psl_trc_genflgl |= PTGN_ALIGN;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_GEN_ALIGN Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_GEN_ERRORS:
        {
            psl_trc.psl_trc_genflgl |= PTGN_ERRORS;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_GEN_ERRORS Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_GEN_TPARSE:
        {
            psl_trc.psl_trc_genflgl |= PTGN_TPARSE;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_GEN_TPARSE Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_GEN_MISC:
        {
            psl_trc.psl_trc_genflgl |= PTGN_MISC;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_GEN_MISC Set On\n");
            /* Output to Trace File */
            break;
        }


    case PTRC_GEN_MEM:
        {
            psl_trc.psl_trc_genflgl |= PTGN_MEM;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_GEN_MEM Set On\n");
            /* Output to Trace File */
            break;
        }



    case PTRC_GEN_PIPES:
        {
            psl_trc.psl_trc_genflgl |= PTGN_PIPES;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_GEN_PIPES Set On\n");
            /* Output to Trace File */
            break;
        }


        /*
         *  Form Driver Options.
         */


    case PTRC_FDR_ENABLED:
        {
            psl_trc.psl_trc_fdrflgl |= PTFD_ENABLED;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_FDR_ENABLED Set On\n");
            /* Output to Trace File */
            break;
        }


    case PTRC_FDR_MONITOR:
        {
            psl_trc.psl_trc_fdrflgl |= PTFD_MONITOR;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_FDR_MONITOR Set On\n");
            /* Output to Trace File */
            break;
        }


    case PTRC_FDR_REPEAT:
        {
            psl_trc.psl_trc_fdrflgl |= PTFD_REPEAT;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_FDR_REPEAT Set On\n");
            /* Output to Trace File */
            break;
        }


        /*
         *  User Options.
         */


    case PTRC_USR_ENABLED:
        {
            psl_trc.psl_trc_usrflgl |= PTUS_ENABLED;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_USR_ENABLED Set On\n");
            /* Output to Trace File */
            break;
        }


    case PTRC_USR_USER1:
        {
            psl_trc.psl_trc_usrflgl |= PTUS_USER1;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_USR_USER1 Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_USR_USER2:
        {
            psl_trc.psl_trc_usrflgl |= PTUS_USER2;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_USR_USER2 Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_USR_USER3:
        {
            psl_trc.psl_trc_usrflgl |= PTUS_USER3;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_USR_USER3 Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_USR_USER4:
        {
            psl_trc.psl_trc_usrflgl |= PTUS_USER4;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_USR_USER4 Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_USR_USER5:
        {
            psl_trc.psl_trc_usrflgl |= PTUS_USER5;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_USR_USER5 Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_USR_USER6:
        {
            psl_trc.psl_trc_usrflgl |= PTUS_USER6;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_USR_USER6 Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_USR_USER7:
        {
            psl_trc.psl_trc_usrflgl |= PTUS_USER7;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_USR_USER7 Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_USR_USER8:
        {
            psl_trc.psl_trc_usrflgl |= PTUS_USER8;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_USR_USER8 Set On\n");
            /* Output to Trace File */
            break;
        }
        /*
         *  TRM Routine Options.
         */
    case PTRC_TRM_ENABLED:
        {
            psl_trc.psl_trc_trmflgl |= PTTR_ENABLED;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_ENABLED Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_TRM_GENRTNS:
        {
            psl_trc.psl_trc_trmflgl |= PTTR_GENRTNS;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_GENRTNS Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_TRM_CLEAR:
        {
            psl_trc.psl_trc_trmflgl |= PTTR_CLEAR;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_CLEAR Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_TRM_OUTPUT:
        {
            psl_trc.psl_trc_trmflgl |= PTTR_OUTPUT;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_OUTPUT Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_TRM_INPUT:
        {
            psl_trc.psl_trc_trmflgl |= PTTR_INPUT;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_INPUT Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_TRM_DRWRTNS:
        {
            psl_trc.psl_trc_trmflgl |= PTTR_DRWRTNS;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_DRWRTNS Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_TRM_HLPRTNS:
        {
            psl_trc.psl_trc_trmflgl |= PTTR_HLPRTNS;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_HLPRTNS Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_TRM_SCROLL:
        {
            psl_trc.psl_trc_trmflgl |= PTTR_SCROLL;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_SCROLL Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_TRM_SPCRTNS:
        {
            psl_trc.psl_trc_trmflgl |= PTTR_SPCRTNS;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_SPCRTNS Set On\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_TRM_ODDRTNS:
        {
            psl_trc.psl_trc_trmflgl |= PTTR_ODDRTNS;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_ODDRTNS Set On\n");
            /* Output to Trace File */
            break;
        }



        /*
         *    Trace the IPC/DTC routines.
         */
    case PTRC_IPC_IPC:
        {
            psl_trc.psl_trc_ipcflgl |= PTIP_IPC;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_IPC_IPC Set On\n");
            /* Output to Trace File */
            break;
        }

    case PTRC_IPC_DTC:
        {
            psl_trc.psl_trc_ipcflgl |= PTIP_DTC;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_IPC_DTC Set On\n");
            /* Output to Trace File */
            break;
        }




    case PTRC_IPC_IPC2:
        {
            psl_trc.psl_trc_ipcflgl |= PTIP_IPC2;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_IPC_IPC2 Set On\n");
            /* Output to Trace File */
            break;
        }

    case PTRC_IPC_DTC2:
        {
            psl_trc.psl_trc_ipcflgl |= PTIP_DTC2;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_IPC_DTC2 Set On\n");
            /* Output to Trace File */
            break;
        }


    default:                /* Invalid Option */
        {
            PSL_OUTPUT_TRACE ("\nInvalid Tracing Option specified : %d\n", option);
            /* Tell the user */
            break;
        }

    }                   /* End of (switch (option)) */

    return;
}


/************************************************************************/
/*                                  */
/*  PSL_SET_TRC_OFF - Cancel Tracing Option             */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     PSL_SET_TRC_OFF (option);                    */
/*                                  */
/*  Where:                              */
/*                                  */
/*     option     = Tracing Option to cancel            */
/*                                  */
/*  This routine Cancels a PSL Tracing option.          */
/*                                  */
/************************************************************************/

void    PSL_SET_TRC_OFF (unsigned long option)
{
    /*
     *  Clear option flag and corresponding Class flag.
     */
    switch (option)
    {
        /*
         *  General Options.
         */
    case PTRC_GEN_COMMAND:
        {
            psl_trc.psl_trc_genflgl &= ~PTGN_COMMAND;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_GEN_COMMAND Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_GEN_TIMING:
        {
            psl_trc.psl_trc_genflgl &= ~PTGN_TIMING;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_GEN_TIMING Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_GEN_ALIGN:
        {
            psl_trc.psl_trc_genflgl &= ~PTGN_ALIGN;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_GEN_ALIGN Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_GEN_ERRORS:
        {
            psl_trc.psl_trc_genflgl &= ~PTGN_ERRORS;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_GEN_ERRORS Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_GEN_TPARSE:
        {
            psl_trc.psl_trc_genflgl &= ~PTGN_TPARSE;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_GEN_TPARSE Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_GEN_MISC:
        {
            psl_trc.psl_trc_genflgl &= ~PTGN_MISC;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_GEN_MISC Set Off\n");
            /* Output to Trace File */
            break;
        }

    case PTRC_GEN_MEM:
        {
            psl_trc.psl_trc_genflgl &= ~PTGN_MEM;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_GEN_MEM Set Off\n");
            /* Output to Trace File */
            break;
        }



    case PTRC_GEN_PIPES:
        {
            psl_trc.psl_trc_genflgl &= ~PTGN_PIPES;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_GEN_PIPES Set Off\n");
            /* Output to Trace File */
            break;
        }


        /*
         *  Form Driver Options.
         */


    case PTRC_FDR_ENABLED:
        {
            psl_trc.psl_trc_fdrflgl &= ~PTFD_ENABLED;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_FDR_ENABLED Set Off\n");
            /* Output to Trace File */
            break;
        }


    case PTRC_FDR_MONITOR:
        {
            psl_trc.psl_trc_fdrflgl &= ~PTFD_MONITOR;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_FDR_MONITOR Set Off\n");
            /* Output to Trace File */
            break;
        }


    case PTRC_FDR_REPEAT:
        {
            psl_trc.psl_trc_fdrflgl &= ~PTFD_REPEAT;
            /* Set Option */
            PSL_OUTPUT_TRACE ("PTRC_FDR_REPEAT Set Off\n");
            /* Output to Trace File */
            break;
        }



        /*
         *  User Options.
         */


    case PTRC_USR_ENABLED:
        {
            psl_trc.psl_trc_usrflgl &= ~PTUS_ENABLED;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_USR_ENABLED Set Off\n");
            /* Output to Trace File */
            break;
        }


    case PTRC_USR_USER1:
        {
            psl_trc.psl_trc_usrflgl &= ~PTUS_USER1;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_USR_USER1 Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_USR_USER2:
        {
            psl_trc.psl_trc_usrflgl &= ~PTUS_USER2;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_USR_USER2 Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_USR_USER3:
        {
            psl_trc.psl_trc_usrflgl &= ~PTUS_USER3;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_USR_USER3 Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_USR_USER4:
        {
            psl_trc.psl_trc_usrflgl &= ~PTUS_USER4;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_USR_USER4 Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_USR_USER5:
        {
            psl_trc.psl_trc_usrflgl &= ~PTUS_USER5;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_USR_USER5 Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_USR_USER6:
        {
            psl_trc.psl_trc_usrflgl &= ~PTUS_USER6;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_USR_USER6 Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_USR_USER7:
        {
            psl_trc.psl_trc_usrflgl &= ~PTUS_USER7;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_USR_USER7 Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_USR_USER8:
        {
            psl_trc.psl_trc_usrflgl &= ~PTUS_USER8;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_USR_USER8 Set Off\n");
            /* Output to Trace File */
            break;
        }
        /*
         *  TRM Routine Options.
         */
    case PTRC_TRM_ENABLED:
        {
            psl_trc.psl_trc_trmflgl &= ~PTTR_ENABLED;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_ENABLED Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_TRM_GENRTNS:
        {
            psl_trc.psl_trc_trmflgl &= ~PTTR_GENRTNS;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_GENRTNS Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_TRM_CLEAR:
        {
            psl_trc.psl_trc_trmflgl &= ~PTTR_CLEAR;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_CLEAR Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_TRM_OUTPUT:
        {
            psl_trc.psl_trc_trmflgl &= ~PTTR_OUTPUT;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_OUTPUT Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_TRM_INPUT:
        {
            psl_trc.psl_trc_trmflgl &= ~PTTR_INPUT;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_INPUT Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_TRM_DRWRTNS:
        {
            psl_trc.psl_trc_trmflgl &= ~PTTR_DRWRTNS;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_DRWRTNS Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_TRM_HLPRTNS:
        {
            psl_trc.psl_trc_trmflgl &= ~PTTR_HLPRTNS;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_HLPRTNS Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_TRM_SCROLL:
        {
            psl_trc.psl_trc_trmflgl &= ~PTTR_SCROLL;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_SCROLL Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_TRM_SPCRTNS:
        {
            psl_trc.psl_trc_trmflgl &= ~PTTR_SPCRTNS;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_SPCRTNS Set Off\n");
            /* Output to Trace File */
            break;
        }
    case PTRC_TRM_ODDRTNS:
        {
            psl_trc.psl_trc_trmflgl &= ~PTTR_ODDRTNS;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_TRM_ODDRTNS Set Off\n");
            /* Output to Trace File */
            break;
        }



        /*
         *    Trace the IPC/DTC routines.
         */
    case PTRC_IPC_IPC:
        {
            psl_trc.psl_trc_ipcflgl &= ~PTIP_IPC;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_IPC_IPC Set Off\n");
            /* Output to Trace File */
            break;
        }

    case PTRC_IPC_DTC:
        {
            psl_trc.psl_trc_ipcflgl &= ~PTIP_DTC;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_IPC_DTC Set Off\n");
            /* Output to Trace File */
            break;
        }


    case PTRC_IPC_IPC2:
        {
            psl_trc.psl_trc_ipcflgl &= ~PTIP_IPC2;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_IPC_IPC2 Set Off\n");
            /* Output to Trace File */
            break;
        }

    case PTRC_IPC_DTC2:
        {
            psl_trc.psl_trc_ipcflgl &= ~PTIP_DTC2;
            /* Clear Option */
            PSL_OUTPUT_TRACE ("PTRC_IPC_DTC2 Set Off\n");
            /* Output to Trace File */
            break;
        }



    default:                /* Invalid Option */
        {
            PSL_OUTPUT_TRACE ("\nInvalid Tracing Option specified : %d\n", option);
            /* Tell the user */
            break;
        }

    }                   /* End of (switch (option)) */

    return;
}


/************************************************************************/
/*                                  */
/*  PSL_TEST_TRACE - Test Tracing option                */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_TEST_TRACE (option);              */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Boolean Routine Return Status           */
/*     option     = Tracing option to test              */
/*                                  */
/*  This routine Tests a PSL Tracing option.            */
/*                                  */
/************************************************************************/

Boolean PSL_TEST_TRACE (unsigned long option)
{
    /*
     *  Local Data.
     */
    Boolean         retval;     /* Routine Return Value */

    /*
     *  Test option flag.
     */
    switch (option)
    {
        /*
         *  General Options.
         */
    case PTRC_GEN_COMMAND:
        {
            if ( (psl_trc.psl_trc_genflgl & PTGN_COMMAND) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_GEN_TIMING:
        {
            if ( (psl_trc.psl_trc_genflgl & PTGN_TIMING) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_GEN_ALIGN:
        {
            if ( (psl_trc.psl_trc_genflgl & PTGN_ALIGN) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_GEN_ERRORS:
        {
            if ( (psl_trc.psl_trc_genflgl & PTGN_ERRORS) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_GEN_TPARSE:
        {
            if ( (psl_trc.psl_trc_genflgl & PTGN_TPARSE) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_GEN_MISC:
        {
            if ( (psl_trc.psl_trc_genflgl & PTGN_MISC) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }


    case PTRC_GEN_MEM:
        {
            if ( (psl_trc.psl_trc_genflgl & PTGN_MEM) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }



    case PTRC_GEN_PIPES:
        {
            if ( (psl_trc.psl_trc_genflgl & PTGN_PIPES) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }


        /*
         *  Form Driver Options.
         */


    case PTRC_FDR_ENABLED:
        {
            if ( (psl_trc.psl_trc_fdrflgl & PTFD_ENABLED) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }



    case PTRC_FDR_MONITOR:
        {
            if ( (psl_trc.psl_trc_fdrflgl & PTFD_MONITOR) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }


    case PTRC_FDR_REPEAT:
        {
            if ( (psl_trc.psl_trc_fdrflgl & PTFD_REPEAT) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }



        /*
         *  User Options.
         */


    case PTRC_USR_ENABLED:
        {
            if ( (psl_trc.psl_trc_usrflgl & PTUS_ENABLED) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }



    case PTRC_USR_USER1:
        {
            if ( (psl_trc.psl_trc_usrflgl & PTUS_USER1) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_USR_USER2:
        {
            if ( (psl_trc.psl_trc_usrflgl & PTUS_USER2) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_USR_USER3:
        {
            if ( (psl_trc.psl_trc_usrflgl & PTUS_USER3) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_USR_USER4:
        {
            if ( (psl_trc.psl_trc_usrflgl & PTUS_USER4) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_USR_USER5:
        {
            if ( (psl_trc.psl_trc_usrflgl & PTUS_USER5) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_USR_USER6:
        {
            if ( (psl_trc.psl_trc_usrflgl & PTUS_USER6) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_USR_USER7:
        {
            if ( (psl_trc.psl_trc_usrflgl & PTUS_USER7) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_USR_USER8:
        {
            if ( (psl_trc.psl_trc_usrflgl & PTUS_USER8) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
        /*
         *  TRM Routine Options.
         */
    case PTRC_TRM_ENABLED:
        {
            if ( (psl_trc.psl_trc_trmflgl & PTTR_ENABLED) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_TRM_GENRTNS:
        {
            if ( (psl_trc.psl_trc_trmflgl & PTTR_GENRTNS) != 0)
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_TRM_CLEAR:
        {
            if ( (psl_trc.psl_trc_trmflgl & PTTR_CLEAR) != 0)
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_TRM_OUTPUT:
        {
            if ( (psl_trc.psl_trc_trmflgl & PTTR_OUTPUT) != 0)
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_TRM_INPUT:
        {
            if ( (psl_trc.psl_trc_trmflgl & PTTR_INPUT) != 0)
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_TRM_DRWRTNS:
        {
            if ( (psl_trc.psl_trc_trmflgl & PTTR_DRWRTNS) != 0)
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_TRM_HLPRTNS:
        {
            if ( (psl_trc.psl_trc_trmflgl & PTTR_HLPRTNS) != 0)
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_TRM_SCROLL:
        {
            if ( (psl_trc.psl_trc_trmflgl & PTTR_SCROLL) != 0)
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_TRM_SPCRTNS:
        {
            if ( (psl_trc.psl_trc_trmflgl & PTTR_SPCRTNS) != 0)
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }
    case PTRC_TRM_ODDRTNS:
        {
            if ( (psl_trc.psl_trc_trmflgl & PTTR_ODDRTNS) != 0)
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }



        /*
         *    Handle the IPC trace options.
         */
    case PTRC_IPC_IPC:
        {
            if ((psl_trc.psl_trc_ipcflgl & PTIP_IPC) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }

    case PTRC_IPC_DTC:
        {
            if ((psl_trc.psl_trc_ipcflgl & PTIP_DTC) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }




    case PTRC_IPC_IPC2:
        {
            if ((psl_trc.psl_trc_ipcflgl & PTIP_IPC2) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }

    case PTRC_IPC_DTC2:
        {
            if ((psl_trc.psl_trc_ipcflgl & PTIP_DTC2) != 0 )
            {
                retval = PTRUE;
            }
            else
            {
                retval = PFALSE;
            }
            break;
        }


    default:                /* Invalid Option */
        {
            PSL_OUTPUT_TRACE ("\nInvalid Tracing Option specified : %d\n", option);
            /* Tell the user */
            retval = PFALSE;
            break;
        }

    }                   /* End of (switch (option)) */

    return retval;
}


/************************************************************************/
/*                                  */
/*  PSL_OUTPUT_TRACE - Output Trace Information         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     PSL_OUTPUT_TRACE (fmt_ptr, ...);             */
/*                                  */
/*  Where:                              */
/*                                  */
/*     fmt_ptr    = Pointer to Format String for use by printf  */
/*     ...        = Arguments required by format string     */
/*                                  */
/*  This routine outputs PSL Tracing information.           */
/*                                  */
/************************************************************************/

void    PSL_OUTPUT_TRACE (const char *fmt_ptr, ...)
{
    /*
     *  Local Data.
     */
    long       status;     /* Return Status */
    va_list    args;       /* Argument List Pointer */

    /*
     *  Just set up the variable argument bits and output the information
     *  via V(F)PRINTF.
     */
    va_start (args, fmt_ptr);       /* Set up the argument pointer */

    if (psl_trc.psl_trc_file_ptrpf != NULL)
        /* If Tracing to a file */
    {
        status = vfprintf (psl_trc.psl_trc_file_ptrpf, fmt_ptr, args);
        /* Output the information */
        (void) fflush (psl_trc.psl_trc_file_ptrpf);
    }
    else                /* Tracing to stdout */
    {
        status = vprintf (fmt_ptr, args);
        /* Output the information */
    }

    va_end (args);          /* Tidy up the argument pointer */

    if (status < 0)         /* If an Error occurred */
    {
        (void) printf ("Error in PSL_OUTPUT_TRACE");
        /* Tell the user */
    }

    return;
}


/************************************************************************/
/*                                  */
/*  PSL_TIME_TRACE - Output Tracing Timestamp           */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     PSL_TIME_TRACE ();                       */
/*                                  */
/*  Where:                              */
/*                                  */
/*                                  */
/*  This routine outputs a Tracing Timestamp.           */
/*                                  */
/************************************************************************/

void    PSL_TIME_TRACE (void)
{
    /*
     *  Local Data.
     */
    unsigned long       date;       /* Current Date */
    unsigned long       time;       /* Current Time */
    char       buffer [132];   /* Output Buffer */

    /*
     *  Get the current Date and Time.
     */
    PSL_GET_DATTIM (&date, &time);

    /*
     *  Format the Date and Time into the output buffer.
     */
    (void) memset (buffer, '\0', sizeof (buffer));
    (void) PSL_FMT_DATTIM (date,    /* Date */
                           time,       /* Time */
                           buffer,     /* Into this buffer */
                           sizeof (buffer),    /* ... which is this big */
                           PSDTYP_PSL,     /* PSL Date Type */
                           PSDFMT_DATE | PSDFMT_TIME | PSDFMT_CENTURY);
    /* Flags */
    /*
     *  Output it.
     */
    PSL_OUTPUT_TRACE ("Tracing Timestamp on %s\n", buffer);

    return;
}





#ifdef ALPHA
/************************************************************************/
/*                                  */
/*  PSL_INIT_ALN_TRC - Initialize Alignment Tracing         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_INIT_ALN_TRC ();                  */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long PSL Routine Return Status     */
/*                                  */
/*  This routine initializes tracing a Alignment Faults.        */
/*                                  */
/************************************************************************/

unsigned long   PSL_INIT_ALN_TRC (void)
{
    /*
     *  Local Data.
     */
    unsigned long       iret;       /* Status Return */

    unsigned long       syserverr = PSL_SYSERVERR;
    /* System Service Error Code */
    unsigned long       fao_count = 1;  /* FAO Count */
    $DESCRIPTOR (err_rtn_dsc, "SYS$START_ALIGN_FAULT_REPORT");
    /* Error Routine Name */

    /*
     *  Start Alignment Fault Reporting using exception reporting.
     */
    iret = sys$start_align_fault_report (AFR$C_EXCEPTION, NULL, 0);
    if ( !(psl_test_status (iret)) )
    {
        PSL_PUTMSG (&syserverr, &fao_count, &err_rtn_dsc, iret);
    }

    return iret;
}

#endif /* ALPHA */







#ifdef ALPHA

/************************************************************************/
/*                                  */
/*  PSL_ALIGN_HDLR - Alignment Tracing Condition Handler        */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_ALIGN_HDLR (sigargs_ptr, mechargs_ptr);       */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret         = Unsigned Long PSL Routine Return Status   */
/*     sigargs_ptr  = Pointer to Signal Arguments Array     */
/*     mechargs_ptr = Pointer to Mechanism Arguments Array      */
/*                                  */
/*  This routine is the Condition Handler used to trap Alignment    */
/*  Faults and report them to the trace file.           */
/*  Any other conditions are resignalled.               */
/*                                  */
/************************************************************************/

unsigned long   PSL_ALIGN_HDLR (void *sigargs_ptr,
                                void *mechargs_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       match;      /* Condition Match Status */
    unsigned long       pc;         /* PC */
    unsigned long       va;         /* Virtual Address */
    unsigned long       *wrk_ptr;       /* Work Pointer */
    unsigned long       align_code = SS$_ALIGN;
    /* Alignment Fault Condition Code */
    unsigned long       *argcnt_ptr;    /* Pointer to arg count */

    /*
     *  Get pointer to the Condition.
     */
    argcnt_ptr = (unsigned long *) sigargs_ptr;
    wrk_ptr = argcnt_ptr + 1;

    /*
     *  See if this is a condition we are interested in.
     *  If so, handle it and continue execution.
     *  If not, send it on its way.
     */
    match = lib$match_cond (wrk_ptr, &align_code);

    if (match == 1)
    {

        /*
         *  Extract requied information from the Signal Arguments Array.
         */
        wrk_ptr = argcnt_ptr + (*argcnt_ptr) - 1;
        pc = *wrk_ptr;

        wrk_ptr = argcnt_ptr + 2;
        va = *wrk_ptr;

        PSL_OUTPUT_TRACE ("Alignment Fault, PC = %X, VA = %X\n", pc, va);
        return SS$_CONTINUE;
    }
    else
    {
        return SS$_RESIGNAL;
    }
}
#endif /* ALPHA */







#ifdef UNIX

/************************************************************************/
/*                                  */
/*  GET_TRACE - Get name of trace file              */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = GET_TRACE (filename_ptr);             */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret         = Unsigned Long PSL Routine Return Status   */
/*            PSL_NORMAL if successful          */
/*            PSL_NOMTCHFOU otherwise           */
/*     filename_ptr = Pointer to buffer to return file name in  */
/*                                  */
/*  This routine returns the name of the Trace File.        */
/*                                  */
/************************************************************************/

static unsigned long GET_TRACE (char *filename_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       iret;       /* Return Status */
    char       *wrk_ptr;       /* Work Pointer */

    /*
     *  Attempt to translate the PSL_TRACE_FILE environment variable.
     */
    wrk_ptr = getenv ("PSL_TRACE_FILE");
    if (wrk_ptr != NULL)
    {
        (void) strcpy (filename_ptr, wrk_ptr);
        iret = PSL_NORMAL;
    }
    else
    {
        iret = PSL_NOMTCHFOU;
    }

    return iret;
}

#endif /* UNIX */







#ifdef VMS

/************************************************************************/
/*                                  */
/*  GET_TRACE - Get name of trace file              */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = GET_TRACE (filename_ptr);             */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret         = Unsigned Long PSL Routine Return Status   */
/*     filename_ptr = Pointer to buffer to return file name in  */
/*                                  */
/*  This routine returns the name of the Trace File.        */
/*  We first attempt to translate the logical name PSL$TRACE_FILE   */
/*  and if that fails we try PSL_TRACE_FILE.            */
/*                                  */
/************************************************************************/

static unsigned long GET_TRACE (char *filename_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       iret;       /* Return Status */
    char       filename [256]; /* Filename */
    unsigned short      length;     /* Length thereof */
    struct   itemlist   itemlist [2];   /* Itemlist */
    struct   dsc$descriptor_s tabdsc;   /* Logical name Table descriptor */
    struct   dsc$descriptor_s namdsc;   /* Logical name descriptor */

    /*
     *  Set up the Itemlist to get Logical Name Equivalence String.
     */
    itemlist [0] .itl_item   = LNM$_STRING;
    /* Get Equivalence String */
    itemlist [0] .itl_buflen = sizeof (filename);
    /* Length of buffer */
    itemlist [0] .itl_bufadr = filename;
    /* Put it here */
    itemlist [0] .itl_lenadr = &length; /* Returned length */

    itemlist [1] .itl_item   = 0;   /* End of Itemlist */
    itemlist [1] .itl_buflen = 0;

    /*
     *  Set up descriptor for any Logical Name Table.
     */
    tabdsc.dsc$b_class   = DSC$K_CLASS_S;
    tabdsc.dsc$b_dtype   = DSC$K_DTYPE_T;
    tabdsc.dsc$a_pointer = "PSL$LNM_TABLE";
    tabdsc.dsc$w_length  = strlen (tabdsc.dsc$a_pointer);

    /*
     *  Set up descriptor for first possible Logical Name.
     */
    namdsc.dsc$b_class   = DSC$K_CLASS_S;
    namdsc.dsc$b_dtype   = DSC$K_DTYPE_T;
    namdsc.dsc$a_pointer = "PSL$TRACE_FILE";
    namdsc.dsc$w_length  = strlen (namdsc.dsc$a_pointer);

    /*
     *  Call $TRNLNM.
     */
    iret  = sys$trnlnm (        /* Translate logical name */
                NULL,           /* No Attributes Required */
                &tabdsc,        /* Logical name table */
                &namdsc,        /* Logical Name */
                NULL,           /* No Mode */
                itemlist);      /* Itemlist */

    /*
     *  If the translation failed, try next possible logical name.
     */
    if (!(psl_test_status (iret)))
    {

        /*
         *  Set up descriptor for second possible Logical Name.
         */
        namdsc.dsc$b_class   = DSC$K_CLASS_S;
        namdsc.dsc$b_dtype   = DSC$K_DTYPE_T;
        namdsc.dsc$a_pointer = "PSL_TRACE_FILE";
        namdsc.dsc$w_length  = strlen (namdsc.dsc$a_pointer);

        /*
         *  Call $TRNLNM again.
         */
        iret  = sys$trnlnm (        /* Translate logical name */
                    NULL,           /* No Attributes Required */
                    &tabdsc,        /* Logical name table */
                    &namdsc,        /* Logical Name */
                    NULL,           /* No Mode */
                    itemlist);      /* Itemlist */

        if (!(psl_test_status (iret))) goto end;
        /* Exit if it failed */
    }



    /*
     *  Null-terminate the string.
     */
    filename [length] = '\0';



    /*
     *  Return it to the caller.
     */
    (void) strcpy (filename_ptr, filename);

    /*
     *  END:
     *  ===
     *
     *  Routine exit point.
     */
end:
    ;

    return iret;

}

#endif /* VMS */







#ifdef WIN32

/************************************************************************/
/*                                  */
/*  GET_TRACE - Get name of trace file              */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = GET_TRACE (filename_ptr);             */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret         = Unsigned Long PSL Routine Return Status   */
/*            PSL_NORMAL if successful          */
/*            PSL_NOMTCHFOU otherwise           */
/*     filename_ptr = Pointer to buffer to return file name in  */
/*                                  */
/*  This routine returns the name of the Trace File.        */
/*                                  */
/************************************************************************/

static unsigned long GET_TRACE (char *filename_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       iret = PSL_NORMAL;  /* Return Status */
#ifndef WIN32
    char       *wrk_ptr;       /* Work Pointer */
#endif

    /*
     *  Attempt to translate the PSL_TRACE_FILE environment variable.
     */
#ifdef WIN32
    if (!GetEnvironmentVariable ("PSL_TRACE_FILE", filename_ptr, 256))
    {
        iret = PSL_NOMTCHFOU;
    }
#else
wrk_ptr = getenv ("PSL_TRACE_FILE");
    if (wrk_ptr != NULL)
    {
        (void) strcpy (filename_ptr, wrk_ptr);
    }
    else
    {
        iret = PSL_NOMTCHFOU;
    }
#endif

    return iret;

}

#endif /* WIN32 */



