/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/* scl_compile.c:
 * ==============
 *
 * This module contains the main scl compilation control routines.
 */

/*
 *  PSL Macros.
 */
#include <psl_macros.h>
#include <vtest_macros.h>

/*
 *  Standard ANSI header files.
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifdef  WIN32
#include <conio.h>
#endif  /* WIN32 */

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>         /* PSL Routines */
#include <trc_rtndef.h>         /* Tracing Routines */
#include <vtest_rtndef.h>       /* V-TEST Routines */
#include "scl_rtndef.h"         /* SCL Routines */

static void     CLR_TSTTHR_TABLE    (struct scl_com *com_ptr);

static void     CLR_TST_TABLE       (struct scl_com *com_ptr);

static void     CLR_VARIABLE_TABLE  (struct scl_com *com_ptr);

static void     SCL_GLOBAL_INIT     (struct scl_com *scl_com_ptr);

static void     SCL_GLOBAL_END      (struct scl_com *com_ptr);

static unsigned long    SCL_SOURCE_INIT     (struct scl_com *scl_com_ptr);

/*
 *  Structure Definitions.
 */
#include <shrmsg.h>         /* PSL Messages     */
#include <confirm_respdef.h>        /* Confirm symbols  */
#include <depprmdef.h>          /* Script dependency parameters */
#include <psldiag_prmdef.h>     /* Diagnostics Constants*/
#include <psl_trc_prmdef.h>     /* Tracing parameters       */
#include <pslerr_prmdef.h>      /* Error Constants  */
#include <psl_vm_prmdef.h>      /* Linked list symbols  */
#include <psl_tpaargdef.h>      /* TPARSE Argument Block*/
#include <psl_tpatabdef.h>      /* TPARSE Table     */
#include "scl_cmd.h"            /* Command Line Data Structure */
#include "scl_comdef.h"         /* SCL Common Data */
#include "scl_tpacom.h"         /* SCL Tparse Data */
#include "scl_tbldef.h"         /* SCL Table definitions */
#include <scpdef.h>         /* SCP Data         */

/*----------------------------------------------------------------------*/
/*                                  */
/*    SCL_COMPILE                           */
/*    ============                          */
/*                                  */
/*    Controling module for the script source file compilation.     */
/*                                  */
/*    Calling Format:                           */
/*  SCL_COMPILE (scl_com)                       */
/*                                  */
/*    Input Arguments:                          */
/*  scl_com - Pointer to SCL command data structure         */
/*                                  */
/*    Outputs:                              */
/*  None                                */
/*                                  */
/*    Return Status:                            */
/*  N/A                             */
/*                                  */
/*----------------------------------------------------------------------*/

void        SCL_COMPILE (struct scl_com *com_ptr)

{
    /*
     *    Local Data
     */
    struct   scl_tpa    *tpa_ptr;
    unsigned long       iret;       /* General Status Return    */
    long       confirm;        /* Confirmation value   */
    struct   scl_cmd    *scl_cmd_ptr;   /* Pointer to Command Line Struct */

    /*
     *    Copy pointers for ease of reference.
     */
    scl_cmd_ptr = (*com_ptr).cmd_ptrps;

    /*
     *    ...and perform global initialization
     */
    SCL_GLOBAL_INIT (com_ptr);      /* ...and global initialization */

    /*
     *    Initialize the linked list pointers to zero as we have not yet
     */
    (*com_ptr).srcfil_ctxps = NULL; /* No source file LL yet */
    (*com_ptr).module_ctxps = NULL; /* No module LL yet */
    (*com_ptr).labels_ctxps = NULL; /* No labels LL yet */
    (*com_ptr).variables_ctxps = NULL;  /* No variables LL yet */

    /*
     *    SCL_COMPILE_NEXT
     *    ----------------
     *
     *    We drop to here before the compilation of each source file.
     *
     *    Now pick up and trim down the name of the next file to compile.
     */
scl_compile_next:
    iret = SCL_GETNEXT_FILE (com_ptr);  /* Pick up next file name   */
    if (!(psl_test_status (iret))) goto scl_compile_end;
    /* ...reporting any errors  */

    /*
     *    If we have requested confirmation of compilation then confirm here.
     */
    if ((*scl_cmd_ptr).confirmn)    /* If confirmation requested    */
    {

        /*
         *    SCL_COMPILE_CNF
         *    ---------------
         *
         *    We get here when we are either to confirm or re-confirm compilation of
         *    the next source file.
         */
scl_compile_cnf:
#ifdef  WIN32
        (void) cprintf ("%s, compile? [n]: ", (*com_ptr).scl_namec256);
#else
(void) fprintf ((*com_ptr).ttyfilpf, "%s, compile? [N]: "
        ,(*com_ptr).scl_namec256);
#endif
        confirm = CONFIRM_RESPONSE ();

        /*
         *    Check the response from this routine.
         */
        switch (confirm)
        {
        case CNF_NO:        /* If response = NO     */
            {
                goto scl_compile_next;  /* Branch to not compile this   */
                /***        break;  ***/
            }

        case CNF_QUIT:      /* ...else if response = QUIT   */
            {
                goto scl_compile_end;   /* Branch to quit compilation   */
                /***        break;  ***/
            }

        case CNF_PROMPT:        /* ...else if response = PROMPT */
            {
                goto scl_compile_cnf;   /* ...branch tyo re-prompt  */
                /***        break;  ***/
            }

        case CNF_ALL:       /* ...else if response = ALL    */
            {
                (*scl_cmd_ptr).confirmn = PFALSE;
                /* ...disable confirmation  */
                break;
            }

        default:            /* Default (yes)    */
            {
                break;          /* ...just continue */
            }
        }
    }

    /*
     *    Another Script Source File to Compile, initialize the script specific
     *    data structures.
     */
    iret = SCL_SOURCE_INIT (com_ptr);
    if (!(psl_test_status (iret))) goto scl_compile_end;
    /* ...reporting any errors  */

    /*
     *    Parse the Script Source File
     */
    iret = SCL_TPARSE (com_ptr);    /* ...compile this script.  */
    if (!(psl_test_status (iret)))  /* if we had an error compiling */
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps,
                     1,
                     iret);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        goto end_source_file;
    }

    /*
     *    ...after we have parsed a script, the pointer to the TPARSE data will
     *    have been set in the COMMON data structure. This is where errror counts
     *    etc will be held. We therefore must de-reference this pointer to check
     *    for errors.
     */
    tpa_ptr = (*com_ptr).tpa_ptrps; /* Dereference pointers     */

    /*
     *  If an Object File is required, then if there were no compilation
     *  errors create one Script Object ('TOF') file for each Thread
     *  definition (if source is a thread script) and a single Script Object
     *  ('TOF') file and a dependency file (if source is a script or test script).
     *  If no object file is required or compilation failed, output appropriate message.
     */
    if ((*scl_cmd_ptr).objectn)     /* If we want object */
    {
        if ((*tpa_ptr).error_countl == 0)
        {
            if ((*com_ptr).type_threadn)
            {
                (void) WRT_THRTOF_FILES (com_ptr);
            }
            else
            {
                iret = WRT_SCRTOF_FILE (com_ptr);   /* Write TOF File */
                if (psl_test_status(iret))      /* If successful */
                {
                    iret = WRT_DEP_FILE (com_ptr);  /* ...write dependency file */
                }
            }
        }
        else                /* ...else if errors in script  */
        {
            PSL_PUTDIAG ((*com_ptr).diag_ptrps,
                         1,
                         PSL_ENDNOOBJ,
                         (*com_ptr).scl_namec256,
                         (*tpa_ptr).error_countl);
            PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        }
    }

    /*
     *    Else if no Object File is required and /LOG was specified then
     *    report a message indicating the success or failure of the
     *    compilation.
     */
    else
    {
        if ((*tpa_ptr).error_countl == 0)
        {
            if ((*scl_cmd_ptr).logn)
            {
                PSL_PUTDIAG ((*com_ptr).diag_ptrps,
                             1,
                             PSL_COMPILED,
                             (*com_ptr).scl_namec256);
                /* Set up diagnostic */
                PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
                /* Output diagnostic */
            }
        }
        else
        {
            PSL_PUTDIAG ((*com_ptr).diag_ptrps,
                         1,
                         PSL_ENDDIAGS,
                         (*com_ptr).scl_namec256,
                         (*tpa_ptr).error_countl);
            /* Set up diagnostic */
            PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
            /* Output diagnostic */
        }
    }

    /*
     *  END_SOURCE_FILE:
     *
     *  Compilation complete:
     *  - Close listing file (if any)
     *  - Cancel output of errors to the Listing;
     *  - Close the source file
     *  - Delete all dynamic storage associated with the test script specific structures.
     */
end_source_file:
    ;

    if ((*scl_cmd_ptr).listn)
    {
        iret = LIST_CLOSE ((*com_ptr).lst_contextps);

        (void) PSL_SETUP_OUDIAG (   /* Setup diagnostics output */
            (*com_ptr).diag_ptrps,
            /* For this primary diagnostic  */
            DIOUTY_LIST,
            /* Output is to listing file    */
            NULL,  /* CANCEL output to listing!    */
            0);    /* CANCEL output to listing!    */

    }

    /*
     *  Close the source file.
     */
    SCL_CLOSE_SOURCE (com_ptr,  SRC_UNIT);
    /* Close the source file */
    /*
     *  If this is a test type script, delete all dynamic storage associated with the
     *  test script specific structures.
     */
    if ((*com_ptr).type_testn)
    {
        CLR_TSTTHR_TABLE (com_ptr);
        CLR_TST_TABLE (com_ptr);
    }

    goto scl_compile_next;      /* ...and loop for next     */

    /*
     *    SCL_COMPILE_END
     *    ---------------
     *
     *    We drop through to this point when we have completed the compilation of
     *    all files requested.
     */
scl_compile_end:
    SCL_GLOBAL_END (com_ptr);      /* Global tidy up */
    return;
}

/******************************************************************************
 * @func    This routine deletes all dynamically allocated storage associated
 *      with the test definitions table.
 *
 * @rdesc   Unsigned long PSL routine return status.
 * @parm    Pointer to the scl common data structure.
 */
static void CLR_TST_TABLE (struct scl_com *com_ptr)
{
    /*
     *  Local data.
     */
    struct scl_tpa   *tpa_ptr;      /* Tparse data */
    struct tsttbl    *tst_ptr;      /* Test symbol table entry */
    struct prm   *prm_ptr;      /* Pointer to test parameter data */
    struct tlh   *tlh_ptr;      /* Pointer to test layout header data */
    unsigned long    tst_cnt;       /* Test count */

    /*
     *  Copy pointers for ease of reference.
     */
    tpa_ptr = com_ptr->tpa_ptrps;   /* Tparse data pointer */

    /*
     *  Loop through the test symbol table and free the space associated
     *  with each entry and clear any pointers to it.
     */
    for (tst_ptr = tpa_ptr->tst_begps,
            tst_cnt = 1;
            tst_cnt <= tpa_ptr->tst_cntl;
            tst_ptr++, tst_cnt++)
    {
        prm_ptr = tst_ptr->tst_prm_ptrps;   /* Test parameter data pointer */
        SCL_FREE_PRM (prm_ptr);
        FREE (prm_ptr);
        tst_ptr->tst_prm_ptrps = NULL;

        tlh_ptr = tst_ptr->tst_tlh_ptrps;   /* Test layout header data pointer */
        FREE (tlh_ptr);
        tst_ptr->tst_tlh_ptrps = NULL;
    }

    return;
}

/******************************************************************************
 * @func    This routine deletes all dynamically allocated storage associated
 *      with the test thread definitions table.
 *
 * @rdesc   Unsigned long PSL routine return status.
 * @parm    Pointer to the scl common data structure.
 */
static void CLR_TSTTHR_TABLE (struct scl_com *com_ptr)
{
    /*
     *  Local data.
     */
    struct scl_tpa   *tpa_ptr;      /* Tparse data */
    struct tstthrtbl *tstthr_ptr;       /* Test thread symbol table entry */
    struct tthr  *tthr_ptr;     /* Pointer to test thread data */
    struct tdf   *tdf_ptr;      /* Pointer to thread definition data */
    unsigned long    tstthr_cnt;        /* Test thread count */

    /*
     *  Copy pointers for ease of reference.
     */
    tpa_ptr = com_ptr->tpa_ptrps;   /* Tparse data pointer */

    /*
     *  Loop through the test thread symbol table and free the space associated
     *  with each entry and clear any pointers to it.
     */
    for (tstthr_ptr = tpa_ptr->tstthr_begps,
            tstthr_cnt = 1;
            tstthr_cnt <= tpa_ptr->tstthr_cntl;
            tstthr_ptr++, tstthr_cnt++)
    {
        tdf_ptr = tstthr_ptr->tstthr_tdf_ptrps;
        if (tdf_ptr != NULL)
        {
            SCL_FREE_TDF (tdf_ptr);
            FREE (tdf_ptr);
            tstthr_ptr->tstthr_tdf_ptrps = NULL;
        }

        tthr_ptr = tstthr_ptr->tstthr_tthr_ptrps;
        if (tthr_ptr != NULL)
        {
            FREE (tthr_ptr);
            tstthr_ptr->tstthr_tthr_ptrps = NULL;
        }
    }

    return;
}

/******************************************************************************
 * @func    This routine deletes all dynamically allocated storage associated
 *      with the variables symbol table.
 *
 * @parm    Pointer to the scl common data structure.
 */
static void CLR_VARIABLE_TABLE (struct scl_com *com_ptr)
{
    /*
     *  Local data.
     */
    struct scl_tpa   *tpa_ptr;      /* Tparse data */
    struct symtbl    *symtbl_ptr;       /* Pointer variable symbol table */
    unsigned long    isub;          /* Loop control */

    /*
     *  Copy pointers for ease of reference.
     */
    tpa_ptr = com_ptr->tpa_ptrps;   /* Tparse data pointer */

    /*
     *  Loop through the variables symbol table and free the space associated
     *  with each entry and clear any pointers to it. The only such item at
     *  the moment is the file name associated with file variables.
     */
    symtbl_ptr = (*tpa_ptr).variable_begps;

    for (isub=0; isub < (*tpa_ptr).var_cntl; isub++, symtbl_ptr++)
    {
        if ((*symtbl_ptr).var_filnam_pc != NULL)
        {
            FREE ((*symtbl_ptr).var_filnam_pc);
            (*symtbl_ptr).var_filnam_pc = NULL;
        }
    }
    return;
}

/************************************************************************/
/*                                  */
/*    DEL_DEP_LL - Delete script dependency linked list         */
/*    ==========                            */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  DEL_DEP_LL (com_ptr);                       */
/*                                  */
/*    Where:                                */
/*                                  */
/*  com_ptr    = Pointer to SCL common data.            */
/*                                  */
/*    This routine deletes any existing script dependency linked list   */
/*    - including all associate memory.                 */
/*                                  */
/************************************************************************/
void DEL_DEP_LL (struct scl_com *com_ptr)
{
    /*
     *  Local data.
     */
    struct sdep  *dep_ptr;      /* Pointer to script dependency */
    unsigned  long   iret;          /* PSL routine status */

    /*
     *  If we have a script dependency linked list, delete it.
     */
    if ((*com_ptr).dep_llps != NULL)
    {
        do
        {
            iret = PSL_VM_REMOVE ((*com_ptr).dep_llps, (char *) &dep_ptr,
                                  NULL, NULL, PSL_VM_POS_BEG);
        }
        while (psl_test_status (iret));     /* Loop while not an error */

        (void) PSL_VM_DELETE ((*com_ptr).dep_llps);
        (*com_ptr).dep_llps = NULL;
    }

    return;
}

/*----------------------------------------------------------------------*/
/*                                  */
/*    SCL_GLOBAL_INIT                           */
/*    ===============                           */
/*                                  */
/*    This Subroutine initializes all the variables that are constant   */
/*    over an execution of the Script Compiler. This currently only */
/*    includes the Command and Control characters held in SCL_CMND_CHR  */
/*    and SCL_CTRL_CHR respectively.                    */
/*                                  */
/*    Input arguments:                          */
/*  scl_com - Pointer to SCL common record              */
/*                                  */
/*    Implicit Inputs:                          */
/*  None                                */
/*                                  */
/*    Outputs:                              */
/*  via SCL common record                       */
/*                                  */
/*    Return Status:                            */
/*  N/A                             */
/*                                  */
/*----------------------------------------------------------------------*/
static void SCL_GLOBAL_INIT (struct scl_com *com_ptr)
{
    /*
     *  Local data.
     */
    struct scl_tpa   *tpa_ptr;      /* Tparse data */

    /*
     *  Copy pointer for ease of reference.
     */
    tpa_ptr = (*com_ptr).tpa_ptrps;   /* Tparse data pointer */

    /*
     *   Set the initial Command and Control characters for each compilation.
     */
    (*com_ptr).cmnd_chr_valuec256 [0] = '~';
    (*com_ptr).cmnd_chr_valuec256 [1] = '\0';

    (*com_ptr).ctrl_chr_valuec256 [0] = '^';
    (*com_ptr).ctrl_chr_valuec256 [1] = '\0';

    /* Allocate memory for the BIG TOF line buffer. */
    (*com_ptr).bigtoflinpc = (char *) MALLOC (CHRBUFSIZ);
    if ((*com_ptr).bigtoflinpc == NULL)
    {
        PSL_SIGDIAG ((*tpa_ptr).diag_ptrps, 2, PSL_SYSERVERR,
            "MALLOC (allocating space for the BIG TOF line buffer)", 0);
    }

    /*
     *    Ensure that the context variables are reset
     */
    (*com_ptr).find_contextps = NULL;

    return;
}

/************************************************************************/
/*                                                                      */
/*    SCL_GLOBAL_END                                                    */
/*    ==============                                                    */
/*                                                                      */
/*    This Subroutine performs all the tidying up (e.g. memory          */
/*    deallocation) required after the Script Compiler has completed    */
/*    compilation of all scripts. It tidies up after all the Global     */
/*    initialization performed in SCL_GLOBAL_INIT.                      */
/*                                                                      */
/*    Input arguments:                                                  */
/*    scl_com - Pointer to SCL common record                            */
/*                                                                      */
/*    Implicit Inputs:                                                  */
/*    None                                                              */
/*                                                                      */
/*    Outputs:                                                          */
/*    via SCL common record                                             */
/*                                                                      */
/*    Return Status:                                                    */
/*    N/A                                                               */
/*                                                                      */
/************************************************************************/
static void SCL_GLOBAL_END (struct scl_com *com_ptr)
{
    /* Deallocate memory for the BIG TOF line buffer. */
    FREE ((*com_ptr).bigtoflinpc);
    (*com_ptr).bigtoflinpc = NULL;

    return;
}

/*----------------------------------------------------------------------*/
/*                                  */
/*    S C L _ S O U R C E  _ I N I T                    */
/*    ==============================                    */
/*                                  */
/*    This Subroutine attempts to open the Script Source file and   */
/*    Script Listing file (if required) ready to compile another    */
/*    Script. If either open fails it reports an error and takes the    */
/*    alternate return.                         */
/*                                  */
/*    Input arguments:                          */
/*  com_ptr - Pointer to SCL common record              */
/*                                  */
/*    Implicit Inputs:                          */
/*  None                                */
/*                                  */
/*    Outputs:                              */
/*  via SCL common record                       */
/*                                  */
/*    Return Status:                            */
/*  N/A                             */
/*                                  */
/*----------------------------------------------------------------------*/
static unsigned long SCL_SOURCE_INIT (struct scl_com *com_ptr)
{

    /*
     *   Local Data
     */
    unsigned long iret;         /* Errors from routines     */
    char listing_version [5];  /* Listing Version (Vn.m)   */
    char subheading1 [256];    /* Subheading 1         */
    char subheading2 [256];    /* Subheading 2         */
    static   char rel_file [256];       /* Related Filename */
    static   char *filename_ptr = rel_file; /* "Filename" pointer   */
    struct   scl_cmd *scl_cmd_ptr;      /* Pointer to Command Line Struct */

    /*
     *    Copy pointers for ease of reference.
     */
    scl_cmd_ptr = (*com_ptr).cmd_ptrps;

    /*
     *    Ensure that we are at the base level of the include stack and also that
     *    the file names linked list Id. is clear. We cannot wait until the
     *    INITIALIZE routine to do this as we require the information whilst
     *    processing the source file.
     */
    (*com_ptr).include_depthl = 0;  /* At Base include depth    */

    /*
     *    Open the script source file.
     */
    iret = SCL_OPEN_SOURCE (com_ptr, SRC_UNIT);
    /* Open source file */
    if (!(psl_test_status (iret))) goto end;
    /* ...if error return it.   */
    /*
     *  We must always initialize the variable symbol table list since we
     *  don't know if this is a mode SQL script.
     */
    if ((*com_ptr).variables_ctxps != NULL)
    {               /* If linked list context not NULL */
        iret = PSL_VM_DELETE ((*com_ptr).variables_ctxps);
        /* ...then delete previous LL   */
        (*com_ptr).variables_ctxps = 0; /* ...and clear the pointer */
    }

    iret = PSL_VM_INITIALIZE (sizeof (struct var_symtbl),
                              &(*com_ptr).variables_ctxps);
    /* Allocate new linked list */
    if (!psl_test_status (iret))    /* If error allocating list */
    {
        PSL_PUTDIAG ((*com_ptr).diag_ptrps, 2, PSL_SYSERVERR,
                     "PSL_VM_INITIALIZE", iret);
        PSL_OUTPUT_DIAG ((*com_ptr).diag_ptrps, 0, DIOUFL_LSTERR);
        goto end;           /* ...and branch to exit    */
    }

    (*com_ptr).srcfil_cntl = 1;     /* Show one source file so far  */

    /*
     *    The source file is Level 0. Initialize the source line number for the
     *    main script.
     */
    (*com_ptr).srclinl10 [(*com_ptr).include_depthl] = 0;
    /* Set liune number at this level = 0 */

    /*
     *    If we have a listing file then open it.
     */
    if ((*scl_cmd_ptr).listn)
    {
        GET_OPENSTA_VERSION (listing_version);

        /*
         *    Set up the subheadings.
         */
        (void) sprintf (subheading1, "Command : %.*s",
                        (int) (sizeof (subheading1) - 11),
                        (*scl_cmd_ptr).command_line_value_ptrpc);

        (void) sprintf (subheading2, "SCL File: %.*s",
                        (int) (sizeof (subheading2) - 11),
                        (*com_ptr).scl_namec256);

        /*
         *    Setup the related file name to use.
         *    On UNIX this is the filename without the 'path'
         *    On WINNT this is the filename without the path.
         */
#ifdef  UNIX
        (void) memcpy (rel_file,    /* Use "source" as "related"    */
                       (*com_ptr).scl_namec256,
                       sizeof((*com_ptr).scl_namec256));

        filename_ptr = (char *) strrchr (rel_file, '/');
        /* Search for LAST "/"      */
        if (filename_ptr == NULL)   /* No "/" (i.e. no path)?   */
        {
            filename_ptr = rel_file;    /* ...use the whole thing   */
        }
        else                /* End of path located      */
        {
            filename_ptr += 1;      /* ...step over the last "/"    */
        }
#elif WIN32
(void) memcpy (rel_file,    /* Use "source" as "related"    */
        (*com_ptr).scl_namec256,
        sizeof((*com_ptr).scl_namec256));

        filename_ptr = (char *) strrchr (rel_file, '\\');
        /* Search for LAST "\"      */
        if (filename_ptr == NULL)   /* No "/" (i.e. no path)?   */
    {
        filename_ptr = rel_file;    /* ...use the whole thing   */
    }
    else                /* End of path located      */
    {
        filename_ptr += 1;      /* ...step over the last "\"    */
    }

#else
#error Platform unsupported at line __LINE__ (Setting up related filename)
#endif

        /*
         *    ...and call the listing initialization routine.
         */
        iret = LIST_INIT ((*com_ptr).lst_contextps,
                          "SCL",    /* Utility Name         */
                          ".prt",   /* Default file     */
                          filename_ptr, /* Related file name    */
                          subheading1,  /* Sub-heading 1        */
                          subheading2,  /* Sub-heading 2        */
                          listing_version);

        if (!(psl_test_status (iret))) goto end;
        /* ...if error branch to return */

        /*
         *    Set up the diagnostics facility so that any diagnostics are output
         *    to the file tidily.
         */
        iret = PSL_SETUP_OUDIAG (   /* Setup diagnostics output */
                   (*com_ptr).diag_ptrps,
                   /* For this primary diagnostic  */
                   DIOUTY_LIST,
                   /* Output is to listing file    */
                   (*com_ptr).lst_contextps,
                   /* here's the listing structure */
                   132);  /* and the max output length    */
        if (!(psl_test_status (iret)))
        {
            PSL_SIGDIAG (       /* Put message on stack     */
                (*com_ptr).diag_ptrps,
                /* Use primary diag stk     */
                2,      /* TWO message segments     */
                PSL_SYSERVERR,  /* 1st message          */
                "PSL_SETUP_OUDIAG",
                /* 1st message para P1      */
                iret);      /* 2nd message          */
            goto end;           /* Return error!        */
        }
    }

    return PSL_NORMAL;

    /*
     *    We have had an error. Return the status to the user.
     */
end:
    return iret;
}
