/* Copyright (C) 2000-2001 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
///////////////////////////////////////////////////////////////////////////////
// @doc     TestInit
// @module  TestInit| This command line utility may be used to start, stop and
//          kill tests.
//
///////////////////////////////////////////////////////////////////////////////
#include <vector>
#include <exception>
#include <iostream>
#include <TestManager.h>
#include <OSTAORB/OSTAORB.h>
#include <process.hpp>

using namespace std;

// Global Variables
CORBAServer* corba;         // Pointer to CORBA Server
string TMname;              // Current Test Manager object name

// Function prototypes.
void usage();
TestManager_var GetTM();
bool TestEnded();
bool TestRunning();

void usage()
{
    std::cout << "usage:  TestInit -start -T test_name -S sample_rate" << std::endl <<
                 "        TestInit -stop" << std::endl <<
                 "        TestInit -kill" << std::endl;
}

int main(int argc, char * argv[])
{
    std::string Test="";
	std::string strSampleRate="";

    bool start=false;
    bool stop=false;
    bool kill=false;

    for(int i=1;i<argc;i++)
    {
        if(strcmp(argv[i],"-T")==0)             // Test name
        {
            if(i+1<argc)Test=argv[++i];
        }
		else if(strcmp(argv[i],"-S")==0)        // Sample Rate
        {
            if(i+1<argc)strSampleRate=argv[++i];
        }
        else if(strcmp(argv[i],"-start")==0)    // Want to start a test
        {
            start=true;
        }
        else if(strcmp(argv[i],"-stop")==0)     // Want to stop a test
        {
            stop=true;
        }
        else if(strcmp(argv[i],"-kill")==0)     // Want to kill a test
        {
            kill=true;
        }
        else
        {
            usage();
            return 1;
        }
    }
    try
    {
        // If 'Start Test' requested.
        if (start)
        {
            if (Test.empty())
            {
                std::cout << "A Test name must be specified" << std::endl;
                usage();
                return 1;
            }
			if (strSampleRate.empty())
            {
                std::cout << "A Sample rate in milliseconds must be specified (0 is no sampling)" << std::endl;
                usage();
                return 1;
            }



            // Check that there are no Tests already running.
            corba = new CORBAServer (argc, argv);
            if (TestRunning())
            {
                std::cout << "Cannot start Test - there is a Test already running";
                return 1;
            }

            // Set up command line to execute Test and issue command.
            string exeCmd = "\"";
            exeCmd += getenv("OPENSTA_HOME");
            exeCmd += "\\Engines\\";
            exeCmd +="TestManager\" -T " + Test;
			exeCmd +=" -S " + strSampleRate;
		    std::cout << "Starting Test " << Test;
            CTProcess proc(exeCmd);
      		proc.start();

            // Check Process created..
		    if (!proc.isRunning())
		    {
                std::cout << std::endl << "Failed to create Test Manager process";
			    return 1;
		    }

            // Wait for Test Manager to be registered.with the naming service.
            while (1)
            {
                if (TestRunning()) break;
                ::Sleep(1000);
            }
            std::cout << std::endl << "Test started";
        }

        // Else if "Stop Test" requested.
        else if (stop)
        {
            // Get Test Manager to Stop.
            corba = new CORBAServer (argc, argv);
            TestManager_var TM = GetTM();
            if (TM == NULL) return 1;

            // Issue 'Stop Test'.
		    std::cout << "Stopping Test " << TM->TestName();
            TM->StopTest(false, 30);

            // Wait for Test to die.
            while (1)
            {
                if (TestEnded()) break;
                ::Sleep(1000);
            }
            std::cout << std::endl << "Test stopped";
        }

        // Else if "Kill Test" requested.
        else if (kill)
        {
            // Get Test Manager to Kill.
            corba = new CORBAServer (argc, argv);
            TestManager_var TM = GetTM();
            if (TM == NULL) return 1;

            // Issue 'Kill Test'.
		    std::cout << "Killing Test " << TM->TestName();
            TM->StopTest(true, 0);

            // Wait for Test to die.
            while (1)
            {
                if (TestEnded()) break;
                ::Sleep(1000);
            }
            std::cout << std::endl << "Test killed";
        }

        else
        {
            usage();
            return 1;
        }
    }
    catch (CORBA::Exception &e)
    {
        std::cout<<std::endl<<"Caught CORBA::"<< e;
    }
    catch (CyrException & e)
    {
        std::cout<<std::endl<<"A CyrException "<<e.what()<<"," <<e.where()<<e.status();
    }
    catch (exception& e)
    {
        std::cout<<std::endl<<"An exception "<<e.what();
    }
    catch(...)
    {
        std::cout<<std::endl<<"Unknown";
    }
    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function looks for a Test Manager. If there is only one a
//          pointer to the corresponding CORBA Test Manager object is returned.
//          If there is more than one, the user is prompted to choose one and
//          a pointer to the selected Test Manager object returned. If there
//          are no Test Managers running, a message is displayed on standard
//          output and a NULL pointer returned.
//
// @rdesc   Test Manager object reference; NULL if none.
//
TestManager_var GetTM()
{
    // Get list of running Test Managers.
    TestManager_var TM;
    CORBA::Object_var obj;
    CORBAServer::NameList_t TMList=corba->GetListOfNames(TestManager::CLASS_NAME);
    int tmcnt = TMList.size();  // Test Manager count
    int tmsub;                  // Test Manager subscript in TM list

    // Now select Test Manager:
    // - If there are 0 Test Managers, report message and return.
    // - Else If there is 1 Test Manager, then auto-select it.
    // - Else prompt the user to select a Test Manager.

    if (tmcnt == 0) // If no running Test Managers
    {
        std::cout<<"There are no Test Managers running"; 
        return NULL; // All done
    }
    else if (tmcnt == 1) // If 1 running Test Manager
    {
        tmsub = 0;
    }
    else
    {
        // Display numbered (from 1) list of Test Managers to select from.
        for (std::vector<std::string>::size_type iter=0; iter<TMList.size(); iter++)  
        {
            string tmnam = TMList[iter].Id;   
            obj=corba->FindObjectWithName(TestManager::CLASS_NAME, tmnam.c_str()); 
            TM=TestManager::_narrow(obj.inout());
		    std::cout << iter + 1 << ". " << tmnam <<
                " running Test " << TM->TestName() << std::endl;
        }
        std::cout << "Select Test Manager:";
        int number;
        char num=getc(stdin);
        number=atoi(&num) - 1;
        if (number > tmcnt)
        {
            std::cout << num << " is invalid" << std::endl;
        }
        else
        {
            tmsub = number;
        }
    }

    TMname = TMList[tmsub].Id;   
    obj=corba->FindObjectWithName(TestManager::CLASS_NAME, TMname.c_str()); 
    TM=TestManager::_narrow(obj.inout());
    return TM;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function checks whether the Test whose Test Manager name is
//          held in TMname has completed. It does so by checking whether the
//          Test Manager is still registered with the naming service.
//
// @rdesc   Return Status: true=Test ended; false=Test NOT ended.
//
bool TestEnded()
{
    try
    {
        CORBA::Object_var obj;
        obj=corba->FindObjectWithName(TestManager::CLASS_NAME, TMname.c_str()); 
    }
    catch (...)
    {
        return true;
    }
    return false;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This function is used to check whether there is a Test running. If
//          there is it returns true, otherwise it returns false. The check is
//          done by checking whether there are any Test Managers running.
//
// @rdesc   Test status: true=there is a Test running; false=No Test running.
//
bool TestRunning()
{
    // Get list of running Test Managers.
    TestManager_var TM;
    CORBA::Object_var obj;
    CORBAServer::NameList_t TMList=corba->GetListOfNames(TestManager::CLASS_NAME);

    // Are there any Test Managers?
    if (TMList.size() == 0)
    {
        return false;       // No Test running
    }
    else
    {
        return true;        // At least one Test running
    }
}