/* Copyright (C) 2000-2001 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// RelayInit.cpp: implementation of the CRelayInit class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "RelayInit.h"
#include <CyrHelper.inl>
#include <definition.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CRelayInit::CRelayInit(CORBAServer* pServer, int nTraceLevel)
{
    // Initialise member variables
    m_nTraceLevel = nTraceLevel;
    m_pServer = pServer;

    // Initialise the CallManager
	m_pCallManager = new CCallManager();
	string strSettings = getenv("OPENSTA_HOME");
	strSettings += "Server\\WebSvr.ini";
	m_pCallManager->LoadServerSettings(strSettings);

	m_pCallProcessor = new XMLCallProcessorImpl(m_pServer);
	m_pRelayMap = new XMLRelayMapImpl(m_pServer, m_pCallManager->m_settings);

    // Load all Providers
    LoadHandlers();
}

CRelayInit::~CRelayInit()
{
    // Unload provider modules
    UnloadHandlers();

	delete m_pRelayMap;
	delete m_pCallManager;
	delete m_pCallProcessor;
}

void CRelayInit::Run()
{
    m_pServer->Run( );
}

void CRelayInit::LoadHandlers()
{
	const char* pattern = "XMLVDK???.dll";
    const char* func    = "XMLVDKInit";

    // Calculate path for libraries
    string path = CyrGetModuleName();
    path = path.substr( 0, path.find_last_of( DIR_SLASH ) + 1 );

    // Construct a list of all libraries to load
    NameList lstName = CyrFindFiles( path.c_str(), pattern );

    // If list is empty then log event and return
    if ( lstName.empty() )
    {
        cout<<"No Provider modules found."<<endl;;
        return;
    }

    // Loop through library list and load them all
    for( NameList::iterator i=lstName.begin();i!=lstName.end();i++)
    {
        CTLibrary* lib = NULL;
        CTProcedure* proc = NULL;

        try
        {
            // Load library and locate relevant function
            const string& libName = *i;
            lib = new CTLibrary( libName.c_str() );
            proc = new CTProcedure( *lib, func );

            // Call initialisation function
            PFN_LIB_INIT pfn = (PFN_LIB_INIT)proc->getAddress();
            (*pfn)( m_pServer, m_pCallManager, m_nTraceLevel);

            // Store module handle
            m_lstLib.push_back( lib );
        }
        catch( ... )
        {
            cout<<"Exception caught while loading library"<<endl;
            if ( lib  != NULL )
                delete lib;
        }

        // Tidy memory
        if ( proc != NULL )
            delete proc;
    }

    // Successfully loaded all Provider modules
    cout<<"All Providers loaded and initialised."<<endl;
}

void CRelayInit::UnloadHandlers()
{
    cout<<"All Providers being unloaded."<<endl;

    // Name of the uninitialisation function within the library
    const char* func = "XMLVDKShutdown";

    // Unload all the previously loaded libraries
    for( LibList::iterator i=m_lstLib.begin();i!=m_lstLib.end();i++)
    {
        const CTLibrary* lib = *i;
        CTProcedure* proc = NULL;

        try
        {
            // Find uninitialisation function in library
            proc = new CTProcedure( *lib, func );

            // Call initialisation function
            PFN_LIB_UNINIT pfn = (PFN_LIB_UNINIT)proc->getAddress();
            (*pfn)();
        }
        catch( ... )
        {
            // Only exception should occur is when function cannot
            // be found in library. Build and log error message
            string error( "Error unloading Provider: " );
            error += lib->getName();
            cout<<error.c_str()<<endl;
        }

        // delete procedure
        if ( proc != NULL )
            delete proc;


        delete lib;
    }

    // Empty module vector
    m_lstLib.clear();

    // Output debug message
    cout<<"All Providers unloaded."<<endl;
}