/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#include <iostream>

// Determine the actual layout of characters in the buffer.
template<typename U>
XmlParse<U>::XML_CHARACTERS XmlParse<U>::GetXMLBufferLayout(const BYTE* pBuffer)
{
    XML_CHARACTERS eRet;

    // Examine the first bytes of the input buffer for possible
    // byte-order bytes.
    if(*(DWORD*)pBuffer == 0xFFFEFFFE)
        eRet = UTF32_BIG_ENDIAN;
    else if(*(DWORD*)pBuffer == 0xFEFFFEFF)
        eRet = UTF32_LITTLE_ENDIAN;
    else if(*(WORD*)pBuffer == 0xFFFE)
        eRet = UTF16_BIG_ENDIAN;
    else if(*(WORD*)pBuffer == 0xFEFF)
        eRet = UTF16_LITTLE_ENDIAN;
    else
        eRet = UTF8;

    return eRet;
}

// Swap all bytes in the buffer 2 by 2.
template<typename U>
void XmlParse<U>::ConvertByteOrder(wchar_t* pBuffer)
{
    BYTE *pB;
    BYTE bTmp;
    while(*pBuffer)
    {
        pB    = (BYTE*)pBuffer;
        bTmp  = *pB;
        *pB   = pB[1];
        pB[1] = bTmp;
        ++pBuffer;
    }
}

// Determine whether an UTF-8 character buffer is only composed of basic latin
// characters (values 0-127).
template<typename U>
BOOL XmlParse<U>::IsBasicLatin(const BYTE* pBuffer)
{
    BOOL bRet = TRUE;
    while(*pBuffer && (*pBuffer & 0x80) == 0)
        ++pBuffer;
    if(*pBuffer)
        bRet = FALSE;   // At least one character exceeds 127.

    return bRet;
}

template<typename U>
BOOL XmlParse<U>::NormalizeBuffer(
    BYTE*& pInputBuf, const char* zCharset,
    XML_BUFFER_TYPE& eBufType, BOOL& bNewBuffer, DWORD& dwOffset)
{
    BOOL bRet  = TRUE;
    bNewBuffer = FALSE;
    dwOffset   = 0;
    eBufType   = WIDECHAR_BUFFER;

    XML_CHARACTERS cl = GetXMLBufferLayout(pInputBuf);
    try
    {
        if(cl == UTF8)
        {
            // UTF-8.
            // Check whether all buffer characters are basic Latin (<128).

            // Check for the presence of an encoded UTF-16 byte-order mark.
            if(pInputBuf[0] == 0xef && pInputBuf[1] == 0xbb && pInputBuf[2] == 0xbf
                ||
               pInputBuf[0] == 0xef && pInputBuf[1] == 0xbf && pInputBuf[2] == 0xbe)
            {
                // Skip it.
                dwOffset = 3;
            }

            if(!IsBasicLatin(pInputBuf + dwOffset))
            {
                // UTF-8 with extended characters.
                // A transformation is necessary.
                // Let's apply the encoding.
                UINT uiCodePage = CP_UTF8;  // By default, use the UTF-8 code page.
                char zEncoding[64];
                if(!zCharset)
                {
                    // Try to retrieve the charset from the encoding declaration
                    // in the XML document.
                    XmlParse<char> parser((const char*)(pInputBuf + dwOffset), TRUE);
                    parser.Parse();
                    if(parser.GetEncoding(zEncoding, sizeof zEncoding * sizeof(char)))
                        zCharset = zEncoding;
                }

                // If we have a charset, try to get the corresponding code page.
                if(zCharset)
                {
                    uiCodePage = ParseCharset(zCharset);
                    bRet = uiCodePage != 0xFFFFFFFF;
                }

                if(bRet)
                {
                    // Finally, allocate a buffer and do the transformation.
                    int iSize = MultiByteToWideChar(uiCodePage, 0, (const char*)pInputBuf, -1, NULL, 0);
                    if(iSize)
                    {
                        wchar_t* pNewBuf = new wchar_t[iSize];
                        if(pNewBuf)
                        {
                            if(MultiByteToWideChar(uiCodePage, 0, (const char*)pInputBuf, -1, pNewBuf, iSize))
                            {
                                bNewBuffer = TRUE;
                                pInputBuf = (BYTE*)pNewBuf;
                                cl = GetXMLBufferLayout(pInputBuf);
                            }
                            else
                                bRet = FALSE;   // MultiByteToWideChar failed.
                        }
                        else
                            bRet = FALSE;   // Memory failure.
                    }
                    else
                        bRet = FALSE;   // MultiByteToWideChar failed.
                }
            }
            else
                eBufType = OCTET_BUFFER;
        }

        if(bRet)
        {
            if(cl == UTF16_LITTLE_ENDIAN || cl == UTF16_BIG_ENDIAN)
            {
                // Skip byte-order mark.
                dwOffset = sizeof(wchar_t);

                if(cl == UTF16_BIG_ENDIAN)
                {
                    // UTF-16 big endian. Converting to little endian.
                    ConvertByteOrder((wchar_t*)(pInputBuf + dwOffset));
                }
            }
            else if(cl == UTF32_LITTLE_ENDIAN || cl == UTF32_BIG_ENDIAN)
                bRet = FALSE;   // Unsupported layout.
        }
    }
    catch(const XmlParse<char>::ParsingError&)
    {
        bRet = FALSE;
    }

    return bRet;
}

template<typename U>
void XmlParse<U>::FreeBuf(BYTE* pBuf)
{
    delete [] pBuf;
}

template<typename U>
UINT XmlParse<U>::ParseCharset(const char* zCharset)
{
    UINT uiCodePage = 0xFFFFFFFF;
    struct charsetArrayEntry
    {
        char* zCharset;
        DWORD dwCodePage;
    };
    static const charsetArrayEntry charsetArray[] =
    {
        // charset (key)   code page
        // -------         ---------
        "ISO-8859-2",       28592,
        "ISO-2022-JP",      50220,
        "EUC-JP",           51932,
        "ISO-8859-1",       28591,
        "SHIFT_JIS",        932,
        "ISO-8859-3",       28593,
        "UTF-8",            CP_UTF8,
        "WINDOWS-1252",     1252,
    };
    static DWORD dwcharsetArraySize = sizeof charsetArray / sizeof (*charsetArray);

    for(DWORD dw = 0; dw < dwcharsetArraySize; ++dw)
    {
        if(lstrcmpi(charsetArray[dw].zCharset, zCharset) == 0)
        {
            uiCodePage = charsetArray[dw].dwCodePage;
            break;
        }
    }

    return uiCodePage;
}

template<typename U>
XmlParse<U>::XmlParse(const U* pcInputBuf, BOOL bFetchEncodingOnly)
    : m_pcInputBuf(pcInputBuf), m_pcCur(pcInputBuf),
      m_dwLine(1), m_dwCol(1), m_eContext(CONTEXT_MARKUP),
      m_eLookahead(TOKEN_XML_ERROR), m_dwLookaheadLine(1), m_dwLookaheadCol(1),
      m_dwDepth(0), m_bInStartTag(FALSE), m_bInDTD(FALSE), m_bEntityValue(FALSE),
      m_bFetchEncodingOnly(bFetchEncodingOnly), m_iCounter(0),
      m_dwPITargetBegin(0), m_dwPITargetEnd(0), m_dwPIBodyBegin(0)
{
}

// char specialization:
template<>
XmlParse<char>::String& XmlParse<char>::String::Assign(const char* pcStr, DWORD dwLen)
{
    m_pcStr = pcStr;
    m_dwLen = dwLen;
    return *this;
}

// wchar_t specialization:
template<>
XmlParse<wchar_t>::String& XmlParse<wchar_t>::String::Assign(const wchar_t* pcStr, DWORD dwLen)
{
    m_pcStr = pcStr;
    m_dwLen = dwLen;
    return *this;
}

// Returns the next XML token.
// Always favors longest match.
// Is context-sensitive (depends on parsing context).
template<typename U>
XmlParse<U>::XML_TOKEN XmlParse<U>::GetNextXMLToken()
{
    XML_TOKEN eToken = TOKEN_XML_ERROR;

    // Update the token coordinates.
    m_dwLookaheadLine = m_dwLine;
    m_dwLookaheadCol  = m_dwCol;

    if(m_eContext == CONTEXT_CDATA_SECTION)
    {
        // We are inside a CDATA section, so necessarily inside
        // the character data part of an element content.
        // Match any sequence of characters up to ]]>.
        const U* pcSave = m_pcCur;
        while(*m_pcCur && eToken == TOKEN_XML_ERROR)
        {
            switch(*m_pcCur)
            {
                case 0:
                    // Premature end of buffer.
                    throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);

                case '\n':
                    ++m_dwLine;
                    m_dwCol = 1;    // New line.
                case '\r':
                    ++m_pcCur;
                    break;

                case ']':
                    if(m_pcCur[1] == ']' && m_pcCur[2] == '>')
                    {
                        // Now leaving the CDATA section and re-entering
                        // the element content.
                        m_eContext = CONTEXT_ELEMENT_CONTENT_CHARDATA;

                        eToken = TOKEN_XML_END_CDATA;
                        m_dwCol += 3;
                        m_pcCur += 3;
                        m_lexeme.Assign(pcSave, m_pcCur - pcSave);
                        break;
                    }
                    // Fall through to default:

                default:
                    ++m_dwCol;
                    ++m_pcCur;
                    break;
            }
        }
    }
    else if(m_eContext == CONTEXT_ATTRIBUTE_VALUE)
    {
        // We are inside an attribute value. If m_bInDTD and m_bEntityValue
        // are TRUE, then we are in an entity value, which follows the same
        // rules as the attribute value, except that it can contain parameter
        // entity references (%xxx;), and it should not contain the character
        // '%' instead of '<'.
        // Match the whole value at once, including the surrounding quotes
        // (simple or double), and any embedded reference.
        U cQuote = 0;
        const U* pcValue = NULL;
        while(*m_pcCur && eToken == TOKEN_XML_ERROR)
        {
            switch(*m_pcCur)
            {
                case 0:
                    throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);

                // Whitespace is permitted whithin an attribute value!
                case '\n':
                    ++m_dwLine;
                    m_dwCol = 1;    // New line.
                case '\r':
                    ++m_pcCur;
                    break;

                case '<':
                    if(m_bInDTD && m_bEntityValue)
                    {
                        // Regular character.
                        ++m_pcCur;
                        ++m_dwCol;
                    }
                    else
                        throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                    break;

                case '%':
                    if(!m_bInDTD || !m_bEntityValue)
                    {
                        // Regular character.
                        ++m_pcCur;
                        ++m_dwCol;
                        break;
                    }
                    // else fall through!

                case '&':
                    // Entering an entity or character reference.
                    // Skip it.
                    {
                        // Save the current position in the input buffer.
                        const U* pcSave = m_pcCur;
                        U c;

                        if(*m_pcCur == '&' && m_pcCur[1] == '#')
                        {
                            // A character reference.
                            if(m_pcCur[2] == 'x')
                            {
                                // A hex character reference.
                                m_pcCur += 3;
                                if(*m_pcCur == ';')
                                {
                                    // Empty reference value.
                                    throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                                }
                                for(c = *m_pcCur; c != ';'; c = *++m_pcCur)
                                {
                                    if(c == 0 || !is_hexdigit(c))
                                    {
                                        // Non-digit in char ref.
                                        throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                                    }
                                }
                            }
                            else
                            {
                                // A character reference.
                                m_pcCur += 2;
                                if(*m_pcCur == ';')
                                {
                                    // Empty reference value.
                                    throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                                }
                                for(c = *m_pcCur; c != ';'; c = *++m_pcCur)
                                {
                                    if(c == 0 || !is_digit(c))
                                    {
                                        // Non-digit in char ref.
                                        throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                                    }
                                }
                            }
                        }
                        else
                        {
                            // An entity reference or parameter entity reference.
                            ++m_pcCur;
                            if(*m_pcCur == ';')
                            {
                                // Empty reference value.
                                throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                            }
                            for(c = *m_pcCur; c != ';'; c = *++m_pcCur)
                            {
                                if(c == 0)
                                {
                                    // Incomplete entity ref.
                                    throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                                }
                            }
                        }
                        ++m_pcCur;
                        m_dwCol += m_pcCur - pcSave;
                    }
                    break;

                case '\'':
                case '"':
                    if(cQuote == 0)
                    {
                        cQuote = *m_pcCur;
                        pcValue = m_pcCur + 1;
                    }
                    else
                    {
                        if(*m_pcCur == cQuote)
                        {
                            // End of value.
                            eToken = TOKEN_XML_ATTVALUE;
                            m_eContext = CONTEXT_MARKUP;    // Reset context to markup.
                            m_lexeme.Assign(pcValue, m_pcCur - pcValue);
                        }
                    }
                    // Fall through to default.

                default:
                    ++m_pcCur;
                    ++m_dwCol;
                    break;
            }
        }
    }
    else if(m_eContext == CONTEXT_ELEMENT_CONTENT_CHARDATA)
    {
        // Save the current position in the input buffer.
        const U* pcSave = m_pcCur;

        // We are inside the character data part of an element content,
        // but not in a CDATA section.
        // Match any character sequence, except <, & and ]]>.
        // Note: unlike rule [14], the TOKEN_XML_CHARDATA token represents
        // non-empty character data.
        BOOL bContinue = TRUE;
        while(bContinue && *m_pcCur && eToken == TOKEN_XML_ERROR)
        {
            switch(*m_pcCur)
            {
                case 0:
                    // Premature end of buffer.
                    throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);

                case '\n':
                    ++m_dwLine;
                    m_dwCol = 1;    // New line.
                case '\r':
                    ++m_pcCur;
                    break;

                case '<':
                    // Leaving the character data part of an element content, and
                    // entering markup. If the character data matched is not empty,
                    // then return it, otherwise, resume scanning the markup.
                    m_eContext = CONTEXT_MARKUP;
                    if(m_pcCur != pcSave)
                    {
                        eToken = TOKEN_XML_CHARDATA;
                        m_lexeme.Assign(pcSave, m_pcCur - pcSave);
                    }
                    else
                        bContinue = FALSE;
                    break;

                case ']':
                    if(m_pcCur[1] == ']' && m_pcCur[2] == '>')
                    {
                        // ]]> is not legal here. Signal an error, using current coordinates.
                        throw LexicalError(m_dwLine, m_dwCol);
                    }
                    else
                    {
                        ++m_dwCol;
                        ++m_pcCur;
                    }
                    break;

                case '&':
                    // Entering an entity or character reference.
                    // If the character data matched is not empty, then return it,
                    // otherwise, scan and return the reference.
                    if(m_pcCur != pcSave)
                    {
                        eToken = TOKEN_XML_CHARDATA;
                        m_lexeme.Assign(pcSave, m_pcCur - pcSave);
                    }
                    else
                    {
                        U c;
                        if(m_pcCur[1] == '#')
                        {
                            if(m_pcCur[2] == 'x')
                            {
                                // A hex character reference.
                                m_pcCur += 3;
                                if(*m_pcCur == ';')
                                {
                                    // Empty reference value.
                                    throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                                }
                                for(c = *m_pcCur; c != ';'; c = *++m_pcCur)
                                {
                                    if(c == 0 || !is_hexdigit(c))
                                    {
                                        // Non-digit in char ref.
                                        throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                                    }
                                }
                            }
                            else
                            {
                                // A character reference.
                                m_pcCur += 2;
                                if(*m_pcCur == ';')
                                {
                                    // Empty reference value.
                                    throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                                }
                                for(c = *m_pcCur; c != ';'; c = *++m_pcCur)
                                {
                                    if(c == 0 || !is_digit(c))
                                    {
                                        // Non-digit in char ref.
                                        throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                                    }
                                }
                            }
                            eToken = TOKEN_XML_CHAR_REF;
                        }
                        else
                        {
                            // An entity reference.
                            ++m_pcCur;
                            if(*m_pcCur == ';')
                            {
                                // Empty reference value.
                                throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                            }
                            for(c = *m_pcCur; c != ';'; c = *++m_pcCur)
                            {
                                if(c == 0)
                                {
                                    // Incomplete entity ref.
                                    throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                                }
                            }
                            eToken = TOKEN_XML_ENTITY_REF;
                        }
                        ++m_pcCur;
                        m_dwCol += m_pcCur - pcSave;
                        m_lexeme.Assign(pcSave, m_pcCur - pcSave);
                    }
                    break;

                default:
                    ++m_dwCol;
                    ++m_pcCur;
                    break;
            }
        }
    }

    if(eToken == TOKEN_XML_ERROR && m_eContext == CONTEXT_MARKUP)
    {
        switch(*m_pcCur)
        {
            case 0:
                eToken = TOKEN_XML_END;
                break;

            case '<':
                switch(m_pcCur[1])
                {
                    case '/':
                        eToken = TOKEN_XML_BEGIN_END_TAG;
                        m_dwCol += 2;
                        m_pcCur += 2;
                        break;

                    case '?':
                        if(m_pcCur[2] == 'x' && m_pcCur[3] == 'm' && m_pcCur[4] == 'l'
                            && (  m_pcCur[5] == ' '
                               || m_pcCur[5] == '\t'
                               || m_pcCur[5] == '\r'
                               || m_pcCur[5] == '\n')
                          )
                        {
                            eToken = TOKEN_XML_BEGIN_XMLDECL;
                            m_dwCol += 5;
                            m_pcCur += 5;
                        }
                        else
                        {
                            // Processing instruction. Match up to ?>.

                            // (('X' | 'x') ('M' | 'm') ('L' | 'l'))' ' forbidden (reserved).
                            if(    (m_pcCur[2] == 'x' || m_pcCur[2] == 'X')
                                && (m_pcCur[3] == 'm' || m_pcCur[3] == 'M')
                                && (m_pcCur[4] == 'l' || m_pcCur[4] == 'L')
                                && (  m_pcCur[5] == ' '
                                   || m_pcCur[5] == '\t'
                                   || m_pcCur[5] == '\r'
                                   || m_pcCur[5] == '\n')
                              )
                              throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);

                            m_pcCur += 2;

                            // Save the current position in the input buffer.
                            const U* pcSave = m_pcCur;

                            while(eToken == TOKEN_XML_ERROR)
                            {
                                switch(*m_pcCur)
                                {
                                    case 0:
                                        // Incomplete processing instruction.
                                        throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);

                                    case '?':
                                        if(m_pcCur[1] == '>')
                                        {
                                            // Check that a processing instruction target exists.
                                            BOOL bHasTarget = FALSE;
                                            int iState = 0;
                                            for(const U* pcT = pcSave; pcT < m_pcCur; ++pcT)
                                            {
                                                if(*pcT == ' ' || *pcT == '\t' || *pcT == '\r' || *pcT == '\n')
                                                {
                                                    if(iState == 0)
                                                        iState = 1; // Leading whitespace.
                                                    else if(iState == 2)
                                                    {
                                                        // Whitespace after non-whitespace.
                                                        bHasTarget = TRUE;
                                                        m_dwPITargetEnd = pcT - pcSave;

                                                        // The processing instruction's body starts after the
                                                        // whitespace:
                                                        ++pcT;
                                                        while(*pcT == ' ' || *pcT == '\t' || *pcT == '\r' || *pcT == '\n')
                                                            ++pcT;
                                                        m_dwPIBodyBegin = pcT - pcSave;
                                                        break;
                                                    }
                                                }
                                                else
                                                {
                                                    if(iState == 0 || iState == 1)
                                                    {
                                                        m_dwPITargetBegin = pcT - pcSave;
                                                        iState = 2; // Entering non-whitespace.
                                                    }
                                                }
                                            }

                                            if(!bHasTarget && iState == 2 && pcT == m_pcCur)    // Only a target; nothing after.
                                            {
                                                bHasTarget = TRUE;
                                                m_dwPIBodyBegin = m_dwPITargetEnd = m_pcCur - pcSave;
                                            }

                                            if(!bHasTarget)
                                            {
                                                // Incomplete processing instruction: missing target.
                                                throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                                            }

                                            eToken = TOKEN_XML_PROCESSING_INSTRUCTION;

                                            // We are exiting a processing instruction.
                                            // If we are not in element content, then we remain
                                            // in markup. Otherwise, we enter character data.
                                            if(m_dwDepth)
                                                m_eContext = CONTEXT_ELEMENT_CONTENT_CHARDATA;

                                            m_lexeme.Assign(pcSave, m_pcCur - pcSave);
                                            m_pcCur += 2;
                                            m_dwCol += 2;
                                        }
                                        else
                                        {
                                            ++m_pcCur;
                                            ++m_dwCol;
                                        }
                                        break;

                                    case '\n':
                                        ++m_dwLine;
                                        m_dwCol = 1;    // New line.
                                    case '\r':
                                        ++m_pcCur;
                                        break;

                                    default:
                                        ++m_pcCur;
                                        ++m_dwCol;
                                        break;
                                }
                            }
                        }
                        break;

                    case '!':
                        if(m_pcCur[2] == '-' && m_pcCur[3] == '-')
                        {
                            // A comment. Match up to -->.

                            m_pcCur += 4;

                            // Save the current position in the input buffer.
                            const U* pcSave = m_pcCur;

                            while(eToken == TOKEN_XML_ERROR)
                            {
                                switch(*m_pcCur)
                                {
                                    case 0:
                                        // Incomplete comment.
                                        throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);

                                    case '-':
                                        if(m_pcCur[1] == '-')
                                        {
                                            if(m_pcCur[2] != '>')
                                                throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                                            else
                                            {
                                                eToken = TOKEN_XML_COMMENT;

                                                // We are exiting a comment.
                                                // If we are not in element content, then we
                                                // remain in markup. Otherwise, we enter character data.
                                                if(m_dwDepth)
                                                    m_eContext = CONTEXT_ELEMENT_CONTENT_CHARDATA;

                                                m_lexeme.Assign(pcSave, m_pcCur - pcSave);
                                                m_pcCur += 3;
                                                m_dwCol += 3;
                                            }
                                        }
                                        else
                                        {
                                            ++m_pcCur;
                                            ++m_dwCol;
                                        }
                                        break;

                                    case '\n':
                                        ++m_dwLine;
                                        m_dwCol = 1;    // New line.
                                    case '\r':
                                        ++m_pcCur;
                                        break;

                                    default:
                                        ++m_pcCur;
                                        ++m_dwCol;
                                        break;
                                }
                            }
                        }
                        else
                        {
                            if(m_bInDTD)
                            {
                                if(
                                       m_pcCur[2] == 'E'
                                    && m_pcCur[3] == 'L'
                                    && m_pcCur[4] == 'E'
                                    && m_pcCur[5] == 'M'
                                    && m_pcCur[6] == 'E'
                                    && m_pcCur[7] == 'N'
                                    && m_pcCur[8] == 'T'
                                       )
                                {
                                    eToken = TOKEN_XML_BEGIN_ELEMENT;
                                    m_dwCol += 9;
                                    m_pcCur += 9;
                                }
                                else if(
                                       m_pcCur[2] == 'A'
                                    && m_pcCur[3] == 'T'
                                    && m_pcCur[4] == 'T'
                                    && m_pcCur[5] == 'L'
                                    && m_pcCur[6] == 'I'
                                    && m_pcCur[7] == 'S'
                                    && m_pcCur[8] == 'T'
                                       )
                                {
                                    eToken = TOKEN_XML_BEGIN_ATTLIST;
                                    m_dwCol += 9;
                                    m_pcCur += 9;
                                }
                                else if(
                                       m_pcCur[2] == 'E'
                                    && m_pcCur[3] == 'N'
                                    && m_pcCur[4] == 'T'
                                    && m_pcCur[5] == 'I'
                                    && m_pcCur[6] == 'T'
                                    && m_pcCur[7] == 'Y'
                                       )
                                {
                                    eToken = TOKEN_XML_BEGIN_ENTITY;
                                    m_dwCol += 8;
                                    m_pcCur += 8;
                                }
                                else if(
                                       m_pcCur[2] == 'N'
                                    && m_pcCur[3] == 'O'
                                    && m_pcCur[4] == 'T'
                                    && m_pcCur[5] == 'A'
                                    && m_pcCur[6] == 'T'
                                    && m_pcCur[7] == 'I'
                                    && m_pcCur[8] == 'O'
                                    && m_pcCur[9] == 'N'
                                       )
                                {
                                    eToken = TOKEN_XML_BEGIN_NOTATION;
                                    m_dwCol += 10;
                                    m_pcCur += 10;
                                }
                            }
                            else if(
                                   m_pcCur[2] == 'D'
                                && m_pcCur[3] == 'O'
                                && m_pcCur[4] == 'C'
                                && m_pcCur[5] == 'T'
                                && m_pcCur[6] == 'Y'
                                && m_pcCur[7] == 'P'
                                && m_pcCur[8] == 'E'
                              )
                            {
                                eToken = TOKEN_XML_BEGIN_DOCTYPE;
                                m_dwCol += 9;
                                m_pcCur += 9;
                            }
                            else if(
                                   m_pcCur[2] == '['
                                && m_pcCur[3] == 'C'
                                && m_pcCur[4] == 'D'
                                && m_pcCur[5] == 'A'
                                && m_pcCur[6] == 'T'
                                && m_pcCur[7] == 'A'
                                && m_pcCur[8] == '['
                                   )
                            {
                                eToken = TOKEN_XML_BEGIN_CDATA;
                                m_eContext = CONTEXT_CDATA_SECTION;
                                m_lexeme.Assign(m_pcCur, 9);
                                m_dwCol += 9;
                                m_pcCur += 9;
                            }
                        }
                        break;

                    default:
                        eToken = TOKEN_XML_BEGIN_START_TAG;
                        m_bInStartTag = TRUE;
                        ++m_dwCol;
                        ++m_pcCur;
                        break;
                }
                break;

            case '>':
                // We are exiting a tag or DTD markupdecl.
                if(m_bInDTD)
                {
                    // Exiting a DTD markupdecl.
                    // Keep the markup context.
                }
                else if(m_bInStartTag)
                {
                    // Exiting a start tag: entering character data.
                    m_bInStartTag = FALSE;
                    m_eContext = CONTEXT_ELEMENT_CONTENT_CHARDATA;
                    ++m_dwDepth;
                }
                else
                {
                    // Exiting an end tag: we enter character data if we are still
                    // in the depths of the root element.
                    if(--m_dwDepth)
                        m_eContext = CONTEXT_ELEMENT_CONTENT_CHARDATA;
                }
                eToken = TOKEN_XML_END_TAG;
                ++m_dwCol;
                ++m_pcCur;
                break;

            case '=':
                eToken = TOKEN_XML_EQUAL;
                ++m_dwCol;
                ++m_pcCur;
                break;

            case '/':
                if(m_pcCur[1] == '>')
                {
                    // We are exiting an empty element tag.
                    m_bInStartTag = FALSE;
                    if(m_dwDepth)
                        m_eContext = CONTEXT_ELEMENT_CONTENT_CHARDATA;
                    eToken = TOKEN_XML_END_EMPTY_ELEMENT_TAG;
                    m_dwCol += 2;
                    m_pcCur += 2;
                }
                break;

            case '[':
                if(m_bInDTD)
                {
                    eToken = TOKEN_XML_LEFT_SQUAREBRACKET;
                    ++m_dwCol;
                    ++m_pcCur;
                }
                break;

            case ']':
                if(m_bInDTD)
                {
                    eToken = TOKEN_XML_RIGHT_SQUAREBRACKET;
                    ++m_dwCol;
                    ++m_pcCur;
                }
                break;

            case '(':
                if(m_bInDTD)
                {
                    eToken = TOKEN_XML_LEFT_PARENTHESIS;
                    ++m_dwCol;
                    ++m_pcCur;
                }
                break;

            case ')':
                if(m_bInDTD)
                {
                    eToken = TOKEN_XML_RIGHT_PARENTHESIS;
                    ++m_dwCol;
                    ++m_pcCur;
                }
                break;

            case ';':
                if(m_bInDTD)
                {
                    eToken = TOKEN_XML_SEMICOLON;
                    ++m_dwCol;
                    ++m_pcCur;
                }
                break;

            case '#':
                if(m_bInDTD)
                {
                    eToken = TOKEN_XML_SHARP;
                    ++m_dwCol;
                    ++m_pcCur;
                }
                break;

            case '|':
                if(m_bInDTD)
                {
                    eToken = TOKEN_XML_PIPE;
                    ++m_dwCol;
                    ++m_pcCur;
                }
                break;

            case '*':
                if(m_bInDTD)
                {
                    eToken = TOKEN_XML_STAR;
                    ++m_dwCol;
                    ++m_pcCur;
                }
                break;

            case '?':
                if(m_bInDTD)
                {
                    eToken = TOKEN_XML_QUESTIONMARK;
                    ++m_dwCol;
                    ++m_pcCur;
                }
                break;

            case '+':
                if(m_bInDTD)
                {
                    eToken = TOKEN_XML_PLUS;
                    ++m_dwCol;
                    ++m_pcCur;
                }
                break;

            case ',':
                if(m_bInDTD)
                {
                    eToken = TOKEN_XML_COMMA;
                    ++m_dwCol;
                    ++m_pcCur;
                }
                break;

            case '%':
                if(m_bInDTD)
                {
                    // Beginning of a parameter entity reference.
                    eToken = TOKEN_XML_PERCENT;
                    ++m_dwCol;
                    ++m_pcCur;
                }
                break;

            case '\'':
            case '"':
                if(m_bInDTD)
                {
                    U cQuote = *m_pcCur++;
                    ++m_dwCol;
                    const U* pcSave = m_pcCur;
                    U c;
                    while((c = *m_pcCur) != 0 && c != cQuote)
                    {
                        if(c == '\n')
                        {
                            ++m_dwLine;
                            m_dwCol = 1;
                        }
                        else if(c != '\r')
                            ++m_dwCol;
                        ++m_pcCur;
                    }
                    if(*m_pcCur == 0)
                        throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);    // Incomplete literal.
                    DWORD dwLen = m_pcCur - pcSave;
                    ++m_pcCur;
                    ++m_dwCol;
                    eToken = TOKEN_XML_SYSTEMLITERAL;
                    m_lexeme.Assign(pcSave, dwLen);
                }
                break;

            case '\n':
                ++m_dwLine;
                m_dwCol = 1;    // New line.
            case ' ' :
            case '\t':
            case '\r':
                if(*m_pcCur == ' ' || *m_pcCur == '\t')
                    ++m_dwCol;
                ++m_pcCur;
                {
                    U c;
                    while((c = *m_pcCur) == ' ' || c == '\t' || c == '\r' || c == '\n')
                    {
                        if(c == '\n')
                        {
                            ++m_dwLine;
                            m_dwCol = 1;    // New line.
                        }
                        else if(*m_pcCur != '\r')
                            ++m_dwCol;
                        ++m_pcCur;
                    }
                    eToken = TOKEN_XML_WHITESPACE;
                }
                break;

            default:
                // Match a Name (rule [5]):
                {
                    const U* pcSave = m_pcCur;
                    if(IS_FIRSTNAMECHAR(*m_pcCur))
                    {
                        ++m_pcCur;
                        while(IS_NAMECHAR(*m_pcCur))
                            ++m_pcCur;
                    }
                    if(m_pcCur != pcSave)
                    {
                        eToken = TOKEN_XML_NAME;
                        DWORD dwLen = m_pcCur - pcSave;
                        m_dwCol += dwLen;
                        m_lexeme.Assign(pcSave, dwLen);
                    }
                    else
                        throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                }
                break;
        }
    }

    return eToken;
}

template<typename U>
void XmlParse<U>::Match(XML_TOKEN eToken)
{
    if(m_eLookahead == eToken)
    {
        // Advance to the next token.
        m_eLookahead = GetNextXMLToken();
    }
    else
        throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
}

template<typename U>
void XmlParse<U>::MatchNoAdvance(XML_TOKEN eToken) const
{
    if(m_eLookahead != eToken)
        throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
}

template<typename U>
void XmlParse<U>::StringMatch(const U* zString)
{
    // The input must match zString. We assume that the string cannot be
    // split among lines.
    U c;
    m_dwLookaheadCol  = m_dwCol;
    const U *pcSave = zString;
    while((c = *zString) != 0 && c == *m_pcCur)
    {
        ++zString;
        ++m_pcCur;
    }
    if(*zString)
    {
        // No match!
        throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
    }
    else
    {
        m_dwCol += zString - pcSave;
    }
}

template<typename U>
void XmlParse<U>::Parse(XMLTree* pOutTree) throw(ParsingError)
{
    // The buffer must not be empty or too small.
    if(m_pcInputBuf[0] == 0 || m_pcInputBuf[1] == 0 ||
       m_pcInputBuf[2] == 0 || m_pcInputBuf[3] == 0)
       throw LexicalError(0, 0);

    // Initialize member variables to point to the beginning of the input buffer.
    m_pcCur      = m_pcInputBuf;
    m_dwLine     = m_dwCol = 1;
    m_eContext   = CONTEXT_MARKUP;
    m_eLookahead = TOKEN_XML_ERROR;
    m_dwDepth    = 0;
    m_bInStartTag= FALSE;
    m_iCounter   = 0;

    // Update m_eLookahead and its coordinates to point to the first token in the stream.
    Match(m_eLookahead);

    // Skip any leading whitespace.
    if(m_eLookahead == TOKEN_XML_WHITESPACE)
        m_eLookahead = GetNextXMLToken();

    // Start parsing the XML document.
    document(pOutTree ? &pOutTree->m_root : NULL);

    // We must have reached the end!
    if(!m_bFetchEncodingOnly && m_eLookahead != TOKEN_XML_END)
        throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
}


    ////////////////////////////// Syntax Rules /////////////////////////////////

// *** XML ***

// Rules [1], [27]:
template<typename U>
void XmlParse<U>::document(XMLTree::NodeRoot* pRootNode)
{
    prolog(pRootNode);
    if(m_bFetchEncodingOnly)
        return;

    XMLTree::NodeElement* pRootElement = element(pRootNode);
    if(pRootNode)
        pRootNode->m_pRootElement = pRootElement;

    while( m_eLookahead == TOKEN_XML_COMMENT
        || m_eLookahead == TOKEN_XML_PROCESSING_INSTRUCTION
        || m_eLookahead == TOKEN_XML_WHITESPACE)
    {
        if(pRootNode)
        {
            // Add the token as a child of the root.
            if(m_eLookahead == TOKEN_XML_COMMENT)
                pRootNode->AddChild(new XMLTree::NodeComment(*pRootNode, m_lexeme, ++m_iCounter));
            else if(m_eLookahead == TOKEN_XML_PROCESSING_INSTRUCTION)
                pRootNode->AddChild(new XMLTree::NodeProcessingInstruction(*pRootNode,
                    m_lexeme.AfterPosition(m_dwPIBodyBegin),
                    m_lexeme.Substring(m_dwPITargetBegin, m_dwPITargetEnd),
                    ++m_iCounter));
        }
        m_eLookahead = GetNextXMLToken();
    }
}

// Rules [22], [27]:
template<typename U>
void XmlParse<U>::prolog(XMLTree::NodeRoot* pRootNode)
{
    if(m_eLookahead == TOKEN_XML_BEGIN_XMLDECL)
        XMLDecl();
    else
        m_lexeme.Clear();
    if(m_bFetchEncodingOnly)
        return; // Encoding not found.

    while( m_eLookahead == TOKEN_XML_COMMENT
        || m_eLookahead == TOKEN_XML_PROCESSING_INSTRUCTION
        || m_eLookahead == TOKEN_XML_WHITESPACE)
    {
        if(pRootNode)
        {
            // Add the token as a child of the root.
            if(m_eLookahead == TOKEN_XML_COMMENT)
                pRootNode->AddChild(new XMLTree::NodeComment(*pRootNode, m_lexeme, ++m_iCounter));
            else if(m_eLookahead == TOKEN_XML_PROCESSING_INSTRUCTION)
                pRootNode->AddChild(new XMLTree::NodeProcessingInstruction(*pRootNode,
                    m_lexeme.AfterPosition(m_dwPIBodyBegin),
                    m_lexeme.Substring(m_dwPITargetBegin, m_dwPITargetEnd),
                    ++m_iCounter));
        }
        m_eLookahead = GetNextXMLToken();
    }

    if(m_eLookahead == TOKEN_XML_BEGIN_DOCTYPE)
    {
        doctypedecl();
        while( m_eLookahead == TOKEN_XML_COMMENT
            || m_eLookahead == TOKEN_XML_PROCESSING_INSTRUCTION
            || m_eLookahead == TOKEN_XML_WHITESPACE)
        {
            if(pRootNode)
            {
                // Add the token as a child of the root.
                if(m_eLookahead == TOKEN_XML_COMMENT)
                    pRootNode->AddChild(new XMLTree::NodeComment(*pRootNode, m_lexeme, ++m_iCounter));
                else if(m_eLookahead == TOKEN_XML_PROCESSING_INSTRUCTION)
                    pRootNode->AddChild(new XMLTree::NodeProcessingInstruction(*pRootNode,
                        m_lexeme.AfterPosition(m_dwPIBodyBegin),
                        m_lexeme.Substring(m_dwPITargetBegin, m_dwPITargetEnd),
                        ++m_iCounter));
            }
            m_eLookahead = GetNextXMLToken();
        }
    }
}

// Rules [23], [80], [81], [32]:
template<typename U>
void XmlParse<U>::XMLDecl()
{
    Match(TOKEN_XML_BEGIN_XMLDECL);
    VersionInfo();
    m_eLookahead = GetNextXMLToken();
    if(m_eLookahead == TOKEN_XML_WHITESPACE)
    {
        // Custom scanning for 'encoding':
        if(*m_pcCur == 'e' || *m_pcCur == 's')
        {
            U c, cQuote;
            const U* pcSave;
            if(*m_pcCur == 'e')
            {
                // Must be 'encoding'
                StringMatch(STR("encoding"));
                m_eLookahead = GetNextXMLToken();
                if(m_eLookahead == TOKEN_XML_WHITESPACE)
                    m_eLookahead = GetNextXMLToken();
                MatchNoAdvance(TOKEN_XML_EQUAL);
                if((c = *m_pcCur) == ' ' || c == '\t' || c == '\r' || c == '\n')
                    m_eLookahead = GetNextXMLToken();   // Collect the whitespace.

                // Scan the EncName with its enclosing quotes.
                m_dwLookaheadLine = m_dwLine;
                m_dwLookaheadCol  = m_dwCol;
                pcSave = m_pcCur;
                cQuote = *m_pcCur++;
                if(cQuote != '\'' && cQuote != '"'
                    || !((c = *m_pcCur) >= 'A' && c <= 'Z' || c >= 'a' && c <= 'z'))
                    throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                ++m_pcCur;
                while((c = *m_pcCur) >= 'A' && c <= 'Z' || c >= 'a' && c <= 'z'
                    || c >= '0' && c <= '9' || c == '.' || c == '_' || c == '-')
                    ++m_pcCur;
                if(*m_pcCur != cQuote)
                    throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                ++m_pcCur;
                m_dwCol += m_pcCur - pcSave;

                if(m_bFetchEncodingOnly)
                {
                    // Pick up the encoding name without the quotes.
                    m_lexeme.Assign(pcSave + 1, m_pcCur - pcSave - 2);
                    return;
                }
            }
            else if(m_bFetchEncodingOnly)
            {
                // No encoding.
                m_lexeme.Clear();
                return;
            }

            if((c = *m_pcCur) == 's' || c == ' ' || c == '\t' || c == '\r' || c == '\n')
            {
                if(c != 's')
                    m_eLookahead = GetNextXMLToken();   // Collect the whitespace.
                if(*m_pcCur == 's')
                {
                    // Must be 'standalone'
                    StringMatch(STR("standalone"));

                    m_eLookahead = GetNextXMLToken();
                    if(m_eLookahead == TOKEN_XML_WHITESPACE)
                        m_eLookahead = GetNextXMLToken();
                    MatchNoAdvance(TOKEN_XML_EQUAL);
                    if((c = *m_pcCur) == ' ' || c == '\t' || c == '\r' || c == '\n')
                        m_eLookahead = GetNextXMLToken();   // Collect the whitespace.

                    // Scan the yes/no with its enclosing quotes.
                    m_dwLookaheadLine = m_dwLine;
                    m_dwLookaheadCol  = m_dwCol;
                    pcSave = m_pcCur;
                    cQuote = *m_pcCur++;
                    if(cQuote != '\'' && cQuote != '"')
                        throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                    ++m_pcCur;
                    if(*m_pcCur == 'y' && m_pcCur[1] == 'e' && m_pcCur[2] == 's')
                        m_pcCur += 3;
                    else if(*m_pcCur == 'n' && m_pcCur[1] == 'o')
                        m_pcCur += 2;
                    else
                        throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                    if(*m_pcCur != cQuote)
                        throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
                    ++m_pcCur;
                    m_dwCol += m_pcCur - pcSave;

                    if((c = *m_pcCur) == ' ' || c == '\t' || c == '\r' || c == '\n')
                        m_eLookahead = GetNextXMLToken();   // Collect the whitespace.
                }
            }
        }
        else if(m_bFetchEncodingOnly)
        {
            // No encoding.
            m_lexeme.Clear();
            return;
        }
    }

    // Match the final ?>
    m_dwLookaheadLine = m_dwLine;
    m_dwLookaheadCol  = m_dwCol;
    if(*m_pcCur == '?' && m_pcCur[1] == '>')
    {
        m_dwCol += 2;
        m_pcCur += 2;
        m_eLookahead = GetNextXMLToken();   // Advance.
    }
    else
        throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
}

// Rule [24]:
template<typename U>
void XmlParse<U>::VersionInfo()
{
    // Custom scanning.
    MatchNoAdvance(TOKEN_XML_WHITESPACE);

    // Must be 'version'
    StringMatch(STR("version"));
    m_eLookahead = GetNextXMLToken();
    if(m_eLookahead == TOKEN_XML_WHITESPACE)
        m_eLookahead = GetNextXMLToken();
    MatchNoAdvance(TOKEN_XML_EQUAL);
    U c;
    if((c = *m_pcCur) == ' ' || c == '\t' || c == '\r' || c == '\n')
        m_eLookahead = GetNextXMLToken();   // Collect the whitespace.

    // Scan the VersionNum with its enclosing quotes.
    m_dwLookaheadLine = m_dwLine;
    m_dwLookaheadCol  = m_dwCol;
    const U* pcSave = m_pcCur;
    U cQuote = *m_pcCur++;
    if(cQuote != '\'' && cQuote != '"')
        throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
    ++m_pcCur;
    while((c = *m_pcCur) >= 'A' && c <= 'Z' || c >= 'a' && c <= 'z'
        || c >= '0' && c <= '9' || c == '.' || c == '_' || c == ':')
        ++m_pcCur;
    if(m_pcCur == pcSave || *m_pcCur != cQuote)
        throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
    ++m_pcCur;
    m_dwCol += m_pcCur - pcSave;
}

// Rules [39], [40], [44]:
template<typename U>
XmlParse<U>::XMLTree::NodeElement* XmlParse<U>::element(XMLTree::NodeParent* pParent)
{
    XMLTree::NodeElement* pNewElement;
    Match(TOKEN_XML_BEGIN_START_TAG);
    if(m_eLookahead == TOKEN_XML_NAME)
    {
        String strName(m_lexeme);
        if(pParent)
        {
            pNewElement = new XMLTree::NodeElement(*pParent, strName, ++m_iCounter);
            pParent->AddChild(pNewElement);
        }
        else
            pNewElement = NULL;
        Match(TOKEN_XML_NAME);
        while(m_eLookahead == TOKEN_XML_WHITESPACE)
        {
            m_eLookahead = GetNextXMLToken();
            if(m_eLookahead == TOKEN_XML_NAME)
                Attribute(pNewElement);
            else
                break;
        }
        if(m_eLookahead == TOKEN_XML_END_EMPTY_ELEMENT_TAG)
            m_eLookahead = GetNextXMLToken();
        else if(m_eLookahead == TOKEN_XML_END_TAG)
        {
            m_eLookahead = GetNextXMLToken();
            content(pNewElement);
            ETag(strName);
        }
    }
    else
        throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);

    return pNewElement;
}

// Rule [42]:
template<typename U>
void XmlParse<U>::ETag(const String& strName)
{
    Match(TOKEN_XML_BEGIN_END_TAG);
    if(m_eLookahead == TOKEN_XML_NAME)
    {
        if(m_lexeme != strName)
            throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
        Match(TOKEN_XML_NAME);
        if(m_eLookahead == TOKEN_XML_WHITESPACE)
            m_eLookahead = GetNextXMLToken();
        Match(TOKEN_XML_END_TAG);
    }
    else
        throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
}

// Rules [43], [67], [18], [19], [20], [21]:
template<typename U>
void XmlParse<U>::content(XMLTree::NodeElement* pParent)
{
    String strText;
    if(m_eLookahead == TOKEN_XML_CHARDATA)
    {
        if(pParent)
            strText = m_lexeme;
        m_eLookahead = GetNextXMLToken();
    }
    BOOL bContinue = TRUE;
    do
    {
        switch(m_eLookahead)
        {
            case TOKEN_XML_BEGIN_START_TAG:
                if(pParent && !strText.IsEmpty())
                {
#ifdef XML_IGNORE_WHITESPACE
                    if(!strText.Trim().IsEmpty())
#endif
                    pParent->AddChild(new XMLTree::NodeText(*pParent, strText, ++m_iCounter));
                    strText.Clear();
                }
                element(pParent);
                break;

            case TOKEN_XML_ENTITY_REF:
            case TOKEN_XML_CHAR_REF:
                if(pParent)
                    strText += m_lexeme;
                m_eLookahead = GetNextXMLToken();
                break;

            case TOKEN_XML_COMMENT:
                if(pParent)
                {
                    if(!strText.IsEmpty())
                    {
#ifdef XML_IGNORE_WHITESPACE
                        if(!strText.Trim().IsEmpty())
#endif
                        pParent->AddChild(new XMLTree::NodeText(*pParent, strText, ++m_iCounter));
                        strText.Clear();
                    }
                    pParent->AddChild(new XMLTree::NodeComment(*pParent, m_lexeme, ++m_iCounter));
                }
                m_eLookahead = GetNextXMLToken();
                break;

            case TOKEN_XML_PROCESSING_INSTRUCTION:
                if(pParent)
                {
                    if(!strText.IsEmpty())
                    {
#ifdef XML_IGNORE_WHITESPACE
                        if(!strText.Trim().IsEmpty())
#endif
                        pParent->AddChild(new XMLTree::NodeText(*pParent, strText, ++m_iCounter));
                        strText.Clear();
                    }
                    pParent->AddChild(new XMLTree::NodeProcessingInstruction(*pParent,
                        m_lexeme.AfterPosition(m_dwPIBodyBegin),
                        m_lexeme.Substring(m_dwPITargetBegin, m_dwPITargetEnd),
                        ++m_iCounter));
                }
                m_eLookahead = GetNextXMLToken();
                break;

            case TOKEN_XML_BEGIN_CDATA:
                m_eLookahead = GetNextXMLToken();
                if(m_eLookahead == TOKEN_XML_END_CDATA)
                {
                    if(pParent)
                        strText += m_lexeme;
                    m_eLookahead = GetNextXMLToken();
                }
                else
                    throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
                break;

            default:
                bContinue = FALSE;
                break;
        }

        if(bContinue && m_eLookahead == TOKEN_XML_CHARDATA)
        {
            if(pParent)
                strText += m_lexeme;
            m_eLookahead = GetNextXMLToken();
        }
    } while(bContinue);

    if(pParent && !strText.IsEmpty())
    {
#ifdef XML_IGNORE_WHITESPACE
        if(!strText.Trim().IsEmpty())
#endif
        pParent->AddChild(new XMLTree::NodeText(*pParent, strText, ++m_iCounter));
    }
}

// Rules [41], [25], [10]:
template<typename U>
void XmlParse<U>::Attribute(XMLTree::NodeElement* pParent)
{
    String strName(m_lexeme);
    m_eLookahead = GetNextXMLToken();
    if(m_eLookahead == TOKEN_XML_WHITESPACE)
        m_eLookahead = GetNextXMLToken();

    // Context will need to be set to "attribute value".
    // This requires care.
    if(m_eLookahead == TOKEN_XML_EQUAL)
    {
        U c = *m_pcCur;
        if(c == ' ' || c == '\t' || c == '\r' || c == '\n')
            m_eLookahead = GetNextXMLToken();   // Collect the whitespace.

        if(*m_pcCur == '\'' || *m_pcCur == '"')
        {
            // The next token is the value: set the context!
            m_eContext = CONTEXT_ATTRIBUTE_VALUE;

            // Collect the attribute value.
            m_eLookahead = GetNextXMLToken();
            if(m_eLookahead == TOKEN_XML_ATTVALUE)
            {
                if(pParent)
                {
                    XMLTree::NodeAttribute* pPrevious = pParent->m_attributes.empty() ? NULL : pParent->m_attributes.back();
                    XMLTree::NodeAttribute* pNewAttribute = new XMLTree::NodeAttribute(*pParent, strName, m_lexeme, ++m_iCounter);
                    pParent->m_attributes.push_back(pNewAttribute);
                    if(pPrevious)
                    {
                        pPrevious->m_pFollowingSibling = pNewAttribute;
                        pNewAttribute->m_pPrecedingSibling = pPrevious;
                    }
                }
                m_eLookahead = GetNextXMLToken();
            }
            else
                throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
        }
    }
    else
        throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
}


// *** Document Type Definition ***

// Rule [28]:
template<typename U>
void XmlParse<U>::doctypedecl()
{
    m_bInDTD = TRUE;    // We are now in the DTD.

    Match(TOKEN_XML_BEGIN_DOCTYPE);
    Match(TOKEN_XML_WHITESPACE);
    Match(TOKEN_XML_NAME);
    if(m_eLookahead == TOKEN_XML_WHITESPACE)
    {
        m_eLookahead = GetNextXMLToken();
        if(m_eLookahead == TOKEN_XML_NAME)
        {
            // Must be 'SYSTEM' or 'PUBLIC'
            ExternalID();
            if(m_eLookahead == TOKEN_XML_WHITESPACE)
                m_eLookahead = GetNextXMLToken();
        }
    }

    if(m_eLookahead == TOKEN_XML_LEFT_SQUAREBRACKET)
    {
        m_eLookahead = GetNextXMLToken();
        while(m_eLookahead != TOKEN_XML_RIGHT_SQUAREBRACKET)
        {
            if(  m_eLookahead == TOKEN_XML_WHITESPACE
              || m_eLookahead == TOKEN_XML_PERCENT)
              DeclSep();
            else
                markupdecl();
        }
        Match(TOKEN_XML_RIGHT_SQUAREBRACKET);
        if(m_eLookahead == TOKEN_XML_WHITESPACE)
            m_eLookahead = GetNextXMLToken();
    }

    m_bInDTD = FALSE;    // We are now exiting the DTD.

    if(m_eLookahead == TOKEN_XML_END_TAG)
        m_eLookahead = GetNextXMLToken();
}

// Rules [28a], [69]:
template<typename U>
void XmlParse<U>::DeclSep()
{
    if(m_eLookahead == TOKEN_XML_WHITESPACE)
        m_eLookahead = GetNextXMLToken();
    else
    {
        Match(TOKEN_XML_PERCENT);
        Match(TOKEN_XML_NAME);
        Match(TOKEN_XML_SEMICOLON);
    }
}

// Rule [29]:
template<typename U>
void XmlParse<U>::markupdecl()
{
    switch(m_eLookahead)
    {
        case TOKEN_XML_BEGIN_ELEMENT:
            elementdecl();
            break;
        case TOKEN_XML_BEGIN_ATTLIST:
            AttlistDecl();
            break;
        case TOKEN_XML_BEGIN_ENTITY:
            EntityDecl();
            break;
        case TOKEN_XML_BEGIN_NOTATION:
            NotationDecl();
            break;
        case TOKEN_XML_PROCESSING_INSTRUCTION:
        case TOKEN_XML_COMMENT:
            m_eLookahead = GetNextXMLToken();
            break;
    }
}

// Rules [70], [71], [72]:
template<typename U>
void XmlParse<U>::EntityDecl()
{
    Match(TOKEN_XML_BEGIN_ENTITY);
    Match(TOKEN_XML_WHITESPACE);
    if(m_eLookahead == TOKEN_XML_NAME)
    {
        // Continuation of rule [71]:
        m_eLookahead = GetNextXMLToken();
        if(m_eLookahead != TOKEN_XML_WHITESPACE)
            throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
        if(*m_pcCur == '\'' || *m_pcCur == '"')
        {
            // The next token is an entity value: set the context!
            m_eContext = CONTEXT_ATTRIBUTE_VALUE;
            m_bEntityValue = TRUE;

            // Collect the entity value.
            m_eLookahead = GetNextXMLToken();
        }
        else
        {
            // Skip the whitespace. Next token should be a TOKEN_XML_NAME.
            m_eLookahead = GetNextXMLToken();
        }

        EntityDef();
    }
    else
    {
        // Continuation of rule [72]:
        Match(TOKEN_XML_PERCENT);
        Match(TOKEN_XML_WHITESPACE);
        Match(TOKEN_XML_NAME);
        if(m_eLookahead != TOKEN_XML_WHITESPACE)
            throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
        if(*m_pcCur == '\'' || *m_pcCur == '"')
        {
            // The next token is an entity value: set the context!
            m_eContext = CONTEXT_ATTRIBUTE_VALUE;
            m_bEntityValue = TRUE;

            // Collect the entity value.
            m_eLookahead = GetNextXMLToken();
        }
        PEDef();
    }

    // End of [71] and [72] factored out:
    if(m_eLookahead == TOKEN_XML_WHITESPACE)
        m_eLookahead = GetNextXMLToken();
    Match(TOKEN_XML_END_TAG);
}

// Rules [73], [76]:
template<typename U>
void XmlParse<U>::EntityDef()
{
    if(m_eLookahead == TOKEN_XML_ATTVALUE)
        m_eLookahead = GetNextXMLToken();
    else
    {
        ExternalID();
        if(m_eLookahead == TOKEN_XML_WHITESPACE)
        {
            // Rule [76]:
            m_eLookahead = GetNextXMLToken();
            if(m_eLookahead == TOKEN_XML_NAME)
            {
                if(m_lexeme != STR("NDATA"))
                    throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
                Match(TOKEN_XML_NAME);
                Match(TOKEN_XML_WHITESPACE);
                Match(TOKEN_XML_NAME);
            }
            else
                throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
        }
    }
}

// Rule [74]:
template<typename U>
void XmlParse<U>::PEDef()
{
    if(m_eLookahead == TOKEN_XML_ATTVALUE)
        m_eLookahead = GetNextXMLToken();
    else
        ExternalID();
}

// Rule [45]:
template<typename U>
void XmlParse<U>::elementdecl()
{
    Match(TOKEN_XML_BEGIN_ELEMENT);
    Match(TOKEN_XML_WHITESPACE);
    Match(TOKEN_XML_NAME);
    Match(TOKEN_XML_WHITESPACE);
    contentspec();
    if(m_eLookahead == TOKEN_XML_WHITESPACE)
        m_eLookahead = GetNextXMLToken();
    Match(TOKEN_XML_END_TAG);
}

// Rule [46]:
template<typename U>
void XmlParse<U>::contentspec()
{
    if(m_eLookahead == TOKEN_XML_NAME)
    {
        if(m_lexeme == STR("EMPTY") || m_lexeme == STR("ANY"))
            Match(TOKEN_XML_NAME);
        else
            throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
    }
    else
    {
        Match(TOKEN_XML_LEFT_PARENTHESIS);
        if(m_eLookahead == TOKEN_XML_WHITESPACE)
            m_eLookahead = GetNextXMLToken();
        if(m_eLookahead == TOKEN_XML_SHARP)
        {
            // Continuation of rule [51]:
            m_eLookahead = GetNextXMLToken();
            if(m_eLookahead == TOKEN_XML_NAME)
            {
                if(m_lexeme != STR("PCDATA"))
                    throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
                m_eLookahead = GetNextXMLToken();
            }
            else
                throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);

            // 9-state automaton to match end of rule [51]:
            BOOL bOK = FALSE;
            DWORD dwNextState = 7;  // Start state.
            while(dwNextState)
            {
                switch(dwNextState)
                {
                    case 1:
                        bOK = TRUE;
                        dwNextState = 0;
                        break;

                    case 2:
                        if(m_eLookahead == TOKEN_XML_PIPE)
                            dwNextState = 8;
                        else if(m_eLookahead == TOKEN_XML_RIGHT_PARENTHESIS)
                            dwNextState = 6;
                        else
                            dwNextState = 0;
                        break;

                    case 3:
                        if(m_eLookahead == TOKEN_XML_NAME)
                            dwNextState = 9;
                        else
                            dwNextState = 0;
                        break;

                    case 4:
                        if(m_eLookahead == TOKEN_XML_PIPE)
                            dwNextState = 8;
                        else if(m_eLookahead == TOKEN_XML_RIGHT_PARENTHESIS)
                            dwNextState = 5;
                        else
                            dwNextState = 0;
                        break;

                    case 5:
                        if(m_eLookahead == TOKEN_XML_STAR)
                            dwNextState = 1;
                        else
                            dwNextState = 0;
                        break;

                    case 6:
                        if(m_eLookahead == TOKEN_XML_STAR)
                            dwNextState = 1;
                        else
                        {
                            bOK = TRUE;
                            dwNextState = 0;
                        }
                        break;

                    case 7:
                        switch(m_eLookahead)
                        {
                            case TOKEN_XML_WHITESPACE:
                                dwNextState = 2; break;
                            case TOKEN_XML_RIGHT_PARENTHESIS:
                                dwNextState = 6; break;
                            case TOKEN_XML_PIPE:
                                dwNextState = 8; break;
                            default:
                                dwNextState = 0; break;
                        }
                        break;

                    case 8:
                        if(m_eLookahead == TOKEN_XML_WHITESPACE)
                            dwNextState = 3;
                        else if(m_eLookahead == TOKEN_XML_NAME)
                            dwNextState = 9;
                        else
                            dwNextState = 0;
                        break;

                    case 9:
                        switch(m_eLookahead)
                        {
                            case TOKEN_XML_PIPE:
                                dwNextState = 8; break;
                            case TOKEN_XML_RIGHT_PARENTHESIS:
                                dwNextState = 5; break;
                            case TOKEN_XML_WHITESPACE:
                                dwNextState = 4; break;
                            default:
                                dwNextState = 0; break;
                        }
                        break;
                }

                // Next token.
                if(dwNextState)
                    m_eLookahead = GetNextXMLToken();
            }

            if(!bOK)
                throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
        }
        else
        {
            // Continuation of rule [47]:
            choice_or_seq_continuation();
            switch(m_eLookahead)
            {
                case TOKEN_XML_QUESTIONMARK:
                case TOKEN_XML_STAR:
                case TOKEN_XML_PLUS:
                    m_eLookahead = GetNextXMLToken();
                    break;
            }
        }
    }
}

// Rules [49], [50] after '(' and S?:
template<typename U>
void XmlParse<U>::choice_or_seq_continuation()
{
    cp();

    // 11-state automaton to match end of rule [47], and rules [49], [50]:
    BOOL bOK = FALSE;
    DWORD dwNextState = 7;  // Start state.
    while(dwNextState)
    {
        switch(dwNextState)
        {
            case 1:
                bOK = TRUE;
                dwNextState = 0;
                break;

            case 2:
                switch(m_eLookahead)
                {
                    case TOKEN_XML_PIPE:
                        dwNextState = 8; break;
                    case TOKEN_XML_RIGHT_PARENTHESIS:
                        dwNextState = 1; break;
                    case TOKEN_XML_COMMA:
                        dwNextState = 11; break;
                    default:
                        dwNextState = 0; break;
                }
                break;

            case 3:
                switch(m_eLookahead)
                {
                    case TOKEN_XML_NAME:
                    case TOKEN_XML_LEFT_PARENTHESIS:
                        cp();
                        dwNextState = 9; break;
                    default:
                        dwNextState = 0; break;
                }
                break;

            case 4:
                switch(m_eLookahead)
                {
                    case TOKEN_XML_RIGHT_PARENTHESIS:
                        dwNextState = 1; break;
                    case TOKEN_XML_PIPE:
                        dwNextState = 8; break;
                    default:
                        dwNextState = 0; break;
                }
                break;

            case 5:
                switch(m_eLookahead)
                {
                    case TOKEN_XML_RIGHT_PARENTHESIS:
                        dwNextState = 1; break;
                    case TOKEN_XML_COMMA:
                        dwNextState = 11; break;
                    default:
                        dwNextState = 0; break;
                }
                break;

            case 6:
                switch(m_eLookahead)
                {
                    case TOKEN_XML_NAME:
                    case TOKEN_XML_LEFT_PARENTHESIS:
                        cp();
                        dwNextState = 10; break;
                    default:
                        dwNextState = 0; break;
                }
                break;

            case 7:
                switch(m_eLookahead)
                {
                    case TOKEN_XML_RIGHT_PARENTHESIS:
                        dwNextState = 1; break;
                    case TOKEN_XML_WHITESPACE:
                        dwNextState = 2; break;
                    case TOKEN_XML_PIPE:
                        dwNextState = 8; break;
                    case TOKEN_XML_COMMA:
                        dwNextState = 11; break;
                    default:
                        dwNextState = 0; break;
                }
                break;

            case 8:
                switch(m_eLookahead)
                {
                    case TOKEN_XML_NAME:
                    case TOKEN_XML_LEFT_PARENTHESIS:
                        cp();
                        dwNextState = 9; break;
                    case TOKEN_XML_WHITESPACE:
                        dwNextState = 3; break;
                    default:
                        dwNextState = 0; break;
                }
                break;

            case 9:
                switch(m_eLookahead)
                {
                    case TOKEN_XML_PIPE:
                        dwNextState = 8; break;
                    case TOKEN_XML_WHITESPACE:
                        dwNextState = 4; break;
                    case TOKEN_XML_RIGHT_PARENTHESIS:
                        dwNextState = 1; break;
                    default:
                        dwNextState = 0; break;
                }
                break;

            case 10:
                switch(m_eLookahead)
                {
                    case TOKEN_XML_COMMA:
                        dwNextState = 11; break;
                    case TOKEN_XML_WHITESPACE:
                        dwNextState = 5; break;
                    case TOKEN_XML_RIGHT_PARENTHESIS:
                        dwNextState = 1; break;
                    default:
                        dwNextState = 0; break;
                }
                break;

            case 11:
                switch(m_eLookahead)
                {
                    case TOKEN_XML_NAME:
                    case TOKEN_XML_LEFT_PARENTHESIS:
                        cp();
                        dwNextState = 10; break;
                    case TOKEN_XML_WHITESPACE:
                        dwNextState = 6; break;
                    default:
                        dwNextState = 0; break;
                }
                break;
        }

        // Next token.
        if(dwNextState && dwNextState != 9 && dwNextState != 10)
            m_eLookahead = GetNextXMLToken();
    }

    if(!bOK)
        throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
}

// Rule [48]:
template<typename U>
void XmlParse<U>::cp()
{
    if(m_eLookahead == TOKEN_XML_NAME)
        m_eLookahead = GetNextXMLToken();
    else
    {
        Match(TOKEN_XML_LEFT_PARENTHESIS);
        if(m_eLookahead == TOKEN_XML_WHITESPACE)
            m_eLookahead = GetNextXMLToken();
        choice_or_seq_continuation();
    }

    switch(m_eLookahead)
    {
        case TOKEN_XML_QUESTIONMARK:
        case TOKEN_XML_STAR:
        case TOKEN_XML_PLUS:
            m_eLookahead = GetNextXMLToken();
            break;
    }
}

// Rules [52], [53]:
template<typename U>
void XmlParse<U>::AttlistDecl()
{
    Match(TOKEN_XML_BEGIN_ATTLIST);
    Match(TOKEN_XML_WHITESPACE);
    Match(TOKEN_XML_NAME);
    while(m_eLookahead == TOKEN_XML_WHITESPACE)
    {
        m_eLookahead = GetNextXMLToken();
        if(m_eLookahead == TOKEN_XML_NAME)
        {
            m_eLookahead = GetNextXMLToken();
            Match(TOKEN_XML_WHITESPACE);
            AttType();
            if(m_eLookahead != TOKEN_XML_WHITESPACE)
                throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
            if(*m_pcCur == '\'' || *m_pcCur == '"')
            {
                // The next token is the value: set the context!
                m_eContext = CONTEXT_ATTRIBUTE_VALUE;
            }

            m_eLookahead = GetNextXMLToken();

            DefaultDecl();
        }
    }
    Match(TOKEN_XML_END_TAG);
}

// Rule [60]:
template<typename U>
void XmlParse<U>::DefaultDecl()
{
    if(m_eLookahead == TOKEN_XML_SHARP)
    {
        m_eLookahead = GetNextXMLToken();
        if(m_eLookahead == TOKEN_XML_NAME)
        {
            if(m_lexeme == STR("FIXED"))
            {
                Match(TOKEN_XML_NAME);
                Match(TOKEN_XML_WHITESPACE);
                Match(TOKEN_XML_ATTVALUE);
            }
            else if(m_lexeme != STR("REQUIRED") && m_lexeme != STR("IMPLIED"))
                throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
            else
                Match(TOKEN_XML_NAME);
        }
    }
    else
        Match(TOKEN_XML_ATTVALUE);
}

// Rules [54], [55], [56]:
template<typename U>
void XmlParse<U>::AttType()
{
    if(m_eLookahead == TOKEN_XML_LEFT_PARENTHESIS
        || m_eLookahead == TOKEN_XML_NAME && m_lexeme == STR("NOTATION"))
        EnumeratedType();
    else if(m_eLookahead == TOKEN_XML_NAME)
    {
        // Rules [55], [56]:
        if(    m_lexeme != STR("CDATA")
            && m_lexeme != STR("ID")
            && m_lexeme != STR("IDREF")
            && m_lexeme != STR("IDREFS")
            && m_lexeme != STR("ENTITY")
            && m_lexeme != STR("ENTITIES")
            && m_lexeme != STR("NMTOKEN")
            && m_lexeme != STR("NMTOKENS")
          )
          throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
        m_eLookahead = GetNextXMLToken();
    }
}

// Rules [57], [58], [59]:
template<typename U>
void XmlParse<U>::EnumeratedType()
{
    if(m_eLookahead == TOKEN_XML_LEFT_PARENTHESIS)
    {
        // Rule [59]:
        m_eLookahead = GetNextXMLToken();
        if(m_eLookahead == TOKEN_XML_WHITESPACE)
            m_eLookahead = GetNextXMLToken();
        Match(TOKEN_XML_NAME);          // Less restrictive than Nmtoken (rule [7]).
        while(m_eLookahead == TOKEN_XML_WHITESPACE)
        {
            m_eLookahead = GetNextXMLToken();
            if(m_eLookahead == TOKEN_XML_PIPE)
            {
                Match(TOKEN_XML_PIPE);
                if(m_eLookahead == TOKEN_XML_WHITESPACE)
                    m_eLookahead = GetNextXMLToken();
                Match(TOKEN_XML_NAME);  // Less restrictive than Nmtoken (rule [7]).
            }
            else
                break;  // Can only match a single trailing whitespace.
        }
    }
    else
    {
        // Rule [58]:
        Match(TOKEN_XML_NAME);
        Match(TOKEN_XML_WHITESPACE);
        Match(TOKEN_XML_LEFT_PARENTHESIS);
        if(m_eLookahead == TOKEN_XML_WHITESPACE)
            m_eLookahead = GetNextXMLToken();
        Match(TOKEN_XML_NAME);
        while(m_eLookahead == TOKEN_XML_WHITESPACE)
        {
            m_eLookahead = GetNextXMLToken();
            if(m_eLookahead == TOKEN_XML_PIPE)
            {
                m_eLookahead = GetNextXMLToken();
                if(m_eLookahead == TOKEN_XML_WHITESPACE)
                    m_eLookahead = GetNextXMLToken();
                Match(TOKEN_XML_NAME);
            }
            else
                break;  // Can only match a single trailing whitespace.
        }
    }

    // Factored out for [58] and [59]:
    Match(TOKEN_XML_RIGHT_PARENTHESIS);
}

// Rules [82], [83], [75]:
template<typename U>
void XmlParse<U>::NotationDecl()
{
    Match(TOKEN_XML_BEGIN_NOTATION);
    Match(TOKEN_XML_WHITESPACE);
    Match(TOKEN_XML_NAME);
    Match(TOKEN_XML_WHITESPACE);

    if(m_eLookahead == TOKEN_XML_NAME)
    {
        if(m_lexeme == STR("SYSTEM"))
        {
            m_eLookahead = GetNextXMLToken();
            Match(TOKEN_XML_WHITESPACE);
            Match(TOKEN_XML_SYSTEMLITERAL);
            if(m_eLookahead == TOKEN_XML_WHITESPACE)
                m_eLookahead = GetNextXMLToken();
        }
        else if(m_lexeme == STR("PUBLIC"))
        {
            m_eLookahead = GetNextXMLToken();
            Match(TOKEN_XML_WHITESPACE);
            if(!m_lexeme.is_pubidliteral()) // The next token must be a PubidLiteral as in rule [12].
                throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
            Match(TOKEN_XML_SYSTEMLITERAL);
            if(m_eLookahead == TOKEN_XML_WHITESPACE)
            {
                m_eLookahead = GetNextXMLToken();
                if(m_eLookahead == TOKEN_XML_SYSTEMLITERAL)
                {
                    m_eLookahead = GetNextXMLToken();
                    if(m_eLookahead == TOKEN_XML_WHITESPACE)
                        m_eLookahead = GetNextXMLToken();
                }
            }
        }
        else
            throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
        Match(TOKEN_XML_END_TAG);
    }
}

// Rule [75]:
template<typename U>
void XmlParse<U>::ExternalID()
{
    if(m_lexeme == STR("SYSTEM"))
    {
        Match(TOKEN_XML_NAME);
        Match(TOKEN_XML_WHITESPACE);
        Match(TOKEN_XML_SYSTEMLITERAL);
    }
    else if(m_lexeme == STR("PUBLIC"))
    {
        Match(TOKEN_XML_NAME);
        Match(TOKEN_XML_WHITESPACE);
        if(!m_lexeme.is_pubidliteral()) // The next token must be a PubidLiteral as in rule [12].
            throw LexicalError(m_dwLookaheadLine, m_dwLookaheadCol);
        Match(TOKEN_XML_SYSTEMLITERAL);
        Match(TOKEN_XML_WHITESPACE);
        Match(TOKEN_XML_SYSTEMLITERAL);
    }
    else
        throw SyntaxError(m_dwLookaheadLine, m_dwLookaheadCol);
}
    ////////////////////////////// End of Syntax Rules ///////////////////////////////

#ifdef XML_DUMP

// Specializations:

template<>
void XmlParse<char>::XMLTree::Dump(XmlParse<char>::XmlParseStream& os)
{
    os << "XML Tree:\n<ROOT>(0)\n";
    DumpChildren(m_root.m_children, os, 2);
}
template<>
void XmlParse<wchar_t>::XMLTree::Dump(XmlParse<wchar_t>::XmlParseStream& os)
{
    os << L"XML Tree:\n";
    DumpChildren(m_root.m_children, os, 2);
}
template<>
void XmlParse<char>::XMLTree::NodeComment::Dump(XmlParse<char>::XmlParseStream& os, int iLevel)
{
    Margin(os, iLevel);
    os << "COMMENT(" << m_dwID << "): " << (const XmlParse<char>::XmlString&)m_strValue << std::endl;
}
template<>
void XmlParse<wchar_t>::XMLTree::NodeComment::Dump(XmlParse<wchar_t>::XmlParseStream& os, int iLevel)
{
    Margin(os, iLevel);
    os << L"COMMENT(" << m_dwID << L"): " << (const XmlParse<wchar_t>::XmlString&)m_strValue << std::endl;
}
template<>
void XmlParse<char>::XMLTree::NodeProcessingInstruction::Dump(XmlParse<char>::XmlParseStream& os, int iLevel)
{
    Margin(os, iLevel);
    os << "PROCESSING INSTRUCTION(" << m_dwID << "): " << (const XmlParse<char>::XmlString&)m_strTarget
       << " " << (const XmlParse<char>::XmlString&)m_strValue << std::endl;
}
template<>
void XmlParse<wchar_t>::XMLTree::NodeProcessingInstruction::Dump(XmlParse<wchar_t>::XmlParseStream& os, int iLevel)
{
    Margin(os, iLevel);
    os << L"PROCESSING INSTRUCTION(" << m_dwID << L"): " << (const XmlParse<wchar_t>::XmlString&)m_strTarget
       << L" " << (const XmlParse<wchar_t>::XmlString&)m_strValue << std::endl;
}
template<>
void XmlParse<char>::XMLTree::NodeText::Dump(XmlParse<char>::XmlParseStream& os, int iLevel)
{
    Margin(os, iLevel);
    os << "TEXT(" << m_dwID << "): " << (const XmlParse<char>::XmlString&)m_strValue << std::endl;
}
template<>
void XmlParse<wchar_t>::XMLTree::NodeText::Dump(XmlParse<wchar_t>::XmlParseStream& os, int iLevel)
{
    Margin(os, iLevel);
    os << L"TEXT(" << m_dwID << L"): " << (const XmlParse<wchar_t>::XmlString&)m_strValue << std::endl;
}
template<>
void XmlParse<char>::XMLTree::NodeElement::Dump(XmlParse<char>::XmlParseStream& os, int iLevel)
{
    Margin(os, iLevel);
    os << "ELEMENT(" << m_dwID << "): " << (const XmlParse<char>::XmlString&)m_strValue << "\n";
    if(!m_attributes.empty())
    {
        Margin(os, iLevel + 4);
        os << "ATTRIBUTES:\n";
        for(Attributes::const_iterator pAtt = m_attributes.begin(); pAtt != m_attributes.end(); ++pAtt)
            DumpAttribute(**pAtt, os, iLevel + 8);
    }
    if(!m_children.empty())
    {
        Margin(os, iLevel + 4);
        os << "CHILDREN:\n";
        DumpChildren(m_children, os, iLevel + 8);
    }
}
template<>
void XmlParse<wchar_t>::XMLTree::NodeElement::Dump(XmlParse<wchar_t>::XmlParseStream& os, int iLevel)
{
    Margin(os, iLevel);
    os << L"ELEMENT(" << m_dwID << L"): " << (const XmlParse<wchar_t>::XmlString&)m_strValue << L"\n";
    if(!m_attributes.empty())
    {
        Margin(os, iLevel + 4);
        os << L"ATTRIBUTES:\n";
        for(Attributes::const_iterator pAtt = m_attributes.begin(); pAtt != m_attributes.end(); ++pAtt)
            DumpAttribute(**pAtt, os, iLevel + 8);
    }
    if(!m_children.empty())
    {
        Margin(os, iLevel + 4);
        os << L"CHILDREN:\n";
        DumpChildren(m_children, os, iLevel + 8);
    }
}
template<>
void XmlParse<char>::XMLTree::DumpAttribute(const NodeAttribute& attribute, XmlParse<char>::XmlParseStream& os, int iLevel)
{
    Margin(os, iLevel);
    os << "(" << attribute.m_dwID << ") " << (const XmlParse<char>::XmlString&)attribute.m_strName
       << " = " << (const XmlParse<char>::XmlString&)attribute.m_strValue << std::endl;
}
template<>
void XmlParse<wchar_t>::XMLTree::DumpAttribute(const NodeAttribute& attribute, XmlParse<wchar_t>::XmlParseStream& os, int iLevel)
{
    Margin(os, iLevel);
    os << L"(" << attribute.m_dwID << L") " << (const XmlParse<wchar_t>::XmlString&)attribute.m_strName
       << L" = " << (const XmlParse<wchar_t>::XmlString&)attribute.m_strValue << std::endl;
}
template<>
void XmlParse<char>::XMLTree::DumpChildren(const Children& children, XmlParse<char>::XmlParseStream& os, int iLevel)
{
    for(Children::const_iterator pCh = children.begin(); pCh != children.end(); ++pCh)
        (*pCh)->Dump(os, iLevel);
}
template<>
void XmlParse<wchar_t>::XMLTree::DumpChildren(const Children& children, XmlParse<wchar_t>::XmlParseStream& os, int iLevel)
{
    for(Children::const_iterator pCh = children.begin(); pCh != children.end(); ++pCh)
        (*pCh)->Dump(os, iLevel);
}

#endif
