/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#include "stdafx.h"
#include "TContext.h"
#include "AsyncVirtualUser.h"
#include "TEHttpRecords.h"
#include <malloc.h>


int TRcrdHttpResponse::ParseXPathExpression(const char* zXPathExpr, const string& strName)
{
    int iRet = 0;
    _eContentType = CONTENT_XML;
    _strVariable = strName;
    try
    {
        if(!_xpathExpr.Parse(zXPathExpr))
            iRet = -1;
    }
    catch(const XPathExpression<char>::LexicalError& e)
    {
        ostringstream os;
        iRet = -1;
        os << "Lexical error in XPath expression '"
           << zXPathExpr << "' at ["
           << e.m_dwLineCur << ","
           << e.m_dwColCur  << "], on token starting at ["
           << e.m_dwLineBeg << ","
           << e.m_dwColBeg  << "]."
           << endl;

        _strError = "Load response_info body: " + os.str();
		MSGERR(_strError);
    }
    catch(const XPathExpression<char>::SyntaxError& e)
    {
        ostringstream os;
        iRet = -1;
        os << "Syntax error in XPath expression '"
           << zXPathExpr << "' on token starting at ["
           << e.m_dwLineBeg << ","
           << e.m_dwColBeg  << "]."
           << endl;

        _strError = "Load response_info body: " + os.str();
		MSGERR(_strError);
    }

    return iRet;
}

BOOL TRcrdHttpResponse::ParseWideXPathExpression(
    XPathExpression<wchar_t>& wXPathExpr, const char* zXPathExpr)
{
    BOOL bRet = TRUE;

    // Transform the expression string to UNICODE.
    int iSizeWideChars = MultiByteToWideChar(CP_ACP, 0, zXPathExpr, -1, NULL, 0);
    wchar_t* pcTmp = (wchar_t*)_alloca(iSizeWideChars * sizeof(wchar_t));
    MultiByteToWideChar(CP_ACP, 0, zXPathExpr, -1, pcTmp, iSizeWideChars);

    try
    {
        if(!wXPathExpr.Parse(pcTmp))
            bRet = FALSE;
    }
    catch(const XPathExpression<wchar_t>::LexicalError& e)
    {
        ostringstream os;
        bRet = FALSE;
        os << "Lexical error in XPath expression '"
           << zXPathExpr << "' at ["
           << e.m_dwLineCur << ","
           << e.m_dwColCur  << "], on token starting at ["
           << e.m_dwLineBeg << ","
           << e.m_dwColBeg  << "]."
           << endl;

        _strError = "Load response_info body: " + os.str();
		MSGERR(_strError);
    }
    catch(const XPathExpression<wchar_t>::SyntaxError& e)
    {
        ostringstream os;
        bRet = FALSE;
        os << "Syntax error in XPath expression '"
           << zXPathExpr << "' on token starting at ["
           << e.m_dwLineBeg << ","
           << e.m_dwColBeg  << "]."
           << endl;

        _strError = "Load response_info body: " + os.str();
		MSGERR(_strError);
    }

    return bRet;
}

// Parses an XML buffer, and resolves an XPath expression.
// CAUTION: the buffer passed in as pInputBuffer may be modified in place.
void TRcrdHttpResponse::ParseXMLContent(TContext& ctxt, int nConid, BYTE* pInputBuffer)
{
    if(!_strError.empty())
    {
        if(ctxt.GetVerbose(TRACEMSG))
            ctxt.log(_strError, TRACEMSG);
        CAsyncVirtualUser& VU = dynamic_cast<CAsyncVirtualUser&>(ctxt);
        VU.LogError("ERROR for VU " + VU.GetThreadID() + ": " + _strError);
        return;
    }

    BOOL  bNewBuffer;
    DWORD dwOffset;
    XmlParse<char>::XML_BUFFER_TYPE eBufferType;
    if(XmlParse<char>::NormalizeBuffer(pInputBuffer, NULL, eBufferType, bNewBuffer, dwOffset))
    {
        if(eBufferType == XmlParse<char>::OCTET_BUFFER)
        {
            // Single-byte character set.
            XmlParse<char> parser((const char*)(pInputBuffer + dwOffset));
            XmlParse<char>::XMLTree tree;
            try
            {
                parser.Parse(&tree);
                XPathExpression<char>::XPathExprValue val;
                _xpathExpr.Evaluate(&tree.m_root, val);
                switch(val.m_eType)
                {
                    case XPathExpression<char>::TYPE_NODESET:
                    case XPathExpression<char>::TYPE_BOOLEAN:
                    case XPathExpression<char>::TYPE_NUMBER:
                        XPathExpression<char>::ToString(val);
                        break;

                    case XPathExpression<char>::TYPE_STRING:
                    case XPathExpression<char>::TYPE_UNKNOWN:
                        break;
                }

                if(ctxt.GetVerbose(TRACEMSG))
        	        ctxt.log("HTTPRESPONSE: Name: " + _strVariable + ", value:" + val.m_string, TRACEMSG);

                // Store the results.
		        int ret = ctxt.set_variable(_strVariable, val.m_string);
		        if(ret == -1)
		        {
			        // Error storing the results.
                    string msg = "Set variable failed(" + _strVariable + ')';
			        ctxt.log(msg, ERRMSG);
                    throw mdbParseException(msg);
		        }

            }
            catch(const XmlParse<char>::LexicalError& e)
            {
                ostringstream os;
                os << "lexical error at [" << e.m_dwLine << "," << e.m_dwCol << "]." << endl;
                if(ctxt.GetVerbose(ERRMSG))
                    ctxt.log("HTTPRESPONSE: XML parser error for connection (" + int_to_string(nConid) + "): " + os.str(), ERRMSG);
            }
            catch(const XmlParse<char>::SyntaxError& e)
            {
                ostringstream os;
                os << "syntax error at [" << e.m_dwLine << "," << e.m_dwCol << "]." << endl;
                if(ctxt.GetVerbose(ERRMSG))
                    ctxt.log("HTTPRESPONSE: XML parser error for connection (" + int_to_string(nConid) + "): " + os.str(), ERRMSG);
            }
        }
        else
        {
            // Double-byte character set.
            XmlParse<wchar_t> parser((const wchar_t*)(pInputBuffer + dwOffset));
            XmlParse<wchar_t>::XMLTree tree;
            try
            {
                XPathExpression<wchar_t> wxpathExpr;
                if(ParseWideXPathExpression(wxpathExpr, _xpathExpr.GetExpressionString()))
                {
                    parser.Parse(&tree);
                    XPathExpression<wchar_t>::XPathExprValue val;
                    wxpathExpr.Evaluate(&tree.m_root, val);
                    switch(val.m_eType)
                    {
                        case XPathExpression<wchar_t>::TYPE_NODESET:
                        case XPathExpression<wchar_t>::TYPE_BOOLEAN:
                        case XPathExpression<wchar_t>::TYPE_NUMBER:
                            XPathExpression<wchar_t>::ToString(val);
                            break;

                        case XPathExpression<wchar_t>::TYPE_STRING:
                        case XPathExpression<wchar_t>::TYPE_UNKNOWN:
                            break;
                    }

                    // Transform the value to a monobyte character string.
                    char* pcTmp;
                    BOOL bHeapAllocated = FALSE;
                    if(val.m_eType != XPathExpression<wchar_t>::TYPE_UNKNOWN)
                    {
                        int iSizeBytes = WideCharToMultiByte(CP_ACP, 0, val.m_string, -1, NULL, 0, NULL, NULL);
                        if(iSizeBytes > 2048)
                        {
                            pcTmp = new char[iSizeBytes];
                            bHeapAllocated = TRUE;
                        }
                        else
                            pcTmp = (char*)_alloca(iSizeBytes);
                        WideCharToMultiByte(CP_ACP, 0, val.m_string, -1, pcTmp, iSizeBytes + 1, NULL, NULL);
                    }
                    else
                        pcTmp = "";

                    if(ctxt.GetVerbose(TRACEMSG))
        	            ctxt.log("HTTPRESPONSE: Name: " + _strVariable + ", value:" + pcTmp, TRACEMSG);

                    // Store the results.
		            int ret = ctxt.set_variable(_strVariable, pcTmp);
                    if(bHeapAllocated)
                        delete [] pcTmp;
		            if(ret == -1)
		            {
			            // Error storing the results.
                        string msg = "Set variable failed(" + _strVariable + ')';
			            ctxt.log(msg, ERRMSG);
                        throw mdbParseException(msg);
		            }
                }
                else
                {
                    if(!_strError.empty())
                    {
                        if(ctxt.GetVerbose(ERRMSG))
                            ctxt.log(_strError, ERRMSG);
                        CAsyncVirtualUser& VU = dynamic_cast<CAsyncVirtualUser&> (ctxt);
                        VU.LogError("ERROR for VU " + VU.GetThreadID() + ": " + _strError);
                    }
                }
            }
            catch(const XmlParse<wchar_t>::LexicalError& e)
            {
                ostringstream os;
                os << "lexical error at [" << e.m_dwLine << "," << e.m_dwCol << "]." << endl;
                if(ctxt.GetVerbose(ERRMSG))
                    ctxt.log("HTTPRESPONSE: XML parser error for connection (" + int_to_string(nConid) + "): " + os.str(), ERRMSG);
            }
            catch(const XmlParse<wchar_t>::SyntaxError& e)
            {
                ostringstream os;
                os << "syntax error at [" << e.m_dwLine << "," << e.m_dwCol << "]." << endl;
                if(ctxt.GetVerbose(ERRMSG))
                    ctxt.log("HTTPRESPONSE: XML parser error for connection (" + int_to_string(nConid) + "): " + os.str(), ERRMSG);
            }
        }
    }

    if(bNewBuffer)
        XmlParse<char>::FreeBuf(pInputBuffer);
}
