/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _OPENSTA_ASYNCDEVICE_H_
#define _OPENSTA_ASYNCDEVICE_H_

////////////////////////////////////////////////////////////////////////////////
// Asynchronous I/O device definition file.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

#include "Iocp.hpp"
#include "Trace.hpp"

// This class acts as a base class for I/O devices bound to an I/O completion port.
// The I/Os involved are asynchronous. Concrete I/O devices are to be implemented as
// derivatives of this class. Such devices include files, sockets, mailslots,
// named pipes, directories, consoles, communications resources and disk devices.
// The completion key is the pointer to the IocpEventDispatcher base interface, from
// which this class derives.
// THREAD UNSAFE. This should not be a problem, since an asynchronous device should
// always be accessed by a single thread at a time. The derivatives should be thread
// unsafe as well.
class CAsyncDevice : public IocpEventDispatcher
{
public:
    //// --------------------------------------------------------------------------------
    //// Construction / Destruction.
    //// --------------------------------------------------------------------------------
    CAsyncDevice(CIOCP& iocp, CTrace* pTrace = NULL)
        : m_hDevice(NULL), m_iocp(iocp), m_pEventTarget(NULL),
          m_pTrace(pTrace)
    {ZeroMemory(&m_overlapped, sizeof(m_overlapped));}
    virtual ~CAsyncDevice() {}

    //// --------------------------------------------------------------------------------
    //// User methods.
    //// --------------------------------------------------------------------------------

    // Opens the device. A source name must be specified, along with
    // the necessary attributes. The device is then bound to the IOCP
    // using the BindToIocp() helper, and pEventTarget is set to the
    // current IOCP event target: upon I/O completion, the worker
    // thread involved will call on the IocpEvent interace through
    // pEventTarget.
    // pdwTime, if not NULL, receives an opening time in milliseconds.
    virtual DWORD Open(LPCSTR zResourceName, DWORD dwAttributes, IocpEvent *pEventTarget, DWORD *pdwTime = NULL) = 0;

    // Close the device.
    // Return value: 0 = success, other = GetLastError() failure code.
    virtual DWORD Close() = 0;

    // Performs an asynchronous read.
    // CAUTION: the buffer must remain valid until the call completes!
    // dwBufferLength must contain the length of the buffer.
    // Return value: 0 = success, other = failure.
    virtual DWORD Read(LPSTR zBuffer, DWORD dwBufferLength) = 0;

    // Performs an asynchronous write.
    // CAUTION: the buffer must remain valid until the call completes!
    // dwBufferLength must contain the length of the buffer.If 0, the
    // length will be computed as for a null-terminated string.
    // Return value: 0 = success, other = failure.
    virtual DWORD Write(LPCSTR zBuffer, DWORD dwBufferLength = 0) = 0;

    // Is the device open?
    virtual BOOL IsOpen() = 0;

    // Changes the event target. Returns the previous event target.
    IocpEvent* SetEventTarget(IocpEvent* pEventTarget)
    {
        IocpEvent* pOldTarget = m_pEventTarget;
        m_pEventTarget = pEventTarget;
        return pOldTarget;
    }

    BOOL DidOverlappedIoComplete()  const   { return HasOverlappedIoCompleted(&m_overlapped); }

    //// --------------------------------------------------------------------------------
    //// IocpEventDispatcher methods.
    //// --------------------------------------------------------------------------------
    IocpEvent* GetIocpEventTarget() {return m_pEventTarget;}

protected:
    //// --------------------------------------------------------------------------------
    //// Helpers.
    //// --------------------------------------------------------------------------------

    // Binds the device to the IOCP. Call only after Open() has
    // created a valid device handle.
    // If the binding succeeded, sets pEventTarget as the event target.
    DWORD BindToIocp(HANDLE hDevice, IocpEvent *pEventTarget)
    {
        DWORD dwRes = 0;
        if(!CreateIoCompletionPort(hDevice, m_iocp, (DWORD)(IocpEventDispatcher*)this, 0))
            dwRes = GetLastError();
        else
            m_pEventTarget = pEventTarget;

        return dwRes;
    }

    //// --------------------------------------------------------------------------------
    //// Data.
    //// --------------------------------------------------------------------------------
    HANDLE              m_hDevice;              // The handle of the device.
    CIOCP&              m_iocp;                 // The associated IOCP.
    OVERLAPPED          m_overlapped;           // Overlapped structure for asynchronous I/O.
    IocpEvent          *m_pEventTarget;         // The event target.
    CTrace             *m_pTrace;               // The ostream trace support object.

private:
    //// --------------------------------------------------------------------------------
    //// Forbidden.
    //// --------------------------------------------------------------------------------
    CAsyncDevice();
    void operator=(const CAsyncDevice&);
};

#endif  // _OPENSTA_ASYNCDEVICE_H_
