/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _OPENSTA_ASYNCTCPSOCKET_H_
#define _OPENSTA_ASYNCTCPSOCKET_H_

////////////////////////////////////////////////////////////////////////////////
// Asynchronous TCP socket definition file.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

// Link the winsock library.
#pragma comment(lib, "ws2_32.lib")

#include "AsyncDevice.hpp"
#include "Iocp.hpp"
#include "Fastlock.hpp"
#include <map>

#ifdef USE_PERF_DATA
#   include "PerfCounterKeys.hpp"
#endif


// This class implements an asynchronous socket bound to an I/O completion port.
// THREAD UNSAFE. This should not be a problem, since an asynchronous device should
// always be accessed by a single thread at a time. The derivatives should be thread
// unsafe as well.
class CAsyncTCPSocket : public CAsyncDevice
{
public:
    //// --------------------------------------------------------------------------------
    //// Construction / Destruction.
    //// --------------------------------------------------------------------------------
    CAsyncTCPSocket(CIOCP& iocp, CTrace* pTrace = NULL);
    virtual ~CAsyncTCPSocket();

    //// --------------------------------------------------------------------------------
    //// User methods.
    //// --------------------------------------------------------------------------------

    // Initialize the Winsock library for Winsock 2.
    static DWORD Init()
    {
        s_hostResService.Init();

        OSVERSIONINFOEX vi = {0};
        vi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
        if(!GetVersionEx((OSVERSIONINFO*)&vi))
        {
            // Failed: try with OSVERSIONINFO.
            vi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
            GetVersionEx((OSVERSIONINFO*)&vi);
        }
        else if(vi.dwPlatformId == VER_PLATFORM_WIN32_NT && vi.dwMajorVersion >= 5)
        {
            // This is Windows 2000 or above.
            m_bIsWindows2000 = TRUE;
        }

        WSADATA WsaData;
        WORD wVersionRequested = MAKEWORD(2,2);
#ifdef USE_PERF_DATA
        m_splConnPerSec = g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_CONNECTIONS_PER_SECOND);
#endif
        return WSAStartup(wVersionRequested, &WsaData);
    }

    // Uninitialize the Winsock library.
    static DWORD Uninit()
    {
        s_hostResService.Uninit();
        return (DWORD)WSACleanup();
    }

    // Create and connect the socket. zServer can be a name or IP address.
    // The socket is then bound to the IOCP.
    // If zServer if an IP address syntax, no name resolution is used.
    // pdwTime, if not NULL, receives the TCP connection time.
    // Return value: 0 = success, other = GetLastError() failure code.
    virtual DWORD Open(LPCSTR zServer, DWORD dwPort, IocpEvent *pEventTarget, DWORD *pdwTime = NULL);

    // Disconnect and close the socket.
    // Return value: 0 = success, other = GetLastError() failure code.
    virtual DWORD Close();

    // Asynchronous socket receive operation.
    // CAUTION: the buffer must remain valid until the call completes!
    // Return value: 0 = success, other = WSAGetLastError() failure code.
    virtual DWORD Read(LPSTR zBuffer, DWORD dwBufferLength);
    DWORD Read(LPWSABUF lpBuffers, DWORD dwBufferCount);

    // Asynchronous socket send operation.
    // CAUTION: the buffer must remain valid until the call completes!
    // Return value: 0 = success, other = GetLastError() failure code.
    virtual DWORD Write(LPCSTR zBuffer, DWORD dwBufferLength = 0);
    DWORD Write(LPWSABUF lpBuffers, DWORD dwBufferCount);

    // Is the device open?
    virtual BOOL IsOpen();

    // Is the name an IP address?
    BOOL IsIPAddr(LPCSTR zName) const
    {
        return inet_addr(zName) != INADDR_NONE;
    }

    // Get the remote address to wich we are connected.
    const SOCKADDR_IN &GetRemoteAddr() const {return m_remoteAddr;}

    // Get the remote host name.
    BOOL GetRemoteHostName(LPSTR pHostName, DWORD dwSize) const;

protected:

    //// --------------------------------------------------------------------------------
    //// Definitions.
    //// --------------------------------------------------------------------------------
    class CHostResolution
    {
    public:
        typedef std::map<std::string, unsigned long> StringToAddr;
        typedef std::map<unsigned long, std::string> AddrToString;
        CHostResolution()   { m_pFlName = m_pFlAddr = NULL; }

        void Init()     {   if(m_pFlName == NULL) m_pFlName = new CFastLock();
                            if(m_pFlAddr == NULL) m_pFlAddr = new CFastLock();
                        }
        void Uninit()   {   if(m_pFlName != NULL) delete m_pFlName; m_pFlName = NULL;
                            if(m_pFlAddr != NULL) delete m_pFlAddr; m_pFlAddr = NULL;
                        }

        BOOL ResolveName(LPCSTR zHostName, IN_ADDR& addr);
        BOOL ResolveIPAddress(const IN_ADDR& addr, std::string &hostName);

    private:
        CFastLock     *m_pFlName, *m_pFlAddr;
        StringToAddr  m_NameToAddr;
        AddrToString  m_AddrToName;
    };

    //// --------------------------------------------------------------------------------
    //// Data.
    //// --------------------------------------------------------------------------------
    WSABUF      m_wsaSendBuf;           // Send buffer. For use only by Write(LPCSTR zBuffer, DWORD dwBufferLength).
    WSABUF      m_wsaRecvBuf;           // Receive buffer.

    SOCKADDR_IN m_remoteAddr;           // Stores the address we are currently connected to.
    static LONG* m_splConnPerSec;       // Pointer to performance counter.
    static CHostResolution  s_hostResService;
    static BOOL m_bIsWindows2000;       // TRUE if the host system is Windows 2000 or above.

private:

    //// --------------------------------------------------------------------------------
    //// Forbidden.
    //// --------------------------------------------------------------------------------
    void operator=(const CAsyncTCPSocket&);
};

#endif  // _OPENSTA_ASYNCTCPSOCKET_H_
