/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _ASYNCVIRTUALUSER_H_
#define _ASYNCVIRTUALUSER_H_

#include "SharedPointer.hpp"
#include "TContext.h"
#include "ScriptCache.h"

#include "AsyncHTTPRequestManager.hpp"
#include "TimerQueue.hpp"
#include "TaskGroup.hpp"

#ifndef CYRANO_TP_STANDALONE
#   include "TestExecuter.h"
#   include "CECSemaphore.h"
#endif

class CAsyncVirtualUserController;
class CAsyncFile;


///////////////////////////////////////////////////////////////////////////////
//
// Class provided for use instead of TContextData in the TContext m_ContextDataStack
// in order to allow script inclusion: stores the current status of a context
// before the call to another script, and restores the caller script's context
// when the called script is over.
//
///////////////////////////////////////////////////////////////////////////////

class CScriptContextData : public TContext::TContextData
{
public:
    CScriptContextData(DWORD dwParallelism);

public: // Public because data is public in TContextData.
    TaskScript *    m_pTaskScript;

    // For use by the script execution routines (TScript::run ...).
    DWORD   m_dwScriptLength;

    // For use by the asynchronous TRecord execute methods.
    DWORD   m_dwTRecordCallStep;    // Stores the position in the execution sequence between each call to
                                    // the TRecord::execute method. Must be reset to zero while moving to
                                    // the next TRecord.
private:
    //// Forbidden:
    CScriptContextData();
    CScriptContextData(const CScriptContextData&);
    operator=(const CScriptContextData&);
};

typedef enum  {
                EXEC_INIT,              // First step in processing.
                EXEC_RUNNING,           // Processing pure-CPU TOF objects.
                EXEC_WAIT_MUTEX,        // Waiting for an asynchronous mutex acquisition.
                EXEC_WAIT_SEMAPHORE,    // Waiting for an asynchronous semaphore acquisition.
                EXEC_WAIT_TIMER,        // Waiting for an asynchronous wait to finish.
                EXEC_WAIT_DELAY,        // Waiting for an asynchronous wait to finish (represents a delay before a task).

                EXEC_HTTP_SYNC,         // Synchronise with an HTTP request before moving to next TOF.
                EXEC_ALL_HTTP_SYNC,     // Synchronise with all HTTP requests before moving to next TOF.

                EXEC_WAIT_DISCONNECTS,  // Wait for several asynchronous connections to disconnect before moving to next TOF.

                EXEC_OVER,              // The whole TOF list was successfully processed.
                EXEC_GOTO               // 
              } EVuExecPhase;


class CAsyncVirtualUser : public CAsyncHTTPRequestManager, public TContext, public IocpEvent, public IocpEventDispatcher
{
#define CURR_SCRIPT_CTXT_DATA ((CScriptContextData*)m_ContextDataStack.top())

public:
    CAsyncVirtualUser();
    ~CAsyncVirtualUser();

    virtual DWORD Init(CThreadPoolController& controller, CTimerQueue* pTimerQueue, 
                            DWORD dwManagerNumber, DWORD dwParallelism, CTrace* pTrace = NULL);

    static  BOOL    InitClass();
    static  void    UninitClass();

    void Initialise(SharedPtr<TaskGroupDefinition>& pTaskGroup);
    virtual void InitContextData() {}

    // Callbacks for use by CAsyncRequestManager derivatives.
    virtual void OnRequestFinished(DWORD dwRequestSlotNumber);      // Called when request is completely over. Mind the re-entering issues.
    virtual void OnRequestCancelled(DWORD dwRequestSlotNumber);
    virtual void OnRequestClosed(DWORD dwRequestNumber);

    // To process IOCP events.
    virtual void OnEvent      (DWORD /*dwNumberOfBytesTransferred*/, BOOL /*bIsIo*/, DWORD dwEventType);
    virtual void OnEventFailed(DWORD dwNumberOfBytesTransferred, BOOL bIsIo, DWORD dwEventType, DWORD dwError);

    ETofResult  ProcessTof();

    //// For use by the script executer.

#ifndef CYRANO_TP_STANDALONE
    BOOL    ManageStepByStep();
    BOOL    StepByStep_NotifyNewRequest(const char* zScriptName, long lLineNumber, long lTaskGroupID, DWORD dwUser, long lConId, const char* zRequest);
    void    StepByStep_NotifyComplete(CAsyncHTTPRequest* pRequest);
    void    StepByStep_Next();
#endif
    CAsyncHTTPRequest*  ObtainRequestStruct(DWORD dwConnectionId);    // Books a new request to execute the requests for dwConnectionId.
                                                                // If dwConnectionId has already been allocated, return the existing one.
    DWORD   DisconnectClient(DWORD dwConnectionId);
    DWORD   DisconnectAllClients();

    void    ReleaseRequest(DWORD dwConnectionId);
    void    ReleaseRequest(CAsyncHTTPRequest* pReq);

    // Asynchronous ops.
    BOOL    SynchroniseHTTPClient(DWORD dwClientId);    // Return TRUE if all the client is already over.
    BOOL    SynchroniseAllHTTPClients();                // Return TRUE if all the clients are already over.
    BOOL    StartEventTimer(DWORD dwNbMilliSeconds);

    // Mutexes and semaphores.
    void    SetMutexStartWaitTime(time_t startTime)         { m_mutexStartWaitTime = startTime; }
    time_t  GetMutexStartWaitTime()                 const   { return m_mutexStartWaitTime;      }

    void    SetSemaphoreStartWaitTime(time_t startTime)         { m_semaphoreStartWaitTime = startTime; }
    time_t  GetSemaphoreStartWaitTime()                 const   { return m_semaphoreStartWaitTime;      }
    BOOL    TryAcquireSemaphore(const std::string &s_name, unsigned l_scope);

    BOOL    MustDumpStats() const;
    void    OnNoteText(const std::string& strNote);

    void    CancelQuery(DWORD dwConnectionId);

    CTrace* GetTrace()  { return m_pTrace; }
    void LogError(const std::string& strError) const;

    int CallScript(const std::string& name);

    void    SetParamList(const TtValList &paramList)    { m_paramsInfo = paramList; }
    int     Entry(const TtValList &parameters);

    // TODO: This service should belong to TContext.
    CECVar *GetVariablePointer(const std::string &szVarName) { std::string szTempVarName = szVarName; return CURR_SCRIPT_CTXT_DATA->varlist.get_variable_ptr(szTempVarName); }

    // Setting, retrieving and freeing authentication info:
    void SetAuthenticationInfo(const CredHandle& hCred, const CtxtHandle& hContext, DWORD dwMaxToken)
    {
        FreeAuthenticationInfo();   // Free previous resources if needed.
        m_hCred      = hCred;
        m_hContext   = hContext;
        m_dwMaxToken = dwMaxToken;
        m_bHasAuthenticationInfo = TRUE;
    }
    void GetAuthenticationInfo(CredHandle& hCred, CtxtHandle& hContext, DWORD& dwMaxToken)
    {
        hCred      = m_hCred;
        hContext   = m_hContext;
        dwMaxToken = m_dwMaxToken;
    }
    void FreeAuthenticationInfo()
    {
        if(m_bHasAuthenticationInfo)
        {
            CAsyncSecureSocket::GetSecurityFunctionTable()->DeleteSecurityContext(&m_hContext);
            CAsyncSecureSocket::GetSecurityFunctionTable()->FreeCredentialsHandle(&m_hCred);
            ZeroMemory(&m_hCred, sizeof(m_hCred));
            ZeroMemory(&m_hContext, sizeof(m_hContext));
            m_dwMaxToken = 0;
            m_bHasAuthenticationInfo = FALSE;
        }
    }

    //// For use by the script execution routines (TScript::run ...).

    void    SetScriptLength(DWORD dwScriptLength)           { CURR_SCRIPT_CTXT_DATA->m_dwScriptLength = dwScriptLength; }
    DWORD   GetScriptLength()                       const   { return CURR_SCRIPT_CTXT_DATA->m_dwScriptLength; }

    void    SetScriptPos(DWORD dwScriptPos);
    DWORD   GetScriptPos();
    const std::string &  GetCurrScriptName()    const   { return CURR_SCRIPT_CTXT_DATA->m_pTaskScript->GetName();   }
    TaskGroupDefinition::LogLev GetLoggingLevel() const   { return m_pTaskGroup->m_iLoggingLevel; }
    int     GetTaskGroupID()  const {return m_pTaskGroup->m_iTaskGroupNumber;}
    DWORD   GetExternalVUID() const {return m_dwExternalVUID;}
    long    GetCurTaskIter()  const {return m_lCurTaskIteration;}
    const std::string& GetNoteText() const {return m_noteText;}
    DWORD   GetHTTPBytesIn()  const {return m_dwHTTPBytesIn;}
    DWORD   GetHTTPBytesOut() const {return m_dwHTTPBytesOut;}
    DWORD   GetNumRequests()  const {return m_dwRequests;}
    DWORD   GetNumFailedRequests() const {return m_dwFailedRequests;}
    DWORD   GetTaskIterationDuration() const;
    void    GetLatencyTimes(DWORD& dwTotalLatency, DWORD& dwLatencyCount,
                DWORD& dwMinLatency, DWORD& dwMaxLatency) const
    {
        dwTotalLatency = m_dwTotalLatency;
        dwLatencyCount = m_dwLatencyCount;
        dwMinLatency   = m_dwMinLatency;
        dwMaxLatency   = m_dwMaxLatency;
    }
    void OnNewConnection(DWORD dwConnectionTime);

    // For use by the asynchronous TRecord execute methods so that they know what to do when called again.
    void    SetTRecordCallStep(DWORD dwCS)          { CURR_SCRIPT_CTXT_DATA->m_dwTRecordCallStep = dwCS;   }
    DWORD   GetTRecordCallStep()            const   { return CURR_SCRIPT_CTXT_DATA->m_dwTRecordCallStep;   }


    //// Data storing temporary info. for async. TOF objects, avoiding to compute the value several times
    //// if that object is processed in several steps.
    void                SetHeaderText(const std::string &szHeaderText)                      { m_szHeaderText = szHeaderText; }
    const std::string & GetHeaderText()                                             const   { return m_szHeaderText; }

    void                SetRemoteServerName(const std::string &szRemoteServerName)          { m_szRemoteServerName = szRemoteServerName; }
    const std::string & GetRemoteServerName()                                       const   { return m_szRemoteServerName; }

    void                SetRemoteServerPort(const std::string &szRemoteServerPort)          { m_szRemoteServerPort = szRemoteServerPort; }
    const std::string & GetRemoteServerPort()                                       const   { return m_szRemoteServerPort; }

    void                SetCurrentUri(const std::string &szCurrentUri)                      { m_szCurrentUri = szCurrentUri; }
    const std::string & GetCurrentUri()                                             const   { return m_szCurrentUri; }

    void                SetProxyServerName(const std::string &szProxyServerName)            { m_szProxyServerName = szProxyServerName; }
    const std::string & GetProxyServerName()                                        const   { return m_szProxyServerName; }

    void                SetProxyServerPort(DWORD dwProxyServerPort)         { m_dwProxyServerPort = dwProxyServerPort; }
    DWORD               GetProxyServerPort()                        const   { return m_dwProxyServerPort; }

    void    SetSecureModeUsed(BOOL bSecureMode)         { m_bSecureModeUsed = bSecureMode; }
    BOOL    GetSecureModeUsed()                 const   { return m_bSecureModeUsed; }

    void    SetCurrentConid(DWORD dwConid)              { m_dwCurrentConid = dwConid; }
    DWORD   GetCurrentConid()                   const   { return m_dwCurrentConid; }

    EVuExecPhase    GetStatus() const       { return m_execPhase; }
    ULONG           GetStartTime()  const   { return m_dwStartTime; }

    // User-defined IOCP events.
    enum { EVENT_TYPE_VIRTUAL_USER_START    = 0x8000,   // Wake-up VUs.
           EVENT_TYPE_VIRTUAL_USER_RESUME   = 0x8001    // Received Next call in steap-by-step mode.
         };

    //// --------------------------------------------------------------------------------
    //// IocpEventDispatcher methods.
    //// --------------------------------------------------------------------------------
    IocpEvent* GetIocpEventTarget() {return this;}

    //// --------------------------------------------------------------------------------
    //// TContext virtual methods.
    //// --------------------------------------------------------------------------------
    virtual void    CleanUp();
    virtual void    SetNoteText(const std::string& msg);
    virtual std::string get_test_name();
    virtual ETofResult StartScript(const std::string& fullscriptname);
    virtual ETofResult EndScript(ETofResult res);
    virtual int     ExitCommand(bool keepalive, const std::string& retvar);
    virtual int     TCterminate(int result, const std::string& desc);
    virtual std::string getThreadVariable(const std::string& strName, const std::string& strIndex) const;
    virtual int     setThreadVariable(const std::string& strName, const std::string& strIndex, const std::string& strValue);

    virtual void    OnFileIoOver()  { ProcessTof(); }
    virtual void    OnFileIoFailed(){ ProcessTof(); }


protected:

    //// --------------------------------------------------------------------------------
    //// Helpers.
    //// --------------------------------------------------------------------------------

    CAsyncRequest*      GetNotConnectedRequest();
    CAsyncHTTPRequest*  GetRequestFromConnId(DWORD dwConnectionId) const;

    virtual void OnAllRequestsFinished()  {/*Neutralize CAsyncRequestManager's one, which is based only on the requests instead of the TOF list.*/}
    virtual void OnAllRequestsCancelled() {/*Neutralize CAsyncRequestManager's one, which is based only on the requests instead of the TOF list.*/}

    // To process timer events (called by OnEvent).
    virtual void OnTimerFinished();
    virtual void OnTimerCancelled();

    // Called by OnEvent.
    virtual void OnStartProcessing();

    ETofResult ReStartScript();
    BOOL    CreateTask(const std::string &szFileName);
    ETofResult BeginTask(const std::string &szFileName);
    void    OnStartTaskIteration();
#ifndef CYRANO_TP_STANDALONE
    void    ExitTaskGroup(TaskGroupStopReason eReason);
#else
    void    ExitTaskGroup();
#endif
    void    ProcessEndOfRequest(DWORD dwRequestSlotNumber, BOOL bWasCancelled);

protected:

    //// --------------------------------------------------------------------------------
    //// Data.
    //// --------------------------------------------------------------------------------

    DWORD           m_dwExternalVUID;       // External VU ID.
    EVuExecPhase    m_execPhase;
    BOOL            m_bFirstTaskIteration;
    typedef std::map<std::string, std::string> ThreadVariables_t;
    ThreadVariables_t m_threadVariables;    // Container for THREAD variable values.
    SharedPtr<TaskGroupDefinition>  m_pTaskGroup; // Pointer to the task group which this VU is executing.
    TaskGroupDefinition::TaskList::const_iterator m_pCurTask;
    long m_lCurTaskIteration;       // The current task iteration.
    long m_lCurTaskGroupIteration;  // The current task group iteration.
    unsigned __int64 m_qwEndTask;   // Timestamp representing the time when the current
                                    // task should end. Ignored if the current task does
                                    // not specify a duration.
    unsigned __int64 m_qwIterationBegin;// Timestamp marking the beginning of each iteration.
    DWORD           m_dwHTTPBytesIn;    // Sums up HTTP bytes received during a task iteration.
    DWORD           m_dwHTTPBytesOut;   // Sums up HTTP bytes sent during a task iteration.
    DWORD           m_dwRequests;       // Sums up HTTP requests during a task iteration.
    DWORD           m_dwFailedRequests; // Sums up HTTP requests returning a 4xx or 5xx code during a task iteration.
    DWORD           m_dwTotalLatency;   // Accumulated latency times in milliseconds.
    DWORD           m_dwLatencyCount;   // Number of latency times collected.
    DWORD           m_dwMinLatency;     // Minimum latency time.
    DWORD           m_dwMaxLatency;     // Maximum latency time.

    //// Info. required by some sync conditions.

    // Info used to synchronise on a specific request.
    DWORD   m_dwWaitedConnId;

#ifndef CYRANO_TP_STANDALONE
    BOOL            m_bWaitingForNext;  // In step-by-step mode, allows to wait for a Next call.

    // Info used to synchronise on a semaphore.
    CECSemaphore    *m_pCurrentSemaphore;
    std::string     m_szCurrentSemaphoreName;
#endif

    CTimerQueue     *m_pTimerQueue;         // Pointer to the timer queue defined in the VU Controller.

    DWORD   m_dwParallelism;    // Max. number of parallel requests.

    list<std::string>   m_Params;       // Current parameter list (passing by value only so stack not required).
    TtValList           m_paramsInfo;   // ParameterListInfo.

    time_t              m_mutexStartWaitTime;       // Used for the mutex lock time out.
    time_t              m_semaphoreStartWaitTime;   // Used for the semaphore try-acquire time out.

    //// Data storing info. for async. TOF objects, avoiding to compute the value several times
    //// if that object is processed in several steps.
    std::string m_szHeaderText;         // Computed header string.
    std::string m_szRemoteServerName;
    std::string m_szRemoteServerPort;
    std::string m_szCurrentUri;

    DWORD   m_dwCurrentConid;           // Current connection being processed. For use by TRcrdRequest.
    BOOL    m_bSecureModeUsed;

    // Connection info for use by TRcrdConnect while stepping from case 1 to case 2.
    std::string m_szProxyServerName;
    DWORD       m_dwProxyServerPort;

    // Authentication info used when processing BASIC/NTLM/KERBEROS authentication:
    CredHandle  m_hCred;
    CtxtHandle  m_hContext;
    BOOL        m_bHasAuthenticationInfo;
    DWORD       m_dwMaxToken;

    DWORD       m_dwStartTime;          // Time in seconds since 1/1/1901.
    std::string m_noteText;
    std::string m_szTestName;

    DWORD   m_dwTestCasePass;
    DWORD   m_dwTestCaseFail;

    DWORD       m_dwDisconnectToWait;   // Number of disconnections we have to wait for.

    static CFastLock *  s_pCacheSl;   // Allow only one VU to access the scripts cache at a time.

private:
    //// --------------------------------------------------------------------------------
    //// Forbidden.
    //// --------------------------------------------------------------------------------

    CAsyncVirtualUser(const CAsyncVirtualUser&);
    CAsyncVirtualUser& operator = (const CAsyncVirtualUser&);
};


inline void CAsyncVirtualUser::SetScriptPos(DWORD dwScriptPos)
{
    CURR_SCRIPT_CTXT_DATA->oi = dwScriptPos;
}

inline DWORD CAsyncVirtualUser::GetScriptPos()
{
    return CURR_SCRIPT_CTXT_DATA->oi;
}

#endif
