/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#include "stdafx.h"
#include "AsyncVirtualUserController.h"

#ifdef USE_PERF_DATA
#   include "PerfCounterKeys.hpp"
#endif

#include "TEMainStructure.hpp"
extern CMainExecuterStructure *g_pTExecuter;


#define OpenErrorStream()   {std::ostringstream errStm; errStm
#define CloseErrorStream()  g_pTExecuter->LogError(errStm.str().c_str());}


///////////////////////////////////////////////////////////////////////////////
//
// Class CAsyncVirtualUserController
//
///////////////////////////////////////////////////////////////////////////////

#ifdef USE_PERF_DATA
LONG *CAsyncVirtualUserController::m_spActiveUsers = NULL;
#endif

CAsyncVirtualUserController::CAsyncVirtualUserController(TScriptRule* pRule, CTrace *pTrace)
    : CThreadPoolController(pTrace), m_scriptCache(pRule), m_pTimerQueue(NULL),
      m_dwFailedUsersCount(0), m_dwSuccessfullUsersCount(0),
      m_dwTotalConnectionTime(0), m_dwConnectionCnt(0)
{}

CAsyncVirtualUserController::~CAsyncVirtualUserController()
{
    if(m_pTimerQueue)
        delete m_pTimerQueue;
}

#ifndef CYRANO_TP_STANDALONE
InjectorControl_ptr CAsyncVirtualUserController::GetInjectorControl() const
{
    return g_pTExecuter->GetInjectorControl();
}
#endif

long CAsyncVirtualUserController::GetCachedTestVUCount() const
{
    long lCount = g_pTExecuter->GetCachedTestVUCount();
#ifndef CYRANO_TP_STANDALONE
    if(lCount == 0)
    {
        // There is probably at least one VU active in this Test Executer!
        lCount = GetTestExecuterVUs();
        g_pTExecuter->SetCachedTestVUCount(lCount);
    }
#endif
    if(m_pTrace && m_pTrace->IsTrace("full"))
    {
        GetTraceStream(m_pTrace) << "Retrieving the cached VU count: "
            << lCount << "\r\n";
        ReleaseTraceStream();
    }
    return lCount;
}

const std::string& CAsyncVirtualUserController::GetExeName() const
{
    return g_pTExecuter->GetExeName();
}

#ifndef CYRANO_TP_STANDALONE
long CAsyncVirtualUserController::GetTimeOffset() const
{
    return g_pTExecuter->GetTimeOffset();
}
#endif

DWORD CAsyncVirtualUserController::Init(DWORD dwThreadPoolSize, DWORD dwNumberOfConcurrentThreadsInIocp, DWORD dwInitialVUCount, DWORD dwVUCountGrowBy, DWORD dwBrowserParallelism)
{
#ifdef USE_PERF_DATA
    m_spActiveUsers = g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_THREADPOOL_ACTIVEUSERS);
#endif
    DWORD res = CThreadPoolController::Init(dwThreadPoolSize, dwNumberOfConcurrentThreadsInIocp);

    if(res == 0)
    {
        m_pTimerQueue = new CTimerQueue(*m_pIocp, INFINITE, m_pTrace);

        if(m_pTimerQueue == NULL)
        {
            if(m_pTrace)
            {
                GetTraceStream(m_pTrace) << "Failed creating the timer queue." << "\r\n";
                ReleaseTraceStream();
            }
            g_pTExecuter->LogError("Failed creating the timer queue.");
            return -1;
        }

        res = m_pTimerQueue->Init();
        if(res == 0)
            res = m_VirtualUsersList.Init(m_pIocp, m_pTimerQueue, this, dwInitialVUCount, dwVUCountGrowBy, dwBrowserParallelism, m_pTrace);
    }

    return res;
}

TaskGroupDefinition::TaskList::const_iterator CAsyncVirtualUserController::GetFirstTask(SharedPtr<TaskGroupDefinition>& pTaskGroup)
{
    return pTaskGroup->m_tasks.begin();
}

// Returns FALSE if we are at the last task of the task group.
BOOL CAsyncVirtualUserController::GetNextTask(SharedPtr<TaskGroupDefinition>& pTaskGroup, TaskGroupDefinition::TaskList::const_iterator& pCurTask)
{
    return (++pCurTask != pTaskGroup->m_tasks.end());
}

DWORD CAsyncVirtualUserController::IncTaskGroupVUs(SharedPtr<TaskGroupDefinition>& pTaskGroup)
{
    return InterlockedIncrement(&pTaskGroup->m_lCurVirtualUsers);
}

// TRUE if all VUs have finished executing the task group.
BOOL CAsyncVirtualUserController::DecTaskGroupVUs(SharedPtr<TaskGroupDefinition>& pTaskGroup)
{
    InterlockedDecrement(&pTaskGroup->m_lCurVirtualUsers);
    return InterlockedIncrement(&pTaskGroup->m_lDoneVirtualUsers) == pTaskGroup->m_lNumVirtualUsers;
}

BOOL CAsyncVirtualUserController::StartTaskGroup(SharedPtr<TaskGroupDefinition>& pTaskGroup)
{
    BOOL bRet = TRUE;

    {
        CLock lock(m_taskGroupsLock);

        // Add the task group to the map.
        if(m_taskGroups.find(pTaskGroup->m_iTaskGroupNumber) != m_taskGroups.end())
            bRet = FALSE;   // Task group already being processed!
        else
            m_taskGroups.insert(make_pair(pTaskGroup->m_iTaskGroupNumber, pTaskGroup));
    }

    if(bRet)
    {
        // Create a timestamp to mark the end of the task group.
        // Do this even if no duration has been specified, because
        // this timestamp is also used in snapshot requests to compute
        // the task group's run time.
        FILETIME ftNow;
        GetSystemTimeAsFileTime(&ftNow);
        pTaskGroup->m_qwEndTime = ftNow.dwHighDateTime;
        pTaskGroup->m_qwEndTime <<= 32;
        pTaskGroup->m_qwEndTime |= ftNow.dwLowDateTime;
        unsigned __int64 qwDuration = UInt32x32To64(pTaskGroup->m_lTaskGroupDuration * 1000, 10000L);    // Convert duration from milliseconds to hundreds of nanoseconds.
        pTaskGroup->m_qwEndTime += qwDuration;

#ifndef CYRANO_TP_STANDALONE
        // Create the snapshots buffer.
        m_snapShotStats.m_taskGroupSnapShots.insert(make_pair(pTaskGroup->m_iTaskGroupNumber, TaskGroupSnapShot(pTaskGroup->m_tasks.size())));
#endif

#ifndef CYRANO_TP_STANDALONE
        // Ramp-up the users.
        bRet = RampUpUsers(pTaskGroup);
#else
        bRet = StartNewUsersGroup(pTaskGroup, 1);
#endif
    }

    return bRet;
}

void CAsyncVirtualUserController::StopTaskGroup(int iTaskGroupNumber)
{
    // Identify the task group.
    TaskGroups::iterator pTG = m_taskGroups.find(iTaskGroupNumber);
    if(pTG != m_taskGroups.end())
        pTG->second->Disable(); // Disable the task group, which should then soon stop.
}

#ifndef CYRANO_TP_STANDALONE
void CAsyncVirtualUserController::OnTaskGroupComplete(int iTaskGroupNumber, TaskGroupStopReason eReason)
#else
void CAsyncVirtualUserController::OnTaskGroupComplete(int iTaskGroupNumber)
#endif
{
#ifndef CYRANO_TP_STANDALONE
    // Delete the task group from the statistics.
    {
        CLock lock(m_taskGroupsStatsLock);
        m_snapShotStats.m_taskGroupSnapShots.erase(iTaskGroupNumber);
    }
#endif

    // Delete the task group.
    {
        CLock lock(m_taskGroupsLock);
        m_taskGroups.erase(iTaskGroupNumber);
    }

#ifndef CYRANO_TP_STANDALONE
    // Notify the Injector Control.
    if(m_pTrace && m_pTrace->IsTrace("full"))
    {
        GetTraceStream(m_pTrace) << "Notifying InjectorControl that task group "
                                 << iTaskGroupNumber << " is over with reason "
                                 << eReason <<".\r\n";
        ReleaseTraceStream();
    }

    // The task group number and the reason are encoded on a single DWORD:
    DWORD dwParam = (DWORD)MAKELONG(iTaskGroupNumber, eReason);
    DWORD dwRes = m_pIocp->PostCompletedIO(dwParam, (IocpEventDispatcher*)this, (OVERLAPPED*)EVENT_TYPE_TASKGROUP_STOPPED);
    if(dwRes)
    {
        if(m_pTrace && m_pTrace->IsTrace("full"))
        {
            GetTraceStream(m_pTrace) << "ERROR: failed to post a stop task group notification for task group "
                << iTaskGroupNumber << "Err# = " << dwRes << "\r\n";
            ReleaseTraceStream();
        }
        OpenErrorStream() << "ERROR: failed to post a stop task group notification for task group "
                << iTaskGroupNumber << ". Err# = " << dwRes;
        CloseErrorStream();
    }
#else
    // Stop the thread pool.
    Stop();
#endif
}

void CAsyncVirtualUserController::StopController()
{
    DWORD dwRes = m_pIocp->PostCompletedIO(0, (IocpEventDispatcher*)this, (OVERLAPPED*)EVENT_TYPE_EXECUTER_STOPPING);
    if(dwRes)
    {
        if(m_pTrace && m_pTrace->IsTrace("full"))
        {
            GetTraceStream(m_pTrace) << "ERROR: failed to post a 'Test Executer stopping' notification. Err# = "
                << dwRes << "\r\n";
            ReleaseTraceStream();
        }
        OpenErrorStream() << "ERROR: failed to post a 'Test Executer stopping' notification. Err# = "
                << dwRes;
        CloseErrorStream();
    }

    Stop();
}

void CAsyncVirtualUserController::OnUserEnterTask(const CAsyncVirtualUser& vu, int iTaskGroupNumber, int iTaskNumber)
{
#ifndef CYRANO_TP_STANDALONE
    CLock lock(m_taskGroupsStatsLock);

    // Update task group snapshot data.
    TaskGroupSnapShots::iterator pTG = m_snapShotStats.m_taskGroupSnapShots.find(iTaskGroupNumber);
    ++pTG->second.m_taskSnaps[iTaskNumber].m_lActiveUsers;

    // Update VU snapshot data.
    SharedPtr<VUSnapShot> pVUSn(new VUSnapShot());
    pVUSn->m_lVUID = (long)vu.GetExternalVUID();
    pVUSn->m_lCurTask = iTaskNumber;
    pVUSn->m_lStartTime = (long)vu.GetStartTime();
    time_t t;
    time(&t);
    pVUSn->m_lRunTime = t - pVUSn->m_lStartTime;
    pVUSn->m_lCompletedIterations = vu.GetCurTaskIter();
    pVUSn->m_strNoteText = vu.GetNoteText();
    VUSnapShots::iterator pVU = pTG->second.m_VUSnapShots.find(pVUSn->m_lVUID);
    if(pVU == pTG->second.m_VUSnapShots.end())
    {
        // Insert the VU.
        pTG->second.m_VUSnapShots.insert(make_pair(pVUSn->m_lVUID, pVUSn));
    }
    else
    {
        // Update the VU data.
        pVU->second = pVUSn;
    }

    TaskKey task(iTaskGroupNumber, iTaskNumber);
    TaskSnapShot::iterator pTask = m_TaskSnapShots.find(task);
    if(pTask == m_TaskSnapShots.end())
    {
        std::pair<TaskSnapShot::iterator,bool> res =
            m_TaskSnapShots.insert(std::make_pair(task, VUSnapShots()));
        if(res.second)
            pTask = res.first;
    }

    if(pTask != m_TaskSnapShots.end())
    {
        pVU = pTask->second.find(pVUSn->m_lVUID);
        if(pVU == pTask->second.end())
            pTask->second.insert(make_pair(pVUSn->m_lVUID, pVUSn));
        else
            pVU->second = pVUSn;
    }
#endif
}

void CAsyncVirtualUserController::OnUserExitTask(const CAsyncVirtualUser& vu, int iTaskGroupNumber, int iTaskNumber, BOOL bExitTaskGroup)
{
#ifndef CYRANO_TP_STANDALONE
    DWORD dwTotalLatency, dwLatencyCount, dwMinLatency, dwMaxLatency;

    CLock lock(m_taskGroupsStatsLock);

    // Update task group snapshot data.
    TaskGroupSnapShots::iterator pTG = m_snapShotStats.m_taskGroupSnapShots.find(iTaskGroupNumber);
    TaskGroupSnapShot& tg = pTG->second;
    --tg.m_taskSnaps[iTaskNumber].m_lActiveUsers;
    ++tg.m_taskSnaps[iTaskNumber].m_lIterations;
    tg.m_taskSnaps[iTaskNumber].m_lRuntime = (long)vu.GetTaskIterationDuration();
    tg.m_lBytesIn  += vu.GetHTTPBytesIn();
    tg.m_lBytesOut += vu.GetHTTPBytesOut();
    tg.m_lRequests += vu.GetNumRequests();
    tg.m_lFailedRequests += vu.GetNumFailedRequests();
    vu.GetLatencyTimes(dwTotalLatency, dwLatencyCount, dwMinLatency, dwMaxLatency);
    if(dwMinLatency < (DWORD)tg.m_lMinRequestLatency || tg.m_lMinRequestLatency == 0)
        tg.m_lMinRequestLatency = dwMinLatency;
    if(dwMaxLatency > (DWORD)tg.m_lMaxRequestLatency)
        tg.m_lMaxRequestLatency = dwMaxLatency;
    tg.m_lTotalRequestLatency += dwTotalLatency;
    tg.m_lCntRequestLatency += dwLatencyCount;
    if(bExitTaskGroup)
        ++tg.m_lCompletedIterations;

    // Update VU snapshot data.
    tg.m_VUSnapShots.erase((long)vu.GetExternalVUID());
    TaskKey task(iTaskGroupNumber, iTaskNumber);
    TaskSnapShot::iterator pTask = m_TaskSnapShots.find(task);
    if(pTask != m_TaskSnapShots.end())
        pTask->second.erase((long)vu.GetExternalVUID());
#endif
}

#ifndef CYRANO_TP_STANDALONE
BOOL CAsyncVirtualUserController::GetStatsSnapShot(TestExecuterSnapShot& sn)
{
    CLock lock(m_taskGroupsStatsLock);

    // Get a timestamp for now.
    FILETIME ftNow;
    GetSystemTimeAsFileTime(&ftNow);
    unsigned __int64 qwNow = ftNow.dwHighDateTime;
    qwNow <<= 32;
    qwNow |= ftNow.dwLowDateTime;

    // Loop on task groups.
    for(TaskGroupSnapShots::iterator pTG = m_snapShotStats.m_taskGroupSnapShots.begin();
        pTG != m_snapShotStats.m_taskGroupSnapShots.end();
        ++pTG)
    {
        // Compute the run time.
        TaskGroups::iterator pTaskGroup = m_taskGroups.find(pTG->first);
        unsigned __int64 qwDuration = UInt32x32To64(pTaskGroup->second->m_lTaskGroupDuration * 1000, 10000L);    // Convert duration from milliseconds to hundreds of nanoseconds.
        unsigned __int64 qwStart = pTaskGroup->second->m_qwEndTime - qwDuration;
        pTG->second.m_lRunTime = (long)((qwNow - qwStart)/10000000L);   // Convert back from hundreds of nanoseconds to seconds.
    }

    sn = m_snapShotStats;

    return TRUE;
}

BOOL CAsyncVirtualUserController::GetTaskSnapshot(const TaskIdList& taskIDs, TaskSnap2List& taskSnapshots)
{
    CLock lock(m_taskGroupsStatsLock);

    // Initialize the output structure.
    _CORBA_ULong ulMax = taskIDs.length();
    taskSnapshots.length(ulMax);

    // Loop on tasks.
    for(_CORBA_ULong ul = 0; ul < ulMax; ++ul)
    {
        taskSnapshots[ul].task = taskIDs[ul];
        TaskKey task(taskSnapshots[ul].task);
        VUSnapList& vus = taskSnapshots[ul].VUSnaps;
        TaskSnapShot::const_iterator pTS = m_TaskSnapShots.find(task);
        if(pTS != m_TaskSnapShots.end())
        {
            vus.length(pTS->second.size());

            // Retrieve the VU records for this task.
            _CORBA_ULong ulVu = 0;
            for(VUSnapShots::const_iterator pVU  = pTS->second.begin();
                                            pVU != pTS->second.end();
                                          ++pVU)
            {
                // Copy the VU data.
                VUSnap& vusnap              = vus[ulVu];
                vusnap.lVUId                = pVU->second->m_lVUID;
                vusnap.lCompletedIterations = pVU->second->m_lCompletedIterations;
                vusnap.lRunTime             = pVU->second->m_lRunTime;
                vusnap.lStartTime           = pVU->second->m_lStartTime;
                vusnap.strNoteText = CORBA::string_dup(pVU->second->m_strNoteText.c_str());

                ++ulVu;
            }
        }
        else
            vus.length(0);
    }

    return TRUE;
}

long CAsyncVirtualUserController::GetTestExecuterVUs() const
{
    CLock lock(m_taskGroupsStatsLock);

    long lRes = 0;

    // Loop on task groups.
    for(TaskGroupSnapShots::const_iterator pTG = m_snapShotStats.m_taskGroupSnapShots.begin();
        pTG != m_snapShotStats.m_taskGroupSnapShots.end();
        ++pTG)
    {
        // Loop on tasks.
        for(TaskSnaps::const_iterator pL = pTG->second.m_taskSnaps.begin();
            pL != pTG->second.m_taskSnaps.end();
            ++pL)
        {
            // Compute the number of active VUs.
            lRes += pL->m_lActiveUsers;
        }
    }

    return lRes;
}
#endif

// This method is responsible for starting the users by small groups.
BOOL CAsyncVirtualUserController::RampUpUsers(SharedPtr<TaskGroupDefinition>& pTaskGroup)
{
    BOOL bRes = TRUE;

    DWORD dwUsersToLaunch  = (DWORD)pTaskGroup->m_lNumVirtualUsers;
    DWORD dwVUsPerBatch    = (DWORD)pTaskGroup->m_lVUsPerBatch;
    if(dwVUsPerBatch == 0)
        dwVUsPerBatch = 1;  // Should not have 0 VUs per batch!
    if(dwUsersToLaunch < dwVUsPerBatch)
        dwVUsPerBatch = dwUsersToLaunch;    // There will only be one batch.
    int   iTaskGroupNumber =        pTaskGroup->m_iTaskGroupNumber;

    while(dwUsersToLaunch)
    {
        if(!RampUpBatch(pTaskGroup, dwVUsPerBatch))
        {
            bRes = FALSE;
            dwUsersToLaunch = 0;
        }
        else
        {
            if(dwUsersToLaunch > dwVUsPerBatch)
            {
                dwUsersToLaunch -= dwVUsPerBatch;
                if(dwUsersToLaunch < dwVUsPerBatch)
                    dwVUsPerBatch = dwUsersToLaunch;    // This is the last batch.
                Sleep((DWORD)pTaskGroup->m_iInterBatchDelay * 1000);
            }
            else
                dwUsersToLaunch = 0;
        }
    }

    return bRes;
}

BOOL CAsyncVirtualUserController::RampUpBatch(SharedPtr<TaskGroupDefinition>& pTaskGroup, DWORD dwNumBatchUsers)
{
    BOOL bRes = TRUE;

    // Let's launch VUs every second.
    // Compute the number of VUs to start in a group.
    DWORD dwBatchRampUpTimeInSeconds = pTaskGroup->m_iBatchRampUpTime;
    DWORD dwNumGroupUsers = 0;
    DWORD dwInterGroupDelayInSeconds = 1;   // By default, one second.
    if(dwBatchRampUpTimeInSeconds > 0)
    {
        if(dwBatchRampUpTimeInSeconds / dwInterGroupDelayInSeconds > dwNumBatchUsers)
        {
            // We want to launch less than one VU per inter-group delay:
            // we need to change the inter-group delay to fit.
            dwInterGroupDelayInSeconds = dwBatchRampUpTimeInSeconds / dwNumBatchUsers;
        }
        dwNumGroupUsers = dwNumBatchUsers * dwInterGroupDelayInSeconds / dwBatchRampUpTimeInSeconds;
        if((dwNumBatchUsers * dwInterGroupDelayInSeconds) % dwBatchRampUpTimeInSeconds)
            ++dwNumGroupUsers;  // The last group will contain less VUs.
    }
    if(dwNumGroupUsers == 0)
        dwNumGroupUsers = dwNumBatchUsers;  // All VUs must be launched at once.

    while(dwNumBatchUsers)
    {
        bRes = StartNewUsersGroup(pTaskGroup, dwNumGroupUsers);
        if(dwNumBatchUsers > dwNumGroupUsers)
        {
            dwNumBatchUsers -= dwNumGroupUsers;
            if(dwNumBatchUsers < dwNumGroupUsers)
                dwNumGroupUsers = dwNumBatchUsers;      // This is the last group.
            Sleep(dwInterGroupDelayInSeconds * 1000);   // Sleep before launching the next group.
        }
        else
            dwNumBatchUsers = 0;
    }

    return bRes;
}

BOOL CAsyncVirtualUserController::StartNewUsersGroup(
    SharedPtr<TaskGroupDefinition>& pTaskGroup, DWORD dwNumUsers)
{
    BOOL bRet = TRUE;

    // Alloc nbUsers clients.
    DWORD dwStartedUsers;

    dwStartedUsers = m_VirtualUsersList.StartUsers(pTaskGroup, dwNumUsers);
    assert(dwStartedUsers == dwNumUsers);

#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
    InterlockedExchangeAdd(m_spActiveUsers, dwStartedUsers);
#   else
    *m_spActiveUsers += dwStartedUsers;
#   endif
#endif

    if(dwStartedUsers != dwNumUsers)
    {
        if(m_pTrace)
        {
            GetTraceStream(m_pTrace) << "Failed starting users." << "\r\n";
            ReleaseTraceStream();
        }
        g_pTExecuter->LogError("Failed starting users.");
        bRet = FALSE;
    }
    else
    {
#ifndef CYRANO_TP_STANDALONE
        CLock lock(m_taskGroupsStatsLock);
        TaskGroupSnapShots::iterator pTG = m_snapShotStats.m_taskGroupSnapShots.find(pTaskGroup->m_iTaskGroupNumber);
        pTG->second.m_lTotalUsers += dwStartedUsers;
#endif
    }

    return bRet;
}

void CAsyncVirtualUserController::CancelAllUsers()
{
    m_VirtualUsersList.CancelAllUsers();
}

void CAsyncVirtualUserController::StepByStep_Next(DWORD dwUser)
{
    if(m_pTrace && m_pTrace->IsTrace("full"))
    {
        GetTraceStream(m_pTrace) << "CAsyncVirtualUserController::StepByStep_Next: user " << dwUser << ".\r\n";
        ReleaseTraceStream();
    }
    m_VirtualUsersList.StepByStep_Next(dwUser);
}

void CAsyncVirtualUserController::OnVirtualUserOver(int iTaskGroupNumber, DWORD dwVUId)
{
#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
    InterlockedDecrement(m_spActiveUsers);
#   else
    --*m_spActiveUsers;
#   endif
#endif

    m_VirtualUsersList.DropVirtualUser(dwVUId);

#ifndef CYRANO_TP_STANDALONE
    {
        CLock lock(m_taskGroupsStatsLock);
        TaskGroupSnapShots::iterator pTG = m_snapShotStats.m_taskGroupSnapShots.find(iTaskGroupNumber);
        --pTG->second.m_lTotalUsers;
    }
#endif

    if(m_pTrace && m_pTrace->IsTrace("full"))
    {
        GetTraceStream(m_pTrace) << "CAsyncVirtualUserController::OnVirtualUserOver: user " << dwVUId << " done.\r\n";
        ReleaseTraceStream();
    }
}

void CAsyncVirtualUserController::OnEvent(DWORD dwNumberOfBytesTransferred, BOOL bIsIo, DWORD dwEventType)
{
    if(bIsIo == TRUE)
    {
        // Should not receive any IOs there.
        assert(0);
    }
    else
    {
        switch(dwEventType)
        {
        case EVENT_TYPE_THREADPOOL_START:
            {
#ifdef __DLLSA__
                //// Start the modelled user.

                // Create a dummy task group.
                TaskGroupDefinition* pTaskGroup = new TaskGroupDefinition(
                    (int)0,     // iTaskGroupNumber
                         0,     // lTaskGroupDuration
                         1,     // lNumVirtualUsers
                         1,     // lNumTaskGroupIterations
                    (int)0,     // iInterBatchDelay
                         0,     // lVUsPerBatch
                    (int)0,     // iBatchRampUpTime
                    TaskGroupDefinition::loglev_none, // iLoggingLevel
                         1);    // bUsePageTimers

                // Add a dummy task.
                TaskGroupDefinition::TaskDefinition task(
                    g_pTExecuter->GetTofFileNameAndPath().c_str(), // Script name.
                                 0,     // Duration.
                                 1,     // Iterations.
                    (int)        0);    // Fixed delay.
                pTaskGroup->m_tasks.push_back(task);

                if(!StartTaskGroup(SharedPtr<TaskGroupDefinition>(pTaskGroup)))
                {
                    if(m_pTrace && m_pTrace->IsTrace("errors"))
                    {
                        GetTraceStream(m_pTrace) << "Could not create the user.\r\n";
                        ReleaseTraceStream();
                    }
                }
#endif
                break;
            }

        case EVENT_TYPE_THREADPOOL_STOP_THREAD:
            if(m_pTrace && m_pTrace->IsTrace("errors"))
            {
                GetTraceStream(m_pTrace) << "CAsyncVirtualUserController::OnEvent: ERROR: do not handle STOP THREAD events YET.\r\n";
                ReleaseTraceStream();
            }
            break;

        case EVENT_TYPE_TIMER:
            if(m_pTrace && m_pTrace->IsTrace("errors"))
            {
                GetTraceStream(m_pTrace) << "CAsyncVirtualUserController::OnEvent: ERROR: do not handle timer events." << "\r\n";
                ReleaseTraceStream();
            }
            break;

        case EVENT_TYPE_TASKGROUP_STOPPED:
            {
#ifndef CYRANO_TP_STANDALONE
                // Decode the task group number and the reason.
                int iTaskGroupNumber = (int)LOWORD(dwNumberOfBytesTransferred);
                TaskGroupStopReason eReason = (TaskGroupStopReason)HIWORD(dwNumberOfBytesTransferred);
                GetInjectorControl()->OnTaskGroupStopped(iTaskGroupNumber, eReason);
#endif
                break;
            }

        case EVENT_TYPE_EXECUTER_STOPPING:
#ifndef CYRANO_TP_STANDALONE
            // Notify the Injector Control that the Test Executer is stopping.
            g_pTExecuter->NotifyTEStopping();
#endif
            break;

        default:
            if(m_pTrace && m_pTrace->IsTrace("errors"))
            {
                GetTraceStream(m_pTrace) << "CAsyncVirtualUserController::OnEvent: ERROR: unknown type." << "\r\n";
                ReleaseTraceStream();
            }
            break;
        };
    }
}

void CAsyncVirtualUserController::OnEventFailed(DWORD dwNumberOfBytesTransferred, BOOL bIsIo, DWORD dwEventType, DWORD dwError)
{
    if(m_pTrace && m_pTrace->IsTrace("errors"))
    {
        GetTraceStream(m_pTrace) << "CAsyncVirtualUserController::OnEventFailed: ERROR " << dwError << "\r\n";
        ReleaseTraceStream();
    }
}


///////////////////////////////////////////////////////////////////////////////
//
// Class CVirtualUsersGroup
//
///////////////////////////////////////////////////////////////////////////////

CVirtualUsersGroup::CVirtualUsersGroup(CTrace* pTrace)
    : m_VirtualUsersArray(NULL), m_dwStartedUsersCount(0L), m_dwVirtualUsersCount(0L), m_pIocp(NULL), m_pTrace(pTrace)
{}

BOOL CVirtualUsersGroup::CreateUsers(DWORD dwNbUsers)
{
    BOOL bCouldCreate = FALSE;

    if(dwNbUsers != 0)
    {
        try
        {
            m_VirtualUsersArray = new CAsyncVirtualUser[dwNbUsers];
        }
        catch(...)
        {
            if(m_pTrace)
            {
                GetTraceStream(m_pTrace) << "CVirtualUsersGroup::CreateUsers: Error allocating the virtual users.\r\n";
                ReleaseTraceStream();
            }
            g_pTExecuter->LogError("Error allocating the virtual users.");
        }

        if(m_VirtualUsersArray != NULL)
        {
            m_dwVirtualUsersCount = dwNbUsers;
            bCouldCreate = TRUE;
        }
        else
        {
            if(m_pTrace)
            {
                GetTraceStream(m_pTrace) << "CVirtualUsersGroup::CreateUsers: Error allocating the virtual users.\r\n";
                ReleaseTraceStream();
            }
            g_pTExecuter->LogError("Error allocating the virtual users.");
            m_dwVirtualUsersCount = 0;
            bCouldCreate = FALSE;
        }
    }
    else
    {
        if(m_pTrace)
        {
            GetTraceStream(m_pTrace) << "CVirtualUsersGroup::CreateUsers: Cannot create a group with no virtual users.\r\n";
            ReleaseTraceStream();
        }
        g_pTExecuter->LogError("Cannot create a group with no virtual users.");
    }

    return bCouldCreate;
}

void CVirtualUsersGroup::CancelAllUsers()
{
    for(DWORD dw=0L ; dw<m_dwVirtualUsersCount ; ++dw)
    {
        m_VirtualUsersArray[dw].CancelAllQueries();
    }
}

// Get all the users in that group ready for start.
BOOL CVirtualUsersGroup::InitUsers(CIOCP *pIocp, CAsyncVirtualUserController& controller, CTimerQueue* pTimerQueue, DWORD dwParallelism)
{
    static DWORD dwManagerID = 0;
    m_pIocp = pIocp;

    for(DWORD dw=0 ; dw<m_dwVirtualUsersCount ; dw++)
    {
        if(m_VirtualUsersArray[dw].Init(controller, pTimerQueue, dwManagerID, dwParallelism, m_pTrace) != 0)
        {
            if(m_pTrace)
            {
                GetTraceStream(m_pTrace) << "CVirtualUsersGroup::InitUsers: Error initialising virtual user number " << dw << ".\r\n";
                ReleaseTraceStream();
            }
            OpenErrorStream() << "Error initialising virtual user number " << dw << ".";
            CloseErrorStream();
            return FALSE;
        }

        dwManagerID++;
    }

    return TRUE;
}

// Start dwNbUsers from the current group. Return the number of users that could NOT be started.
DWORD CVirtualUsersGroup::StartMoreUsers(SharedPtr<TaskGroupDefinition>& pTaskGroup, DWORD dwNbUsers)
{
    while(dwNbUsers > 0 && m_dwStartedUsersCount < m_dwVirtualUsersCount)
    {
        m_VirtualUsersArray[m_dwStartedUsersCount].Initialise(pTaskGroup);

        DWORD dwRes = m_pIocp->PostCompletedIO(0, (IocpEventDispatcher*) &m_VirtualUsersArray[m_dwStartedUsersCount], (OVERLAPPED*)CAsyncVirtualUser::EVENT_TYPE_VIRTUAL_USER_START);
        if(dwRes != 0)
        {
            assert(0);
            if(m_pTrace && m_pTrace->IsTrace("full"))
            {
                GetTraceStream(m_pTrace) << "Start users ERROR." << "\r\n";
                ReleaseTraceStream();
            }
            OpenErrorStream() << "Failed to start users. Error " << dwRes;
            CloseErrorStream();

            break;
        }

        m_dwStartedUsersCount++;
        dwNbUsers--;
    }

    return dwNbUsers;
}

// Delete all the users from that group.
void CVirtualUsersGroup::DropAllUsers()
{
    if(m_VirtualUsersArray != NULL)
    {
        delete [] m_VirtualUsersArray;
        m_VirtualUsersArray = NULL;
        m_dwVirtualUsersCount = 0;
    }
}

BOOL CVirtualUsersGroup::StepByStep_Next(DWORD dwUser)
{
    BOOL bRet = FALSE;
    for(DWORD dw=0L ; dw<m_dwVirtualUsersCount ; ++dw)
    {
        if(m_VirtualUsersArray[dw].GetManagerNumber() == dwUser)
        {
            m_pIocp->PostCompletedIO(0, m_VirtualUsersArray+dw, (OVERLAPPED*)CAsyncVirtualUser::EVENT_TYPE_VIRTUAL_USER_RESUME);
            bRet = TRUE;
            break;
        }
    }
    return bRet;
}

// Append the pipe-separated list of all unique identifiers for that group's users (formerly Threads).
// Code based on TestExecuterImpl::GetThreadIdents().
void CVirtualUsersGroup::AppendThreadIdents(string &str) const
{
    for(DWORD dw=0 ; dw<m_dwStartedUsersCount ; dw++)
    {
        str += m_VirtualUsersArray[dw].GetThreadID() + '|';
    }
}

DWORD CVirtualUsersGroup::GetMatchingThreadsCount(const string &sname) const
{
    DWORD dwThreadsCount = 0;

    for(DWORD dw=0 ; dw<m_dwStartedUsersCount ; dw++)
    {
        if(m_VirtualUsersArray[dw].ScriptName() == sname)
            dwThreadsCount++;
    }

    return dwThreadsCount;
}

// Return thread current scriptname based on unique ID.
BOOL CVirtualUsersGroup::GetVuScriptName(const char *ident, string &name) const
{
    // Browse all the threads in that group in search for the 'ident' thread and return its ScriptName.
    for(DWORD dw=0 ; dw<m_dwStartedUsersCount ; dw++)
    {
        if(m_VirtualUsersArray[dw].GetThreadID() == ident)
        {
            name = m_VirtualUsersArray[dw].ScriptName();
            return TRUE;
        }
    }

    return FALSE;
}

// Return a thread's execution phase based on its unique ID.
BOOL CVirtualUsersGroup::GetVuStatus(const char *ident, EVuExecPhase &phase) const
{
    for(DWORD dw=0 ; dw<m_dwStartedUsersCount ; dw++)
    {
        if(m_VirtualUsersArray[dw].GetThreadID() == ident)
        {
            phase = m_VirtualUsersArray[dw].GetStatus();
            return TRUE;
        }
    }

    return FALSE;
}

BOOL CVirtualUsersGroup::GetVuStartTime(const char *ident, ULONG &ulTime) const
{
    for(DWORD dw=0 ; dw<m_dwStartedUsersCount ; dw++)
    {
        if(m_VirtualUsersArray[dw].GetThreadID() == ident)
        {
            ulTime = m_VirtualUsersArray[dw].GetStartTime();
            return TRUE;
        }
    }

    return FALSE;
}


// ////////////////////////////////////////////////////////////////////////////
//
// Class CVirtualUsersList
//
// ////////////////////////////////////////////////////////////////////////////

CVirtualUsersList::CVirtualUsersList()
    : m_pIocp(NULL), m_pTimerQueue(NULL), m_pTrace(NULL), m_pVUController(NULL), m_dwVirtualUsersCount(0L), 
    m_dwParallelism(0L), m_dwVUCountGrowBy(0L), m_currGroup(0)
{}

CVirtualUsersList::~CVirtualUsersList()
{
    DeleteGroups();
}

DWORD CVirtualUsersList::Init(CIOCP* pIocp, CTimerQueue* pTimerQueue, CAsyncVirtualUserController *pVUController, DWORD dwInitialVUCount, DWORD dwVUCountGrowBy, DWORD dwParallelism, CTrace *pTrace)
{
    m_pIocp             = pIocp;
    m_pTimerQueue       = pTimerQueue;
    m_pTrace            = pTrace;
    m_pVUController     = pVUController;
    m_dwParallelism     = dwParallelism;
    m_dwVUCountGrowBy   = dwVUCountGrowBy;

    if(dwInitialVUCount != 0)
    {
        // Pre-allocate users.
        if(CreateGroup(dwInitialVUCount) == FALSE)
        {
            if(m_pTrace)
            {
                GetTraceStream(m_pTrace) << "CVirtualUsersList::Init: error creating a new group with " << dwInitialVUCount << " users.\r\n";
                ReleaseTraceStream();
            }
            OpenErrorStream() << "Error creating a new VU group with " << dwInitialVUCount << " users.";
            CloseErrorStream();
   
            return (DWORD)-1;
        }
    }
    m_currGroup = 0L;

    return 0L;
}

// Protected method that creates a group of standing-by users.
BOOL CVirtualUsersList::CreateGroup(DWORD dwNbUsers)
{
    CVirtualUsersGroup *pNewGroup = new CVirtualUsersGroup(m_pTrace);

    if(pNewGroup != NULL)
    {
        if(pNewGroup->CreateUsers(dwNbUsers) == FALSE)
            return FALSE;
        if(pNewGroup->InitUsers(m_pIocp, *m_pVUController, m_pTimerQueue, m_dwParallelism) == FALSE)
            return FALSE;

        CLock guard (m_userGroupListFl);
        m_GroupsList.push_back(pNewGroup);
        m_currGroup = m_GroupsList.size() - 1;

        return TRUE;
    }
    else
    {
        if(m_pTrace)
        {
            GetTraceStream(m_pTrace) << "Failed creating a new group with " << dwNbUsers << " users.\r\n";
            ReleaseTraceStream();
        }
        return FALSE;
    }
}

// Return the number of users that were started actually.
DWORD CVirtualUsersList::StartUsers(SharedPtr<TaskGroupDefinition>& pTaskGroup, DWORD dwNbUsers)
{
    if(dwNbUsers == 0)
    {
        assert(0);
        return 0;
    }

    CLock guard (m_userGroupListFl);

    if(m_GroupsList.size() == 0)
    {
        // Must create the first VU group.
        if(CreateGroup(__max(dwNbUsers, m_dwVUCountGrowBy)) == FALSE)
            return 0;
        assert(m_GroupsList.size() != 0);
    }

    DWORD dwRemaining = m_GroupsList[m_currGroup]->StartMoreUsers(pTaskGroup, dwNbUsers);
    if(dwRemaining != 0)
    {
        if(CreateGroup(__max(dwRemaining, m_dwVUCountGrowBy)) == FALSE)
        {
            InterlockedExchangeAdd((LONG*)&m_dwVirtualUsersCount, (LONG)(dwNbUsers-dwRemaining));
            return dwNbUsers-dwRemaining;
        }

        dwRemaining = m_GroupsList[m_currGroup]->StartMoreUsers(pTaskGroup, dwRemaining);
        assert(dwRemaining == 0);
    }

    InterlockedExchangeAdd((LONG*)&m_dwVirtualUsersCount, (LONG)dwNbUsers);

    return dwNbUsers;
}

void CVirtualUsersList::CancelAllUsers()
{
    if(m_pTrace && m_pTrace->IsTrace("full"))
    {
        GetTraceStream(m_pTrace) << "Entering CVirtualUsersList::CancelAllUsers()...\r\n";
        ReleaseTraceStream();
    }
    CLock lock(m_userGroupListFl);

    for(vector<CVirtualUsersGroup*>::iterator i=m_GroupsList.begin() ; i != m_GroupsList.end() ; ++i)
    {
        (*i)->CancelAllUsers();
    }
}

DWORD CVirtualUsersList::DropVirtualUser(DWORD dwVuId)
{
    // No need to delete that user.
    return InterlockedDecrement((LONG*)&m_dwVirtualUsersCount);
}

void CVirtualUsersList::DeleteGroups()
{
    CLock lock(m_userGroupListFl);

    for(vector<CVirtualUsersGroup*>::iterator i=m_GroupsList.begin() ; i!=m_GroupsList.end() ; ++i)
    {
        (*i)->DropAllUsers();
        delete *i;
    }

    m_GroupsList.clear();
}

void CVirtualUsersList::StepByStep_Next(DWORD dwUser)
{
    for(vector<CVirtualUsersGroup*>::iterator i=m_GroupsList.begin() ; i != m_GroupsList.end() ; ++i)
    {
        if((*i)->StepByStep_Next(dwUser))
            break;
    }
}

// Get pipe-separated list of all unique identifiers for users (formerly Threads) in this Test executer.
// Code based on TestExecuterImpl::GetThreadIdents().
void CVirtualUsersList::GetThreadIdents(string &str) const
{
    str.erase();
    str.reserve(1024);

    CLock guard (m_userGroupListFl);

    // Ask all the groups to append their working threads UIDs to the end of 'str'.
    for(vector<CVirtualUsersGroup*>::const_iterator i=m_GroupsList.begin() ; i!=m_GroupsList.end() ; ++i)
    {
        (*i)->AppendThreadIdents(str);
    }

    str += '|';
}

DWORD CVirtualUsersList::GetMatchingThreadsCount(const string &sname) const
{
    DWORD dwThreadsCount = 0;

    CLock guard (m_userGroupListFl);

    // Ask all the groups to return their working threads UIDs to the end of 'str'.
    for(vector<CVirtualUsersGroup*>::const_iterator i=m_GroupsList.begin() ; i!=m_GroupsList.end() ; ++i)
    {
        dwThreadsCount += (*i)->GetMatchingThreadsCount(sname);
    }

    return dwThreadsCount;
}


// Return thread current scriptname based on unique ID.
string CVirtualUsersList::GetVuScriptName(const char *ident) const
{
    string scriptName;

    CLock guard (m_userGroupListFl);

    // Browse all the groups in search for the 'ident' thread and return its ScriptName.
    for(vector<CVirtualUsersGroup*>::const_iterator i=m_GroupsList.begin() ; i!=m_GroupsList.end() ; ++i)
    {
        if((*i)->GetVuScriptName(ident, scriptName) == TRUE)
            return scriptName;
    }

    // Could not find that user.
    assert(0);
    return string();
}

// Browse the groups until we find 'ident' and return its status.
EVuExecPhase CVirtualUsersList::GetVuStatus(const char *ident) const
{
    EVuExecPhase execPhase;

    CLock guard (m_userGroupListFl);

    for(vector<CVirtualUsersGroup*>::const_iterator i=m_GroupsList.begin() ; i!=m_GroupsList.end() ; ++i)
    {
        if((*i)->GetVuStatus(ident, execPhase) == TRUE)
            return execPhase;
    }

    // Could not find that user.
    assert(0);
    return EXEC_OVER;
}

// Browse all the groups in search for the 'ident' thread and return its ScriptName.
ULONG CVirtualUsersList::GetVuStartTime(const char *ident) const
{
    ULONG time;

    CLock guard (m_userGroupListFl);

    for(vector<CVirtualUsersGroup*>::const_iterator i=m_GroupsList.begin() ; i!=m_GroupsList.end() ; ++i)
    {
        if((*i)->GetVuStartTime(ident, time) == TRUE)
            return time;
    }

    // Could not find that user.
    assert(0);
    return 0;
}
