/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef __MEMBUFFHANDLER_H__
#define __MEMBUFFHANDLER_H__

#include <assert.h>
#include <iosfwd>

#include "Fastlock.hpp"

#pragma warning (disable: 4127 4097 4663)


/* #########################################################################

    MemoryBufferHandler.hpp

    Defines classes used to store and iterate through the data received on
    the injector sockets until/while processing. The data stored is in HTTP  
    format.

   ######################################################################### */




/* #########################################################################

    CDataBuff

    Template mapped on pages from the virtual memory allocated by 
    CGlobalDataBuffHandler to simplify access to the data.
    That class should not be used by any other classes than those described 
    here after in order for the buffers to be safely released and allocated.
    This is because we chose not to do reference counting and buffer
    duplication might be expensive.

    Serialization: performed from client objects.

   ######################################################################### */

class CDataBuff
{
public:
    void    Init(CDataBuff *pNext = NULL)   { m_pNextBuff = pNext; m_dwUsedDataSize = 0; }

    CHAR *      GetDataPtr()                                { return m_acDataArray; }
    const CHAR *GetDataPtr()                        const   { return m_acDataArray; }

    DWORD   GetUsedDataSize()                       const   { return m_dwUsedDataSize; }
    void    SetUsedDataSize(DWORD dwNewDataSize)            { m_dwUsedDataSize = dwNewDataSize; }

    void        SetNextBuff(CDataBuff *pNextBuff)           { m_pNextBuff = pNextBuff; }
    CDataBuff * GetNextBuff()                       const   { return m_pNextBuff; }

    CDataBuff * operator+(DWORD dwInc);     // Return item stored in current + dwInc
                                            // position or NULL if it does
                                            // not exist.
private:
    CDataBuff * m_pNextBuff;
    DWORD       m_dwUsedDataSize;
    CHAR        m_acDataArray[1];   // Array of data. The size isn't
                                    // relevant since CDataBuff will
                                    // only be used in casts ie not
                                    // instantiated.

    //// Forbidden operations.
    const CDataBuff &operator=(const CDataBuff &);
    CDataBuff();    // Cannot create CDataBuffs because of the unpredictable data array size.
                    // Can only use it in casts.
    CDataBuff(const CDataBuff&);
};

std::ostream &operator << (std::ostream& os, const CDataBuff &dataBuff);


/* #########################################################################

    CGlobalDataBuffHandler

    Class used to allocate the data buffers that store the data received on 
    the connections. Buffers are allocated from a virtual memory area, 
    devided into buffers the size of the PC memory page. Each buffer is mapped
    to a CDataBuff structure that handles chaining.
    Shared by several CDataBuffHandler s, which request and release any number
    of buffers, depending on the size of data they want to store.

    Serialization: synchronizes the updates of the available buffers list 
    and the allocation of the memory mapped file.

   ######################################################################### */

class CGlobalDataBuffHandler
{
    // Default values.
    enum { DEF_MAX_DATA_BUFF_COUNT=64000 }; // Multiplied by the page size to compute
                                            // the MAXIMUM global buffer length (256 Mb).
                                            // If devided by the max number of users,
                                            // gives the average number of buffers 
                                            // available to each user.

    enum { DEF_GROWING_BUFF_COUNT=1000 };   // Number of buffers that are committed at once.
                                            // Each commit reserves memory for 
                                            // DEF_GROWING_BUFF_COUNT/WSABUF_COUNT requests slot.
                                            // Ideally, should be WSABUF_COUNT*(max number of
                                            // concurrent requests) = WSABUF_COUNT*4*(max nb VU).

public:
    CGlobalDataBuffHandler();
    ~CGlobalDataBuffHandler();

    static BOOL Init(DWORD dwMaxDataBuff=DEF_MAX_DATA_BUFF_COUNT, DWORD dwGrowingBuffCount=DEF_GROWING_BUFF_COUNT);
    static void UnInit();

    BOOL    RequestBuffs(DWORD dwBuffCount, CDataBuff *&pNewList, CDataBuff *&pLastItem); // Returns a list of dwBuffCount free buffers.
    void    ReturnBuffs(CDataBuff *);
    void    ReturnBuffs(CDataBuff *pFirstBuff, CDataBuff *pLastBuff, DWORD dwRetBuffCount);

    void    SetGrowBy(DWORD dwGrowBy)   { CLock lock(m_pFl); s_dwBytesGrownBy = dwGrowBy; }

    DWORD   GetDataBuffUsableSize() const   { return s_dwDataBuffUsableSize; }

#ifdef USE_PERF_DATA

public:
    // Pointers to performance counters.
    static LONG *   s_pRecvBuffAllocatedMem;
    static LONG *   s_pRecvBuffUsedMem;
    static LONG *   s_pDataStoredCounterRef;

#endif

protected:

    //// HELPERS. Serialization performed by calling methods.
    static BOOL AllocMemoryBuffer();                    // Create the global memory buffer.
    BOOL    CommitMemoryBuffer(DWORD dwNbBuffRequired); // Grows the number of memory buffers 
                                                        // available to m_freeDataBuffsList.

private:

    //// DATA.
    static  CFastLock * m_pFl;                  // Method-level protection against concurrent accesses.

    static  DWORD   s_dwMaxBuffCount;
    static  DWORD   s_dwGrowingBuffCount;

    static LPVOID   s_pvVirtualMemBuff;         // Pointer to the virtual memory buffer.

    CDataBuff *     m_freeDataBuffsList;        // Pointer to the first free buffer in the list.
    CDataBuff *     m_lastFreeDataBuffer;       // Pointer to the last free buffer in the list.

    DWORD           m_dwAllocatedBuffsCount;    // Whole number of buffers allocated.
    DWORD           m_dwFreeBuffsCount;         // Whole number of free buffers.

    static DWORD    s_dwMaxPhysBuffLength;      // Size of the hardware memory pages.
    static DWORD    s_dwDataBuffExtraInfoSize;  // Size used to store the buffer
                                                // info (CDataBuff private data).
    static DWORD    s_dwDataBuffUsableSize;     // Bytes in the buffers that are usable.

    static DWORD    s_dwBytesGrownBy;           // Size of each commit operation.
                                                // Multiple of the hardware memory
                                                // page size and GROWING_BUFF_COUNT.

    DWORD           m_dwCommittedMem;           // Bytes in the global virtual memory buffer
                                                // that are reserved.

    //// Forbidden operations.
    operator=(const CGlobalDataBuffHandler &);
    CGlobalDataBuffHandler(const CGlobalDataBuffHandler&);
};


class CdBLCA {};    // Data buff list chunk accessor template-specialising dummy class.
class CdBLA {};     // Data buff list accessor template-specialising dummy class.

template <class T>
class CDataBuffListBaseAccessor;
class CDataBuffHandler;

/* #########################################################################

    CDataBuffList

    Class used to handle a list of data buffers.

    Serialisation: not needed unless the same CDataBuffList is used by several
    threads.

   ######################################################################### */

class CDataBuffList
{
    friend CDataBuffListBaseAccessor<CdBLCA>;
    friend CDataBuffListBaseAccessor<CdBLA>;
    friend CDataBuffHandler;
    friend std::ostream &operator << (std::ostream& os, const CDataBuffList &dataBuffList);

    // All of these classes are provided with high level rights. Avoids other
    // not-really-aware classes to put the mess in there.

public:
    CDataBuffList();
    ~CDataBuffList()    {}

    DWORD   GetDataBuffUsableSize() const   { return m_globalDataBuffHandler.GetDataBuffUsableSize(); }
    BOOL    AllocateBuffs(DWORD dwNbItems); // Allocates dwNbItems buffers.

    BOOL    IsEmpty()   const   { return m_pDataBuffList == NULL; }

protected:
    // For use by the iterators.
    CDataBuff *         GetBuffList()           { return m_pDataBuffList; }
    const CDataBuff *   GetBuffList()   const   { return m_pDataBuffList; }

    void    ReleaseBuffList();

    void    AddBuffList(const CDataBuffList &buffList, CDataBuff *pLastBuff, DWORD dwNbItemsToAdd, CDataBuffList &unusedBuffList);
    const CDataBuffList &operator+=(const CDataBuffList &buffList); // Add an existing buffer list.

private:
    //// Helpers.

    //// Data.
    static CGlobalDataBuffHandler   m_globalDataBuffHandler;    // Structure the buffers are
                                                                // allocated from.

    CDataBuff * m_pDataBuffList;    // Pointer to the first data buffer in the list.
    CDataBuff * m_pLastItem;        // Pointer to the last data buffer in the list.
    DWORD       m_dwItemsCount;     // Number of items stored in the list.

    //// Forbiddden operators.
    const CDataBuffList &operator=(const CDataBuffList &);  // Dangerous use since may
                                                            // require updates
                                                            // on the CGlobalDataBuffHandler.
    CDataBuffList(CDataBuffList &);                         // Dangerous for the same
                                                            // reasons.
};

std::ostream &operator << (std::ostream& os, const CDataBuffList &dataBuffList);


/* #########################################################################

    CDataBuffListBaseAccessor

    Class used to browse through the CDataBuffs contained in a
    CDataBuffList. Defines string operation methods that will help
    perform common operations such as browsing and parsing.

    That class assumes that no empty buffers are stored in the list.

    CDataBuffListBaseAccessor will be instantiated as CDataBuffListAccessor 
    and CDataBuffListChunkAccessor, the first being used to parse standard
    HTTP buffers, the second capable for seemlessly retrieving the real data
    out of the chunked-formatted data.

    Serialisation: not needed as long as it is used by one thread at a time.

    CDataBuffListBaseAccessor has no use but to implement CDataBuffListChunkAccessor
    and CDataBuffListAccessor. Differentiation is obtained by specializing some of
    CDataBuffListBaseAccessor's methods. This avoids using virtual functions.
    CdBLCA and CdBLA are void classes that differentiate the chunk and the classic
    accessors.

   ######################################################################### */


template <class T>
class CDataBuffListBaseAccessor
{
    friend std::ostream &operator << (std::ostream& os, const CDataBuffListBaseAccessor<CdBLCA> &dataBuffList);
    friend std::ostream &operator << (std::ostream& os, const CDataBuffListBaseAccessor<CdBLA> &dataBuffList);
    friend CDataBuffListBaseAccessor<CdBLCA>;
public:
    CDataBuffListBaseAccessor(CDataBuffList *pDataBuffList = NULL);
    CDataBuffListBaseAccessor(const CDataBuffListBaseAccessor& dba);

    BOOL    AdjustIterator();   // Adjust the iterator position.

    //// Parsing methods.
    DWORD Skip(DWORD dwItems);  // Try to move the iterator dwItems items further. Returns the number of items it skipped actually.
    BOOL SkipWS();              // Return TRUE if managed to move to something else than a white space.
    BOOL SkipTillWS();          // Return TRUE if found a WS after a word.
    BOOL SkipCRLF();            // Return TRUE if a CRLF was found, and move one char further.
    BOOL SkipChar(CHAR c);      // Return TRUE if char c was found, and move one char further.
    BOOL SkipTillAny(CHAR* str, DWORD dwStrLen);
    void SkipEnd();

    BOOL NextChar(CHAR &nextChar) const;
    BOOL MoveNextChar();
    BOOL FindChar(CHAR c, DWORD &dwPos) const;                      // Return the number of chars separating current pos from "c".
    BOOL FindAny(LPCSTR str, DWORD dwStrLen, DWORD &dwPos) const;   // Return the number of chars separating current pos from the first char in "str".

    BOOL MatchFirstWord(LPCSTR str, DWORD strSize) const;           // Case sensitive.
    BOOL MatchFirstWordToLower(LPCSTR str, DWORD strSize) const;    // Where str is lower case.
    BOOL IsCRLF();              // Return TRUE if current char is CR and next is LF, and move to the next char.

    BOOL IsEndOfBuffs();

    DWORD   ComputeDataLeft() const;
    DWORD   GetCurrentChunkSizeLeft()   const   { return m_dwCurrentChunkSizeLeft; }

    BOOL ReadDWORD(DWORD &res)  const;      // Parse the next word found into a DWORD.
    BOOL ReadHexDWORD(DWORD &res)  const;   // Parse the next word found into a DWORD, as an hexadecimal number.

    BOOL ForEach(BOOL (__cdecl *callBackFct)(CHAR, void *param), void *callBackFctParam);   // Applies callBackFct to all items while 
                                                                                            // callBackFct returns TRUE.

    // Depending on bCaseSensitive, performs a case-sensitive or insensitive search of str in buffers delimitted by this and IterStopAt.
    BOOL ContainString(const CHAR *str, DWORD dwStrLen, const CDataBuffListBaseAccessor<T> &IterStopAt, BOOL bCaseSensitive)    const;

    // Assuming that the data at the current pos is single byte characters,
    // copy dwCharCount chars into destStr starting from current pos.
    BOOL BuildString(std::string& destStr, DWORD dwCharCount) const;

    // Assuming that the data at the current pos is bytes of binary data,
    // get the size of the data available starting from current pos.
    DWORD GetDataSize() const;

    // Assuming that the data at the current pos is bytes of binary data,
    // copy dwByteCount bytes into pbDest starting from current pos.
    BOOL BuildData(BYTE* pbDest, DWORD dwByteCount) const;

    BOOL SkipStr(LPCSTR str, DWORD dwStrLen, BOOL bCaseSensitive);      // Skip next occurrence of string str.
    BOOL MoveToStr(LPCSTR str, DWORD dwStrLen, BOOL bCaseSensitive);    // Skip all chars till the str string.

    // Implemented only for CDataBuffListAccessor:
    void Rewind();
    BOOL ReadNextWord(std::string &destStr)  const;                     // Copy the next word found (delimitted by spaces, tabs ...) into destStr.
    BOOL ReadNextWord(DWORD &dwDest)  const;                            // Copy the next DWORD value (delimitted by spaces, tabs ...) into dwDest.
    BOOL MoveNextBuff();

    //// Operators.
    const CDataBuffListBaseAccessor<T> &operator+=(DWORD dwInc);
    const CDataBuffListBaseAccessor<T>  operator+ (DWORD dwInc)    const;   // Return item stored in current + dwInc
                                                                            // position or NULL if it does not exist.
    DWORD operator- (const CDataBuffListBaseAccessor<T> &nextIter) const;   // Computes the number of items separating the two iterators.

    BOOL operator!=(const CDataBuffListBaseAccessor<T> &)  const;
    BOOL operator==(const CDataBuffListBaseAccessor<T> &)  const;
    BOOL operator< (const CDataBuffListBaseAccessor<T> &)  const;
    BOOL operator<=(const CDataBuffListBaseAccessor<T> &)  const;
    BOOL operator> (const CDataBuffListBaseAccessor<T> &)  const;
    BOOL operator>=(const CDataBuffListBaseAccessor<T> &)  const;

    BOOL operator!=(void *) const;
    BOOL operator==(void *) const;
    void *operator= (void *);

    void MoveIterTo(const CDataBuffListBaseAccessor<CdBLA> &);

    const CDataBuffListBaseAccessor<T> &operator= (const CDataBuffListBaseAccessor<T> &);

    CDataBuffListBaseAccessor<T> &operator++(); // WARNING: no guarantee that will point to a valid data.
    CHAR        operator*();

    CDataBuff * GetCurrDataBuff()   { return m_pCurrDataBuff; }

protected:
    //// Helpers.
    BOOL    IsBefore(const CDataBuffListBaseAccessor<T> &other) const;
    DWORD   ComputeDistance(const CDataBuffListBaseAccessor<T> &first, const CDataBuffListBaseAccessor<T> &second) const;
    typedef enum { SKIP, FIND } EFindSkip;
    BOOL    FindSkipStr(LPCSTR str, DWORD dwStrLen, EFindSkip findSkip, BOOL bCaseSensitive);

    // Methods that iterate through the buffers.
    void    CurrentChar(CHAR &curChar) const;

protected:
    CDataBuffList * m_pDataBuffList;  // The CDataBuffList we are working on.
    CDataBuff *     m_pCurrDataBuff;
    DWORD           m_dwPosInCurrDataBuff;

protected:
    // Specific to chunks:
    BOOL    ckReadChunkSize(BOOL bSkipLastCkCRLF = TRUE);
    BOOL    ckMoveNextChunk();

protected:
    // Chunked data parsing info. For use only by CDataBuffListChunkAccessor.
    DWORD   m_dwCurrentChunkSizeLeft;   // Data left to parse for current chunk.
};  // class CDataBuffListBaseAccessor

typedef CDataBuffListBaseAccessor<CdBLCA>    CDataBuffListChunkAccessor;
typedef CDataBuffListBaseAccessor<CdBLA>     CDataBuffListAccessor;

inline std::ostream &operator <<(std::ostream &os, const CDataBuffListChunkAccessor &iterMem)
{
    // Write till the end of current buffer.
    os.write(&(iterMem.m_pCurrDataBuff->GetDataPtr()[iterMem.m_dwPosInCurrDataBuff]), iterMem.m_pCurrDataBuff->GetUsedDataSize()-iterMem.m_dwPosInCurrDataBuff);

    // Write the remaining buffers in the list.
    if(iterMem.m_pCurrDataBuff->GetNextBuff() != NULL)
        os << *iterMem.m_pCurrDataBuff->GetNextBuff();

    return os;
}

inline std::ostream &operator <<(std::ostream &os, const CDataBuffListAccessor &iterMem)
{
    // Write till the end of current buffer.
    os.write(&(iterMem.m_pCurrDataBuff->GetDataPtr()[iterMem.m_dwPosInCurrDataBuff]), iterMem.m_pCurrDataBuff->GetUsedDataSize()-iterMem.m_dwPosInCurrDataBuff);

    // Write the remaining buffers in the list.
    if(iterMem.m_pCurrDataBuff->GetNextBuff() != NULL)
        os << *iterMem.m_pCurrDataBuff->GetNextBuff();

    return os;
}


/* #########################################################################

    CDataBuffHandler

    Class that will handle the memory buffers used by a specific socket.
    The request will use buffers allocated by CDataBuffHandler to perform
    receive operations, store them if needed into CDataBuffHandler and 
    then make that CDataBuffHandler available to the code in charge with the 
    processing of the results.

    Serialisation: no need since one CDataBuffHandler is associated with 
    one socket.

   ######################################################################### */

class CDataBuffHandler
{
public:
    enum { WSABUF_COUNT=5 };    // The number of buffers available for a receive operation.
                                // Here, 5 is experimental: it allows for the maximum SSL message
                                // (16k) to be held.

    CDataBuffHandler() : m_dwRemainingBuffsInMappedList(0)  {}
    ~CDataBuffHandler() { ClearBuffer(); }

    WSABUF*     AllocWsaBuffList();
    void        ReturnWsaBuffList(DWORD dwBytesStored);

    BOOL        IsEmpty()   const   { return m_dataBuffList.IsEmpty(); }

    const CDataBuffList &   GetWholeBuffList()  const { return m_dataBuffList; }
    CDataBuffList &         GetWholeBuffList()  { return m_dataBuffList; }

    void    ClearBuffer();

    DWORD   GetDataBuffUsableSize() const   { return m_dataBuffList.GetDataBuffUsableSize(); }

private:
    CDataBuffList   m_dataBuffList;                 // List of the data buffers associated with that connection.
    WSABUF          m_wsaBuffArray[WSABUF_COUNT];   // For use in scatter-gather receive operations
                                                    // instead of the CDataBuff s which are not 
                                                    // WSARead-compliant.
    CDataBuffList   m_wsaMappedBuffList;            // List of the data buffers mapped to m_wsaBuffArray.
    DWORD           m_dwRemainingBuffsInMappedList; // Number of buffers remaining in m_wsaMappedBuffList from the
                                                    // last ReturnWsaBuffList and can be reused.

    //// Forbidden.
    CDataBuffHandler(const CDataBuffHandler&);
    const CDataBuffHandler &operator = (const CDataBuffHandler &dataBuffHandler);
};

std::ostream &operator<<(std::ostream &os, CDataBuffHandler &);

#ifndef _DEBUG
// To avoid unending builds each time we touch to a method.
#   define WILL_INLINE __inline
#   include "MemoryBufferHandler.inl"
#else
#   define WILL_INLINE
#endif

#include "MemoryBufferHandler.tpp"

#endif
