/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/


#include "stdafx.h"

#include "TEMainStructure.hpp"
#include "TagParsing.hpp"
#include "CyrHelper.inl"
#include "mutex.hpp"
#include "Common.hpp"           // For BeginThreadEx.
#include "TEHttpRecords.h"

#ifndef CYRANO_TP_STANDALONE
#   include "RemoteRepository.h"
CRemoteRepository* Repos = NULL;
#endif

#include "ipmutex.hpp"


#ifdef USE_PERF_DATA
#   include "PerfCounters.hpp"
    using namespace OpenSTA_Performance;
#   include "PerfCounterMap.hpp"
#endif

#ifndef CYRANO_TP_STANDALONE
#	include <GlobalVariable.h>  // For CyrGlobalVariableException
#endif

#include "TestExecuter_i.h"

#define	ODLFILE "TEHttp.odl"
const string TPTEXT	= "TestExecuter_web";

LogTrace* g_pLogTrace = NULL;

#ifdef USE_PERF_DATA
OpenSTA_Performance::CPerfSource* g_pPerformance = NULL;
#endif

#ifndef CYRANO_TP_STANDALONE
    static TestExecuterImpl *g_pTExecuterImpl = NULL;
    TestExecuter_ptr TExecuter_ptr = NULL;
#endif

string exedir;      // TODO: change this! For use as an external variable in CORBARepository, CECScriptFile
string filevardir;  // TODO: change this! For use as an external variable in CORBARepository.cpp
CAsyncVirtualUserController* _threadPoolController = NULL;  // TODO: change this! For use in TContext and TestExecuter_i.
BOOL g_bTCPKeepAlives = FALSE;          // Flag controlling the presence of TCP keepalives.
unsigned long g_ulKeepAliveInterval = 0;// TCP keepalive interval in milliseconds.

const char g_successMsg[] = "Script Completed Successfully.";
const char g_failureMsg[] = "Script Execution Failed.";

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

CMainExecuterStructure::CMainExecuterStructure()
    : m_pServer(NULL), m_pThreadPoolController(NULL), m_pRule(NULL),
      m_bInitialised(FALSE), m_pTrace(NULL), m_lCachedTestVUCount(0),
      m_pLogTrace(NULL)
#ifndef __DLLSA__
    , m_cmdLine(0, NULL, ""), m_lOffset(0), m_bStepByStep(FALSE), m_bFirstTaskGroup(TRUE)
#else
    , m_hThreadPoolHandlerThread(NULL)
#endif
{
    m_configFile.SetFile("TestExecuter_web.ini");
}


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

CMainExecuterStructure::~CMainExecuterStructure()
{
    if(m_pThreadPoolController)
    {
        delete m_pThreadPoolController;
        m_pThreadPoolController = _threadPoolController = NULL;
    }

    if(m_bInitialised)
        UnInit();

    if(m_pTrace)
        delete m_pTrace;

    if(m_pLogTrace)
        delete m_pLogTrace;
}


// ----------------------------------------------------------------------------
// Initialising methods. Should be called before each Run.
// ----------------------------------------------------------------------------

#ifdef __DLLSA__

BOOL CMainExecuterStructure::Init(HWND hWnd, char *TofFile, char *OdlDir)
{
    if(m_bInitialised == FALSE)
    {
        if(hWnd == NULL)
        {
            // Detected error while checking args
            LogError("Handle for window is NULL!");
            return FALSE;
        }

        TCHAR szDirAndName [MAX_PATH];
        GetModuleFileName(NULL, szDirAndName, sizeof(szDirAndName));
        exedir = filename_dir(szDirAndName);
        m_zExeName = filename_base(szDirAndName);

        m_hWnd = hWnd;

        if(CommonInit() == FALSE)
            return FALSE;

        m_zTofFileNameAndPath = TofFile;
        m_zTofFileDir = TofFile;

        // Remove TOF file name
        string::size_type lastSlash = m_zTofFileDir.find_last_of('\\');
        if(lastSlash != string::npos)
            m_zTofFileDir.erase(lastSlash);
        else
            m_zTofFileDir = "";

        filevardir = m_zTofFileDir;

        m_zCurrentDir = OdlDir;

        m_bInitialised = TRUE;
    }

    return TRUE;
}

#else   // __DLLSA__ not defined.

BOOL CMainExecuterStructure::Init(int argc, char* argv[])
{
    if(m_bInitialised == FALSE)
    {
        if(ParseCommandLine(argc, argv) == FALSE)
            return FALSE;

        // Install/uninstall the performance counters. Must be done before most of other classes' InitClass.
        if(m_cmdLine.option_supplied("-u"))
        {
#ifdef USE_PERF_DATA
            return InitPerfCounters(PrfCtrsUninstall);
#else
            TraceMsg("Performance counters not supported.");
            cout << "Performance counters not supported." << endl;
            return FALSE;
#endif  // USE_PERF_DATA
        }
        else
        {
            if(m_cmdLine.option_supplied("-i"))
            {
#ifdef USE_PERF_DATA
                return InitPerfCounters(PrfCtrsInstall);
#else
                TraceMsg("Performance counters not supported.");
                cout << "Performance counters not supported." << endl;
                return FALSE;
#endif  // USE_PERF_DATA
            }
        }

        // Remember the command line args.
        m_zTestName         = m_cmdLine.get_value_for("-n");
        m_strRepositoryNode = m_cmdLine.get_value_for("-r", "localhost");
        m_strTestRun        = m_cmdLine.get_value_for("-t");

#ifdef USE_PERF_DATA
        if(InitPerfCounters(PrfCtrsUse) == FALSE)
            return FALSE;
#endif

        TCHAR szDirAndName[MAX_PATH];
        GetModuleFileName(NULL, szDirAndName, MAX_PATH);
        exedir = m_zCurrentDir = filename_dir(szDirAndName);
        m_zExeName = filename_base(szDirAndName);

        if(CommonInit() == FALSE)
            return FALSE;

        if(InitialiseCORBA(argc, argv) == FALSE)
            return FALSE;

        m_bInitialised = TRUE;
    }

    return TRUE;
}

InjectorControl_ptr CMainExecuterStructure::GetInjectorControl() const
{
    return g_pTExecuterImpl->GetInjectorControl();
}

const char* CMainExecuterStructure::GetRegisterName() const
{
    return g_pTExecuterImpl->GetRegisterName();
}

#endif  // __DLLSA__

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

BOOL CMainExecuterStructure::CommonInit()
{
    // Get the process ID.
    m_dwProcessId = GetCurrentProcessId();

    CFastLock::Init();

    string zTraceFileName;

#ifdef __DLLSA__
    // Use max trace options.
    int nTraceVal = TRCLEV_HIGH;
    m_configFile.Get("TraceFileName", zTraceFileName, (exedir + "\\TModeler_Web.log").c_str(), "FILES");
#else
    try
    {
        m_zUnderscoredHostIpAddress = GetHostIPAddressAsUnderscoreString();
    }
    catch(...)
    {
        LogError("Failed to resolve host name to IP address.");
        return FALSE;
    }
    m_configFile.Get("TraceFileName", zTraceFileName, (exedir + "\\TExecuter_htp_" + m_zUnderscoredHostIpAddress + ".log").c_str(), "FILES");

    long nTraceVal = m_cmdLine.get_int_value_for("-v", 0);
    long lLevel;
    m_configFile.Get("TraceLevel", lLevel, 0, "FILES");
    if(nTraceVal < lLevel)
        nTraceVal = lLevel; // Use the highest of trace levels.
#endif

    CreateTrace(nTraceVal, zTraceFileName.c_str());

    if(!TRecord::InitClass())
    {
        TraceMsg("ERROR initializing the TRecord derivatives.");
        return FALSE;
    }
    if(CAsyncVirtualUser::InitClass() == FALSE)
    {
        TraceMsg("ERROR initializing the CAsyncVirtualUser class.");
        return FALSE;
    }

#ifndef __DLLSA__
    m_zStatBaseFileName = strip_extension(CyrGetStatFileName(m_zUnderscoredHostIpAddress.c_str(), true));
#endif

    // Init class internals.
    long lMaxDataBuff, lGrowingBuffCount, lMaxConcurrentSSLReq, lSSLGrowingBuffCount, lTCPKeepAlives;

    m_configFile.Get("TCP_KeepAlives",                  lTCPKeepAlives, 0,              "SOCKET");
    m_configFile.Get("KeepAliveInterval",               (long&)g_ulKeepAliveInterval, 300000, "SOCKET");
    if(lTCPKeepAlives)
    {
        g_bTCPKeepAlives = TRUE;
        TraceMsg("Using TCP keepalives.");
    }
    else
        TraceMsg("Not using TCP keepalives.");
    m_configFile.Get("MaxSocketDataBuffersCount",       lMaxDataBuff, 64000,            "SOCKET");
    m_configFile.Get("SocketDataBuffersGrowingCount",   lGrowingBuffCount, 2000,        "SOCKET");
    m_configFile.Get("MaxSSLConcurrentReq",             lMaxConcurrentSSLReq, 2000*4,   "SOCKET");
    m_configFile.Get("SSLGrowingBuffCount",             lSSLGrowingBuffCount, 1000,     "SOCKET");

    if(CAsyncTCPSocket::Init() != 0 
        || CAsyncSecureSocket::Init() != NO_ERROR 
        || CGlobalDataBuffHandler::Init(lMaxDataBuff, lGrowingBuffCount) == FALSE  // Use default values.
#ifndef __DLLSA__
        || CStatistics::Init(m_zStatBaseFileName + ".stat", m_zStatBaseFileName + ".urls") == FALSE
#else
        || CStatistics::Init("", "") == FALSE
#endif
        || CAsyncHTTPRequest::InitClass(lMaxConcurrentSSLReq, lSSLGrowingBuffCount) == FALSE)   // Max number of users times max requests per user.
    {
        TraceMsg("Error initialising structures.");
        return FALSE;
    }

    // Initialize the parser.
    InitTagParsing();

    return TRUE;
}

#ifndef CYRANO_TP_STANDALONE
void CMainExecuterStructure::SetTimeOffset(long lOffset)
{
    m_lOffset = lOffset;
    CStatistics::SetTimeOffset(m_lOffset);
}
#endif

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void CMainExecuterStructure::UnInit()
{
#ifndef __DLLSA__
    CStatistics::Release();
#endif
    CAsyncVirtualUser::UninitClass();
    CAsyncHTTPRequest::UninitClass();

    CGlobalDataBuffHandler::UnInit();
    CAsyncSecureSocket::Uninit();
    CAsyncTCPSocket::Uninit();

    if(m_pServer)
    {
        delete m_pServer; // End of CORBA communication.
        m_pServer = NULL;
    }

    TRecord::UninitClass();

    m_bInitialised = FALSE;
}


void CMainExecuterStructure::CreateTrace(int nTraceVal, LPCSTR zTraceFileName)
{
    LPCSTR zTraceList   = LogTrace::LevelToTraceList(nTraceVal);
    DWORD dwViewSize    = 10485760U;

    if(zTraceFileName != NULL)
    {
        string s;
        istringstream is(zTraceList);
        BOOL bContinue = TRUE;

        m_pTrace = new CTrace(zTraceFileName, dwViewSize);
        if(m_pTrace)
        {
            GetTraceStream(m_pTrace) << "Trace started at level " << nTraceVal << ".\r\n";
            ReleaseTraceStream();

            while(bContinue)
            {
                is >> s;

                if(is.fail() == FALSE)
                {
                    m_pTrace->AddTrace(s.c_str());
                }
                else
                {
                    bContinue = FALSE;
                }
            }
        }
    }

#ifdef CYRANO_TP_STANDALONE
    m_pLogTrace = new LogTrace(m_pTrace, m_hWnd);
#else
    m_pLogTrace = new LogTrace(m_pTrace);
#endif
    if(m_pLogTrace)
        g_pLogTrace = m_pLogTrace;
}


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

BOOL CMainExecuterStructure::Run()
{
    TraceMsg("OpenSTA starting");

    try
    {
        if(RunThreadPoolController() == FALSE)
            return FALSE;
    }
    catch(const string &es)
    {
        // Detected error
        TraceMsg(es);
        LogError(es.c_str());
        return FALSE;
    }
    catch(const mdbIOException& e)
    {
        TraceMsg(e.What() + ": " + e.ioname());
        LogError((e.What() + ": " + e.ioname()).c_str());
        return FALSE;
    } 
    catch(const mdbParseException& e) 
    {
        TraceMsg(e.What());
        LogError(e.What().c_str());
        return FALSE;
    }
    catch(const CyrRepositoryException& e) 
    {
        TraceMsg(e.what());
        LogError(e.what().c_str());
        return FALSE;
    }
    catch(const exception& e)
    {
        TraceMsg(e.what());
        LogError(e.what());
        return FALSE;
    }
#ifndef CYRANO_TP_STANDALONE
    catch(const CORBA::Exception& e)
    {
        TraceMsg("Caught CORBA::" + CORBAExString(e));
        LogError(("Caught CORBA::" + CORBAExString(e)).c_str());
        return FALSE;
    }
#endif
    catch(...)
    {
        TraceMsg("Caught exception while executing the thread pool.");
        LogError("General exception while running the thread pool.");
        return FALSE;
    }

    return TRUE;
}

#ifndef __DLLSA__

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

BOOL CMainExecuterStructure::ParseCommandLine(int argc, char* argv[])
{
    // Parse the command line.
    try
    {
        m_cmdLine.init(argc, argv, "TestExecuter: ");

        m_cmdLine.define_option("-u",  CLO_NOVAL);      // Uninstall.
        m_cmdLine.define_option("-i",  CLO_STRINGVAL);  // Install DLL full path.
        m_cmdLine.define_option("-v",  CLO_INTVAL);     // Trace level.
        m_cmdLine.define_option("-d",  CLO_NOVAL);      // Debug mode.
        m_cmdLine.define_option("-r",  CLO_STRINGVAL);  // Node where the CORBA repository lives.
        m_cmdLine.define_option("-n",  CLO_STRINGVAL);  // Test name.
        m_cmdLine.define_option("-t",  CLO_STRINGVAL);  // Test run string.
        m_cmdLine.define_option("-s",  CLO_STRINGVAL);  // Signal name.

	    m_cmdLine.validate();
    }
    catch(const string &err)
    {
        TraceMsg("Parameter error: " + err);
        return FALSE;
    }

    return TRUE;
}

#   ifdef USE_PERF_DATA

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

BOOL CMainExecuterStructure::InitPerfCounters(ePrfCtrsAction installType) const
{
    g_pPerformance = new OpenSTA_Performance::CPerfSource(L"OpenSTA_TE", L"OpenSTA_Open", L"OpenSTA_Collect", L"OpenSTA_Close");
    if(g_pPerformance == NULL)
    {
        TraceMsg("ERROR: could not access performance counters.");
        return FALSE;
    }

    if(g_pPerformance->ParseMap(OpenSTAPerf, MAP_SIZE(OpenSTAPerf)) != ERROR_SUCCESS)
    {
        TraceMsg("ERROR: failed to parse performance map.");
        delete g_pPerformance;
        g_pPerformance = NULL;
        return FALSE;
    }

    switch(installType)
    {
    case PrfCtrsUse:
        {
            if(g_pPerformance->AdjustTitleIndexes() != ERROR_SUCCESS)
            {
                TraceMsg("ERROR: failed to adjust performance counter numbers.");
                delete g_pPerformance;
                g_pPerformance = NULL;
                return FALSE;
            }

            DWORD dwRes = g_pPerformance->Initialize();
            if(dwRes != ERROR_SUCCESS && dwRes != ERROR_ALREADY_EXISTS)
            {
                TraceMsg("ERROR: failed to initialise performance source.");
                delete g_pPerformance;
                g_pPerformance = NULL;
                return FALSE;
            }

            TraceMsg("Performance counters successfully initialized.");
        }

        break;

    case PrfCtrsInstall:
        {
            // Install the counters.
            WCHAR lpMultiByteStr [256];
            MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, m_cmdLine.get_value_for("-i").c_str(), -1, lpMultiByteStr, 256);
            if(g_pPerformance->Install(lpMultiByteStr) != ERROR_SUCCESS)
            {
                TraceMsg("Failed installing the performance counters.");
                cout << "Failed installing the performance counters." << endl;
                return FALSE;
            }
            else
            {
                TraceMsg("Performance counters installed.");
                cout << "Performance counters installed." << endl;
                return TRUE;
            }
        }
        break;

    case PrfCtrsUninstall:
        {
            g_pPerformance->Uninstall();
            TraceMsg("Performance counters removed.");
            cout << "Performance counters removed." << endl;
        }
        break;
    };

    return TRUE;
}

void CMainExecuterStructure::ReleasePerfCounters() const
{
    delete g_pPerformance;
    g_pPerformance = NULL;
}

#   endif  // USE_PERF_DATA


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

BOOL CMainExecuterStructure::InitialiseCORBA(int argc, char* argv[])
{
    // Initialise CORBA.
    try
    {
        assert(m_pServer == NULL);
        m_pServer = new CORBAServer(argc, argv);
    }
    catch(const CORBA::Exception &e)
    {
        TraceMsg("Caught CORBA::" + CORBAExString(e));
        return FALSE;
    }
    catch(...)
    {
        // Note TPLog::Setup was not called yet.
        TraceMsg("Error connecting to the CORBA server.");
        return FALSE;
    }

#ifndef CYRANO_TP_STANDALONE
    // Create a remote repository object.
    try
    {
        // Create a remote repository object.
        Repos = new CRemoteRepository(m_pServer, m_strRepositoryNode);
    }
    catch(const CORBA::Exception &e)
    {
        TraceMsg("Caught CORBA::" + CORBAExString(e));
        LogError(("Caught CORBA::" + CORBAExString(e)).c_str());
        return FALSE;
    }
    catch(...)
    {
        TraceMsg("Caught Exception on startup, is Nameservice Running?");
        LogError("Caught Exception on startup, is Nameservice Running?");
        return FALSE;
    }
#endif

    return TRUE;
}

BOOL CMainExecuterStructure::ReleaseCORBA()
{
    // Catch any exceptions during cleanup. This could happen if
    // a thread was hard killed, leaving some object in an undefined 
    // state.

    try
    {
#ifndef CYRANO_TP_STANDALONE
        if(Repos)
        {
            delete Repos;
            Repos = NULL;
        }

        if(g_pTExecuterImpl)
        {
            // Disable new incoming calls.
            g_pTExecuterImpl->Deactivate();

            CORBA::release(TExecuter_ptr);
            g_pTExecuterImpl->_remove_ref();    // Should call delete this.
            g_pTExecuterImpl = NULL;
            TExecuter_ptr = NULL;
        }
#endif
    }
#ifndef CYRANO_TP_STANDALONE
    catch(const CyrGlobalVariableException& e) 
    {
        TraceMsg(string((const char*)e.what));
        return FALSE;
    }
    catch(const CORBA::Exception& e)
    {
        string msg = "Caught CORBA::" + CORBAExString(e) + " in clean-up";
        TraceMsg(msg);
        return FALSE;
    }
#endif
    catch(const CyrRepositoryException& e) 
    {
        string logstr = e.what() + ',' + e.where() + " Status: " + int_to_string(e.status()).c_str();
        TraceMsg(logstr);
        return FALSE;
    }
    catch(const exception &e)
    {
        string msg = string("Caught exception in clean-up :") + string(e.what());
        TraceMsg(msg);
        return FALSE;
    }
    catch(...)
    {
        TraceMsg(string("Caught exception during clean-up."));
        return FALSE;
    }

    return TRUE;
} // End of ReleaseCORBA.

#endif // ndef __DLLSA__


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

BOOL CMainExecuterStructure::RunThreadPoolController()
{
    //// Read in odl file to create the Script Rules.

    string zOdlFile = m_zCurrentDir;
    if(!m_zCurrentDir.empty() && *m_zCurrentDir.rbegin() != '\\')
        zOdlFile += FN_PATHSEP;
    zOdlFile += ODLFILE;

    TraceMsg("Using odl file: " + zOdlFile);

    try
    {
        m_pRule = new TScriptRule(zOdlFile);    // Note: m_pRule is used solely by the VU controller. Could be created there?
    }
    catch(const mdbParseException &e)
    {
        TraceMsg("ERROR while creating the script rule: " + e.What());
        return FALSE;
    }
    catch(const mdbIOException &e)
    {
        TraceMsg(e.What() + ": " + e.ioname());
        return FALSE;
    } 

#ifndef CYRANO_TP_STANDALONE
    // Create an instance of the Test Executer implementation object.
    // If NOT in standalone mode, a corresponding CORBA object is also created.
    StepByStepScriptExecuterImpl* pSBS = new StepByStepScriptExecuterImpl(m_pServer);
    if(pSBS == NULL)
    {
        TraceMsg("ERROR: Could not create the Step-by-step Test Executer CORBA object.");
        return FALSE;
    }
    g_pTExecuterImpl = new TestExecuterImpl(m_pServer, pSBS);
    if(g_pTExecuterImpl == NULL)
    {
        TraceMsg("ERROR: Could not create the Test Executer CORBA object.");
        delete pSBS;
        return FALSE;
    }

    // Query the step-by-step control to see if we are in step-by-step mode.
    StepByStepControl_var& sbs = g_pTExecuterImpl->GetStepByStep()->GetStepByStepControl();
    if(!CORBA::is_nil(sbs))
    {
        try {
            m_bStepByStep = sbs->IsStepByStep();
        } catch(...) {
            m_bStepByStep = FALSE;
        }
    }
    else
        m_bStepByStep = FALSE;

	/*
	//use IC as it can cope across the web
    // Get the time offset.
	//THIS CODE ISN'T USED

    CECInetInfo inet;
	CORBA::Object_var obj = m_pServer->FindObjectWithName(TimeService::CLASS_NAME,
        inet.GetIPAddressAsString(m_strRepositoryNode).c_str());
	TimeService_var TS = TimeService::_narrow(obj);
	CTTime tm;
	tm.now();
	*/
    SetTimeOffset(GetInjectorControl()->GetTimeOffset());

    TExecuter_ptr = g_pTExecuterImpl->_this();

    if(m_cmdLine.option_supplied("-d")) // If run in debug mode, register as such in order for the test manager to use the current debugging process instead of starting a new one.
    {
        if(RegisterCurrentDebugExecuter() == FALSE)
            return FALSE;
    }
#endif

    // Start the thread pool controller.

    _threadPoolController = m_pThreadPoolController = new CAsyncVirtualUserController(m_pRule, m_pTrace);

    if(m_pThreadPoolController == NULL)
    {
        TraceMsg("ERROR: could not create the thread pool controller.");
        return FALSE;
    }

    DWORD dwThreadPoolSize, dwNbConcurrentThreads, dwInitialVUCount, dwVUCountGrowBy, dwBrowserParallelism;
    EvaluateTPParams(dwThreadPoolSize, dwNbConcurrentThreads, dwInitialVUCount, dwVUCountGrowBy, dwBrowserParallelism);

    if(m_pThreadPoolController->Init(dwThreadPoolSize, dwNbConcurrentThreads, dwInitialVUCount, dwVUCountGrowBy, dwBrowserParallelism) != 0)
    {
        TraceMsg("ERROR: could not initialize the thread pool controller.");
        delete m_pThreadPoolController; _threadPoolController = m_pThreadPoolController = NULL;
        return FALSE;
    }

    // Note that in the modeler mode, the VU controller will automatically start a user uppon EVENT_TYPE_THREADPOOL_START receival.
    // The threadpool is started either in current thread (standalone exe mode) or in a new thread (modeler's dll version).
#ifdef __DLLSA__
    if(StartThreadPoolHandlerThread() == FALSE)
    {
        TraceMsg("ERROR: could not start the thread pool thread.");
        delete m_pThreadPoolController; _threadPoolController = m_pThreadPoolController = NULL;
        return FALSE;
    }
#else
    // Name the stats files to retrieve after the test.
    string strFile = strip_extension(CyrGetStatFileName(m_zUnderscoredHostIpAddress.c_str(), false));
    string strFileWithExt = strFile + ".stat";
    TraceMsg("Registering results file: " + strFileWithExt, "full");
    GetInjectorControl()->SetResultsFile(strFileWithExt.c_str());
    strFileWithExt = strFile + ".urls";
    TraceMsg("Registering results file: " + strFileWithExt, "full");
    GetInjectorControl()->SetResultsFile(strFileWithExt.c_str());
    strFileWithExt = filename_base(m_pTrace->GetTraceFileName());
    TraceMsg("Registering log file: " + strFileWithExt, "full");
    GetInjectorControl()->SetResultsFile(strFileWithExt.c_str());

    // Let the local daemon know that the object has been registered.
    CIPSignal condition;
    if(  !condition.attach(m_cmdLine.get_value_for("-s").c_str())
       ||!condition.isValid()
       ||!condition.signal())
    {
        TraceMsg("ERROR: failed to notify the daemon that the T.E. object was registered.");
        delete m_pThreadPoolController; _threadPoolController = m_pThreadPoolController = NULL;
        return FALSE;
    }

    ThreadPoolHandler(this);

    // Test commander requested to stop the executer.
    if(m_pThreadPoolController)
        m_pThreadPoolController->Stop();
#endif

    return TRUE;
}

struct StartTaskGroupThreadData
{
    StartTaskGroupThreadData(CMainExecuterStructure* pThis, SharedPtr<TaskGroupDefinition>& pTaskGroup)
        : m_pThis(pThis), m_pTaskGroup(pTaskGroup) {}
    StartTaskGroupThreadData(const StartTaskGroupThreadData& stgtd)
        : m_pThis(stgtd.m_pThis), m_pTaskGroup(stgtd.m_pTaskGroup) {}
    void operator=(const StartTaskGroupThreadData& stgtd)
    {
        m_pThis      = stgtd.m_pThis;
        m_pTaskGroup = stgtd.m_pTaskGroup;
    }

    CMainExecuterStructure*         m_pThis;
    SharedPtr<TaskGroupDefinition>  m_pTaskGroup;
};

BOOL CMainExecuterStructure::StartTaskGroup(SharedPtr<TaskGroupDefinition>& pTaskGroup)
{
    BOOL bRet = TRUE;

#ifndef CYRANO_TP_STANDALONE
    if(m_bFirstTaskGroup)
        InterlockedExchange((LPLONG)&m_bFirstTaskGroup, 0);
    else
    {
        if(m_bStepByStep)
            return FALSE;   // Sorry, cannot start more than 1 task group in step-by-step mode!
    }

    // Actually create a dedicated thread which will take in charge the ramp-up.
    StartTaskGroupThreadData* pStgd = new StartTaskGroupThreadData(this, pTaskGroup);
    HANDLE hThread = BeginThreadEx(
        NULL,
        0,
        StartTaskGroupThreadFunction,
        (void *)pStgd,
        0,
        NULL
    );

    if(hThread == 0)
    {
        bRet = FALSE;
        char zMsg[128];
        wsprintf(zMsg, "ERROR: failed to spawn a thread to start task group #%d", pTaskGroup->m_iTaskGroupNumber);
        TraceMsg(zMsg);
        LogError(zMsg);
    }
    else
    {
        CLock lock(m_otherThreadsLock);
        m_otherThreads.push_back(hThread);
    }
#else
    // Start the task group synchronously: no need to spin-off a thread.
    bRet = pTaskGroup->m_tasks.size() > 0     // Task group must have at least 1 task.
        && m_pThreadPoolController->StartTaskGroup(pTaskGroup);
#endif

    return bRet;
}

DWORD WINAPI CMainExecuterStructure::StartTaskGroupThreadFunction(LPVOID pContext)
{
    StartTaskGroupThreadData* pData = (StartTaskGroupThreadData*)pContext;
    StartTaskGroupThreadData stgd = *pData;
    delete pData;
    DWORD dwRet  = (DWORD)(stgd.m_pTaskGroup->m_tasks.size() > 0     // Task group must have at least 1 task.
        && stgd.m_pThis->m_pThreadPoolController->StartTaskGroup(stgd.m_pTaskGroup));

    char zMsg[128];
    wsprintf(zMsg, "Ramp-up thread %ld exiting...", GetCurrentThreadId());
    stgd.m_pThis->TraceMsg(zMsg, "full");

    return dwRet;
}

void CMainExecuterStructure::StopTaskGroup(int iTaskGroupNumber)
{
    m_pThreadPoolController->StopTaskGroup(iTaskGroupNumber);
}

#ifndef CYRANO_TP_STANDALONE
BOOL CMainExecuterStructure::GetStatsSnapShot(CAsyncVirtualUserController::TestExecuterSnapShot& sn)
{
    return m_pThreadPoolController->GetStatsSnapShot(sn);
}

BOOL CMainExecuterStructure::GetTaskSnapshot(const TaskIdList& taskIDs, TaskSnap2List& taskSnapshots)
{
    return m_pThreadPoolController->GetTaskSnapshot(taskIDs, taskSnapshots);
}

long CMainExecuterStructure::GetTestExecuterVUs() const
{
    return m_pThreadPoolController->GetTestExecuterVUs();
}
#endif

BOOL CMainExecuterStructure::Complete()
{
    TraceMsg("Entering CMainExecuterStructure::Complete()...", "full");
#ifndef CYRANO_TP_STANDALONE
    CStatistics::Release();

#endif

    // All done - perform any cleanup required

    BOOL bExecutionRes = FALSE;

	if(m_pThreadPoolController)
	{
        bExecutionRes = (_threadPoolController->GetFailedUsersCount() == 0);

	    m_pThreadPoolController->DropUsers();
		delete m_pThreadPoolController;
		_threadPoolController = m_pThreadPoolController = NULL;
	}

	if(m_pRule)
    {
        delete m_pRule;
        m_pRule = NULL;
    }

#ifndef CYRANO_TP_STANDALONE
    // Wait for other threads.
    HANDLE hThread;
    do
    {
        {
            CLock lock(m_otherThreadsLock);
            if(!m_otherThreads.empty())
            {
                hThread = m_otherThreads.back();
                m_otherThreads.pop_back();
            }
            else
                hThread = NULL;
        }
        if(hThread != NULL)
        {
            WaitForSingleObject(hThread, 20000);
            CloseHandle(hThread);
        }
    } while(hThread != NULL);
#endif

#ifndef __DLLSA__
    ReleaseCORBA();
#endif

    UnInit();

#ifdef __DLLSA__
    if(m_hWnd != NULL)
    {
        PostMessage(m_hWnd, WM_APP+3,
            (WPARAM)m_hThreadPoolHandlerThread,
            (LPARAM)(new std::string(bExecutionRes ? g_successMsg : g_failureMsg)));
    }
#endif

#ifdef USE_PERF_DATA
    ReleasePerfCounters();
#endif

    return TRUE;
}


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

// If run in debug mode, register as such in order for the test manager to use the current debugging process instead of starting a new one.
BOOL CMainExecuterStructure::RegisterCurrentDebugExecuter()
{
#ifndef CYRANO_TP_STANDALONE
    // Test Manager registration in order to force our debug version to be used.
    string ObjectClass  = TestExecuter::CLASS_NAME;
    //TODO   string debug_reg    = "DEBUG." + zHostIpAddress + '.' + strip_extension(zExeName);
    string debug_reg    = "DEBUG_" + m_zUnderscoredHostIpAddress + '_' + strip_extension("TExecuter_htp.exe");
    string trcmsg       = "Debug requested, register as " + debug_reg;

    CORBA::String_var dbg_reg = CORBA::string_dup(debug_reg.c_str());
    CORBA::String_var obj_reg = CORBA::string_dup(ObjectClass.c_str());

    assert(m_pServer);
    m_pServer->NameObject(TExecuter_ptr, obj_reg.inout(), dbg_reg.inout());
#endif

    return TRUE;
}


// ----------------------------------------------------------------------------
// Trace messages to the log if created.
// ----------------------------------------------------------------------------

void CMainExecuterStructure::TraceMsg(const string &zMsg, const char* zLevel) const
{
    if(m_pTrace)
    {
        if(zLevel == NULL || m_pTrace->IsTrace(zLevel))
        {
            GetTraceStream(m_pTrace) << zMsg << "\r\n";
            ReleaseTraceStream();
        }
    }
}

#ifdef __DLLSA__

// ----------------------------------------------------------------------------
// Start the thread pool handler in a thread different from the Modeller's one.
// ----------------------------------------------------------------------------

BOOL CMainExecuterStructure::StartThreadPoolHandlerThread()
{
    m_hThreadPoolHandlerThread = BeginThreadEx(NULL, 0, ThreadPoolHandler, (PVOID)this, 0, NULL);
    return m_hThreadPoolHandlerThread != NULL;
}

#endif  // __DLLSA__

// ----------------------------------------------------------------------------
// Execute the thread pool controller. Will be started in a separate thread 
// when used in the modeller.
// ----------------------------------------------------------------------------

unsigned CMainExecuterStructure::ThreadPoolHandler(void *pVoidParams)
{
    BOOL bErr = FALSE;

    CMainExecuterStructure *pTestExecuter = (CMainExecuterStructure *)pVoidParams;
    CAsyncVirtualUserController *pThreadPoolController = pTestExecuter->GetThreadPoolController();
    try
    {
        if(pThreadPoolController->Run() != 0)
        {
            pTestExecuter->TraceMsg("ERROR: CAsyncVirtualUserController::Run() failed.");
            pTestExecuter->LogError("ERROR: could not execute the thread pool.");
            bErr = TRUE;
        }
    }
    catch(const mdbRuntimeError &err)
    {
        pTestExecuter->TraceMsg("ERROR in Run: " + err.What());
        pTestExecuter->LogError(("ERROR in Run: " + err.What()).c_str());
        bErr = TRUE;
    }

    pTestExecuter->Complete();  // If DLL, will post a message to the Modeler.

    if(bErr)
    {
        assert(0);
        // TODO: in case of error, cancel everything!
    }

    return 1;
}

void CMainExecuterStructure::Stop()
{
    if(m_pThreadPoolController)
    {
        //// Asynchronous stop.

        // Cancel all VUs.
        m_pThreadPoolController->CancelAllUsers();

        // Causes the pThreadPoolController->Run() call in ThreadPoolHandler()
        // to return. The ThreadPoolHandler() function then calls pTestExecuter->Complete(),
        // which posts a message to the Modeler.
        m_pThreadPoolController->StopController();
    }
}

void CMainExecuterStructure::EvaluateTPParams(DWORD &dwThreadPoolSize, DWORD &dwNbConcurrentThreads, DWORD &dwInitialVUCount, DWORD &dwVUCountGrowBy, DWORD &dwBrowserParallelism)
{
    m_configFile.Get("BrowserParallelism",          (long&)dwBrowserParallelism,    4, "TEST");

#ifdef __DLLSA__
    dwThreadPoolSize = dwNbConcurrentThreads = dwInitialVUCount = dwVUCountGrowBy = 1;
#else
    m_configFile.Get("ThreadPoolSize",              (long&)dwThreadPoolSize,        0,      "THREAD POOL");
    m_configFile.Get("ThreadPoolConcurrentThreads", (long&)dwNbConcurrentThreads,   0,      "THREAD POOL");
    m_configFile.Get("InitialVirtualUserCount",     (long&)dwInitialVUCount,        1000,   "TEST");
    m_configFile.Get("VirtualUserGrowBy",           (long&)dwVUCountGrowBy,         20,     "TEST");

    if(dwNbConcurrentThreads == 0)
    {
        // Automatically compute the thread pool parameters.
        SYSTEM_INFO sI;
        GetSystemInfo(&sI);
        dwNbConcurrentThreads = sI.dwNumberOfProcessors;
    }

    if(dwThreadPoolSize  == 0)
    {
        dwThreadPoolSize = dwNbConcurrentThreads * 25;
    }
#endif
}

void CMainExecuterStructure::LogError(const char* zErrMsg) const
{
#ifndef CYRANO_TP_STANDALONE
    CECErrorLog out(m_pThreadPoolController->GetTimeOffset(), m_zTestName.c_str(), TestExecuter::CLASS_NAME, zErrMsg);
    GetInjectorControl()->AddTestOutput(CECErrorLog::OUTPUT_TYPE, out.ToString().c_str());
#endif
}

void CMainExecuterStructure::SetCachedTestVUCount(long lTestVUCount)
{
    InterlockedExchange(&m_lCachedTestVUCount, lTestVUCount);
}

long CMainExecuterStructure::GetCachedTestVUCount() const
{
    return m_lCachedTestVUCount;
}

#ifndef CYRANO_TP_STANDALONE

BOOL CMainExecuterStructure::StepByStep_NotifyNewRequest(
    const char* zScriptName, long lLineNumber, long lTaskGroupID,
    DWORD dwUser, long lConId, const char* zRequest)
{
    g_pTExecuterImpl->GetStepByStep()->GetStepByStepControl()->NextRequest(
        zScriptName, (CORBA::Long)lLineNumber, (CORBA::Long)lTaskGroupID,
        (CORBA::Long)dwUser, (CORBA::Long)lConId, zRequest);

    return TRUE;
}

void CMainExecuterStructure::StepByStep_NotifyComplete(
    long lTaskGroupID, DWORD dwUser, long lConId, BOOL bIsBinaryData,
    const char* zHeaders, const char* zBody, long lBodyLength)
{
    g_pTExecuterImpl->GetStepByStep()->GetStepByStepControl()->RequestComplete(
        (CORBA::Long)lTaskGroupID, (CORBA::Long)dwUser, (CORBA::Long)lConId,
        (CORBA::Boolean)(bIsBinaryData != FALSE), zHeaders, zBody,
        (CORBA::Long)lBodyLength);
}

void CMainExecuterStructure::NotifyTEStopping()
{
    // Notify the step-by-step control that the Test Executer is stopping.
    StepByStepControl_var& sbs = g_pTExecuterImpl->GetStepByStep()->GetStepByStepControl();
    if(!CORBA::is_nil(sbs))
    {
        try {
            sbs->SetTestExecuter(StepByStepScriptExecuter::_nil());
        } catch(...) {}
    }

    // Notify the Injector Control that the Test Executer is stopping.
    GetInjectorControl()->OnExecuterStopping(GetRegisterName());
}

#endif
