/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
                        /***********************************
                        *              PHOENIX             *
                        * Base class for lexical analyzer. *
                        * Class implementation.            *
                        ***********************************/



#include "loptions.h"                   // Compilation options. Generated by parser generator or lexical generator.
                                        // Includes choice of target platform and array packing

// User code and support routines, Part 1 and Part 3 of the lex source program
#define PH_NO_SCAN_TABLES
#include "wslexrt.h"
#include "winscantref.h"
#include <ctype.h>



//////////////////////////////// Phoenix lexical analyzer ///////////////////////////



template <typename U>
inline PhLexer<U>::PhLexer(U zInput, std::ostream& os)
    : pcInput_(zInput),
      lColumn_(1), lLine_(1), lCurCol_(1), lCurLine_(1),
      lLexErrorType_(0), os_(os),
      pcCurChar_(zInput), pcFirstChar_(zInput),
      bInTag_(FALSE), bInScript_(FALSE)
{
}

// Compute the next state in the automaton, using packed arrays, given
// an input character and a current state.
// Arguments :	ulS (IN): the current state.
//              ulA (IN): the input character.
// Return    :  the new state. 0 is the error state.
template <typename U>
inline unsigned long PhLexer<U>::PackedNextState(unsigned long ulS, unsigned long ulA)
{
    while(PCheck[PBase[ulS] + ulA] != ulS)
        ulS = PDefault[ulS];
    return PNext[PBase[ulS] + ulA];
}

// Get the next token and perform associated action.
// Returns the token number. -1 means fatal lexical error.
template <typename U>
inline long PhLexer<U>::FindToken(unsigned long& ulLine, unsigned long& ulCol, CSymbolAttributes<U>* pTokenAttributes)
{
    long           lRetVal = -1;    // Default: fatal lexical error.
    unsigned long  ulState = DFA_START, ulSavedState;
    long           lToken, lSavedToken = 0;
    U              pcSavedChar = NULL;
    BOOL           bMore, bRecoveryPhase = FALSE;
    unsigned char  ucSavedChar = 0;
    unsigned char  ucCurChar = 0;

    bMore = FALSE;
    
    lLexErrorType_ = 0;

    for(;;)
    {
        if(bMore == FALSE)
        {
            // New Token
            pcFirstChar_ = pcCurChar_;

            // Set current line and column to begining of next token, for the parser.
            lCurCol_  = lColumn_;
            lCurLine_ = lLine_;

            // Read next token, using longest match method.
            ulState = DFA_START;    // Initial state = start state. This constant is in wslexrt.h.
            lSavedToken = 0;
            pcSavedChar = pcFirstChar_;
        }
        else bMore = FALSE;         // bMore = TRUE when PH_MORE has been called by user.

        BOOL bError = FALSE;
        do      // Loop to find longest match for a token.
        {
            ucCurChar = *pcCurChar_;
            if((lToken = TokenTable[ulState]) != 0)
            {
                // Save the token found in this state, and the related input buffer index value.
                ulSavedState = ulState;
                lSavedToken = lToken;
                pcSavedChar = pcCurChar_;
                ucSavedChar = ucCurChar;
            }
            ulState = PackedNextState(ulState, ucCurChar);
            if(ulState)
            {
                UpdateLineCol(ucCurChar);
                ++pcCurChar_;
            }
        } while(ulState);

        if(bError)
            break;

        // Token recognized.
        if(lSavedToken != 0)
        {
            pcCurChar_ = pcSavedChar;
            ucCurChar = ucSavedChar;

            if(pTokenAttributes)
            {
                // Set the common fields of the token attribute variable.
                SetCommonTokenAttributes(*pTokenAttributes);
            }

            if(lSavedToken > 0)
            {
                if(pTokenAttributes)
                {
                    // Set the default token attribute variable's value.
                    SetDefaultTokenAttributes(*pTokenAttributes);
                }

                // Simple return action
                lRetVal = lSavedToken;
                break;
            }
            else    // lSavedToken < 0
            {
                // User defined action.
                lSavedToken = -lSavedToken;
                long lTheToken = OnToken(lSavedToken, pTokenAttributes);
                if(lTheToken != -2)
                {
                    lRetVal = lTheToken;
                    break;
                }
            }
        }
        else    // lSavedToken == 0
        {
            lCurLine_ = lLine_;
            lCurCol_  = lColumn_;
            if(ucCurChar == 0)
            {
                // End of input
                lRetVal = ENDMARKER;
                break;
            }
            else
            {
                // Call error routine.
                if(!OnError(lCurLine_, lCurCol_, bRecoveryPhase))
                {
                    lRetVal = -1;   // Fatal error
                    break;
                }

                // Lexical error: try to resume scanning at character following the first
                // character in the incomplete "token".
                OnEnterErrorRecovery(bRecoveryPhase);
                bRecoveryPhase = TRUE;
                ucCurChar = *pcCurChar_;
                UpdateLineCol(ucCurChar);
                pcCurChar_ = pcFirstChar_;
                ++pcCurChar_;
            }
        }
    }

    // Report any fatal error.
    if(lRetVal == -1)
        OnFatalError(bRecoveryPhase);

    // Report any successful error recovery.
    if(bRecoveryPhase)
        OnErrorRecovered();

    ulLine = lCurLine_;
    ulCol  = lCurCol_;

#ifdef PHTRACE
    if(lRetVal != -1)
    {
        os_ << "FindToken: ";
        DumpLexeme(os_);
        os_ << std::endl;
    }
#endif

    return lRetVal;
}

// Called when a lexical error is detected.
// bDuringRecovery: TRUE if error was detected while attempting to recover
// from a previous error.
// Return: if FALSE, then a fatal error is produced, otherwise,
// the lexical analyzer skips the current character and attempts
// to recover from the error.
template <typename U>
inline BOOL PhLexer<U>::OnError(long lLine, long lCol, BOOL bDuringRecovery)
{
#ifdef PHTRACE
    os_ << "Lexical Error at [" << lLine << "," << lCol << "]";
    if(bDuringRecovery)
        os_ << " while attempting to recover from a previous error";
    os_ << std::endl;
#else
    // To avoid "unreferenced formal parameter" warning.
    lLine; lCol; bDuringRecovery;
#endif
    return TRUE;    // By default, attempt to recover from error.
}

// Called after a lexical error, when attempting error recovery.
// bDuringRecovery: TRUE if error recovery was already in progress
// for a previous error when this function was called.
template <typename U>
inline void PhLexer<U>::OnEnterErrorRecovery(BOOL bDuringRecovery)
{
#ifdef PHTRACE
    os_ << "Trying to recover from lexical error";
    if(bDuringRecovery)
        os_ << " while already attempting to recover from a previous error";
    os_ << std::endl;
#else
    // To avoid "unreferenced formal parameter" warning.
    bDuringRecovery;
#endif
}

// Called on an unrecovered lexical error.
// bDuringRecovery: TRUE if error recovery was already in progress
// for a previous error when this function was called.
template <typename U>
inline void PhLexer<U>::OnFatalError(BOOL bDuringRecovery)
{
#ifdef PHTRACE
    os_ << "Fatal error";
    if(bDuringRecovery)
        os_ << " during error recovery phase.";
    os_ << std::endl;
#else
    // To avoid "unreferenced formal parameter" warning.
    bDuringRecovery;
#endif
}

#ifdef PHTRACE
template <typename U>
inline void PhLexer<U>::DumpLexeme(std::ostream& os)
{
    for(U puc = pcFirstChar_; puc < pcCurChar_; ++puc)
        os << (char)*puc;
}
#endif

template <typename U>
inline 
void PhLexer<U>::CollectComment(CSymbolAttributes<U>& sa)
{
    // Skip to the end of the comment (the character just after).
    int iState = 0;     // 0 =
                        // 1 = -
                        // 2 = --
                        // 3 = -->
    while(*pcCurChar_ && iState != 3)
    {
        switch(*pcCurChar_)
        {
            case '-':
                if(iState == 0)
                    iState = 1;
                else if(iState == 1 || iState == 2)
                    iState = 2;
                break;

            case '>':
                if(iState == 2)
                    iState = 3;
                else
                    iState = 0;
                break;

            default:
                iState = 0;
                break;
        }
        PH_SKIP;
    }

    if(iState == 3)
    {
        sa.label_.pcBegin_ = pcFirstChar_;
        sa.label_.ulSize_  = pcCurChar_ - pcFirstChar_;
    }
}

template <typename U>
inline
void PhLexer<U>::CollectScript(CSymbolAttributes<U>& sa)
{
    // Skip characters until </script (not included).
    U pcBegin = pcCurChar_;
    U pcEnd   = pcBegin;
    int iState = 0;     // 0 =
                        // 1 = <
                        // 2 = </
                        // 3 = </s
                        // 4 = </sc
                        // 5 = </scr
                        // 6 = </scri
                        // 7 = </scrip
                        // 8 = </script
    while(*pcCurChar_ && iState != 8)
    {
        switch(*pcCurChar_)
        {
            case '<':
                iState = 1;
                pcEnd = pcCurChar_;
                break;

            case '/':
                if(iState == 1)
                    iState = 2;
                else
                    iState = 0;
                break;

            case 'S':
            case 's':
                if(iState == 2)
                    iState = 3;
                else
                    iState = 0;
                break;

            case 'C':
            case 'c':
                if(iState == 3)
                    iState = 4;
                else
                    iState = 0;
                break;

            case 'R':
            case 'r':
                if(iState == 4)
                    iState = 5;
                else
                    iState = 0;
                break;

            case 'I':
            case 'i':
                if(iState == 5)
                    iState = 6;
                else
                    iState = 0;
                break;

            case 'P':
            case 'p':
                if(iState == 6)
                    iState = 7;
                else
                    iState = 0;
                break;

            case 'T':
            case 't':
                if(iState == 7)
                    iState = 8;
                else
                    iState = 0;
                break;

            default:
                iState = 0;
                break;
        }

        // Skip without updating line and column.
        if(iState != 8)
            ++pcCurChar_;
    }

    if(iState == 8)
    {
        if(pcBegin != pcEnd)
        {
            sa.label_.pcBegin_ = pcBegin;
            sa.label_.ulSize_  = pcEnd - pcBegin;

            // Update the line and column.
            for(U pc = pcBegin; pc != pcEnd; ++pc)
                UpdateLineCol(*pc);
        }
        pcCurChar_ = pcEnd;
    }
}

template <typename U>
inline 
void PhLexer<U>::CollectText(CSymbolAttributes<U>& sa)
{
    // Skip to the end of the text.
    U pcBegin = pcCurChar_;
    U pcEnd   = pcBegin;
    int iState = 0;     // 0 =
                        // 1 = <
                        // 2 = </
                        // 3 = </a or <a  or  <! or end of input (where a is an alphabetic character).
    while(*pcCurChar_ && iState != 3)
    {
        switch(*pcCurChar_)
        {
            case '<':
                iState = 1;
                pcEnd = pcCurChar_;
                break;

            case '/':
                if(iState == 1)
                    iState = 2;
                else
                    iState = 0;
                break;

            case '!':
                if(iState == 1)
                    iState = 3;
                else
                    iState = 0;
                break;

            default:
                if((iState == 1 || iState == 2) && isalpha(*pcCurChar_))
                    iState = 3;
                else
                    iState = 0;
                break;
        }

        // Skip without updating line and column.
        if(iState != 3)
            ++pcCurChar_;
    }

    if(*pcCurChar_ == 0)
        pcEnd = pcCurChar_;
    else
        pcCurChar_ = pcEnd;

    sa.label_.pcBegin_ = pcFirstChar_;
    sa.label_.ulSize_  = pcEnd - pcFirstChar_;

    // Update the line and column.
    for(U pc = pcBegin; pc != pcEnd; ++pc)
        UpdateLineCol(*pc);
}

template <typename U>
inline
BOOL PhLexer<U>::InScript() const
{
    BOOL bRes = FALSE;
    U pc = pcFirstChar_;
    if(*pc == '<')
    {
        ++pc;
        if(*pc == 's' || *pc == 'S')
        {
            ++pc;
            if(*pc == 'c' || *pc == 'C')
            {
                ++pc;
                if(*pc == 'r' || *pc == 'R')
                {
                    ++pc;
                    if(*pc == 'i' || *pc == 'I')
                    {
                        ++pc;
                        if(*pc == 'p' || *pc == 'P')
                        {
                            ++pc;
                            if(*pc == 't' || *pc == 'T')
                            {
                                bRes = TRUE;
                            }
                        }
                    }
                }
            }
        }
    }

    return bRes;
}

template <typename U>
inline
BOOL PhLexer<U>::RecognizeTag(CSymbolAttributes<U>& sa) const
{
    BOOL bRet = FALSE;
    U pbegin = pcFirstChar_;    // Should be '<'.
    ++pbegin;
    if(*pbegin == '/')
        ++pbegin;

    // pbegin now points to the beginning of the tag's name.
    // Identify the tag.
    char zTagName[16];
    unsigned char *pName = (unsigned char*)zTagName;
    unsigned int uiNameLen = sizeof(zTagName) / sizeof(char);
    unsigned int uiC = 0;
    U pc = pbegin;
    unsigned char c;
    while(uiC < uiNameLen - 1   // Don't write beyond the end of zTagName.
          && pc != pcCurChar_   // Don't go beyond the end of the token.
          && (c = *pc) != '>'
          && c != ' '
          && c != '\t'
          && c != '\r'
          && c != '\n')
    {
        *pName++ = (char)(islower(c) ? _toupper(c) : c);
        ++pc;
        ++uiC;
    }
    *pName = '\0';

    NameToTag::const_iterator it = nameToTag.find(zTagName);
    bRet = it != nameToTag.end();
    if(bRet)
        sa.eTag_ = it->second;

#ifdef PHTRACE
    os_ << "RECOGNIZING TAG '" << zTagName << "' -> "
        << sa.eTag_ << endl;
#endif

    return bRet;
}

template <typename U>
inline
long PhLexer<U>::OnToken(long lToken, CSymbolAttributes<U>* pTokenAttributes)
{
    switch(lToken)
    {
      #include "wslexac.h"
    }
    return -2;
}
