/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//==================================================================< Pragmas >
#ifdef  WIN32
#  pragma warning(disable : 4710)
#endif  // WIN32


//=================================================================< Includes >
#include <ctime.hpp>

#ifndef __GATEWAY_LISTEN_CLIENT__
#  include "gatewlc.hpp"
#endif  // __GATEWAY_LISTEN_CLIENT__


//====================================================================< Class >

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////// Class CGGatewayListen //
///////////////////////////////////////////////////////////////////////////////

//============================================< Constructor(s) and destructor >
//-----------------------------------------------------------------------------
// CGGatewayListen::CGGatewayListen
//
// Constructor.
//
// Arguments : task : task associeted.
// Return    : none.
//-----------------------------------------------------------------------------
CGGatewayListen::CGGatewayListen (CGTask& task)
        : task_ (task), gatewayInfo_ (*( (CGGatewayInfo*) (task_.getInitInfo () ) ) ), pLibrary_ (NULL)
{
    // Set the task data.
    task_.setData (this) ;

    // Save this pointer.
    CTCurrentThread::setData (this) ;

    // Set the abort routine.
    CTCurrentThread::setAbortRoutine (abort) ;
} // CGGatewayListen::CGGatewayListen

//-----------------------------------------------------------------------------
// CGGatewayListen::~CGGatewayListen
//
// Destructor.
//
// Arguments : none.
// Return    : none.
//-----------------------------------------------------------------------------
CGGatewayListen::~CGGatewayListen ()
{
    if (GpProcessMain->getTraceMode () )
    {
        CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
        GpProcessMain->getLogFileList () << task_.getTaskName () << ": listener <" << listen_.getInetAddress ().getHostAddress () << ":" << listen_.getPort () << "> closed." << endl ;
    }
    else
    {
        CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
        GpProcessMain->getLogFileList () << "Close listener on <" << listen_.getInetAddress ().getHostAddress () << ":" << listen_.getPort () << ">." << endl ;
    }

    // Close the listen socket.
    listen_.close () ;

    if (pLibrary_)
    {   // Unload library.
        delete pLibrary_ ;
        pLibrary_ = NULL ;
    }
} // CGGatewayListen::~CGGatewayListen

//=========================================================< Control methodes >
//-----------------------------------------------------------------------------
// CGGatewayListen::init
//
// Listen for the gateway clients to connect.
//
// Arguments : none.
// Return    : none.
//-----------------------------------------------------------------------------
void CGGatewayListen::init (void)
{
    string  libraryName ;

    // Get the library.
    if (gatewayInfo_.getLibrary () != "")
    {
        libraryName = gatewayInfo_.getLibrary () ;
        if (GpProcessMain->getTraceMode () )
        {
            CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
            GpProcessMain->getLogFileList () << task_.getTaskName () << ": try to load the library '" << CTLibrary::getSystemFileName (CTLibrary::getBaseFileName (libraryName) ) << "'." << endl ;
        }
        try
        {
            pLibrary_ = new CTLibrary (libraryName) ;

            // Determinate if there is a GWParse version variable.

            if (GpProcessMain->getTraceMode () )
            {
                CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
                GpProcessMain->getLogFileList () << task_.getTaskName () << endl ;
            }
        }
        catch (CTxalloc&)
        {   // Error from new.
            throw CGTaskException (string ("init: fail to load the library '") + CTLibrary::getSystemFileName (CTLibrary::getBaseFileName (libraryName) ) + "': new CTLibrary (libraryName) failed.", __FILE__, __LINE__) ;
        }
        catch (CTException& e)
        {
            throw CGTaskException (string ("init: fail to load the library '") + CTLibrary::getSystemFileName (CTLibrary::getBaseFileName (libraryName) ) + "'.\n" + e.what (), __FILE__, __LINE__) ;
        }

        try
        {
            // Pass the library command line arguments.
            CTProcedure                 procSetCmdLineArg (*pLibrary_, SETCMDLINEARG) ;
            CGWParse::fnSetCmdLineArg   setCmdLineArg = (CGWParse::fnSetCmdLineArg) (procSetCmdLineArg.getAddress () ) ;

            setCmdLineArg (GpProcessMain->getLibraryCommandLineArguments () ) ;
        }
        catch (CTException&)
        {}

    }

    if (GpProcessMain->getTraceMode () )
    {
        CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
        GpProcessMain->getLogFileList () << task_.getTaskName () << ": try to start listener on port " << gatewayInfo_.getPort () << "." << endl ;
    }

    port_t  tListenPort ;

    // Get the gateway port.
    tListenPort = gatewayInfo_.getPort () ;

    // Specify if reuse of address is allow.
    listen_.reuseAddress (GpProcessMain->canReuseAddress () ) ;

    // Binding the listen socket to the gateway port.
    listen_.bindToPort (tListenPort) ;

    // Mark the listen socket as a listen socket.
    listen_.listen (10) ;

    if (GpProcessMain->getTraceMode () )
    {
        CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
        char szAddress [12] ;
        char szLine [80] ;
        if (pLibrary_)
        {
            sprintf (szLine, "<target listener       <- '%-9.9s' <- application client   >", CTLibrary::getSystemFileName (CTLibrary::getBaseFileName (libraryName) ).c_str () ) ;
        }
        else
        {
            sprintf (szLine, "<target listener       <- gateway     <- application client   >") ;
        }
        GpProcessMain->getLogFileList () << task_.getTaskName () << ": " << szLine << endl ;
        sprintf (szAddress, "xxxxx/%hu", listen_.getPort () ) ;
        sprintf (szLine, "<%-21.21s <- %-11.11s <- aaa.bbb.ccc.ddd:xxxxx>", gatewayInfo_.getTargetAddress (), szAddress) ;
        GpProcessMain->getLogFileList () << task_.getTaskName () << ": " << szLine << endl ;
        if (gatewayInfo_.getTargetAddressAutoDetectionMode () )
        {
            GpProcessMain->getLogFileList () << task_.getTaskName () << ": target address auto detect mode is actif." << endl ;
        }
    }
    else
    {
        CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
        GpProcessMain->getLogFileList () << "Listening " << (char*) (gatewayInfo_.getTargetAddressAutoDetectionMode () ? "default " : "") << "on <" << listen_.getInetAddress ().getHostAddress () << ":" << listen_.getPort () << "> for <" << gatewayInfo_.getTargetAddress () << "> using " << CTLibrary::getSystemFileName (CTLibrary::getBaseFileName (libraryName) ) << "." << endl ;
    }

    // Tell the control task that initialization is terminated.
    task_.initializationTerminated () ;
} // CGGatewayListen::init

//-----------------------------------------------------------------------------
// CGGatewayListen::listenForConnections
//
// Listen for the gateway clients to connect.
//
// Arguments : none.
// Return    : none.
//-----------------------------------------------------------------------------
void CGGatewayListen::listenForConnections (void)
{
    try
    {
        CTSocket::CTSelection event ;

        do
        {
            if (task_.hasToBeReinitialized () )
            {
                task_.reinitialized () ;
            }

            // Listen for the gateway clients to connect.
            event = listen_.poll (_S_RECEIVE_ | _S_EXCEPTION_) ;
            switch (event)
            {
            case CTSocket::RECEIVE :
                // A client wants to connect.
                newClient () ;
                break ;
            case CTSocket::TIMEDOUT :
                // No client is waiting.
                break ;
            case CTSocket::EXCEPTION :
                // Error on socket listen_.
                throw CTSocketException ("listenForConnections: Exception event on socket listen_.", __FILE__, __LINE__) ;
                break ;
            default :
                // Unexpected event on socket listen_.
                throw CTSocketException ("listenForConnections: Unexpected event on socket listen_.", __FILE__, __LINE__) ;
            }
            CTCurrentThread::sleep (1) ;
        }
        while ( (event == CTSocket::TIMEDOUT) || (event == CTSocket::RECEIVE) ) ;
    }
    catch (CTSocketException& e)
    {
#ifdef  WIN32
        if (e.getNumber () != WSAEINTR)
#endif  // WIN32
#ifdef  UNIX
            if (e.getNumber () != EINTR)
#endif  // UNIX
            {
                throw e ;
            }
    }
} // CGGatewayListen::listenForConnections

//-----------------------------------------------------------------------------
// CGGatewayListen::newClient
//
// Accept a new client connection.
//
// Arguments : none.
// Return    : none.
//-----------------------------------------------------------------------------
void CGGatewayListen::newClient (void)
{
    try
    {
        // Create an gateway client task.
        task_.getTasksList ().newTask (CGTask::TASK_GATEWAY_CLIENT, (void*) this, task_) ;
    }
    catch (CGTaskException&)
    {}

} // CGGatewayListen::newClient

//-----------------------------------------------------------------------------
// CGGatewayListen::abort
//
// Abort Routine.
//
// Arguments : pStatus : pointer to the exit status.
// Return    : none.
//-----------------------------------------------------------------------------
void CGGatewayListen::abort (void*)
{
    // Delete the CGGatewayListen pointer.
    ( (CGGatewayListen*) CTCurrentThread::getData () )->~CGGatewayListen () ;
} // CGGatewayListen::abort

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////// End of Class CGGatewayListen //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////// Globals Functions //
///////////////////////////////////////////////////////////////////////////////

void gatewayListenTask (CGTask& task)
{
    try
    {
        if (GpProcessMain->getTraceMode () )
        {
            CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
            GpProcessMain->getLogFileList () << task.getTaskName () << ": new gateway listen task." << endl ;
        }

        // Destruction of this object must occur before killing the task.
        CGGatewayListen gatewayListen (task) ;
        gatewayListen.init () ;
        gatewayListen.listenForConnections () ;

        if (GpProcessMain->getTraceMode () )
        {
            CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
            GpProcessMain->getLogFileList () << task.getTaskName () << ": gateway listen task terminated." << endl ;
        }
    }
    catch (CTException& e)
    {
        if (GpProcessMain->getTraceMode () )
        {
            CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
            GpProcessMain->getLogFileList () << task.getTaskName () << ": !! " << e.what () << endl ;
            GpProcessMain->getLogFileList () << task.getTaskName () << ": gateway listen task terminated." << endl ;
        }
        else
        {
            CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
            GpProcessMain->getLogFileList () << "!! " << e.what () << endl ;
        }
    }
    catch (...)
    {
        if (GpProcessMain->getTraceMode () )
        {
            CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
            GpProcessMain->getLogFileList () << task.getTaskName () << ": !! unknown exception." << endl ;
            GpProcessMain->getLogFileList () << task.getTaskName () << ": gateway listen task terminated." << endl ;
        }
        else
        {
            CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
            GpProcessMain->getLogFileList () << "!! unknown exception." << endl ;
        }
    }
    if (task.initializationIsTerminated () )
    {   // Remove the task from the list.
        task.invalidate () ;

        for (int iIndex = 0; iIndex < (int)task.getTasksList ().getTasksList ().size(); iIndex++)
        {
            if (task.getTasksList ().getTasksList ()[iIndex] == &task)
            {
                task.getTasksList ().killTask (iIndex) ;
                break ;
            }
        }
    }
    else
    {   // Tell the control task that initialization is terminated.
        task.initializationTerminated (false) ;
    }
} // gatewayListenTask


///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////// End of Globals Functions //
///////////////////////////////////////////////////////////////////////////////
